/* Conditions of use:                                                        */
/* 1. donlp2 is under the exclusive copyright of P. Spellucci                */
/*    (e-mail:spellucci@mathematik.tu-darmstadt.de)                          */
/*    "donlp2" is a reserved name                                            */
/* 2. donlp2 and its constituent parts come with no warranty, whether ex-    */
/*    pressed or implied, that it is free of errors or suitable for any      */
/*    specific purpose.                                                      */
/*    It must not be used to solve any problem, whose incorrect solution     */
/*    could result in injury to a person , institution or property.          */
/*    It is at the users own risk to use donlp2 or parts of it and the       */
/*    author disclaims all liability for such use.                           */
/* 3. donlp2 is distributed "as is". In particular, no maintenance, support  */
/*    or trouble-shooting or subsequent upgrade is implied.                  */
/* 4. The use of donlp2 must be acknowledged, in any publication which       */ 
/*    contains                                                               */
/*    results obtained with it or parts of it. Citation of the authors name  */
/*    and netlib-source is suitable.                                         */
/* 5. The free use of donlp2 and parts of it is restricted for research      */
/*    purposes                                                               */
/*    commercial uses require permission and licensing from P. Spellucci.    */

/* d o n l p  2                                                              */

/*    version 28/11/2001 (*)                                                 */
/*    tauqp dependent on scalres only                                        */
/*    weights computed in a modified version in the singular case            */
/*    some comparisons relative to zero changed                              */
/*    error-return for function evaluation as added feature                  */
/*    termination of QP-solver changed (d not set to zero)                   */
/*    new version of BFGS: if nr  =  0 take Powell's update                  */
/*    no suppression of update beforehand (with exception dg = 0)            */
/*    plus some minor corrections                                            */

/*    for consistency reasons variable names cf and cgf changed into         */
/*    icf and icgf                                                           */
/*    added feature numerical differentiation of order 1,2,6                 */
/*    requires new parameters epsfcn, taubnd, difftype                       */
/*    added feature of external blockwise evaluation of functions            */
/*    rather than individual evaluation                                      */
/*    requires new parameter bloc                                            */
/*    new feature of user-supplied scaling of x                              */

/* ************************************************************************* */
/* (*) using a c-version of donlp2 as of 1998 made by                        */
/*     by S. Schoeffert, ASB, Bourges, France                                */
/*     dynamic memory allocation added by k.s. Cover                         */
/*     copyrights and right of commercial exploitation remaining by contract */
/*     with P. Spellucci                                                     */
/* ************************************************************************  */
/*  new function format                                                      */
/*  developed in 2002/2003 by P. Spellucci copyrigth P. Spellucci            */
/*  The problem format here is                                               */
/*     min_x f(x)                                                            */ 
/*                                                                           */
/*     subject to                                                            */
/*       o8low(i) <= o8x(i)   <=   o8up(i) , i=1,..,o8n                              */
/*     o8low(i+o8n) <= (Ax)_i <= o8up(i+o8n) , i=1,o8nlin     o8nlin=0 is allowed        */
/*o8low(i+o8nlin+o8n) <= c_i(o8x) <= o8up(i+o8n+o8nlin)  i=1,o8nonlin, o8nonlin=0 allowe       */
/*                                                                           */
/*                                                                           */
   
/*  o8low(i)=o8up(i) is allowed .   In this case only o8one constraint is used     */ 
/*  o8low(i)=-o8big, o8up(i)=o8big is allowed, o8big is userdefined                    */
/*  these are a total of 2*(o8n+o8nlin+o8nonlin) constraints, numbered             */
/*  consecutively                                                            */
/*  such that odd numbers correspond to the lower and even numbered to the   */
/*  upper                                                                    */
/*  bound.  A is to be stored in the o8nlin first columns of o8gres.             */
/*  the user evaluation routines for the constraints only deal with the      */
/*  nonlinear part and the objective function.                               */
/*  gradients of the bound constraints are never explicitly stored .         */
/*  the relevant parts of the code have been changed to reflect this.        */
/*  the parameters                                                           */ 
/*   o8n,o8nlin,o8nonlin,o8big                                                       */
/*   o8low                                                                     */
/*   o8up                                                                      */
/*   O8A  = o8gres[][i],   i=1,...,o8nlin                                          */
/*   are to be set (or read from a file) in the basic routine user_init      */
/*   the parameters o8tau0, o8del0, o8delmin etc keep there meaning and usage.     */
/*                                                                           */
/*  A new form of evaluation interface is used for the gradients:            */
/*  only the gradients of the nonlinear constraints must be evaluated,       */
/*  for a given list of indizes, that is grad_c_i , i in list.               */
/*  Also the nonlinear functions c_i(o8x) are evaluated blockwise              */ 
/*  bug fix: mes-file written before opened (correction of initial value)    */
/*  bug fix: dead loop in o8qpdu in case of degenerate and illconditioned    */
/*      primal solution                                                      */
/*   bug fix:   log(o8b2n) now only for o8b2n>0                                  */
/*   bug fix in user_eval.c: o8nres replaced by o8nonlin                         */
/*  fix of a problem with glibc : o8bind and o8bind0 renamed to o8bind and o8bind0*/
/*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>

/* added this one for DBL_MIN, DBL_EPSILON */
#include <float.h>

#include "o8gene.h"
#include "donlp.h"

/* ARRAY_BORDER_CHECK: Check array borders for indices writing off
   the end of arrays.  Turn off by commenting out the line
   "#define ARRAY_BORDER_CHECK".  It is recommended that  ARRAY_BORDER_CHECK
   be turned off during regular use and  turned on for debugging
   and testing only.  The function "checkArrayBorders" checks if there has been
   any array border violations o8up to point where it is called.  It is currently
   called at the beginning and end of donlp2 if border checking is turned on.
   To find the location of the first border violation insert additional 
   calls to the function narrowing the bracketing on the initial border
   violation occurance.  Additionally, the values of the border counters
   (d_borders_count, i_borders_count, l_borders_count) can be printed
   out at various locations in global_mem_malloc to find out
   which array has been assigned which index. */
/*
#define ARRAY_BORDER_CHECK  
*/

#ifdef ARRAY_BORDER_CHECK
#define ABC 1
#else
#define ABC 0
#endif

#ifdef ARRAY_BORDER_CHECK 

#define ABC_NUM_1DARRAYS 100000
DOUBLE  *o8d_border[2*ABC_NUM_1DARRAYS];
INTEGER *o8i_borders[2*ABC_NUM_1DARRAYS];
LOGICAL *o8l_borders[2*ABC_NUM_1DARRAYS];
const DOUBLE  o8d_unique = 1.23413936727;
const INTEGER o8i_unique = 972897426;
const LOGICAL o8l_unique = 953712449;
INTEGER o8d_border_count = 0;
INTEGER o8i_borders_count = 0;
INTEGER o8l_borders_count = 0;
extern void checkArrayBorders(char *str); 
#endif


/* Global variables only used locally */

DOUBLE   *donlp2_yy; 
DOUBLE   *o8bfgs_dg,*o8bfgs_adx,*o8bfgs_ltdx, 
         *o8bfgs_gtdx,*o8bfgs_updx,*o8bfgs_updz;
DOUBLE   *o8opti_qtx;
DOUBLE   *o8opti_yy,*o8opti_yx,*o8opti_trvec,*o8opti_con;
INTEGER  *o8opti_delist,*o8opti_bindba;
DOUBLE   *o8eval_con;
DOUBLE   *o8unim_step;
DOUBLE   *o8dec_qri,*o8dec_qri0;
DOUBLE   *o8elim_qri,*o8elim_y,*o8elim_rhsscal;
INTEGER  *o8elim_col;
DOUBLE   *o8sol_xl;
DOUBLE   *o8upd_sdiag,*o8upd_rn1,*o8upd_w;
DOUBLE   *o8qpdu_g0,*o8qpdu_ci0,*o8qpdu_cii,
                    *o8qpdu_cei,
                    *o8qpdu_xd,*o8qpdu_s,*o8qpdu_z,*o8qpdu_vr;
DOUBLE   *o8qpdu_xdold,*o8qpdu_udold;
INTEGER  *o8qpdu_ai,*o8qpdu_iai,*o8qpdu_iaexcl,*o8qpdu_aiold;
INTEGER  *o8qpdu_eqlist,*o8qpdu_iqlist;
DOUBLE   *o8qpdu_y,*o8qpdu_mult;
INTEGER  *o8qpdu_qpdel;
LOGICAL  *escongrad_errloc;
DOUBLE   *escongrad_fhelp1,*escongrad_fhelp2,
         *escongrad_fhelp3,*escongrad_fhelp4,
         *escongrad_fhelp5,*escongrad_fhelp6;


void donlp2(void){

    void    o8st   (void);
    void    o8fin  (void);
    void    o8opti (void);
    DOUBLE  o8vecn (INTEGER nl,INTEGER nm,DOUBLE o8x[]);
    void    esgradf(DOUBLE o8x[],DOUBLE o8gradf[]);
    
    void    setup0 (void);
    void    setup  (void);
    void user_init_size (void);
    void user_init (void);
    void    o8msg(INTEGER num);
    void    o8elim(void);
    static DOUBLE   term ;
    static INTEGER  i,j,k,iz;
    static char     fil[13],xxx[9] = "xxxxxxxx",name1[9];    
    
/* --------------->  step o8one of problem initialization by user               */
    /*  user_init_size must initialize  o8n, o8nlin, o8nonlin, o8iterma, o8nstep.       */
    /*  These values may not be changed.                                      */

    user_init_size();
    o8ndualm = 2*o8n+o8nlin+o8nonlin;
    o8mdualm = 2*(o8n+o8nlin+o8nonlin);


#ifdef ARRAY_BORDER_CHECK
    checkArrayBorders("donlp2: after global_mem_malloc"); 
#endif
    
    /* default settings of new parameters */
    
    o8bloc     = FALSE;
    o8analyt   = TRUE;
    o8valid    = FALSE;
    o8epsfcn   = 1.e-16;
    o8difftype = 3;
    o8taubnd   = 1.;
    for (i = 1 ; i <= o8n ; i++) {
        o8xsc[i] = o8one;
        o8xtr[i] = o8zero;
    }
    o8epsdif = o8tm8;
    for (i = 0 ; i <= o8nlin+o8nonlin ; i++) {
        o8val[i]    = FALSE;
        if ( i > 0 ) o8gresn[i] = o8one;
    }
    o8ffuerr = FALSE;

    /* some standard initialization which may be overwritten by */
    /* user_init or setup                                       */

    o8silent = FALSE;
    
    /* the interactive input feature is no longer supported here. for the sake */
    /* of easy revision the variable is supplied here however                  */
    /* if o8intakt is set TRUE, output to protocol file is copied to stdout in   */
    /* addition                                                                */
    
    o8intakt = FALSE;
    o8te0    = TRUE;   /* control running optimizer on stdout */
    o8te1    = FALSE;   /* information about iteration in compressed form */
    o8te2    = FALSE;   /* detailed output in case of trouble */
    o8te3    = FALSE;   /* print gradients and Hessian        */
    o8cold   = TRUE;
    o8big    = 1.e20 ; 

/* ->   here user initialization continued                                    */

    
    /*  user_init must initialize o8analyt, o8epsdif, o8del0, o8tau0 ,                */
    /*  o8low, o8up, and the gradients of the linear constraints, if any          */
    /*  (stored in o8gres[1:o8n][j], j=1,o8nlin)                                   */
    /*  o8bloc                                                                  */
    /*  o8analyt and if o8analyt = FALSE then also o8epsfcn , o8taubnd , viobnd ,     */
    /*  o8difftype                                                              */
    /*  and the initial value for o8x                                           */
    /* may also change settings of all variables initialized above            */

    user_init();


/*   open files for output if wanted                                          */
    
    j = 0;
    while ( o8name[j] == ' ' ) {
        j = j+1;
    }
    if ( o8name[j] == '\0' ) {
        strcpy(name1,xxx);
    } else {
        k = j+1;
        while ( o8name[k] != ' ' && o8name[k] != '\0' && k-j < 8 ) {
            k = k+1;
        }
        strncpy(name1,&o8name[j],k-j);
        name1[k-j] = '\0';
        
        for (i = 0 ; i <= k-j-1 ; i++) {
            iz = name1[i];
            if ( iz < 48 || ( iz > 57 && iz < 65 )
            || ( iz > 90 && iz < 97 ) || iz > 122 ) name1[i] = 'x';
        }
        if ( k - j < 8 ) strncat(name1,xxx,8-k+j);
    }
    if ( ! o8silent ) 
    {
        strcpy(fil,name1);
        o8meu  = fopen(strcat(fil,".mes"),"w");
        strcpy(fil,name1);
        o8prou = fopen(strcat(fil,".pro"),"w");
    }




/* ************************************************************************** */
/* ***** automatic correction of o8del0                                         */
/******************************************************************************/

    for ( i = 1 ; i <= o8n ; i++ )
    {
        if ( o8xsc[i] == o8zero ) 
        {
            Rprintf("scaling variable %i is o8zero\n",i);
            exit(1);
        }
    }
    o8nres = o8n+o8nlin+o8nonlin ; 
    for ( i = 1 ; i <= o8nres ; i++ )
    {
       if ( i <= o8n ) 
       {
         term = o8xsc[i] ;
       }
       else
       {
         term = o8one ;
       }
       if (! (o8low[i] == o8up[i]) )
       { 
          o8del0 = min ( o8del0 , (o8up[i]-o8low[i])/(o8four*term) );
       }
    }
    o8delmin = o8tm6;

    for (i = 1 ; i <= o8n ; i++) 
    {
        
        o8ug[i] = o8low[i]; 
        if ( o8ug[i] <= -o8big )    
        {
           o8llow[i] = FALSE ; 
        }   
        else
        {
           o8llow[i] = TRUE ;  
        }   
        o8og[i] = o8up[i];
        if ( o8og[i] >= o8big ) 
        {
           o8lup[i] = FALSE ; 
        }
        else 
        {
           o8lup[i] = TRUE ;  
        }
           
    }      
    
    for (i=1; i<=o8n; i++)
    {
       if ( o8x[i] < o8low[i] || o8x[i] > o8up[i] )
       {
          o8corr = TRUE ;
          if ( o8llow[i] && o8lup[i]  ) o8x[i] = (o8up[i]+o8low[i])/2.0;
          if ( o8llow[i] && !o8lup[i] ) o8x[i] = o8low[i]+2.0*o8del0;
          if ( o8lup[i]  && !o8llow[i]) o8x[i] = o8up[i]-2.0*o8del0; 
       }
    }   

    if ( o8corr && ! o8silent ) o8msg(13);

    for (i = 1 ; i <= o8n ; i++) {
        o8xst[i] = o8x[i];
        o8x[i] = o8x[i]/o8xsc[i];
        o8opti_yy[i] = o8xsc[i];

    }
    for (i = 1 ; i <= o8n ; i++) {
    
        /* o8ug and o8og have been evaluted for the original variables */
        /* here we use them internally for the scaled ones         */
        /* o8ug = o8low o8up to scaling etc                              */
        o8ug[i] = o8ug[i]/o8xsc[i];
        o8og[i] = o8og[i]/o8xsc[i];
    }
    for ( i = 1 ; i <= o8nlin ; i++ )
    { 
 /*   rescale the gradients of the linear constraints */
      o8gresn[i] = o8zero ;
      for ( j = 1 ; j <= o8n ; j++ )
      {
          o8gres[j][i] = o8xsc[j]*o8gres[j][i] ;
          o8gresn[i] += pow(o8gres[j][i],2);
      }
      o8gresn[i] = max ( o8one , sqrt(o8gresn[i]) ) ;
      o8cgres[i] = 1 ;
      o8val[i]   = TRUE ;
    }
/*    the sign of the gradient is determined from the current constraint */
/*    and stored in o8gres[0][i]. done afterwards                          */
    o8nreset = o8n;

    o8st();
    
    /* setup may change standard settings of parameters  */
    /* and add some computations in the user environment */
    o8qpitma = o8nres  ;
    setup();
    
    if ( o8taubnd <= 0 ) {
      Rprintf("o8taubnd le o8zero is not allowed ");
      exit(1);
    }
    for ( i = 1 ; i<= o8n ; i++ ) {
      if ( o8opti_yy[i] != o8xsc[i] ) {
        Rprintf("setup has changed o8xsc, not allowed");
        exit(1);
      }
    }
    /* preevaluation of gradients of linear functions    */
    /* done now in user_init                             */


    o8runtim = clock();
    /* check for redundant linear equality constraints   */
    o8elim();
    

    /* call the optimizer */
    
    o8opti();
    

    o8runtim = (clock()-o8runtim)/CLOCKS_PER_SEC;

    /* do final solution check and output */

    o8fin();

#ifdef ARRAY_BORDER_CHECK
    checkArrayBorders("donlp2: before global_mem_free"); 
#endif

    return;
}

/* **************************************************************************** */
/*        initialization program , standard parameter settings done here        */
/* **************************************************************************** */
void o8st(void) {

    void o8msg    (INTEGER num);
    void user_eval(DOUBLE xvar[],INTEGER mode);

    static INTEGER  i,j,k;
    static DOUBLE   tol1 ,bd0,infiny,gxi,hxi,term;
    static time_t   tim;

    /* 11/1/06 took this out for compatibility with R
       where we can't use -ffloat-store
    o8epsmac = pow(o8two,-20);
    
    L100:

    o8epsmac = o8epsmac/o8two;
    term   = o8one+o8epsmac;
    
    if ( term != o8one ) goto L100;
    
    o8epsmac = o8epsmac+o8epsmac;
    o8tolmac = o8epsmac;
    
    L200:

    tol1   = o8tolmac;
    o8tolmac = o8tolmac/o8twop4;
    
    if ( o8tolmac != o8zero ) goto L200;
    
    o8tolmac = tol1;
    
    */

    o8epsmac = DBL_EPSILON;
    o8tolmac = DBL_MIN;

    /* o8epsmac machine precision, o8tolmac smallest machine number */
    /* larger than o8zero (approximately , base 16 for exponent   */
    /* therefore division by 16 assumed)                        */

    /*                        ***** warning *****                        */
    /* on some machines the computation of o8tolmac may result in an error */
    /* because underflow is not accepted as o8zero as is assumed here      */

    /* warning: the computation of o8epsmac may give faulty results if     */
    /* compiler uses value residing in register instead of the one stored */
    /* back to memory. with gcc use the switch  -ffloat-store in order */
    /* to prevent this                                                 */

    if ( o8tau0 == o8zero ) o8tau0 = o8one;
    if ( o8del0 == o8zero ) o8del0 = o8tau0*o8p5;
    
    if ( o8nreset > o8n  ) o8nreset = o8n;
    if ( o8nreset <= 4 ) o8nreset = 4;
    
    /* standard initialization */
    
    o8lastch = 0;
    o8lastdw = 0;
    o8lastup = 0;
    o8level  = o8one;
    o8tau    = o8tm1;
    o8epsx   = o8tm5;
    o8sigsm  = sqrt(o8epsmac);
    o8smalld = o8tm1;
    
    /* formerly o8tm2. o8smalld has much influence on the Maratos-effect */
    
    o8smallw = exp(o8two*log(o8epsmac)/o8three);
    o8rho    = o8tm6;
    o8rho1   = o8tm10;
    o8del01  = o8del0/o8tp1;
    o8delmin = min(o8del01,max(o8tm6*o8del0,o8smallw));
    if ( ! o8analyt ) o8delmin = min(o8del01,max(o8epsdif,o8delmin));

    o8c1d    = o8tm2;
    o8scfmax = o8tp4;
    o8taufac = o8tp1;
    o8taumax = pow(o8scfmax,2);
    o8updmy0 = o8tm1;
    
    /* take o8del0 and o8tau0 from  user_init in the user function suite       */
    /* may be modified by subsequent call of setup                         */

    
    infiny = o8epsmac/o8tolmac;
    o8fx     = o8zero;
    o8b2n    = o8zero;
    o8b2n0   = o8zero;
    o8nres   = o8n+o8nlin+o8nonlin;
/*  but remember there are formally 2*o8nres constraints */
    if ( o8cold ) 
    {
    /*  initialize the quasi Newton update                 */
      for (i = 1 ; i <= o8n ; i++) 
      {
         for (j = 1 ; j <= o8n ; j++) 
         {
            o8a[i][j] = o8zero;
         }
         o8a[i][i]  = o8one;
         o8diag0[i] = o8one;
      }
      o8matsc = o8one ;
    }
      for (i = 1 ; i <= o8n ; i++) 
      {
        o8diag[i] = o8zero;
      }
      for ( i = 1 ; i <= o8nres ; i++ )
      {
        for (j = 1 ; j <= o8n ; j++) 
        {
          o8qr[j][i]   = o8zero;
        }
      }
      for ( i = o8nlin+1 ; i <= o8nlin+o8nonlin ; i++ )
      {
         for ( j = 1 ; j <= o8n ; j++ )
         {
           o8gres[j][i] = o8zero;
         }
         o8gres[0][i] = o8one ;
      }
    
    
    /* nonlinear gradients not yet evaluated */
    for ( i = o8nlin+1 ; i<=o8nlin+o8nonlin ; i++ ) o8val[i] = FALSE ;    
  
    if ( o8bloc )  
    {    
       o8valid = FALSE ;
       /* user_eval must reset o8valid */
       user_eval(o8x,1);
    }
    /* this call gives all function and gradient values in o8fu[] and o8fugrad[][]*/
    /* later call of escon and escongrad simply gives a move in this case     */


    for ( i = 1 ; i <= 2*o8nres ; i ++) 
    {
        o8bind[i]  = 0;
        o8bind0[i] = 0;
        o8u[i]     = o8zero;
        o8u0[i]    = o8zero;
        /* initial weights of the penalty-term */
        if ( o8cold ) o8w[i] = o8one;
    }   

    for ( i = 1 ; i <= o8nlin+o8nonlin ; i++ )
    {
        o8cres[i] = 0 ;
        o8cgres[i] = 0 ;
    }

    o8clow = o8one;
    o8ny   = o8two;

    /* o8scf  = weight factor for objective function    */
    /* o8scf0 = damping factor for tangential direction */
    
    o8scf    = o8one;
    o8scf0   = o8one;
    o8sigla  = o8twop11;
    o8beta   = o8four;  /* formerly o8two */
    o8alpha  = o8tm1;
    o8delta1 = o8p9;
    o8delta  = o8tm3;   /* o8delta = o8tm2  formerly */
    o8theta  = o8p9;    /* o8theta = 0.99 formerly */
    o8icf    = 0;
    o8icgf   = 0;
    if ( ! o8silent ) {
        fprintf(o8prou,"donlp2-intv 21/06/04, copyright P. Spellucci\n");
        
        time(&tim);
        
        fprintf(o8prou,"%s",ctime(&tim));
        fprintf(o8prou,"%s\n", o8name);
        
        fprintf(o8meu, "donlp2-intv 21/06/04, copyright P. Spellucci\n");
        fprintf(o8meu, "%s",ctime(&tim));
        fprintf(o8meu, "%s\n", o8name);
    }
    return;
}

/* **************************************************************************** */
/*                        do final solution check and output                    */
/* **************************************************************************** */
void o8fin(void) {

    void    solchk(void);

    static INTEGER  i,j,k,ih1,ih2,ih3,ih5,ih6,ih7,ih8,ih9;
    static DOUBLE   umin,term;
    static INTEGER  nsing,crtot,cgrtot,nupreg,nbfgs1,nresta;
    static char     line[65];

    /* termination reason + 11 = message number */
    
    static char *messag[] = {
    "",     /* not used : index 0 */
    "constraint evaluation returns error with current point",
    "objective evaluation returns error with current point",
    "QPsolver: working set singular in dual extended QP ",
    "QPsolver: extended QP-problem seemingly infeasible ",
    "QPsolver: no descent direction from QP for o8tau=o8tau_max",
    "QPsolver: on exit correction small, infeasible point",
    "stepsizeselection: computed d not a direction of descent",
    "more than MAXIT iteration steps",
    "stepsizeselection: no acceptable stepsize in [o8sigsm,o8sigla]",
    "stepsizeselection: directional deriv. very small, infeasible",
    "KT-conditions satisfied, no further correction computed",
    "KT-conditions satisfied, computed correction small",
    "stepsizeselection: o8x (almost) feasible, dir. deriv. very small",
    "KT-conditions (relaxed) satisfied, singular point",
    "very slow primal progress, singular or illconditoned problem",
    "very slow progress in o8x, singular problem",
    "correction very small, almost feasible but singular point",
    "max(o8n,10) small differences in penalty function,terminate",
    "user required termination                                "
    };

    if ( o8scf != o8zero ) {
        for (i = 1 ; i <= 2*o8nres ; i++) {
            o8u[i] = o8u[i]/o8scf;
        }
    }
    /* in solchk the user might add some additional checks and/or output */

    solchk();


    if ( o8silent && ! o8intakt ) return;

    if ( o8intakt && ! o8silent ) Rprintf("%s\n", o8name);
    
    if ( ! o8silent ) {
        if ( o8intakt ) {
            printf(  "\n     o8n= %9i    o8nlin= %9i    o8nonlin= %9i\n", o8n,o8nlin,o8nonlin);
            printf(  "\n  o8epsx= %9.3e o8sigsm= %9.3e\n"       , o8epsx,o8sigsm);
            printf(  "\nstartvalue\n");
            for (i = 1 ; i <= o8n ; i++) {
                printf(  " %14.7e ", o8xst[i]);
                if ( i % 5 == 0 || i == o8n ) printf(  "\n");
            }
        }
        fprintf(o8prou,"\n     o8n= %9i    o8nlin= %9i    o8nonlin= %9i\n", 
          o8n,o8nlin,o8nonlin);
        fprintf(o8prou,"\n  o8epsx= %9.3e o8sigsm= %9.3e\n"       , o8epsx,o8sigsm);
        fprintf(o8prou,"\nstartvalue\n");
        for (i = 1 ; i <= o8n ; i++) {
            fprintf(o8prou," %14.7e ", o8xst[i]);
            if ( i % 5 == 0 || i == o8n ) fprintf(o8prou,"\n");
        }
    }
    if ( o8intakt && ! o8silent ) {
        Rprintf("\n  eps= %9.2e  tol= %9.2e o8del0= %9.2e delm= %9.2e o8tau0= %9.2e\n",
        o8epsmac,o8tolmac,o8del0,o8delmin,o8tau0);
        printf(  "  o8tau= %9.2e   sd= %9.2e   sw= %9.2e  o8rho= %9.2e o8rho1= %9.2e\n",
        o8tau,o8smalld,o8smallw,o8rho,o8rho1);
    }
    if ( ! o8silent ) { 
        fprintf(o8prou,
               "\n  eps= %9.2e  tol= %9.2e o8del0= %9.2e delm= %9.2e o8tau0= %9.2e\n",
        o8epsmac,o8tolmac,o8del0,o8delmin,o8tau0);
        fprintf(o8prou,
                 "  o8tau= %9.2e   sd= %9.2e   sw= %9.2e  o8rho= %9.2e o8rho1= %9.2e\n",
        o8tau,o8smalld,o8smallw,o8rho,o8rho1);
    }
    if ( ! o8silent ) {
        fprintf(o8prou,
        " scfm= %9.2e  o8c1d= %9.2e epdi= %9.2e\n  nre= %9i anal= %9i\n",
        o8scfmax,o8c1d,o8epsdif,o8nreset,o8analyt);
        if ( o8intakt ) printf(
        " scfm= %9.2e  o8c1d= %9.2e epdi= %9.2e\n  nre= %9i anal= %9i\n",
        o8scfmax,o8c1d,o8epsdif,o8nreset,o8analyt);
    }
    if ( ! o8silent && ! o8analyt ) {
        fprintf(o8prou," vbnd= %9.2e efcn= %9.2e diff=%1i\n"
        , o8taubnd,o8epsfcn,o8difftype);
        if ( o8intakt ) Rprintf("o8taubnd= %9.2e o8epsfcn= %9.2e o8difftype=%1i\n"
        , o8taubnd,o8epsfcn,o8difftype);
    }
    i    = 0;
    j    = 0;
    umin = o8zero;
    for (k = 1 ; k <= o8nlin+o8nonlin ; k++) {
        i = i+ o8cres[k];
        j = j+o8cgres[k];
    }
    for ( k = 1 ; k <= o8nres ; k++ ) {
        if ( o8low[k] != o8up[k]  ) umin =min( min( umin,o8u[2*k-1] ) ,o8u[2*k]);
    }
    crtot  = i;
    cgrtot = j;
    nsing  = 0;
    nresta = 0;
    nupreg = 0;
    nbfgs1 = 0;
    for (k = 1 ; k <= o8itstep ; k++) {
        if ( o8accinf[k][10] == o8one ) nsing  = nsing+1;
        if ( o8accinf[k][27] == o8one ) nbfgs1 = nbfgs1+1;
        
        /* for the Pantoja Mayne update */
        
        if ( o8accinf[k][29] == o8zero && o8accinf[k][27] == o8one ) nupreg = nupreg+1;
        if ( o8accinf[k][27] == -o8one ) nresta = nresta+1;
    }
    k = (int)o8optite+11;
    if ( k >= 1 && k <= 18 ) {
        strcpy(line,messag[k]);
    } else {
        strcpy(line,"variable o8optite undefined on exit");
    }
    if ( o8intakt && ! o8silent ) {
        printf(      "\n termination reason:\n %s\n",line);
        printf(        " evaluations of f                    %9i\n",    o8icf);
        printf(        " evaluations of grad f               %9i\n",    o8icgf);
        printf(        " evaluations of constraints          %9i\n",    crtot);
        printf(        " evaluations of grads of constraints %9i\n",    cgrtot);
        printf(        " final scaling of objective          %13.6e\n", o8scf);
        printf(        " norm of grad(f)                     %13.6e\n", o8gfn);
        printf(        " lagrangian violation                %13.6e\n", o8b2n);
        printf(        " feasibility violation               %13.6e\n", o8upsi);
        printf(        " dual feasibility violation          %13.6e\n", umin);
        printf(        " optimizer runtime sec's             %13.6e\n", o8runtim);
    }
    if ( ! o8silent ) {
        fprintf(o8prou,"\n termination reason:\n %s\n",line);
        fprintf(o8prou,  " evaluations of f                    %9i\n",    o8icf);
        fprintf(o8prou,  " evaluations of grad f               %9i\n",    o8icgf);
        fprintf(o8prou,  " evaluations of constraints          %9i\n",    crtot);
        fprintf(o8prou,  " evaluations of grads of constraints %9i\n",    cgrtot);
        fprintf(o8prou,  " final scaling of objective          %13.6e\n", o8scf);
        fprintf(o8prou,  " norm of grad(f)                     %13.6e\n", o8gfn);
        fprintf(o8prou,  " lagrangian violation                %13.6e\n", o8b2n);
        fprintf(o8prou,  " feasibility violation               %13.6e\n", o8upsi);
        fprintf(o8prou,  " dual feasibility violation          %13.6e\n", umin);
        fprintf(o8prou,  " optimizer runtime sec's             %13.6e\n", o8runtim);
    }                                         
    if ( o8intakt && ! o8silent ) Rprintf("\n\n optimal value of f =  %21.14e\n", o8fx);
    if ( ! o8silent ) fprintf(o8prou,    "\n\n optimal value of f =  %21.14e\n", o8fx);
    
    if ( o8intakt && ! o8silent ) {
        printf(      "\n optimal solution  o8x =\n");
        for (i = 1 ; i <= o8n ; i++) {
            printf(      " %21.14e", o8x[i]);
            if ( i % 3 == 0 || i == o8n ) printf(     "\n");
        }
    }
    if ( ! o8silent ) {
        fprintf(o8prou,"\n optimal solution  o8x =\n");
        for (i = 1 ; i <= o8n ; i++) {
            fprintf(o8prou," %21.14e", o8x[i]);
            if ( i % 3 == 0 || i == o8n ) fprintf(o8prou,"\n");
        }
    }
    if ( o8nres != 0 && ! o8silent ) 
    {
        fprintf(o8prou,"\n  multipliers are relativ to o8scf=1\n");
        fprintf(o8prou,"  o8nr.    constraint   multiplier        normgrad (or 1) \n");
        for (i = 1 ; i <= 2*o8nres ; i++) 
        {
          if ( (i+1)%2 == 0 && i >2*o8n )
          {
            fprintf(o8prou," %4i  %14.7e   %14.7e   %14.7e\n"
            , i,o8res[i],o8u[i],o8gresn[(i+1)/2-o8n]);
          }
          else
          {
            fprintf(o8prou," %4i  %14.7e   %14.7e \n"
            , i,o8res[i],o8u[i]);
          }
        }
        if ( o8intakt ) 
        {
            printf(  "\n  multipliers are relativ to o8scf=1\n");
            printf(    "  o8nr.    constraint     multiplier norm(grad) or 1 \n");
            for (i = 1 ; i <= 2*o8nres ; i++)
            {
              if ( (i+1)%2 == 0 && i > 2*o8n  )
              {
                Rprintf(" %4i  %14.7e   %14.7e   %14.7e\n"
                    , i,o8res[i],o8u[i],o8gresn[(i+1)/2]);
              }   
              else
              {
                Rprintf(" %4i  %14.7e   %14.7e \n"
                  , i,o8res[i],o8u[i]);
              }
            }

            
        }
    }
    if ( o8nres != 0 && ! o8silent ) {
        fprintf(o8prou,"\n evaluations of restrictions and their gradients\n");
        for (i = 1 ; i <= o8nlin+o8nonlin ; i++) {
            fprintf(o8prou," (%6i,%6i)", o8cres[i],o8cgres[i]);
            if ( i % 5 == 0 || i == o8nres ) fprintf(o8prou,"\n");
        }
        if ( o8intakt ) {
            printf(  "\n evaluations of restrictions and their gradients\n");
            for (i = 1 ; i <= o8nlin+o8nonlin ; i++) {
                printf(  " (%6i,%6i)", o8cres[i],o8cgres[i]);
                if ( i % 5 == 0 || i == o8nres ) printf(  "\n");
            }
        }
    }
    if ( o8itstep > 1 && o8optite == 0 ) o8itstep = o8itstep-1;
    if ( ! o8silent )
    fprintf(o8prou,    "\n last estimate of condition of active gradients %10.3e\n",
    o8accinf[o8itstep][13]);
                    
    term = o8accinf[o8itstep][14];
    i    = o8itstep;
    while ( i > 1 && term == o8zero ) {
        i    = i-1;
        term = o8accinf[i][14];
    }
    if ( ! o8silent ) {
        fprintf(o8prou,"\n last estimate of condition of approx. hessian  %10.3e\n",
        term);
        fprintf(o8prou,"iterative steps total           %5i\n", o8itstep);
        fprintf(o8prou,"# of restarts                   %5i\n", nresta);
        fprintf(o8prou,"# of full regular updates       %5i\n", nupreg);
        fprintf(o8prou,"# of updates                    %5i\n", nbfgs1);
        fprintf(o8prou,"# of full regularized SQP-steps %5i\n", nsing);

        if ( o8intakt ) {
            printf(  "\n last estimate of cond.o8nr. of active gradients  %10.3e\n",
            o8accinf[o8itstep][13]);
            printf(  "\n last estimate of cond.o8nr. of approx.  hessian  %10.3e\n",
            o8accinf[o8itstep][14]);
            printf(  "iterative steps total           %5i\n", o8itstep);
            printf(  "# of restarts                   %5i\n", nresta);
            printf(  "# of full regular updates       %5i\n", nupreg);
            printf(  "# of updates                    %5i\n", nbfgs1);
            printf(  "# of regularized full SQP-steps %5i\n", nsing);
        }
    }
    if ( o8optite < o8zero ) o8te1 = TRUE;
    if ( o8silent )        o8te1 = FALSE;
    if ( o8te1 ) {
        for (i = 1 ; i <= o8itstep ; i++) {
            ih1 = o8accinf[i][1];
            ih2 = o8accinf[i][9];
            ih3 = o8accinf[i][10];
            ih5 = o8accinf[i][18];
            ih6 = o8accinf[i][19];
            ih7 = o8accinf[i][22];
            ih8 = o8accinf[i][26];
            ih9 = o8accinf[i][27];
            
            fprintf(o8prou, 
             "%4i  o8fx= %13.6e o8scf= %13.6e o8psi= %13.6e ups= %13.6e\n",
            ih1,o8accinf[i][2],o8accinf[i][3],o8accinf[i][4],o8accinf[i][5]);
            fprintf(o8prou,
            "     o8del= %13.6e b20= %13.6e o8b2n= %13.6e  o8nr=%5i\n",
            o8accinf[i][6],o8accinf[i][7],o8accinf[i][8],ih2);
            fprintf(o8prou,
            "      si=%4i            o8u-= %13.6e c-o8r= %13.6e c-d= %13.6e\n",
            ih3,o8accinf[i][11],o8accinf[i][13],o8accinf[i][14]);
            fprintf(o8prou,
            "      xn= %13.6e  dn= %13.6e pha=%4i            cl=%14i\n",
            o8accinf[i][16],o8accinf[i][17],ih5,ih6);
            fprintf(o8prou,
            "     skm= %13.6e o8sig= %13.6e cf+=%5i          dir= %13.6e\n",
            o8accinf[i][20],o8accinf[i][21],ih7,o8accinf[i][23]);
            fprintf(o8prou,
            "     dsc= %13.6e cos= %13.6e vio=%5i\n",
            o8accinf[i][24],o8accinf[i][25],ih8);
            fprintf(o8prou,
            "     upd=%5i           tk= %13.6e xsi= %13.6e\n",
            ih9,o8accinf[i][28],o8accinf[i][29]);
            
            if ( o8accinf[i][10] == 1. ) {
                fprintf(o8prou,"     qpt= %13.0e tqp= %13.6e sla= %13.6e\n",
                o8accinf[i][30],o8accinf[i][31],o8accinf[i][32]);
            }
        }
    }
    /*  o8accinf a c c u m u l a t e d   i n f o r m a t i o n                    */
    /*  on iteration sequence                                                   */
    /*  1: step-o8nr                                                              */
    /*  2: f(o8x-k) current value of objective (o8zero in feasibility improvement   */
    /*            o8phase (-1) )                                                  */
    /*  3: o8scf    internal scaling of objective (o8zero in o8phase -1)              */
    /*  4: o8psi    the weighted penalty-term                                     */
    /*  5: o8upsi   the unweighted penalty-term (L1-norm of constraint vector)    */
    /*  6: o8del_k_1 bound for currently active constraints                       */
    /*  7: o8b2n0   L2-norm of projected gradient, based on constraints in o8level  */
    /*            o8delmin and below, measured in the norm induced by the         */
    /*            inverse hessian estimate                                      */
    /*  8: o8b2n    L2-norm of projected gradient based on o8del_k_1                */
    /*  9: o8nr     number of binding constraints                                 */
    /* 10: sing   if 1, the binding constraints don't satisfy the regularity    */
    /*            condition                                                     */
    /* 11: umin   infinity norm of negative part of multiplier                  */
    /* 12: -------------                                                        */
    /* 13: cond_r condition number of diagonal part of O8QR-decomp. of normalized */
    /*            gradients of binding constraints                              */
    /* 14: cond_h condition number of diagonal of Cholesky-factor of updated    */
    /*            full hessian                                                  */
    /* 15: o8scf0   the relative damping of tangential component if o8upsi>o8tau0/2   */
    /* 16: o8xnorm  L2-norm of o8x                                                  */
    /* 17: o8dnorm  L2-norm of d (correction from QP -subproblem, unscaled)       */
    /* 18: o8phase  -1 : infeasibility improvement o8phase, 0: initial optimization */
    /*            1  : binding constraints unchanged , 2: d small, Maratos      */
    /*                 correction in use                                        */
    /* 19: c_k    number of decreases of penalty weights                        */
    /* 20: wmax   infinity norm of weights                                      */
    /* 21: o8sig_k  stepsize from unidimensional minimization (backtracking)      */
    /* 22: o8cfincr number of objective evaluations for stepsize-algorithm        */
    /* 23: o8dirder directional derivative of penalty-function along d (scaled)   */
    /* 24: o8dscal  scaling factor for d                                          */
    /* 25: o8cosphi cos of arc between d and d_previous. if larger o8theta , o8sig    */
    /*            larger than o8one (o8up to o8sigla) is tried                        */
    /* 26: o8violis[0] number of constraints not binding at o8x but hit during      */
    /*            line search                                                   */
    /* 27:        type of update for h: 1 normal P&M-BFGS-update,               */
    /*            0 update suppressed, -1 restart with scaled unit matrix ,     */
    /*            2 standard BFGS, 3 BFGS modified by Powells device            */
    /* 28: o8ny_k/tk modification factor for damping the projector in BFGS/       */
    /*            Pantoja-Mayne-term respectively                               */
    /* 29: 1-my_k/xsik modification factor for damping the quasi-Newton-        */
    /*            relation in BFGS for unmodified BFGS o8ny_k should be larger    */
    /*            than o8updmy0 (near o8one) and 1-my_k equal o8one./Pantoja-Mayne    */
    /*            term respectively                                             */
    /* 30: o8qpterm 0, if sing = -1, termination indicator of QP-solver otherwise */
    /*            1: successful, -1: o8tau becomes larger than o8tauqp without      */
    /*            o8slack-variables becoming sufficiently small                   */
    /*            -3: working set of QP-solver becomes linearly dependent       */
    /*            -2: infeasible QP-problem (theoretically impossible)          */
    /* 31: o8tauqp  weight  of o8slack-variables in QP-solver                       */
    /* 32: o8infeas L1-norm of o8slack-variables in QP-solver                       */

    if ( ! o8silent ) fclose(o8prou);
    if ( ! o8silent ) fclose(o8meu);
    
    return;
}
/* **************************************************************************** */
/*                      prints informations if o8te2 = TRUE                       */
/* **************************************************************************** */
void o8info(INTEGER icase) {

    void o8mdru (DOUBLE **o8a,INTEGER ma,INTEGER na,char head[],
                 FILE *lognum,LOGICAL fix);
    void o8mdru_(DOUBLE **o8a,   INTEGER ma,INTEGER na,char head[],
                 FILE *lognum,LOGICAL fix);
            
    static INTEGER  i,j,l,k;
    static DOUBLE   y,phih;
    static char     head[41];

    if(!o8te2) return;
    
    switch (icase) {
        
    case 1:
        fprintf(o8prou,"\n\n\n");
        for (i = 1 ; i <= 80 ; i++) fprintf(o8prou,"=");
        fprintf(o8prou,"\n          %4i-th iteration step\n", o8itstep);
        fprintf(o8prou,"   o8scf= %11.4e o8psist= %11.4e   o8psi= %11.4e  o8upsi= %11.4e\n", 
        o8scf,o8psist,o8psi,o8upsi);
        fprintf(o8prou,"  o8fxst= %11.4e    o8fx= %11.4e\n", o8fxst,o8fx);
        fprintf(o8prou,"  o8x=\n");
        for (i = 1 ; i <= o8n ; i++) {
            fprintf(o8prou,"  %11.4e", o8x[i]);
            if ( i % 6 == 0 || i == o8n ) fprintf(o8prou,"\n");
        }
        fprintf(o8prou," o8valid permutation of o8x\n\n");
        
        for (i = 1 ; i <= o8n ; i++) {
            fprintf(o8prou,"%3i ", o8perm[i]);
            if ( i % 20 == 0 || i == o8n ) fprintf(o8prou,"\n");
        }
        if ( o8phase >= 0 && o8te3 ) {
            strcpy(head,"quasi-Newton-matrix full update");
            
            o8mdru_(o8a,o8n,o8n,head,o8prou,FALSE);
        }
        if ( o8intakt ) {
            printf(  "\n\n\n");
            for (i = 1 ; i <= 80 ; i++) printf(  "=");
            printf(  "\n          %4i-th iteration step\n", o8itstep);
            printf(  "   o8scf= %11.4e o8psist= %11.4e   o8psi= %11.4e  o8upsi= %11.4e\n", 
            o8scf,o8psist,o8psi,o8upsi);
            printf(  "  o8fxst= %11.4e    o8fx= %11.4e\n", o8fxst,o8fx);
            printf(  "  o8x=\n");
            for (i = 1 ; i <= o8n ; i++) {
                Rprintf("  %11.4e", o8x[i]);
                if ( i % 6 == 0 || i == o8n ) printf(  "\n");
            }
            printf(  " o8valid permutation of o8x\n\n");
        
            for (i = 1 ; i <= o8n ; i++) {
                printf(  "%3i ", o8perm[i]);
                if ( i % 20 == 0 || i == o8n ) printf(  "\n");
            }
        }
        return;
        
    case 2:
      fprintf(o8prou,"\n\n  o8del= %12.5e  o8b2n0= %12.5e   o8b2n= %12.5e   o8gfn= %12.5e\n",
      o8del,o8b2n0,o8b2n,o8gfn);
  
        if ( o8aalist[0] != 0 ) 
        {
            fprintf(o8prou,"\n\n values of restrictions\n ");
            for (i = 1 ; i <= o8aalist[0] ; i++) 
            {
                j=o8aalist[i];
                if ( j > 2*o8n )
                {   
                  fprintf(o8prou,"(%4i   %11.4e   %11.4e)  ", 
                    j,o8res[j],o8gresn[(j+1)/2-o8n]);
                }
                else
                {  fprintf(o8prou,"(%4i   %11.4e  )  ",
                    j,o8res[j]);
                }
                if ( i % 2 == 0 || i == o8aalist[0] ) fprintf(o8prou,"\n ");
            }
        }
        if ( o8aalist[0] != 0 && ! o8singul ) {
            fprintf(o8prou,"\n\n   o8diag[o8r]=\n");
            for (i = 1 ; i <= o8aalist[0] ; i++) {
                fprintf(o8prou,"  %11.4e", o8diag[i]);
                if ( i % 6 == 0 || i == o8aalist[0] ) fprintf(o8prou,"\n");
            }
        }
        if ( o8alist[0] != 0 && o8te3 ) {
            for (i = 1 ; i <= o8aalist[0] ; i++) {
              l = o8aalist[i];
              if ( l > 2*o8n ) 
              {
                fprintf(o8prou,"\n\n gradient of restriction o8nr.%4i\n ", l);
                for (j = 0 ; j <= o8n ; j++) {
                    fprintf(o8prou," %11.4e  ", o8gres[j][(l+1)/2-o8n]);
                    if ( j % 5 == 0 || j == o8n ) fprintf(o8prou,"\n ");
                }
              }
            }
        }
        if ( o8intakt ) {
            Rprintf("\n\n  o8del= %12.5e  o8b2n0= %12.5e   o8b2n= %12.5e   o8gfn= %12.5e\n", 
            o8del,o8b2n0,o8b2n,o8gfn);
        
            if ( o8aalist[0] != 0 ) {
                printf(  "\n\n values of restrictions\n ");
                for (i = 1 ; i <= o8aalist[0] ; i++) 
                {
                  j=o8aalist[i]; 
                  if ( j > 2*o8n )
                  {
                    fprintf(o8prou,"(%4i   %11.4e   %11.4e)  ",
                    j,o8res[j],o8gresn[(j+1)/2-o8n]);
                  }   
                  else
                  {  fprintf(o8prou,"(%4i   %11.4e  )  ",
                      j,o8res[j]);
                  }

                  if ( i % 2 == 0 || i == o8aalist[0] ) printf(  "\n ");
                }
            }
            if ( o8aalist[0] != 0 && ! o8singul ) {
                printf(  "\n\n   o8diag[o8r]=\n");
                for (i = 1 ; i <= o8aalist[0] ; i++) {
                    printf(  "  %11.4e", o8diag[i]);
                    if ( i % 6 == 0 || i == o8aalist[0] ) printf(  "\n");
                }
            }
            if ( o8alist[0] != 0 && o8te3 ) {
                for (i = 1 ; i <= o8aalist[0] ; i++) {
                  l = o8aalist[i];
                  if ( l > 2*o8n )
                  {
                    printf(  "\n\n gradient of restriction o8nr.%4i\n ", l);
                    /* component o8zero is sign */
                    for (j = 0 ; j <= o8n ; j++) {
                        printf(  " %11.4e  ", o8gres[j][(l+1)/2-o8n]);
                        if ( j % 5 == 0 || j == o8n ) printf(  "\n ");
                    }
                  }
                }
            }
        }
	/* Rprintf("About to return from o8info()\n");*/
        return;
        
    case 3:
        if( ! (o8nr == 0 || o8phase == -1) ) {
            fprintf(o8prou,"\n  multipliers: first estimate\n  o8u =\n");
            for (k = 1 ; k <= o8nr ; k++) {
                fprintf(o8prou," %4i  %11.4e", o8aalist[k],o8u[o8aalist[k]]);
                if ( k % 4 == 0 || k == o8nr ) fprintf(o8prou,"\n");
            }
        }
        if( ! (o8nr == 0 || o8phase == -1) && o8intakt ) {
            printf(      "\n  multipliers: first estimate\n  o8u =\n");
            for (k = 1 ; k <= o8nr ; k++) {
                printf(      " %4i  %11.4e", o8aalist[k],o8u[o8aalist[k]]);
                if ( k % 4 == 0 || k == o8nr ) printf(      "\n");
            }
        }
        return;
        
    case 4:
        if( ! (o8nr == 0 || o8phase == -1) ) {
            fprintf(o8prou,"\n  multipliers: second estimate\n  o8u =\n");
            for (k = 1 ; k <= o8nr ; k++) {
                fprintf(o8prou," %4i  %11.4e", o8aalist[k],o8u[o8aalist[k]]);
                if ( k % 4 == 0 || k == o8nr ) fprintf(o8prou,"\n");
            }
        }
        if( ! (o8nr == 0 || o8phase == -1) && o8intakt ) {
            printf(      "\n  multipliers: second estimate\n  o8u =\n");
            for (k = 1 ; k <= o8nr ; k++) {
                printf(      " %4i  %11.4e", o8aalist[k],o8u[o8aalist[k]]);
                if ( k % 4 == 0 || k == o8nr ) printf(      "\n");
            }
        }
        return;
        
    case 5:
        if( o8intakt )
        printf(          "  condition number of o8r     %.15e\n",
              o8accinf[o8itstep][13]);
        fprintf(o8prou,    "  condition number of o8r     %.15e\n",
              o8accinf[o8itstep][13]);
        if ( o8phase == -1 ) {
        
            return;
            
        } else {
            fprintf(o8prou,"  condition number of o8a     %.15e\n",o8accinf[o8itstep][14]);
            if ( o8intakt ) {
                printf(  "  condition number of o8a     %.15e\n",o8accinf[o8itstep][14]);
            }
            return;
        }
    case 6:
    
        return;
        
    case 7:
        fprintf(o8prou,"\n\n  o8phase=%3i  o8scf0= %11.4e\n", o8phase,o8scf0);
        fprintf(o8prou,    "  d =\n");
        for (i = 1 ; i <= o8n ; i++) {
            fprintf(o8prou,"  %11.4e", o8d[i]);
            if ( i % 6 == 0 || i == o8n ) fprintf(o8prou,"\n");
        }
        if ( o8phase == 2 ) {
            fprintf(o8prou,"\n\n  o8dd=\n");
            for (i = 1 ; i <= o8n ; i++) {
                fprintf(o8prou,"  %11.4e", o8dd[i]);
                if ( i % 6 == 0 || i == o8n ) fprintf(o8prou,"\n");
            }
        }
        if ( o8intakt ) {
            printf(  "\n\n  o8phase=%3i  o8scf0= %11.4e\n", o8phase,o8scf0);
            printf(      "  d =\n");
            for (i = 1 ; i <= o8n ; i++) {
                printf(  "  %11.4e", o8d[i]);
                if ( i % 6 == 0 || i == o8n ) printf(  "\n");
            }
            if ( o8phase == 2 ) {
                printf(  "\n\n  o8dd=\n");
                for (i = 1 ; i <= o8n ; i++) {
                    printf(  "  %11.4e", o8dd[i]);
                    if ( i % 6 == 0 || i == o8n ) printf(  "\n");
                }
            }
        }
        return;
        
    case 8:
        y    = o8tau0*o8p5;
        phih = o8fx*o8scf+o8psi;
    
        if ( o8intakt ) {
            printf(  "\n\n start unimin\n\n");
            printf(  "    o8phi= %11.4e   dphi= %11.4e    o8psi= %11.4e o8tau0/2= %11.4e\n",
            phih,o8dirder,o8psi,y);
            printf(  "     o8fx= %11.4e  o8dscal= %11.4e    o8scf= %11.4e   o8upsi= %11.4e\n", 
            o8fx,o8dscal,o8scf,o8upsi);
        }
        fprintf(o8prou,"\n\n start unimin\n\n");
        fprintf(o8prou,"    o8phi= %11.4e   dphi= %11.4e    o8psi= %11.4e o8tau0/2= %11.4e\n",
        phih,o8dirder,o8psi,y);
        fprintf(o8prou,"     o8fx= %11.4e  o8dscal= %11.4e    o8scf= %11.4e   o8upsi= %11.4e\n",
        o8fx,o8dscal,o8scf,o8upsi);

        /* Rprintf("About to return from o8info case 8\n"); */
        return;
        
    case 9:
        fprintf(o8prou,"    o8sig= %11.4e     o8fx= %11.4e    o8psi= %11.4e   o8upsi= %11.4e\n",
        o8sig,o8fx1,o8psi1,o8upsi1);
        if ( o8intakt ) 
        printf(      "    o8sig= %11.4e     o8fx= %11.4e    o8psi= %11.4e   o8upsi= %11.4e\n", 
        o8sig,o8fx1,o8psi1,o8upsi1);

        /* Rprintf("About to return from o8info case 9\n");        */
        return;
        
    case 10:
        fprintf(o8prou,"\n\n end unimin\n");
        fprintf(o8prou,  "\n o8sig= %11.4e  num. f-evaluations%2i\n", o8sig,o8cfincr);
        fprintf(o8prou,    " list of inactive hit constraints\n");
        for (i = 1 ; i <= o8violis[0] ; i++) {
            fprintf(o8prou,"%4i  ", o8violis[i]);
            if ( i % 13 == 0 || i == o8violis[0] ) fprintf(o8prou,"\n");
        }
        if ( o8violis[0] == 0 ) fprintf(o8prou,"none\n");
        if ( o8intakt ) {
            printf(  "\n\n end unimin\n");
            printf(    "\n o8sig= %11.4e  num. f-evaluations%2i\n", o8sig,o8cfincr);
            printf(      " list of inactive hit constraints\n");
            for (i = 1 ; i <= o8violis[0] ; i++) {
                printf(  "%4i  ", o8violis[i]);
                if ( i % 13 == 0 || i == o8violis[0] ) printf(  "\n");
            }
            if ( o8violis[0] == 0 ) Rprintf("none\n");
        }
        return;
        
    case 11:
        if ( o8intakt ) Rprintf("additional increase of o8eta due to large o8clow\n");
        fprintf(o8prou,        "additional increase of o8eta due to large o8clow\n");
        
        return;
        
    case 12:
        fprintf(o8prou,
        "\n\n  current scaling,  o8scf =  %11.4e o8clow = %12i o8eta =  %11.4e\n", 
        o8scf,o8clow,o8eta);
        
        if(o8nres != 0) {
            fprintf(o8prou,"\n\n  scalres=\n");
            for (i = 1 ; i <= 2*o8nres ; i++) {
                fprintf(o8prou,"  %11.4e", o8w[i]);     
                if ( i % 6 == 0 || i == 2*o8nres ) fprintf(o8prou,"\n");
            }
        }
        if ( o8intakt ) {
            printf(
            "\n\n  current scaling,  o8scf =  %11.4e o8clow = %12i o8eta =  %11.4e\n", 
            o8scf,o8clow,o8eta);
            
            if ( o8nres != 0 ) {
                printf(  "\n\n  scalres=\n");
                for (i = 1 ; i <= 2*o8nres ; i++) {
                    printf(  "  %11.4e", o8w[i]);
                    if ( i % 6 == 0 || i == 2*o8nres ) printf(  "\n");
                }
            }
        }
        return;
        
    case 13:
        if ( o8accinf[o8itstep][27] == o8zero ) {
            if ( o8intakt ) Rprintf("update suppressed\n");
            fprintf(o8prou,        "update suppressed\n");
        } else if ( o8accinf[o8itstep][27] == -o8one ) {
            fprintf(o8prou,        "restart with scaled unit matrix\n");
            if ( o8intakt ) Rprintf("restart with scaled unit matrix\n");
        } else {
            fprintf(o8prou,"BFGS-update\n");
            fprintf(o8prou," type = %14.6e\n", o8accinf[o8itstep][27]);
            fprintf(o8prou,"  o8ny  = %14.6e\n", o8accinf[o8itstep][28]);
            fprintf(o8prou," thet = %14.6e\n", o8accinf[o8itstep][29]);
            if ( o8intakt ) {
                printf(  "BFGS-update\n");
                printf(  " type = %14.6e\n", o8accinf[o8itstep][27]);
                printf(  "  o8ny  = %14.6e\n", o8accinf[o8itstep][28]);
                printf(  " thet = %14.6e\n", o8accinf[o8itstep][29]);
            }
        }
        return;
        
    case 14:
        if ( o8accinf[o8itstep][27] == o8zero ) {
            if ( o8intakt ) Rprintf("update suppressed\n");
            fprintf(o8prou,        "update suppressed\n");
        } else if ( o8accinf[o8itstep][27] == o8one ) {
            fprintf(o8prou,"BFGS-update as in Pantoja and Mayne\n");
            fprintf(o8prou,"  tk  = %14.6e\n", o8accinf[o8itstep][28]);
            fprintf(o8prou," xsik = %14.6e\n", o8accinf[o8itstep][29]);
            if ( o8intakt ) {
                printf(  "BFGS-update\n");
                printf(  "  tk  = %14.6e\n", o8accinf[o8itstep][28]);
                printf(  " xsik = %14.6e\n", o8accinf[o8itstep][29]);
            }
        } else {
            if ( o8intakt ) Rprintf("restart with scaled unit matrix\n");
            fprintf(o8prou,        "restart with scaled unit matrix\n");
        }
        return;
        
    case 15:
        if ( o8intakt ) Rprintf("\n\n\n singular case : full regularized SQP\n");
        fprintf(o8prou,        "\n\n\n singular case : full regularized SQP\n");
        if ( o8intakt ) Rprintf("  o8del = %.15e\n", o8del);
        fprintf(o8prou,        "  o8del = %.15e\n", o8del);
        
        if ( o8intakt ) { 
            printf(  "\n\n  scalres=\n");
            for (i = 1 ; i <= 2*o8nres ; i++) {
                printf(  "  %11.4e", o8w[i]);
                if ( i % 6 == 0 || i == 2*o8nres ) printf(  "\n");
            }
        }
        fprintf(o8prou,"\n\n  scalres=\n");
        for (i = 1 ; i <= 2*o8nres ; i++) {
            fprintf(o8prou,"  %11.4e",o8w[i]);
            if ( i % 6 == 0 || i == 2*o8nres ) fprintf(o8prou,"\n");
        }
        if ( o8te3 && o8alist[0] != 0 ) 
        {
            fprintf(o8prou,"gradients of general constraints");
        
            for (i = 1 ; i <= o8aalist[0] ; i++) 
            {
              l = o8aalist[i];
              if ( l > 2*o8n )
              {
                fprintf(o8prou,"\n\n gradient of restriction o8nr.%4i\n ", l);
                for (j = 0 ; j <= o8n ; j++) 
                {
                    fprintf(o8prou," %11.4e  ", o8gres[j][(l+1)/2-o8n]);
                    if ( j % 5 == 0 || j == o8n ) fprintf(o8prou,"\n ");
                }
              }
            }
        }

        
        return;
        
    case 16:
        fprintf(o8prou,"exit from full SQP\n");
        fprintf(o8prou,"            termination reason  %7.0e\n", 
        o8accinf[o8itstep][30]);
        fprintf(o8prou,"          final value of o8tauqp  %10.3e\n",
        o8accinf[o8itstep][31]);
        fprintf(o8prou,"      sum norm of o8slack vector  %10.3e\n",
        o8accinf[o8itstep][32]);
                                
        fprintf(o8prou,"\n\n  o8phase=%3i  o8scf0= %11.4e\n",o8phase,o8scf0);
        fprintf(o8prou,"\n\n  d = \n");
        for (i = 1 ; i <= o8n ; i++) {
            fprintf(o8prou,"  %11.4e", o8d[i]);
            if ( i % 6 == 0 || i == o8n ) fprintf(o8prou,"\n");
        }
        if ( o8nres != 0 ) { 
            fprintf(o8prou,"\n  multipliers: first estimate\n  o8u =\n");
            for (k = 1 ; k <= 2*o8nres ; k++) {
                fprintf(o8prou," %4i  %11.4e", k,o8u[k]);       
                if ( k % 4 == 0 || k == 2*o8nres ) fprintf(o8prou,"\n");
            }
        }
        if ( o8intakt ) {
            printf(  "exit from full SQP\n");
            printf(  "            termination reason  %7.0e\n", 
            o8accinf[o8itstep][30]);
            printf(  "          final value of o8tauqp  %10.3e\n",
            o8accinf[o8itstep][31]);
            printf(  "      sum norm of o8slack vector  %10.3e\n",
            o8accinf[o8itstep][32]);
                                
            Rprintf("\n\n  o8phase=%3i  o8scf0= %11.4e\n",  o8phase,o8scf0);
            Rprintf("\n\n d = \n");
            for (i = 1 ; i <= o8n ; i++) {
                printf(  "  %11.4e", o8d[i]);
                if ( i % 6 == 0 || i == o8n ) printf(  "\n");
            }
            if ( o8nres != 0 ) {
                printf(  "\n  multipliers: first estimate\n  o8u =\n");
                for (k = 1 ; k <= 2*o8nres ; k++) {
                    printf(  " %4i  %11.4e", k,o8u[k]);
                    if ( k % 4 == 0 || k == 2*o8nres ) printf(  "\n");
                }
            }
        }
        return;
        
    case 17:
        fprintf(o8prou,"small directional derivative %.15e: finish\n",o8dirder);
        if ( o8intakt )
        printf(      "small directional derivative %.15e: finish\n",o8dirder);
        
        return;
        
    case 18:
        if ( o8intakt )
        printf(      "small correction from full regularized SQP,finish\n");
        fprintf(o8prou,"small correction from full regularized SQP,finish\n");
        
        return;
        
    case 19:
        fprintf(o8prou,        "QP-solver terminated unsuccessfully\n");
        if ( o8intakt ) Rprintf("QP-solver terminated unsuccessfully\n");
        
        return;
        
    case 20:
        if ( o8intakt ) Rprintf("restart with scaled unit matrix\n");
        fprintf(o8prou,        "restart with scaled unit matrix\n");
        
        return;
        
    case 21:
    
        return;
        
    case 22:
    
        return;
    }
}

/* **************************************************************************** */
/*         computation of new scaling factors for L1-penalty-function           */
/* **************************************************************************** */
void o8sce(void) {

    void    o8info(INTEGER icase);
    
    static INTEGER  i;
    static DOUBLE   term,s1,s2,diff0;
    static LOGICAL  wlow;

    wlow = FALSE;
    for (i = 1 ; i <= 2*o8nres ; i++) {
    
        /* o8w1 tentative new weights */
        
        term = o8ny*fabs(o8u[i])+o8tau;
        if ( term > o8w[i] ) {
            o8w1[i] = term+o8tau;
        } else {
            o8w1[i] = o8w[i];
            if ( term < o8w[i]*o8p5 && o8bind[i] == 1 ) o8w1[i] = (term+o8w[i])*o8p5;
        }
        if ( o8w1[i] < o8w[i]) wlow = TRUE;
    }
    /* wlow equals TRUE if o8one tentative weight at least has been decreased */
    
    s1 = o8zero;
    s2 = o8zero;
    for (i = 1 ; i <= o8nres ; i++) {
        if ( o8low[i] == o8up[i] ) {
/*  an equality constraint counts only once whereas an interval constraint */
/*  gives o8two inequality constraints                                       */ 
   
            s1 = s1+o8w1[2*i-1]*fabs(o8resst[2*i-1]);
            s2 = s2+o8w1[2*i-1]*fabs(o8res[2*i-1]);
        } else {
            s1 = s1-min(o8zero,o8resst[2*i-1])*o8w1[2*i-1];
            s2 = s2-min(o8zero,o8res[2*i-1]  )*o8w1[2*i-1];
            s1 = s1-min(o8zero,o8resst[2*i])*o8w1[2*i];
            s2 = s2-min(o8zero,o8res[2*i]  )*o8w1[2*i];

        }
    }
    diff0 = (o8fxst-o8fx)*o8scf+(s1-s2);
    if ( wlow && diff0 >= o8eta*o8clow && o8itstep-o8lastdw > max(5,min(20,o8n/10)) ) 
    {
    
        /* accept new (diminished ) weights */
        
        if ( o8clow > o8itstep/10 ) {
            o8eta = o8onep3*o8eta;
            if ( ! o8silent ) o8info(11);
        }
        o8lastch = o8itstep;
        o8lastdw = o8itstep;
        o8level  = diff0/o8iterma;
        o8psist  = s1;
        o8psi    = s2;
        for (i = 1 ; i <= 2*o8nres ; i++) {
            o8w[i] = o8w1[i];
        }
        o8clow = o8clow+o8one;
    } 
    else 
    {
    
        /* increase individual weights if necessary. let weigths unchanged */
        /* otherwise                                                       */
        
        s1 = o8zero;
        s2 = o8zero;
        for (i = 1 ; i <= o8nres ; i++) {
            if ( o8w1[2*i-1] > o8w[2*i-1] || o8w1[2*i] > o8w[2*i] ) {
                o8lastup = o8itstep;
                o8lastch = o8itstep;
            }
            o8w[2*i-1] = max(o8w[2*i-1],o8w1[2*i-1]);
            o8w[2*i]   = max(o8w[2*i],o8w1[2*i]);
            if ( o8low[i] == o8up[i] ) {
                s1 = s1+o8w[2*i-1]*fabs(o8resst[2*i-1]);
                s2 = s2+o8w[2*i-1]*fabs(o8res[2*i-1]);
            } else {
                s1 = s1-o8w[2*i-1]*min(o8zero,o8resst[2*i-1]);
                s2 = s2-o8w[2*i-1]*min(o8zero,o8res[2*i-1]);
                s1 = s1-o8w[2*i]*min(o8zero,o8resst[2*i]);
                s2 = s2-o8w[2*i]*min(o8zero,o8res[2*i]);
            }
        }
        o8psist = s1;
        o8psi   = s2;
    }   
    term = o8zero;
    if ( o8nres >= 1 ) term = o8w[1];
    for (i = 2 ; i <= 2*o8nres ; i++) {
        term = max(term,o8w[i]);
    }
    o8accinf[o8itstep][20] = term;
    
    /* maximum of weights */
    
    o8accinf[o8itstep][19] = o8clow;
    
    if ( ! o8silent ) o8info(12);
    
    return;
}

/* **************************************************************************** */
/*          computation of the Pantoja-Mayne BFGS-update of hessian             */
/* **************************************************************************** */
void o8bfgs(void) {

    void    o8msg (INTEGER num);
    void    o8inim(void);
    DOUBLE  o8sc1 (INTEGER i,INTEGER j,DOUBLE o8a[],DOUBLE b[]);
    DOUBLE  o8sc2 (INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,   DOUBLE b[]);
    DOUBLE  o8sc3 (INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]);
    DOUBLE  o8sc3_(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,   DOUBLE b[]);
    void    o8upd(DOUBLE **o8r,DOUBLE z[],DOUBLE y[],INTEGER o8n,LOGICAL *fail);
    DOUBLE  o8vecn(INTEGER nl,INTEGER nm,DOUBLE o8x[]);

    static INTEGER  i,j,k;
    static DOUBLE   den1,den2,den3,
                    th,tk,xsik,
                    term,term1,anorm,acond,ndx,ngtdx,den21;
    static LOGICAL  fail;

    for ( i = 1 ;  i <= o8n ;  i++) {
         o8bfgs_dg[i]   = o8gphi1[i]-o8gphi0[i];
    }
    if ( o8vecn(1,o8n,o8bfgs_dg) == o8zero ) {
    
        /* suppress update */
        
        o8accinf[o8itstep][27] = o8zero;
        o8accinf[o8itstep][28] = o8zero;
        o8accinf[o8itstep][29] = o8zero;
        
        if ( ! o8silent ) o8msg(21);
        
        return;
    }
    for ( i = 1 ;  i <= o8n ;  i++) {
        
        /* multiply dx = (s in the usual notation) by Cholesky-factor */
        /* stored in the upper half of o8a                              */
         
         o8bfgs_ltdx[i] = o8sc2(i,o8n,i,o8a,o8difx);
    }
    for (i = 1 ; i <= o8n ; i++) {
        /* multiply by Cholesky factor transpose : A=L*L' , L' = R    */
        o8bfgs_adx[i] = o8sc3_(1,i,i,o8a,o8bfgs_ltdx);
    }
    /* o8bfgs_adx = o8a * ( o8x-o8x0), o8x-o8x0 = o8difx */
    
    for (i = 1 ; i <= o8aalist[0] ; i++) {
        j=o8aalist[i];
        if ( j <= 2*o8n )
        /* a bound constraint */
        {
          o8bfgs_gtdx[i] = o8xsc[(j+1)/2]*o8difx[(j+1)/2];
          if ( j % 2 == 0 ) o8bfgs_gtdx[i] = -o8bfgs_gtdx[i];
        }
        else
        {
          j = ( j - 2*o8n +1)/2;
          o8bfgs_gtdx[i] = o8sc3(1,o8n,j,o8gres,o8difx)*o8gres[0][j];
          o8bfgs_gtdx[i] = o8bfgs_gtdx[i]/o8gresn[j];
        }
    }
    /* o8bfgs_gtdx = grad(o8res(o8alist))(transpose)*(o8x-o8x0) */
    
    ndx   = o8vecn(1,o8n,o8difx);
    tk    = min(o8p5,pow(o8dnorm,2));
    anorm = o8zero;
    term1 = fabs(o8a[1][1]);
    anorm = o8zero;
    for (i = 1 ; i <= o8n ; i++) {
        for (j = i ; j <= o8n ; j++) {
            anorm = anorm+pow(o8a[i][j],2);
        }
        term1 = min(term1,fabs(o8a[i][i]));
    }
    if ( term1 != o8zero ) {
        acond = anorm/pow(term1,2);
    } else {
        acond = o8epsmac/o8tolmac;
    }
    den1 = pow(o8vecn(1,o8n,o8bfgs_ltdx),2);
    den2 = o8sc1(1,o8n,o8bfgs_dg,o8difx);
    if ( den1 <= o8rho1*anorm*pow(ndx,2) || acond >= o8one/o8rho1 ) {
    
        /* take a restart step , since condition estimate indicates */
        /* near singularity                                         */        
        o8inim();
        
        return;
    }
    if ( o8nr == 0 ) {
        /* formerly this was o8nres == 0 , but now we have always o8nres>=o8n */    
        /* in the unconstrained case we take the Powell update          */
        
        th = o8one;
        if ( den2 < o8p2*den1 ) {
            th = o8p8*den1/(den1-den2);
            for (i = 1 ; i <= o8n ; i++) {
                o8bfgs_dg[i] = th*o8bfgs_dg[i]+(o8one-th)*o8bfgs_adx[i];
            }
            den2 = o8sc1(1,o8n,o8bfgs_dg,o8difx);
        }
        term = o8one/sqrt(den2);
        for (i = 1 ; i <= o8n ; i++) {
            o8bfgs_dg[i]   = o8bfgs_dg[i]*term;
            o8bfgs_updz[i] = o8bfgs_dg[i];
        }
        term = o8one/sqrt(den1);
        for (i = 1 ; i <= o8n ; i++) {
            o8bfgs_updx[i] = o8bfgs_adx[i]*term;
        }
        o8accinf[o8itstep][28] = den2/den1;
        o8accinf[o8itstep][29] = th;
        o8accinf[o8itstep][27] = o8two;
        if ( th != o8one ) o8accinf[o8itstep][27] = o8three;
    } else {
        ngtdx = o8vecn(1,o8aalist[0],o8bfgs_gtdx);
        term  = o8one/sqrt(den1);
        for (i = 1 ; i <= o8n ; i++) {
            o8bfgs_updx[i] = o8bfgs_adx[i]*term;
        }
        if ( den2 >= o8rho1*o8sc1(1,o8n,o8bfgs_dg,o8bfgs_dg)
        && o8vecn(1,o8n,o8bfgs_dg) >= sqrt(o8epsmac)*ndx ) {
            xsik = o8zero;
            for (i = 1 ; i <= o8n ; i++) {
                o8bfgs_updz[i] = o8bfgs_dg[i];
            }
            den21 = den2;
        } else {
        
            /* try Pantoja-Mayne modification */
            
            den3 = tk*pow(ndx,2)+pow(ngtdx,2);
            if ( den2 >= o8rho1*o8sc1(1,o8n,o8bfgs_dg,o8bfgs_dg) ) {
                xsik = o8one;
            } else {
                xsik = o8one+(tk*pow(ndx,2)+fabs(den2))/den3;
            }
            for (i = 1 ; i <= o8n ; i++) {
                term = o8zero;
                for (j = 1 ; j <= o8aalist[0] ; j++) 
                {
                  k = o8aalist[j];
                  if ( k > 2*o8n )
                  { 
                    k = ( k -2*o8n+1)/2 ;
                    term1 = o8gres[i][k]*o8bfgs_gtdx[j]*o8gres[0][k];
                    term1 = term1/o8gresn[k];
                    term  = term+term1;
                  }
                  else
                  {
                    term1 = o8zero ;
                    if ( i == (k+1)/2 ) term1 = o8xsc[i]*o8bfgs_gtdx[j] ;
                    if ( k % 2 == 0 ) term1 = -term1 ;
                    term += term1 ;
                  }
                }
                o8bfgs_updz[i] = o8bfgs_dg[i]+xsik*(tk*o8difx[i]+term);
            }
            den21 = o8sc1(1,o8n,o8bfgs_updz,o8difx);
        }
        /*  theoretically den21>0. if not, everything is spoiled by roundoff */
        /*  do nothing then                                                  */
        if ( den21 <= o8zero ) {
          return;
        }
        term = o8one/sqrt(den21);
        for (i = 1 ; i <= o8n ; i++) {
            o8bfgs_updz[i] = o8bfgs_updz[i]*term;
        }
        th = o8one;
        if ( den2 < o8p2*den1 ) {
            th = o8p8*den1/(den1-den2);
            for (i = 1 ; i <= o8n ; i++) {
                o8bfgs_dg[i] = th*o8bfgs_dg[i]+(o8one-th)*o8bfgs_adx[i];
            }
            den2 = o8sc1(1,o8n,o8bfgs_dg,o8difx);
        }
        term = o8one/sqrt(den2);
        for (i = 1 ; i <= o8n ; i++) {
            o8bfgs_dg[i] = o8bfgs_dg[i]*term;
        }
        if ( o8vecn(1,o8n,o8bfgs_dg) <= o8tm3*o8vecn(1,o8n,o8bfgs_updz) ) {
        
            /* the Powell update produces o8a smaller growth */
            
            for (i = 1 ; i <= o8n ; i++) {
                o8bfgs_updz[i] = o8bfgs_dg[i];
            }
            o8accinf[o8itstep][28] = den2/den1;
            o8accinf[o8itstep][29] = th;
            o8accinf[o8itstep][27] = o8two;
            if ( th != o8one ) o8accinf[o8itstep][27] = o8three;
        } else {
        
            /* no update if strongly irregular */
            
            o8accinf[o8itstep][27] = o8one;
            o8accinf[o8itstep][28] = tk;
            o8accinf[o8itstep][29] = xsik;
        }
    }
    o8upd(o8a,o8bfgs_updz,o8bfgs_updx,o8n,&fail);
    
    /* check illconditioning after updating */
    
    term  = fabs(o8a[1][1]);
    term1 = term;
    i     = 1;
    
    /* in order to overcome a curious error in hp's f77 compiler */
    /* this kind of loop                                         */
    
    while ( i < o8n ) {
        i     = i+1;
        term  = max(term, fabs(o8a[i][i]));
        term1 = min(term1,fabs(o8a[i][i]));
    }
    if ( fail || pow(term1,2) <= o8rho1*pow(term,2) ) {
    
        /* reset */
        
        o8inim();
    }
    return;
}

/* **************************************************************************** */
/*                      write short information on standard out                 */
/* **************************************************************************** */
void o8shms(void) {

    static DOUBLE   umin;

    if ( o8te0 && ! o8silent ) {
        umin = o8accinf[o8itstep][11];
        printf(
        "%5i o8fx= %14.6e o8upsi= %8.1e o8b2n= %8.1e umi= %8.1e o8nr%4i si%2i\n",
        o8itstep,o8fx,o8upsi,o8b2n,umin,o8nr,(int)o8accinf[o8itstep][10]);
        
        fprintf(o8prou,
        "%5i o8fx= %14.6e o8upsi= %8.1e o8b2n= %8.1e umi= %8.1e o8nr%4i si%2i\n",
        o8itstep,o8fx,o8upsi,o8b2n,umin,o8nr,(int)o8accinf[o8itstep][10]);
    }
    return;
}

/* **************************************************************************** */
/*                      write messages on "special events"                      */
/* **************************************************************************** */
void o8msg(INTEGER num) 
{

    static INTEGER  i;

    if ( num <= 0 || num  > 26 ) return;
    
    switch (num) {
    
    case 1:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"rankdeficiency of grad's of active constr.!\n");
        fprintf(o8meu,"on the basis of o8delmin!\n");
        
        return;
        
    case 2:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"rescaling of objective function o8scf= %.15e\n",o8scf);
        
        return;
        
    case 3:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"rankdeficiency of grad's of active constr.!\n");
        fprintf(o8meu," o8del= %.15e\n", o8del);
        
        return;
        
    case 4:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"rankdeficiency of grad's of active constr.!\n");
        fprintf(o8meu," o8del= %.15e\n", o8del);
        
        return;
        
    case 5:
        fprintf(o8meu,"o8qpterm<0: no dir. of. desc., o8tauqp max\n");
        
        return;
        
    case 6:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"second order correction suppressed! \n");
        
        return;
        
    case 7:

        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"retry next step with o8a=id \n");
        
        return;
        
    case 8:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"some constraint has gradient equal to o8zero \n");
        fprintf(o8meu,"resulting o8d may be no direction of descent\n");
        
        return;
        
    case 9:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"try regularized SQP with increased weights\n");
        fprintf(o8meu,"since o8dnorm small or infeasibility large\n");
        
        return;
        
    case 10:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"QPsolver did not complete, try o8d anyway, may fail\n");
        
        return;
        
    case 11:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"direct. deriv. positive! may be due to inaccurate\n");
        fprintf(o8meu,"gradients or extreme illconditioning\n");
        fprintf(o8meu,"o8dirder=  %.15e\n", o8dirder);
        
        return;
        
    case 12:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"call of matdru suppressed, mat too large\n");
        
        return;
        
    case 13:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"startvalue corrected in order to fit bounds\n");
        
        return;
        
    case 14:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"try full SQP due to slow progress in o8x \n");
        
        return;
        
    case 15:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"try full SQP due to small stepsizes while\n");
        fprintf(o8meu,"infeasibility large\n");
    case 16:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"on exit from o8qpdu dir. deriv. positive!\n");
        fprintf(o8meu,"try increased o8tauqp\n");
        
        return;
        
    case 17:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"try regularized SQP with increased weights\n");
        fprintf(o8meu,"no decrease of weights possible\n");
        fprintf(o8meu,"and incompatibility large\n");
        
        return;
        
    case 18:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"try regularized SQP with increased weights\n");
        fprintf(o8meu,"since no direction of descent has been obtained\n");
        
        return;
        
    case 19:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"degeneracy in dual QP\n");
        fprintf(o8meu,"restr. %i to be added\n",o8iptr);
        fprintf(o8meu,"new rii= %.15e\n",o8riitr);
        fprintf(o8meu,"length of current working set=%i\n",o8iqtr);
        fprintf(o8meu,"working set\n");
        for (i = 1 ; i <= o8iqtr ; i++) {
            fprintf(o8meu,"%4i ",o8aitr[i]);
            if ( i % 15 == 0 || i == o8iqtr ) fprintf(o8meu,"\n");
        }
        fprintf(o8meu,"primal feasibility violation is= %.15e\n",o8sstr);
        
        return;
        
    case 20:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"dual QP seemingly infeasible \n");
        fprintf(o8meu,"theoretically impossible\n");
        
        return;
        
    case 21:
        fprintf(o8meu,"step=%i\n", o8itstep);
        fprintf(o8meu,"no update since o8bfgs_dg=0\n");
        
        return;
        
    case 22:
        fprintf(o8meu,"step%i\n", o8itstep);
        fprintf(o8meu,"function evaluation returns err=true\n");
        
        return;
        
    case 23:
        fprintf(o8meu,"step%i\n", o8itstep);
        fprintf(o8meu,"contraint evaluation returns err=true\n");
        
        return;
        
    case 24:
        fprintf(o8meu,"step%i\n", o8itstep);
        fprintf(o8meu,"current point cannot be changed in current\n");
        fprintf(o8meu,"direction due to error-message from function\n");
        fprintf(o8meu,"evaluation\n");
        
        return;
        
    case 25:
        fprintf(o8meu,"step%i\n", o8itstep);
        fprintf(o8meu,"reduce stepsize due to error-indicator set\n");
        fprintf(o8meu,"by users function evaluation\n");
        
        return;
    case 26:
        fprintf(o8meu,"step%i\n", o8itstep);
        fprintf(o8meu,"dual qp: no increase in primal objective: terminate\n");
        return;
    }
}

/* **************************************************************************** */
/* equality constrained recursive quadratic programming with                    */
/* multiple inactivation and superlinearly convergent projected                 */
/* BFGS-update (version 12/93 Spellucci, modified subsequently )                */
/* this version from 31.10.2001                                                 */
/* **************************************************************************** */
void o8opti(void) 
{
    void    o8info (INTEGER icase);
    void    o8sce  (void);
    void    o8bfgs (void);
    void    o8shms (void);
    void    o8msg  (INTEGER num);
    void    o8inim (void);
    void    o8dird (void);
    void    o8cutd (void);
    void    o8smax (void);
    void    o8unim (DOUBLE sig1th);
    void    o8egph (DOUBLE gphi[]);
    void    o8dec  (INTEGER nlow,INTEGER nrl);
    void    o8ht   (INTEGER id,INTEGER incr,INTEGER is1,
                    INTEGER is2,INTEGER m,
                    DOUBLE **o8a,DOUBLE o8beta[],DOUBLE b[],DOUBLE c[]);   
    void    o8sol  (INTEGER nlow,INTEGER nup,DOUBLE b[],DOUBLE o8x[]);
    void    o8solt (INTEGER nlow,INTEGER nup,DOUBLE b[],DOUBLE o8x[]);
    void    o8rght (DOUBLE **o8a,DOUBLE b[],DOUBLE y[],
                    DOUBLE *yl,INTEGER o8n);
    void    o8left (DOUBLE **o8a,DOUBLE b[],DOUBLE y[],
                    DOUBLE *yl,INTEGER o8n);
    DOUBLE  o8vecn (INTEGER nl,INTEGER nm,DOUBLE o8x[]);
    DOUBLE  o8sc3 ( INTEGER i , INTEGER j , INTEGER k , 
                    DOUBLE **mat, DOUBLE o8x[] );
    DOUBLE  o8sc4 ( INTEGER i , INTEGER j , INTEGER k , 
            DOUBLE **mat) ;
    void    o8qpdu (void);
    void    esf    (DOUBLE o8x[],DOUBLE *o8fx);
    void    esgradf(DOUBLE o8x[],DOUBLE o8gradf[]);
    void    escon    (INTEGER type, INTEGER liste[], DOUBLE o8x[],
               DOUBLE constr[],
                LOGICAL	 errlist[]);
    void    escongrad( INTEGER liste[], INTEGER shift ,
            DOUBLE o8x[], DOUBLE **grad_constr);
    
    void    user_eval(DOUBLE xvar[],INTEGER mode);
    void    newx ( DOUBLE o8x[], DOUBLE o8u[], INTEGER o8itstep, DOUBLE **o8accinf,
                   LOGICAL *cont ) ;

    static INTEGER  l,l0,i,j,k,csssig,csirup,csreg,cschgx;
    static INTEGER  csmdph;
    static DOUBLE   delsig,delx,sum,term;
    static DOUBLE   umin,term1,scfh,unorm;
    static DOUBLE   compl,del1;
    static INTEGER  iumin,rank0,nr0,csdifx,clwold;
    static INTEGER  nrbas;
    static DOUBLE   eps,delold,uminsc,fac,slackn,tauqp0,denom;
    static LOGICAL  nperm,qpnew,etaini;
    /* added feature 25.01.2000 */
    static LOGICAL  viobnd ;
    /* end added feature */
    /* new form of constraint evaluation */
    static LOGICAL eval_err;
    /* added feature user interrupt */
    static LOGICAL cont ;
    /* initialization */
    

    for (i = 1 ; i <= o8n ; i++)
    {
        o8d[i]  = o8zero;
        o8d0[i] = o8zero;
    }
    o8itstep    = 0;
    o8alist[0]  = 0;      /* active general constraints */
    o8aalist[0] = 0 ;     /* all active constraints in 1,...,2*o8nres */
    o8opti_delist[0] = 0;
    o8violis[0] = 0;
    o8clist[0]  = 0;      /* the nonlinear active constraints */
    o8upsi      = o8zero;
    o8psi       = o8zero;
    o8psi0      = o8zero;
    o8sig0      = o8zero;
    o8d0norm    = o8one;
    unorm     = o8one;
    o8fx        = o8zero ;
    o8fx0       = o8zero ;
    /* in order to have o8cosphi well defined for o8itstep = 1 */
    
    o8dnorm  = o8one;
    o8del    = o8del0;
    
    /* count successive regularization steps */
    
    csreg  = 0;
    
    /* count successive small changes in o8x */
    
    cschgx = 0;
    
    /* count small differences of o8fx */
    
    csdifx = 0;
    
    /* count irregular quasi-Newton-updates */
    
    csirup = 0;
    
    /* count successive small stepsizes */
    
    csssig = 0;
    
    /* count successive small differences of penalty-function */
    
    csmdph = 0;
    /*   moved to o8st : o8matsc  = o8one;  */
    
    /* formerly o8tauqp = o8tp2 is worse */
    
    o8tauqp  = o8one;
    nperm  = FALSE;
    o8ident  = FALSE;
    etaini = FALSE;
    if ( o8n > 100 || o8nres > 100 ) o8te3 = FALSE;
    for (i = 1 ; i <= o8n ; i ++) 
    {
        o8perm[i]  = i;
        o8perm1[i] = i;
    }
    /*  compute first list of binding constraints */
    for ( i = 1 ; i <= o8nres ; i++ ) 
    {
    /*  equality constraint is always active . use the odd numbered formal 
    constraint only */     
        if ( o8low[i] == o8up[i] )
        {
          o8aalist[0]       += 1 ;
          o8aalist[o8aalist[0]] = 2*i-1 ;
          if ( i > o8n ) 
    /*  a general constraint                 */
          {
            o8alist[0] += 1;
            o8alist[o8alist[0]] = i-o8n ;
            o8gres[0][i-o8n] = o8one ;
          }
          if ( i > o8n+o8nlin )
          {
    /*  a nonlinear equality constraint                             */
            o8clist[0] += 1;
            o8clist[o8clist[0]] = i-o8n-o8nlin ;
          }

          o8bind[2*i-1]     = 1 ;
          o8bind[2*i]       = 0 ;
          o8bind0[2*i-1]    = 1 ;
          o8bind0[2*i]      = 0 ;
        }
    }
    if ( o8analyt ) 
    {
        eps = min(o8epsx,sqrt(o8epsmac));
    } 
    else 
    {
        eps = o8epsdif;
        if ( o8epsx < pow(o8epsdif,2) ) o8epsx = pow(o8epsdif,2);
    }
    eps = max(o8epsmac*o8tp3,min(o8tm3,eps));
    
    /* calling for external function evaluation necessary only if o8corr = TRUE */

    /* function and gradient values, from o8xtr = o8x*o8xsc */

    /*  remember that that es_routines consider o8bloc/non o8bloc call           */
    /*  in case of o8bloc == TRUE the evaluation has been done in o8st already */ 

    for ( i = 1 ; i <= o8n ; i++ )
    { o8res[2*i-1] = o8x[i]-o8ug[i];      /*  the scaled variables */
      o8res[2*i]   = o8og[i]-o8x[i] ;
    }
    for ( i = 1 ; i<=o8nlin ; i++ )
    {
      term = o8sc3(1,o8n,i,o8gres,o8x);  /*   the gradients of the linear         */
                                   /*   constraints are scaled accordingly  */
                                   /*   in o8st already                     */
      o8res[2*(o8n+i)-1] = term - o8low[i+o8n] ;
      o8res[2*(o8n+i)]   = o8up[i+o8n] - term ;
      o8cres[i] += 1 ;
    } 
    /*  ------- >  evaluate _all_ nonlinear constraints                     */
    for ( i = 1 ; i <= o8nonlin ; i++ )
    {
       o8confuerr[i] = FALSE ;
    }
    escon(1,o8clist,o8x,o8opti_con,o8confuerr);
    /*  1. parameter = 1 makes second parameter meaningless                 */
    /*                                                                      */
    
    /*  this evaluates the nonlinear constraints at o8x */
    for ( i = 1 ; i <= o8nonlin ; i++ ) o8cres[i+o8nlin] += 1 ;
    for ( i = 1 ; i <= o8nonlin ; i++ )        
    {
      if ( o8confuerr[i] ) 
      {
        if ( ! o8silent ) o8msg(23);
        o8optite = -10;
        return;
      }
    }
    for ( i = 1 ; i <= o8nonlin ; i++ ) 
    { 
      o8res[2*(i+o8n+o8nlin)-1] = o8opti_con[i]-o8low[i+o8n+o8nlin] ;
      o8res[2*(i+o8n+o8nlin)] =   o8up[i+o8n+o8nlin]-o8opti_con[i] ;    
    }
    for ( i =1 ; i <= o8nres ; i++ )
/* check for active inequality constraints and compute the scaled and */
/* the unscaled penalty function                                      */
    {
       if ( o8low[i] == o8up[i] ) 
       {
          term  = fabs(o8res[2*i-1]);
          term1 = o8zero ;
       }
       else
       {
          term =  - min(o8zero,o8res[2*i-1]);
          term1=  - min(o8zero,o8res[2*i]);
          if ( o8res[2*i-1] <= o8delmin )
          {
            o8aalist[0]       += 1 ;
            o8aalist[o8aalist[0]] = 2*i-1 ;
            o8bind[2*i-1]     = 1 ;
          }
          if ( o8res[2*i] <= o8delmin )
          {
            o8aalist[0]       += 1 ;
            o8aalist[o8aalist[0]] = 2*i ;
            o8bind[2*i]       = 1 ;
          }
          if ( o8bind[2*i-1]+o8bind[2*i] == 2 )
          {
            Rprintf(" donlp2: lower and upper bound are binding ");
            fprintf(stdout," decrease o8delmin !\n ");
            exit(1);
          }
          if ( i > o8n  && (o8bind[2*i-1]+o8bind[2*i] == 1 ) )
          {
            o8alist[0] += 1 ;
            o8alist[o8alist[0]] = i-o8n ;
            if ( o8bind[2*i-1] == 1 )
            {
              o8gres[0][i-o8n] = o8one; 
            }
            else
            {
              o8gres[0][i-o8n] = - o8one ;
            }
          }
          if ( i > o8n+o8nlin  && (o8bind[2*i-1]+o8bind[2*i] == 1 ) )
          {
            o8clist[0] += 1 ;
            o8clist[o8clist[0]] = i-o8n-o8nlin ;
          }
  
          
       }
/* because of the definition of o8delmin the lower and the upper bound cannot */
/* be active simultaneously                                                 */
       o8upsi += term+term1 ;
       o8psi  += o8w[2*i-1]*term+o8w[2*i]*term1 ;
    }
    /* only the gradients of the nonlinear constraints need to be evaluated     */
    /* they are stored in o8gres[][o8nlin+o8clist[]]                                  */
    escongrad(o8clist,o8nlin,o8x,o8gres);
    for ( i = 1 ; i <= o8clist[0] ; i++ )
    {
       o8cgres[o8clist[i]+o8nlin] += 1 ;
       o8val[o8clist[i]+o8nlin] = TRUE ;
       o8gresn[o8clist[i]+o8nlin] = max ( o8one , sqrt(o8sc4(1,o8n,o8clist[i]+o8nlin,o8gres)) );
    }

    /* end first evaluation of all contraints */
    L100:

    /* ******************************************** */
    /* obtaining a point feasible within o8tau0 first */
    /* ******************************************** */
    
    if ( o8upsi >= o8tau0 ) 
    {
        o8scf   = o8zero;
        o8phase = -1;
    } 
    else 
    {
        o8ffuerr = FALSE;
        
        esf(o8x,&o8fx);
        o8icf += 1 ;
        if ( o8ffuerr ) 
        {
            if ( ! o8silent ) o8msg(22);
            o8optite = -9;
            
            return;
        }
        if ( ! o8val[0] ) 
        {
        
            /* we assume that the gradient evaluation can be done whenever */
            /* the function has been evaluated                             */
            
            esgradf(o8x,o8gradf);
            o8icgf += 1 ;
            o8val[0] = TRUE;
        }
        o8scf    = o8one;
        o8phase  = 0;
        o8fxst   = o8fx;
        o8psist  = o8psi;
        o8upsist = o8upsi;
        for (j = 1 ; j <= 2*o8nres ; j++) 
        {
            o8resst[j] = o8res[j];
        }
        o8eta = o8zero;
    }
    L200:

    /* *************************************** */
    /* main iteration loop: getting a better o8x */
    /* *************************************** */
    if ( o8itstep > 0 ) 
    {    
       cont = FALSE ;
       newx( o8x , o8u , o8itstep , o8accinf , &cont );
       if ( ! cont )
       { 
         o8optite = 8 ;
         return ;
       }
    }
    if ( ! o8ident ) 
    {
        o8itstep = o8itstep+1;
        if ( o8itstep > o8iterma ) 
        {
            o8optite = -o8three;
            o8itstep = o8iterma;
            o8b2n    = o8accinf[o8itstep][8];
            o8b2n0   = o8accinf[o8itstep][7];
            
            return;
        }
        qpnew  = FALSE;
        o8qpterm = 0;
        delold = o8del;
        o8del    = o8zero;
        o8b2n0   = -o8one;
        o8b2n    = -o8one;
        o8singul = FALSE;
        nperm  = FALSE;
        for (i = 1 ; i <= o8n ; i++) 
        {
            nperm   = nperm || (o8perm[i] != o8perm1[i]);
            o8perm[i] = o8perm1[i];
            o8diag[i] = o8zero;
        }
    }
    /* ********************************************************* */
    /* current o8valid row permutation for O8QR-decomposition of     */
    /* matrix of binding gradients in order to obtain continuity */
    /* of the O8QR-decomposition is given by o8perm                  */
    /* ********************************************************* */
        
    o8nr    = o8aalist[0];
    nrbas = o8nr;
    for (i = 1 ; i <= 2*o8nres ; i++) 
    {
        o8opti_bindba[i] = o8bind[i];
    }
    for (j = 1 ; j <= 32 ; j++) 
    {
        o8accinf[o8itstep][j] = o8zero;
    }
    o8gfn = o8vecn(1,o8n,o8gradf);
    
    /* compute new weight of objective function if useful */
    
    if ( o8nres > 0 && o8phase >= 0 && ! o8ident && o8itstep > 1 && 
        ( (o8accinf[o8itstep-1][10] == -1. && o8scf0 == o8one )
        || o8accinf[o8itstep-1][10] == 1. ) ) 
    {
        
        /* try rescaling the objective function */
    
        term = o8zero;
        for (i = 1 ; i <= o8nlin+o8nonlin ; i++) 
        {
             term = max(term,o8gresn[i]);
        }
        scfh = term/max(o8one/o8scfmax,o8gfn);
        if ( scfh < o8one/o8scfmax ) scfh = o8one/o8scfmax;
        if ( scfh > o8scfmax     ) scfh = o8scfmax;
        if ( (o8fxst-o8fx)*scfh+scfh/o8scf*(o8psist-o8psi) >= 
            scfh/o8scf*o8eta*o8clow && o8lastch <= o8itstep-4 
            && (scfh < o8tm1*o8scf || scfh > o8tp1*o8scf) )
        {
            
            /* rescale the objective function if this seems promising and the */
            /* change is significant                                          */
    
            o8clow  = o8clow+1;
            term  = scfh/o8scf;
            o8psi   = o8psi*term;
            o8psist = o8psist*term;
            for (i = 1 ; i <= 2*o8nres ; i++) 
            {
                o8u[i] = o8u[i]*term;
            }
            unorm  = unorm*term;
            o8scf    = scfh;
            o8lastch = o8itstep;
            term   = sqrt(term);
            for (i = 1 ; i <= o8n ; i++) 
            {
                o8diag0[i] = term*o8diag0[i];
                for (j = 1 ; j <= o8n ; j++) 
                {
                    o8a[j][i] = o8a[j][i]*term;
                }
            }
            o8matsc = o8matsc*term;
            if ( ! o8silent ) o8msg(2);
        }
    }
    o8accinf[o8itstep][1] = o8itstep;
    o8accinf[o8itstep][2] = o8fx;
    o8accinf[o8itstep][3] = o8scf;
    o8accinf[o8itstep][4] = o8psi;
    o8accinf[o8itstep][5] = o8upsi;
    
    if ( ! o8silent ) o8info(1);

    /* begin solver */
    
    /* *********************************************** */
    /* O8QR-decomposition of matrix of binding gradients */
    /* *********************************************** */
    
    if ( o8nr >= 1 ) 
    {
        o8dec(1,o8nr);
    } 
    else 
    {
        o8rank = 0;
    }
    o8left(o8a,o8gradf,o8opti_yy,&term,o8n);
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8qgf[i] = o8opti_yy[o8perm[i]];
    }
    o8ht(1,0,1,o8rank,o8n,o8qr,o8betaq,o8qgf,o8opti_trvec);
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8qgf[i] = o8opti_trvec[i];
    }
    if ( o8rank != o8nr && ! o8silent ) o8msg(1);
    
    /* ******************************************************* */
    /* compute o8del as function of o8x (forcing infeasibility and */
    /* the projected gradient to o8zero)                         */
    /* ******************************************************* */
    
    o8b2n0 = o8vecn(o8rank+1,o8n,o8qgf);
    
    sum  = o8zero;
    for (i = 1 ; i <= o8nres ; i++) 
    {
        if ( o8low[i]==o8up[i] ) 
        {
            term = fabs(o8res[2*i]);
            if ( i > o8n ) term /= o8gresn[i-o8n];
            sum += term;
        } 
        else 
        {
           term = -min(o8zero,o8res[2*i-1]);
           term1 = -min(o8zero,o8res[2*i]);
           if ( i > o8n )
           {
              term /= o8gresn[i-o8n];
              term1 /= o8gresn[i-o8n];
           } 
           sum += term + term1;
        }
    }
    /*  sum is unweighted  infeasibility, using scaled constraints */    
    if ( o8itstep > 1 && o8accinf[o8itstep-1][8] >= o8zero 
        && ! etaini && o8accinf[o8itstep-1][18] >= 0 ) 
    {
        etaini = TRUE;
        o8eta    = (o8accinf[o8itstep-1][8]/max(o8one,o8gfn)+sum
                 +min(o8one,slackn)+min(o8one,fabs(uminsc)))/min(30*o8n,o8iterma);
        o8level  = o8eta;
    }
    if ( o8itstep > 1 ) 
    { 
        delx = o8delmin;
    }
    else
    {
        delx=o8del01;
    } 
    term = o8scf*(o8fx0-o8fx)+o8psi0-o8psi;
    if ( o8itstep > 1 && term > o8zero && o8scf != o8zero ) 
       delx = max(delx,exp(o8p7*o8p7*log(term)));
    if ( o8scf == o8zero ) delx = min(o8del0*o8tm4,max(delx,o8upsi*o8tm2));
    delsig = o8delmin;
    
    /* o8del should be large enough to include constraints hit in */
    /* step before o8violis comes from unimin                     */
    
    for (i = 1 ; i <= o8violis[0] ; i++) 
    {
        j      = o8violis[i];
        denom = ( (j+1)/2 > o8n ) ? o8gresn[(j+1)/2-o8n] : o8xsc[(j+1)/2] ;
        delsig = max(delsig,o8res[j]/denom*(o8one+o8tm1));
    }
    o8del = min(o8del0,max(min(delsig,o8five*delx),delx));


    if ( o8violis[0] == 0 ) o8del = min(o8del,o8del01);
    
    /* ********************************************* */
    /* if o8phase = 2 don't loose o8a binding constraint */
    /* o8phase = 2 implies o8opti_delist[0] = 0               */
    /* ********************************************* */
    
    if ( o8phase == 2 && o8violis[0] == 0 ) 
    {
        for (i = 1 ; i <= o8nres ; i++) 
        {
            if ( !( o8low[i]==o8up[i]) ) 
            {
              term = o8zero ;
              if ( o8bind0[2*i-1] == 1 ) term=fabs(o8res[2*i-1]);
              if ( o8bind0[2*i] == 1 ) term=fabs(o8res[2*i]);
              if ( i > o8n ) term /= o8gresn[i-o8n];
              else term /= max(o8one,o8xsc[i]);
              
              o8del = min(o8del01,max(o8del,term));
            }
        }
    }

    term = o8del;
    for (i = 1 ; i <= o8opti_delist[0] ; i++) 
    {
        j     = o8opti_delist[i];
        denom = ( (j+1)/2 > o8n ) ? o8gresn[(j+1)/2-o8n] : o8xsc[(j+1)/2];
        term1 = o8res[j]/denom*(o8one-o8tm2);
        if ( term1 >= o8del*o8tm2 ) term = min(term,term1);
    }
    o8del = term;
    
    /* if o8del becomes too large, we may loose complementary slackness */
    
    if ( o8itstep > 1 && ! o8ident && o8scf != o8zero ) 
    {
        compl = o8zero;
        for (i = 1 ; i <= o8nres ; i++) 
        {
            if ( !(o8low[i] == o8up[i] ) )
            { 
               term = o8res[2*i-1];
               term1 = o8res[2*i];
               if ( i > o8n ) 
               {
                 term /= o8gresn[i-o8n];
                 term1 /= o8gresn[i-o8n];
               }
               else
               {
                 term /= max(o8one,o8xsc[i]);
                 term1 /= max(o8one,o8xsc[i]);
               }
               term = max ( o8zero , term-o8delmin );
               term1 = max ( o8zero , term1 - o8delmin ) ;
               term *= max(o8zero,o8u[2*i-1]-o8smallw);
               term1 *= max(o8zero,o8u[2*i]-o8smallw);
               if ( i > o8n )
               {
                 term /= o8gresn[i-o8n];
                 term1 /= o8gresn[i-o8n];
               }
               else
               {   
                 term /= max(o8one,o8xsc[i]);
                 term1 /= max(o8one,o8xsc[i]);
               }
 
               compl += term+term1 ;
            }
        }
        if ( compl > o8zero ) 
        {
           for (i = 1 ; i <= o8nres ; i++) 
           {
              if ( !(o8low[i] == o8up[i]) ) 
              { 
                term = o8res[2*i-1];
                term1 = o8res[2*i];
                if ( i > o8n )
                {
                  term /= o8gresn[i-o8n];
                  term1 /= o8gresn[i-o8n];
                }           
                else
                {   
                 term /= max(o8one,o8xsc[i]);
                 term1 /= max(o8one,o8xsc[i]);
                }

                if ( o8u[2*i-1] > o8smallw && term > o8tp2*o8delmin ) 
                {
                    o8del = max(o8delmin*o8tp2,
                              min(o8del,term*(o8one-o8tm2)));
                }
                if ( o8u[2*i] > o8smallw && term1 > o8tp2*o8delmin )
                {
                    o8del = max(o8delmin*o8tp2,
                              min(o8del,term1*(o8one-o8tm2)));
                }
              }
           }
        }
    }
    /* if the current step was singular and not successful,try o8a greater o8del */
    /* the same, if o8qpterm in the last step did signal trouble, but          */
    /* stepsize selection was nevertheless successful                        */
    
    if ( o8itstep > 1 && o8accinf[o8itstep-1][30] < 0. ) o8del = min(o8tp1*delold,o8del0);
    
    /* ********************************************* */
    /* include nearly binding inequality constraints */
    /* ********************************************* */
    o8clist[0] = 0 ;
    /* o8clist is the list of nonlinear constraints to be included additionally */
    for (i = 1 ; i <= o8nres ; i++) 
    {
       if ( !(o8low[i] == o8up[i] ) )
       {
        term = o8res[2*i-1];
        term1 = o8res[2*i];
        if ( i > o8n )
        {
          term /= o8gresn[i-o8n];
          term1 /= o8gresn[i-o8n];
        }
        else
        {
          term /=max(o8one,o8xsc[i]);
          term1 /=max(o8one,o8xsc[i]);
        } 
        if ( min(term,term1) <= o8del && (o8bind[2*i-1]+o8bind[2*i] ==0 ) )
        {
          j=2*i-1;
          if ( term1 < term ) j=2*i;
          /* it may be useful to include constraints>0 if near its boundary */
          /* but avoid it, if the constraint was in the old delete-list     */
          /* constraint j is a candidate for inclusion in the active set    */
          /* check previous delete_list                                     */
/* inactive  
          for ( k = 1 ; k <= o8opti_delist[0] ; k++ )
          {
             if ( j == o8opti_delist[k] ) j = -1;
          }
end inactive */
          if ( j >= 1 )   
          /*  that means the value of j was not found in o8opti_delist */
          {
             o8bind[j]         = 1;
             o8aalist[0]       += 1;
             o8aalist[o8aalist[0]] = j;
             if ( j > 2*o8n )
             {
               o8alist[0] += 1 ;
               k = (j-2*o8n+1)/2 ;
               o8alist[o8alist[0]] = k ;
               if ( j % 2 == 0 ) 
               {
                 o8gres[0][k] = -o8one ;
               }
               else
               {
                 o8gres[0][k] = o8one ;
               }
             }
             if ( j > 2*(o8n+o8nlin) )
             {
               o8clist[0]       += 1 ;
               o8clist[o8clist[0]] = i-o8n-o8nlin ;
             }
          }
        }
       }
    }
    if ( o8clist[0] > 0 )
    {
       escongrad(o8clist,o8nlin,o8x,o8gres);
       for ( i = 1 ; i <= o8clist[0] ; i++ )
       {
         o8cgres[o8clist[i]+o8nlin] += 1 ;
         o8val[o8clist[i]+o8nlin]    = TRUE ; 
         o8gresn[o8clist[i]+o8nlin] = max(o8one,sqrt(o8sc4(1,o8n,o8clist[i]+o8nlin,o8gres)));
       }
     
    }

    rank0 = o8rank;
    nr0   = o8nr;
    o8nr    = o8aalist[0] ;
    
    o8dec(nr0+1,o8nr);
    
    if ( o8rank != o8nr && ! o8silent ) o8msg(3);
    
    o8ht(1,0,rank0+1,o8rank,o8n,o8qr,o8betaq,o8qgf,o8opti_trvec);
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8qgf[i] = o8opti_trvec[i];
    }
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8opti_yy[i] = -o8qgf[i]*o8scf;
    }
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8yu[i] = o8zero;
    }
    /* ********************************************* */
    /* first computation of lagrangian multipliers o8u */
    /* ********************************************* */
    
    o8sol(1,o8rank,o8opti_yy,o8yu);
    
    umin  = o8zero;
    unorm = o8zero;
    for (i = 1 ; i <= 2*o8nres ; i++) 
    {
        o8u[i] = o8zero;
    }
    iumin  = 0;
    uminsc = o8zero;
    for (i = 1 ; i <= o8rank ; i++) 
    {
        unorm = max(unorm,fabs(o8yu[i]));
        k     = o8aalist[o8colno[i]];
        o8u[k]  = -o8yu[i];
        term = o8one ;
        if ( k > 2*o8n ) term=o8gresn[(k-2*o8n+1)/2];
        if ( !(o8low[(k+1)/2] == o8up[(k+1)/2]) ) 
        {
            if ( -o8yu[i]/term < uminsc ) 
            {
                iumin  = k;
                uminsc = -o8yu[i]/term;
            }
        }
    }
    if ( o8scf != o8zero ) 
    {
        for (i = 1 ; i <= o8n ; i++) 
        {
            o8opti_yx[i] = o8scf*o8gradf[i];
            for (j = 1 ; j <= o8nlin+o8nonlin ; j++) 
            {
                o8opti_yx[i] = o8opti_yx[i]-o8gres[i][j]*(o8u[2*j-1+2*o8n]-o8u[2*j+2*o8n]);
            }
        }
        for (i = 1 ; i <= o8n ; i++ )
        {
                o8opti_yx[i] -= o8xsc[i]*(o8u[2*i-1]-o8u[2*i]) ;
        } 
    /*  the l2-norm of the original unscaled gradient of the lagrangian */        
        o8b2n = o8vecn(1,o8n,o8opti_yx)/o8scf;
    } 
    else 
    {
        o8b2n = -o8one;
    }
    if ( ! o8silent ) o8info(2);
    if ( ! o8silent ) o8info(3);
    
    /* compute new o8delta */

    del1 = o8del;
    if ( o8b2n > o8zero ) del1 = max(o8del,o8tm1*min(o8del0,
    exp(o8p7*log( fabs(o8b2n)/(o8gfn+o8one) + max(o8zero,-uminsc) + sum)) ));
                
    /* exclude constraints which were candidates for inactivating in the */
    /* previous step if useful                                           */
    
    for (i = 1 ; i <= o8opti_delist[0] ; i++) 
    {
        j     = o8opti_delist[i];
        term1 = o8res[j]*(o8one-o8tm2) ;
        if ( j > 2*o8n )
        {
           term1 /= o8gresn[(j-2*o8n+1)/2];
        }
        else
        {
           term1 /= max(o8one,o8xsc[(j+1)/2]);
        }
        if ( term1 >= del1*o8tm2 ) del1 = max(o8delmin,min(del1,term1));
    }

    slackn = o8zero;
    for (i = 1 ; i <= o8nres ; i++) 
    {
        term = o8res[2*i-1];
        term1 = o8res[2*i];
        denom = o8one ;
        if ( i > o8n ) denom = o8gresn[i-o8n];
        slackn += 
           max(o8zero,term/denom-o8delmin)*max(o8zero,o8u[2*i-1]-o8smallw)/denom
          +max(o8zero,term1/denom-o8delmin)*max(o8zero,o8u[2*i]-o8smallw)/denom;
    }
    if ( o8upsi <= o8delmin && o8b2n <= o8epsx*(o8gfn+o8one)
        && o8b2n != -o8one  && uminsc >= -o8smallw &&
        slackn <= o8delmin*o8smallw*o8nres ) 
    {
        
        /* sufficient accuracy in Kuhn-Tucker conditions */
    
        o8optite = o8zero;
        
        return;
    }
    /* include additional constraints if necessary */
    
    l0 = o8aalist[0];
    o8clist[0]=0;
    for (i = 1 ; i <= o8nres ; i++) 
    {
      if ( o8low[i] != o8up[i] )
      {
        term = o8res[2*i-1];
        term1 = o8res[2*i];
        if ( i > o8n ) 
        {
          term /= o8gresn[i-o8n];
          term1 /= o8gresn[i-o8n];
        }
        else
        {
          term /= max(o8one,o8xsc[i]);
          term1 /= max(o8one,o8xsc[i]);
        }
        if ( min(term,term1) <= del1 && (o8bind[2*i-1]+o8bind[2*i] == 0 ))
        {
         
            if ( term <= term1 ) 
            {
              o8bind[2*i-1]         = 1;
              o8aalist[0]        = o8aalist[0]+1;
              o8aalist[o8aalist[0]] = 2*i-1;
              if ( i > o8n ) 
              {
                o8alist[0] += 1 ;
                o8alist[o8alist[0]] = i-o8n ;
              }
              if ( i > o8n+o8nlin )
              {
                 o8clist[0] +=1 ;
                 o8clist[o8clist[0]] = i-o8n-o8nlin;
              }
            }
            else 
            {   
              o8bind[2*i]         = 1;
              o8aalist[0]        = o8aalist[0]+1;
              o8aalist[o8aalist[0]] = 2*i;
            
              if ( i > o8n )
              {
                o8alist[0] += 1 ;
                o8alist[o8alist[0]] = i-o8n ;
              }
              if ( i > o8n+o8nlin )
              {
                 o8clist[0] +=1 ;
                 o8clist[o8clist[0]] = i-o8n-o8nlin;
              } 
            } 
        }        
      } 
    } /* end for i */

    o8del = del1;
    o8accinf[o8itstep][6]  = o8del;
    o8accinf[o8itstep][7]  = o8b2n0;
    o8accinf[o8itstep][9]  = o8aalist[0];
    o8accinf[o8itstep][10] = -1.;
    o8nr = o8aalist[0];
    if ( o8clist[0] != 0 )
    {
       escongrad(o8clist,o8nlin,o8x,o8gres);
       for ( i = 1 ; i <= o8clist[0] ; i++ )
       {
         o8cgres[o8clist[i]+o8nlin] += 1 ;   
         o8val[o8clist[i]+o8nlin]    = TRUE ;
         o8gresn[o8clist[i]+o8nlin] = max(o8one,sqrt(o8sc4(1,o8n,o8clist[i]+o8nlin,o8gres)));
       }
    }
    
    if ( l0 != o8nr ) 
    {
        rank0 = o8rank;
        
        o8dec(l0+1,o8nr);
        
        o8ht(1,0,rank0+1,o8rank,o8n,o8qr,o8betaq,o8qgf,o8opti_trvec);
        
        for (i = 1 ; i <= o8n ; i++) 
        {
            o8qgf[i] = o8opti_trvec[i];
        }
    }

    if ( o8rank != o8nr ) 
    {
        if ( ! o8silent ) o8msg(4);
        
        goto L400;
    }
    /* ******************************************************** */
    /* second solution for multipliers, o8rank may have changed ! */
    /* ******************************************************** */
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8opti_yy[i] = -o8qgf[i]*o8scf;
    }
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8yu[i] = o8zero;
    }
    o8sol(1,o8rank,o8opti_yy,o8yu);
    
    /* remember the column interchanges in o8qr! o8yu[i] corresponds to */
    /* o8u[o8aalist[o8colno[i]]]                                          */
    
    umin  = o8zero;
    unorm = o8zero;
    for (i = 1 ; i <= 2*o8nres ; i++) 
    {
        o8u[i] = o8zero;
    }
    iumin  = 0;
    uminsc = o8zero;
    for (i = 1 ; i <= o8rank ; i++) 
    {
        unorm = max(unorm,fabs(o8yu[i]));
        k     = o8aalist[o8colno[i]];
        o8u[k]  = -o8yu[i];
        term = o8one ;
        term =  ( k > 2*o8n )? o8gresn[(k-2*o8n+1)/2] : o8xsc[(k+1)/2];
        if ( !(o8low[(k+1)/2] == o8up[(k+1)/2]) ) 
        {
            if ( -o8yu[i]/term < uminsc ) 
            {
                iumin  = k;
                uminsc = -o8yu[i]/term;
            }
        }

    }
    if ( o8scf != o8zero ) 
    {

        for (i = 1 ; i <= o8n ; i++)
        {
            o8opti_yx[i] = o8scf*o8gradf[i];
            for (j = 1 ; j <= o8nlin+o8nonlin ; j++)
            {
                o8opti_yx[i] -= o8gres[i][j]*(o8u[2*j-1+2*o8n]-o8u[2*j+2*o8n]);
            }
        }
        for (i = 1 ; i <= o8n ; i++ )
        {
                o8opti_yx[i] -= o8xsc[i]*(o8u[2*i-1]-o8u[2*i]) ;
        }
        o8b2n = o8vecn(1,o8n,o8opti_yx)/o8scf;
    }
    o8accinf[o8itstep][8]  = o8b2n;
    o8accinf[o8itstep][11] = umin;
    
    o8shms();
   
    
    if ( ! o8silent ) 
    {  
      o8info(4);
      if ( l0 != o8nr ) o8info(2);
    }
    
    o8opti_delist[0] = 0;
    if ( o8phase >= 0 && o8b2n != -o8one ) 
    {
        if ( fabs(uminsc) >= max(o8smallw,fabs(o8b2n)/(o8gfn+o8one)*o8c1d)) 
        {
            for (i = 1 ; i <= o8nr ; i++) 
            {
                k = o8aalist[o8colno[i]];
                term = o8one ;
                if ( k > 2*o8n ) term = o8gresn[(k-2*o8n+1)/2];
                if ( !(o8low[(k+1)/2] == o8up[(k+1)/2]) )
                { 
                   if ( -o8yu[i]/term <= -o8smallw) 
                   {
                     o8opti_delist[0]         = o8opti_delist[0]+1;
                     o8opti_delist[o8opti_delist[0]] = k;
                   }
                }
            }
        }
    }
    /* *********************************************************** */
    /* the new o8opti_delist doesn't influence the current o8d but only the */
    /* computation of the next o8del                                 */
    /* *********************************************************** */
    
    o8eqres = TRUE;
    for (i = 1 ; i <= 2*o8nres ; i ++) 
    {
        o8eqres = o8eqres && ( o8bind[i] == o8bind0[i] );
    }
    /* compute condition number estimators of o8diag-o8r and o8diag of */
    /* Cholesky-decomposition of b                               */
    
    if ( o8nr > 1 ) 
    {
        term  = o8zero;
        term1 = o8one;
        for (i = 1 ; i <= o8nr ; i++) 
        {
            term  = max(term, fabs(o8diag[i]));
            term1 = min(term1,fabs(o8diag[i]));
        }
        o8accinf[o8itstep][13] = term/term1;
    } 
    else if ( o8nr == 1 ) 
    {
        o8accinf[o8itstep][13] = 1.;
    } 
    else 
    {
        o8accinf[o8itstep][13] = -1.;
    }
    term  = fabs(o8a[1][1]);
    term1 = fabs(o8a[1][1]);
    i     = 2;
    while ( i <= o8n ) 
    {
        term  = max(term, fabs(o8a[i][i]));
        term1 = min(term1,fabs(o8a[i][i]));
        i     = i+1;
    }
    o8accinf[o8itstep][14] = pow(term/term1,2);
    
    if ( ! o8silent ) o8info(5);
    
    /* since a represents the Cholesky-factor, this square */
    slackn = o8zero;
    for (i = 1 ; i <= o8nres ; i++) 
    {
        term = o8res[2*i-1];
        term1 = o8res[2*i]; 
        denom = (i > o8n ) ? o8gresn[i-o8n] : max(o8one,o8xsc[i]);
        slackn +=
           max(o8zero,term/denom-o8delmin)*max(o8zero,o8u[2*i-1]-o8smallw)/denom
          +max(o8zero,term1/denom-o8delmin)*max(o8zero,o8u[2*i]-o8smallw)/denom;
    }

    if ( umin >= -o8smallw &&
        slackn <= o8delmin*o8smallw*o8nres &&
        o8upsi <= o8nres*o8delmin && o8upsi0 <= o8nres*o8delmin
        && fabs(o8fx-o8fx0) <= eps*(fabs(o8fx)+o8one) &&
        o8b2n != -o8one &&
        o8b2n <= o8tp2*o8epsx*(o8gfn+o8one) ) 
    {
        
        csdifx = csdifx+1;
    } 
    else 
    {
        csdifx = 0;
    }
    if ( o8phase >= 0 && (o8accinf[o8itstep][14] > o8tp3 || ! o8analyt ) && csdifx > o8n ) 
    {
        o8optite = o8four;
            
        /* to avoid possible slow convergence with singular    */
        /* projected hessian or inaccurate numerical gradients */
            
        return;
    }
    /* compute damping factor for tangential component if o8upsi>o8tau0/2 */

    o8scf0 = o8one;
    if ( o8phase >= 0 && o8upsi > o8tau0*o8p5 )
         o8scf0 = max(o8one/o8scfmax,(o8two*(o8tau0-o8upsi)/o8tau0)*o8upsi*o8tm1/max(o8one,o8gfn) )/o8scf;
    o8accinf[o8itstep][15] = o8scf0;

    /* **************************** */
    /* compute tangential component */
    /* **************************** */
    
    for (i = o8nr+1 ; i <= o8n ; i++) 
    {
        o8opti_qtx[i] = o8opti_yy[i]*o8scf0;
    }
    /* o8opti_qtx[o8nr+1],..,o8opti_qtx[o8n] is s2 */
    
    /* **************************************************** */
    /* compute right hand side and vertical component       */
    /* use damping for inactivation direction if very large */
    /* no indirect inactivation if infeasibility large and  */
    /* we are not almost stationary on the current manifold */
    /* **************************************************** */
    
    fac = o8one;
    if ( -umin*o8c1d > o8b2n+o8upsi && o8b2n != -o8one ) fac = o8c1d;
    if ( o8upsi > o8tau0*o8p5 ) fac = o8zero;
    for (i = 1 ; i <= o8nr ; i++) 
    {
        k    = o8aalist[o8colno[i]];
        term = o8res[k];
        if ( !(o8low[(k+1)/2] == o8up[(k+1)/2]) )
        {
          if (  -o8yu[i] < o8zero && term > o8zero ) term = -term;
          if (  -o8yu[i] < o8zero ) term = term-o8yu[i]*fac;
        }
        o8opti_yx[i] = -term;
    }
    o8solt(1,o8nr,o8opti_yx,o8opti_qtx);
    
    /* o8opti_qtx is transformed direction of descent for o8phi */
    
    o8ht(-1,0,1,o8nr,o8n,o8qr,o8betaq,o8opti_qtx,o8opti_yx);
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8opti_qtx[o8perm[i]] = o8opti_yx[i];
    }
    
    /* solve l(transp)*o8d = o8opti_qtx, l = o8a = Cholesky-factor of b */
    
    o8rght(o8a,o8opti_qtx,o8d,&term,o8n);
    
    /* end solver */
    
    /* compute new penalty weights : regular case */
    
    clwold = o8clow;
    if ( o8phase >= 0 ) o8sce();
    if ( o8clow > clwold ) 
    {
    
        /* o8tau_qp depends on the (new) weights */
        
        term = o8w[1];
        for (i = 1 ; i <= 2*o8nres ; i++) 
        {
            term = max(term,o8w[i]);
        }
        o8tauqp = max(o8one,min(o8tauqp,term));
    }
    /* compute parameter o8phase and stopping criterion */

    if ( uminsc < -o8smallw )     o8phase = min(1,o8phase);
    if ( ! o8eqres )              o8phase = min(0,o8phase);
    if ( o8eqres && o8upsi < o8tau0 ) o8phase = max(1,o8phase);
    
    /* rescale and project o8d if appropriate */
    
    o8cutd();
    
    /* compute the directional derivative o8dirder */
    
    o8dird();
    
    /* terminate if correction is small */
    
    if ( o8dnorm <= o8epsx*(o8xnorm+o8epsx) && o8upsi <= o8delmin
        && o8b2n != -o8one && uminsc >= -o8smallw && o8b2n <= o8epsx*(o8gfn+o8one) ) 
    {
        
        o8optite = o8one;
        
        return;
    }
    L350:

    /* reenter from the singular case: o8dirder has been computed already */
    
    o8accinf[o8itstep][16] = o8xnorm;
    o8accinf[o8itstep][17] = o8dnorm;
    o8accinf[o8itstep][18] = o8phase;
    
    /* compute stepsize */
    
    o8cfincr = o8icf;

    /* if no descent direction is obtained, check whether restarting the method */
    /* might help                                                               */
    
    if ( o8dirder >= o8zero ) 
    {
    
        /* no direction of descent */
        
        if ( ! o8silent ) o8msg(11);
        o8stptrm = -o8two;
        o8sig    = o8zero;
        
        goto L360;
    }
    /* if directional derivative correct but very small, terminate */
    /* since no further progress might be possible                 */
    
    if ( -o8dirder <= o8epsmac*o8tp2*(o8scf*fabs(o8fx)+o8psi+o8one) )
    {
        if ( o8upsi > o8delmin*o8nres ) 
        {
            o8optite = -o8one;
            o8stptrm = -o8one;
        } 
        else 
        {
            o8optite = o8two;
            o8stptrm = o8one;
        }
        o8sig = o8zero;
        
        goto L360;
    }
    /* o8phase = 2 : we may hope to obtain superlinear convergence */
    /* switch to Maratos-correction is on then                   */
    /* return to o8phase = 1 if first order correction large       */
    
    if ( o8phase >= 1 && o8dnorm <= o8smalld*(o8xnorm+o8smalld) && o8scf0 == o8one &&
        uminsc >= -o8smallw && ! o8singul) o8phase = 2;
        
    /* return to o8phase 1 since correction large again */
    
    if ( o8phase == 2 && o8dnorm > (o8xnorm+o8smalld) ) o8phase = 1;

    o8smax();
    
    /* o8stmaxl is the maximal stepsize such that point on projected */
    /* ray changes with sigma, but o8sigla at most                   */
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8dd[i] = o8zero;
        o8x1[i] = o8x[i]+o8d[i];
    }
    /* compute second order correction of infeasibility if useful    */
    /* if o8x1 already violates the bounds by viobnd, suppress it      */
    /* also suppress it if becomes too large compared with the first */
    /* order correction                                              */
    viobnd = FALSE ;     
    for ( i = 1 ; i<= o8n ; i++ ) 
    {
      if ( (o8llow[i] && o8x1[i] < o8ug[i]-o8taubnd) || 
           (o8lup[i] && o8x1[i] > o8og[i]+o8taubnd )  )
      { viobnd = TRUE ; 
      }
    }

    if ( o8phase == 2 && o8dnorm > sqrt(o8xnorm*o8epsmac)  && ! o8singul 
         && ! viobnd && o8nres > 0 ) 
    {
    /* compute the Maratos correction using a difference approximation */    
        if ( o8bloc ) 
        {
          o8valid = FALSE ;
          /* user_eval must reset o8valid */
          user_eval(o8x1,-1);
        }
        
        /* only function values, from o8xtr = o8xsc*o8x1 */
        o8clist[0] = 0 ; 
        for (i = 1 ; i <= o8aalist[0] ; i++) 
        {
        /* here only the active constraints are evaluated  and */
        /* stored in o8opti_yx                                        */
             o8opti_yx[i] = o8zero;
             k = o8aalist[i];
             if ( k > 2*(o8n+o8nlin) )
             /* for the linear constraints the function value at o8x1 is o8zero */
             /* anyway                                                      */
             {
               o8clist[0] += 1 ;
               o8clist[o8clist[0]] = (k+1)/2-o8n-o8nlin;
             }
        }
  /*  evaluate nonlinear active constraints from o8clist only         */
        for ( j = 1 ; j <= o8clist[0] ; j++ )
        {
           o8confuerr[o8clist[j]] = FALSE ;
           o8cres[o8clist[j]+o8nlin] += 1 ;
        }
        escon(2,o8clist,o8x1,o8opti_con,o8confuerr);
  /*                                                                */
        
        for ( j = 1 ; j <= o8clist[0] ; j++ )
        {
          if ( o8confuerr[o8clist[j]] ) goto L355;
        }
  /* if no error occured, computation of o8opti_yx is now to be completed */
        for ( i = 1 ; i <= o8aalist[0] ; i++ ) 
        {
           k = o8aalist[i];
           if ( k > 2*(o8n+o8nlin) )
           {
             if ( k % 2 == 0 ) 
             {
               o8opti_yx[i] =  o8up[(k+1)/2]-o8opti_con[(k+1)/2-o8n-o8nlin];
             }
             else
             {
               o8opti_yx[i] = o8opti_con[(k+1)/2-o8n-o8nlin]-o8low[(k+1)/2];
             }
           }
        }     
        for (i = 1 ; i <= o8aalist[0] ; i++) 
        {
            o8opti_yy[i] = -o8opti_yx[o8colno[i]];
        }
        o8solt(1,o8nr,o8opti_yy,o8dd);
        
        for (i = o8nr+1 ; i <= o8n ; i++) 
        {
            o8dd[i] = o8zero;
        }
        o8ht(-1,0,1,o8nr,o8n,o8qr,o8betaq,o8dd,o8opti_yx);
        
        for (i = 1 ; i <= o8n ; i++) 
        {
            o8dd[o8perm[i]] = o8opti_yx[i];
        }
        o8rght(o8a,o8dd,o8dd,&term,o8n);
        
        if ( sqrt(term) > o8p5*o8dnorm ) 
        {
        
            /* second order correction almost as large as first order o8one: */
            /* not useful                                                  */
            
            for (i = 1 ; i <= o8n ; i++) 
            {
                o8dd[i] = o8zero;
            }
            if ( ! o8silent ) o8msg(6);
        }
    }
    L355:

    if ( ! o8silent ) o8info(7);

    o8sig = min(o8one,o8stmaxl);
    
    o8unim(o8sig);

    L360:

    o8cfincr = o8icf-o8cfincr;
    if ( ! o8silent ) o8info(10);
    
    /* count successive small steps */
    
    term = o8scf*(o8fx0-o8fx)+o8psi0-o8psi;
    if ( fabs(term) <= o8epsmac*o8tp3*(o8scf*fabs(o8fx)+o8psi) ) 
    {
        csmdph = csmdph+1;
    } 
    else 
    {
        csmdph = 0;
    }
    /* csmdph counts contiguous small differences of penalty function o8phi */
    
    if ( csmdph > max(o8n,10) ) 
    {
        o8optite = o8seven;
        
        return;
    }
    if ( o8sig <= o8five*o8tm2 ) 
    {
        if ( o8sig0 <= o8five*o8tm2 ) csssig += 1;
    } 
    else 
    {
        csssig = 0;
    }
    /* csssig counts the number of successive small o8sig's */

    o8accinf[o8itstep][21] = o8sig;
    o8accinf[o8itstep][22] = o8cfincr;
    o8accinf[o8itstep][23] = o8dirder;
    o8accinf[o8itstep][24] = o8dscal;
    o8accinf[o8itstep][25] = o8cosphi;
    o8accinf[o8itstep][26] = o8violis[0];
    
    if ( o8sig == o8zero && o8stptrm == o8one && o8optite == o8two ) 
    {
    
        /* no further significant progress possible */
        
        if ( ! o8silent ) o8info(17);
        
        return;
    }
    if ( o8stptrm == o8one && o8sig <= o8tm4 && o8accinf[o8itstep][13] > o8tp4 && ! o8singul
        && o8nres > 0) 
    {
        
        /* try o8a regularized step, hopefully this will give a better o8d */
        /* and larger o8sig                                              */
        
        if ( o8accinf[o8itstep][14] > o8tp4 ) o8inim();
        o8ident  = TRUE;
        o8singul = TRUE;
        
        goto L400;
    }
    if ( o8stptrm < o8zero ) 
    {
    
        /* stepsize selection failed */
        
        if ( ! o8ident ) 
        {
        
            /* try restart with o8a = identity scaled */
            
            if ( ! o8silent ) o8msg(7);
            o8ident     = TRUE;
            o8opti_delist[0] = 0;
            o8violis[0] = 0;
            csreg     = 0;
            csssig    = 0;
            csirup    = 0;
            
            o8inim();
            
            o8aalist[0] = nrbas;
            for (i = 1 ; i <= 2*o8nres ; i++) 
            {
                o8bind[i] = o8opti_bindba[i];
            }
            if ( o8upsi >= o8tau0 ) 
            {
            
                goto L100;
                
            } 
            else 
            {
            
                goto L200;
            }
        }
        if ( ! o8singul && o8ident && o8accinf[o8itstep][13] > o8tp4 && o8nres > 0 ) 
        {
        
            /* try the full SQP-direction               */
            /* this may be the third try for this point */
            
            o8singul = TRUE;
            o8ident  = TRUE;
            
            goto L400;
        }
        if ( o8stptrm == -o8two ) 
        {
            o8optite = -o8four;
            
            return;
        }
        if ( o8sig == o8zero && o8optite == -o8one ) return;
        
        /* unidimensional search unsuccessfully terminated */
    
        o8optite = -o8two;
        
        return;
    }
    if ( o8singul && o8itstep > o8n && fabs(o8fx-o8fx0) <= eps*(fabs(o8fx)+o8one) && o8phase >= 0
        && o8upsi <= o8nres*o8delmin && o8upsi0 <= o8nres*o8delmin
        && slackn <= o8delmin*o8smallw*o8nres && o8infeas <= o8upsi && ! o8ident ) 
    {
        
        /* since multipliers may be incorrect for o8infeas != o8zero be careful */
    
        o8optite = o8four;
        
        /* avoid slow progress in case of singular constraints */
    
        return;
    }
    /* relaxed termination criteria in the singular case */
    
    if ( o8singul && o8upsi <= o8delmin*o8nres && o8upsi0 <= o8delmin*o8nres
         && o8b2n != -o8one && o8b2n <= (o8gfn+o8one)*o8epsx*o8tp2 && o8phase >= 0
         && slackn <= o8delmin*o8smallw*o8nres && o8infeas <= o8upsi ) 
    {
         
        /* since multipliers may be incorrect for o8infeas != o8zero be careful */
        
        o8optite = o8three;
        
        return;
    }
    k = 0;
    for (i = 1 ; i <= o8n ; i++) 
    {
        if ( fabs(o8difx[i]) >= o8epsx*(fabs(o8x[i])+o8tm2) ) k = 1;
    }
    if ( k == 0 ) 
    {
        cschgx = cschgx+1;
    } 
    else 
    {
        cschgx = 0;
    }
    if ( cschgx > o8nreset && o8singul ) 
    {
    
        /* very slow progress in o8x in the singular case. terminate */
        
        o8optite = o8five;
        
        return;
    }
    /* new value of o8x has been accepted */
    
    o8xnorm = o8vecn(1,o8n,o8x);
    o8ident = FALSE;
    
    o8egph(o8gphi0);
    
    for (i = 1 ; i <= o8nonlin ; i++) 
    {
      o8val[i+o8nlin] = FALSE;
    }
    o8val[0] = FALSE ;
    /* in o8bloc mode , all constraints are evaluated in the user interface */
    /* the subsequent escongrad simply is then move from o8fugrad to o8gres   */


  /* ****************************************                               */
    if ( o8bloc ) 
    {
      o8valid = FALSE ;
      /* user_eval must reset o8valid */
      user_eval(o8x,2);
      
    }
  /* ****************************************                               */

    
    /* evaluate gradients only, since function values are already */
    /* o8valid from unidimensional minimization                     */
    /* argument is o8xtr = o8xsc*o8x                                    */
    
    if ( o8phase >= 0  ) 
    {
        o8val[0] = TRUE;
        esgradf(o8x,o8gradf);
        o8icgf += 1;
    }
  /* evaluate gradients of nonlinear constraints  at the new point   */
  /* in case of a o8bloc evaluation this is simply a move from o8fugrad  */
    o8clist[0] = 0 ; 
    for ( i = 1 ; i <= o8aalist[0] ; i++ ) 
    {
      k = o8aalist[i];
      if ( k > 2*(o8n+o8nlin) )
      {
         o8clist[0] += 1 ;
         o8clist[o8clist[0]] = (k+1)/2-o8n-o8nlin;
         if ( k % 2 == 0 )
         {
           o8gres[0][(k+1)/2-o8n] = -o8one ;
         }
         else
         {
           o8gres[0][(k+1)/2-o8n] = o8one ;
         }
      }
    }      
    if ( o8clist[0] != 0 )
    {
  /*  this evaluates the gradients of the nonlinear constraints which were    */
  /*  active at the old point also at the new point (necessary for the update)*/

       escongrad(o8clist,o8nlin,o8x,o8gres);
       for ( i = 1 ; i <= o8clist[0] ; i++ )
       {
         o8cgres[o8clist[i]+o8nlin] += 1 ;   
         o8val[o8clist[i]+o8nlin]    = TRUE ;
         o8gresn[o8clist[i]+o8nlin] = max(o8one,sqrt(o8sc4(1,o8n,o8clist[i]+o8nlin,o8gres)));
       }
    }
    o8egph(o8gphi1);
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8opti_yx[i] = o8x[i]-o8x0[i];
        o8opti_yy[i] = o8gphi1[i]-o8gphi0[i];
    }
    /* since a represents the Cholesky-factor, this sqrt */
    
    term = sqrt(o8vecn(1,o8n,o8opti_yy)/o8vecn(1,o8n,o8opti_yx));
    if ( term != o8zero && o8phase >= 0 ) o8matsc = max(o8one/o8scfmax,min(o8scfmax,term/2));
    
    /* current scaling of identity in case of restart */

    o8aalist[0] = 0;
    o8clist[0] = 0 ;
    o8alist[0] = 0 ;
    /* o8clist is the list of nonlinear active constraints whose gradients */
    /* need to be evaluated                                              */
    for (i = 1 ; i <= o8nres ; i++) 
    {
        o8u0[2*i-1]    = o8u[2*i-1];
        o8bind0[2*i-1] = o8bind[2*i-1];
        o8res0[2*i-1]  = o8res[2*i-1];
        o8u0[2*i]    = o8u[2*i];
        o8bind0[2*i] = o8bind[2*i];
        o8res0[2*i]  = o8res[2*i];
        if ( o8low[i] == o8up[i] ) 
        {
            o8aalist[0] += 1 ;
            o8aalist[o8aalist[0]] = 2*i-1 ;
            o8bind[2*i-1] = 1;
            o8bind[2*i]   = 0;
            if ( i > o8n+o8nlin  && ! o8val[i-o8n] )
            {
              o8clist[0] += 1 ;
              o8clist[o8clist[0]] = i-o8n-o8nlin ;
            }
            if ( i > o8n )
            {
              o8alist[0] +=1 ;
              o8alist[o8alist[0]] = i-o8n;
            }
        } 
        else 
        {
            o8bind[2*i-1] = 0;
            o8bind[2*i]   = 0;
            term = o8res[2*i-1];
            term1 = o8res[2*i] ;
            if ( i > o8n ) 
            {
              term /= o8gresn[i-o8n];
              term1 /= o8gresn[i-o8n];
            }
            else
            {
              term /= max(o8one,o8xsc[i]);
              term1 /= max(o8one,o8xsc[i]);
            }
            if ( min(term,term1) <= o8delmin )
            { 
  /* an active inequality constraint at the new point    */
  /* may be the gradient has been evaluated already (for computing the update)*/  
              if ( term <= term1 )
              {
                 o8bind[2*i-1] = 1;
                 o8aalist[0] += 1;
                 o8aalist[o8aalist[0]] = 2*i-1 ;
                 if ( i > o8n )
                 {
                   o8alist[0] +=1 ;
                   o8alist[o8alist[0]] = i-o8n ;
                   o8gres[0][i-o8n] = o8one ;
                   if ( i > o8n+o8nlin && ! o8val[i-o8n] )
                   { 
                     o8clist[0] += 1;
                     o8clist[o8clist[0]] = i-o8n-o8nlin ;
                   }
                 }
              }
              else
              {
                 o8bind[2*i] = 1 ;
                 o8aalist[0] +=1 ;
                 o8aalist[o8aalist[0]] = 2*i ;
                 if ( i > o8n )
                 {
                   o8alist[0] +=1 ;
                   o8alist[o8alist[0]] = i-o8n ;
                   o8gres[0][i-o8n] = -o8one ;
                   if ( i > o8n+o8nlin && ! o8val[i-o8n] )
                   {
                     o8clist[0] += 1;
                     o8clist[o8clist[0]] = i-o8n-o8nlin ;
                   }
                 }
              }
            } /* end an active constraint */ 
        } /* end inequality case */     
    } /* end for i */

    if ( o8clist[0] != 0 )
    {
  /*  this evaluates the gradients of the nonlinear constraints which were    */
  /*  active at the new point but not at the old point                        */
       
       escongrad(o8clist,o8nlin,o8x,o8gres);
       for ( i = 1 ; i <= o8clist[0] ; i++ )
       {
         o8cgres[o8clist[i]+o8nlin] += 1 ;   
         o8val[o8clist[i]+o8nlin]    = TRUE ;
         o8gresn[o8clist[i]+o8nlin] = max(o8one,sqrt(o8sc4(1,o8n,o8clist[i]+o8nlin,o8gres)));
       }
    }

    /* **** o8bind now corresponds to the state of the new point               */
    /* **** but there may be gradients evaluated at the new point            */
    /* **** not yet put to o8bind                                              */
    /* ****                                                                  */
    /* **** update the unprojected quasi-Newton-matrix anyway                */
    /* ****                                                                  */

    if ( o8scf != o8zero ) 
    {
        if ( csirup > o8nreset || csssig > o8nreset || csreg  > o8nreset ) 
        {
            csreg  = 0;
            csssig = 0;
            csirup = 0;
         
            o8inim();
            
        } 
        else 
        {
        
            o8bfgs();
         
            /* for projected update:  if ( ! o8silent ) o8info(13) */
            /* for Pantoja&Mayne update:                         */
        
            if ( ! o8silent ) o8info(14);
        }
    }
    /* proceed */

    if ( o8accinf[o8itstep][27] == o8one ) 
    {
        if ( o8itstep > 1 && 
            o8accinf[o8itstep-1][29] != o8zero && o8accinf[o8itstep][29] != o8zero ) 
        {

            /* c ount s uccessive ir regular o8up dates */
            
            csirup = csirup+1;
        } 
        else 
        {
            csirup = 0;
        } 
    }
    /* o8accinf[o8itstep][27] = 1   update Pantoja&Mayne                   */
    /*                    = 0   noupdate                               */
    /*                    = -1  restart                                */
    /*                    = 2   normal BFGS (o8nr = 0)                   */
    /*                    = 3   Powell's modified BFGS (o8nr = 0)        */
    /* o8accinf[o8itstep][28] = modification term tk/ den2/den1  resp.     */
    /* o8accinf[o8itstep][29] = modification factor th/xsik resp.          */
    /* csirup counts the number of successive irregular updating steps */
    
    if ( o8phase == -1 ) 
    {
    
        goto L100;
        
    }
    else
    {
    
        goto L200;
    }
    L400:

    o8singul = TRUE;
    o8phase  = min(o8phase,0);
    o8accinf[o8itstep][10] = o8one;
    
    /* try to compute a descent direction using                       */
    /* an extended quadratic program with                             */
    /* individual o8slack variable for any constraint                   */
    /* compute damping factor for tangential component if o8upsi>o8tau0/2 */
    /* by rescaling f if possible                                     */
    
    o8scf0 = o8one;
    if ( o8phase >= 0 && o8upsi > o8tau0*o8p5 ) 
    {
        scfh = max(o8one/o8scfmax,
               min(o8scfmax,(o8two*(o8tau0-o8upsi)/o8tau0)*o8upsi*o8tau/max(o8one,o8gfn) ) );
        if ( (o8fxst-o8fx)*scfh+scfh/o8scf*(o8psist-o8psi) >= 
             scfh/o8scf*o8eta*o8clow && o8lastch <= o8itstep-4
             && (scfh < o8tm1*o8scf || scfh > o8tp1*o8scf) )
        {
             
            /* rescale the objective function if this seems promising and */
            /* the change is significant                                  */
            
            o8clow  = o8clow+1;
            term  = scfh/o8scf;
            o8scf0  = term ;
            o8psi   = o8psi*term;
            o8psist = o8psist*term;
            for (i = 1 ; i <= 2*o8nres ; i++) 
            {
                o8u[i] = o8u[i]*term;
            }
            unorm  = unorm*term;
            o8scf    = scfh;
            o8lastch = o8itstep;
            o8accinf[o8itstep][15] = o8scf;
            term = sqrt(term);
            for ( i = 1 ; i <= o8n ; i++) 
            {
                o8diag0[i] = term*o8diag0[i];
                for ( j = 1 ; j <= o8n ; j++) 
                {
                    o8a[j][i] = o8a[j][i]*term;
                }
            }
            o8matsc = o8matsc*term;
            if ( ! o8silent ) o8msg(2);
        }
    }
    /* o8slack is o8upsi at most */
    
    o8accinf[o8itstep][32] = o8upsi;
    if ( ! o8silent ) o8info(15);

    o8accinf[o8itstep][13] = -o8one;
    term  = fabs(o8a[1][1]);
    term1 = term;
    i     = 2;
    while ( i <= o8n ) 
    {
        term  = max(term ,fabs(o8a[i][i]));
        term1 = min(term1,fabs(o8a[i][i]));
        i     = i+1;
    }
    o8accinf[o8itstep][14] = pow(term/term1,2);
    if ( ! o8silent ) o8info(5);
    clwold = o8clow;
    
    /* save for restart */
    
    tauqp0 = o8tauqp;
    for (i = 1 ; i <= 2*o8nres ; i++) 
    {
        o8u[i] = o8zero;
    }
    /* no second order correction here */
    for (i = 1 ; i <= o8n ; i++) 
    {
        o8dd[i] = o8zero;
    }
    /* solve full qp */
    o8qpdu();
    
    if ( o8dnorm == o8zero && o8qpterm == 1 && o8optite == o8three ) return;
    if ( o8dnorm <= o8epsx*(min(o8xnorm,o8one)+o8epsx) && o8qpterm < 0 ) 
    {   /* QP solver indicated trouble */
    
        /* may be it failed because of illconditioning */
    
        if ( o8upsi >= o8nres*o8delmin && qpnew ) 
        {
        
            /* restarting the method has been done already: game is over */
            
            o8optite = -o8one;
            if ( ! o8silent ) o8info(18);
            
            return;
        }
        if ( qpnew ) 
        {
            o8optite = o8qpterm-o8five;
            if ( ! o8silent ) o8info(18);
            
            return;
        }
        /* try o8a = id */
        
        qpnew = TRUE;
        for (i = 1 ; i <= 2*o8nres ; i++) 
        {
            o8w[i] = o8one;
        }
        o8lastch = o8itstep;
        
        o8inim();
            
        o8ident = TRUE;
        
        /* formerly o8tauqp = tauqp0 */
        
        o8tauqp    = o8one;
        o8aalist[0] = nrbas;
        for (i = 1 ; i <= 2*o8nres ; i++) 
        {
            o8bind[i] = o8opti_bindba[i];
        }
        if ( o8scf == o8zero ) 
        {
        
            goto L100;
            
        } 
        else 
        {
        
            goto L200;
        }
    }
    o8accinf[o8itstep][11] = o8zero;
    o8opti_delist[0] = 0;
    umin      = o8zero;
    
    /* o8b2n is defined also internally in o8qpdu */
    
    o8shms();
    
    if ( (o8qpterm >= 0 || o8qpterm == -3) && o8scf != o8zero ) 
    {
        unorm = fabs(o8u[1]);
        for (i = 2 ; i <= 2*o8nres ; i++) 
        {
            unorm = max(unorm,fabs(o8u[i]));
        }
        for (i = 1 ; i <= o8n ; i++)
        {
            o8opti_yx[i] = o8scf*o8gradf[i];
            for (j = 1 ; j <= o8nlin+o8nonlin ; j++)
            {
                o8opti_yx[i] -= o8gres[i][j]*(o8u[2*j-1+2*o8n]-o8u[2*j+2*o8n]);
            }
        }
        for (i = 1 ; i <= o8n ; i++ )
        {
            o8opti_yx[i] -= o8xsc[i]*(o8u[2*i-1]-o8u[2*i]) ;
        }
        o8b2n = o8vecn(1,o8n,o8opti_yx)/o8scf;
    } 
    else 
    {
        o8b2n = -o8one;
        
        /* signals "undefined" here */
    }
    if ( ! o8silent ) o8info(2);
    if ( ! o8silent ) o8info(16);
    
    /* to avoid termination */
    
    if ( o8b2n == -o8one ) o8b2n = o8epsmac/o8tolmac;
    if ( o8qpterm >= 0 && o8dnorm <= o8tm2*o8epsx*(o8epsx+min(o8one,o8xnorm))) 
    {
        if ( o8upsi <= o8nres*o8delmin ) 
        {
            o8optite = o8six;
        }
        else 
        {
            o8optite = -o8five;
        }
        return;
    }
    /* check whether QPsolver terminated unsuccessfully */
    
    if ( o8qpterm < 0 ) 
    {
    
        /* we have a unfeasible solution for QP. try it */
        /* but don't complain if it fails               */
        
        if ( ! o8silent ) o8msg(10);
    }
    if ( o8clow > clwold ) 
    {
        term = o8one;
        for (i = 1 ; i <= 2*o8nres ; i++) 
        {
            term = max(term,o8w[i]);
        }
        o8tauqp = max(o8one,term);
    }
    if ( o8tauqp > pow(o8taufac,3)*tauqp0 ) o8tauqp = pow(o8taufac,3)*tauqp0;
    
    /* no change of o8tauqp otherwise */
    
    o8b2n0   = o8b2n;
    umin   = o8zero;
    uminsc = o8zero;
    if ( o8qpterm >= 1 ) 
    {
      slackn = o8zero;
      for (i = 1 ; i <= o8nres ; i++)  
      { 
        term = o8res[2*i-1];
        term1 = o8res[2*i];
        denom = o8one ;
        if ( i > o8n ) denom = o8gresn[i-o8n];
        slackn +=
            max(o8zero,term/denom-o8delmin)*max(o8zero,o8u[2*i-1]-o8smallw)/denom
           +max(o8zero,term1/denom-o8delmin)*max(o8zero,o8u[2*i]-o8smallw)/denom;
      }

    } 
    else 
    {
    
        /* o8slack is undefined, since multipliers are undefined */
        /* use this value to prevent premature termination     */
        
        slackn = o8one;
    }
    goto L350;
}

void o8inim(void) 
{
/* ************************************************************************ */
/*      initialize the quasi Newton update with a multiple of the identity  */
/* ************************************************************************ */


    void o8info(INTEGER icase);
    
    static INTEGER i,j;

    for (i = 1 ; i <= o8n ; i++) 
    {
        for (j = 1 ; j <= o8n ; j++) 
        {
            o8a[j][i] = o8zero;
        }
        o8a[i][i]  = o8matsc;
        o8diag0[i] = o8matsc;
    }
    o8accinf[o8itstep][27] = -o8one;
    o8accinf[o8itstep][14] = o8one;
    if ( ! o8silent ) o8info(20);
    
    return;
}
void o8dird(void)
{
/* **************************************************************************** */
/*          compute the directional derivative of the L1-penalty-function       */
/* **************************************************************************** */
    
    DOUBLE o8sc1(INTEGER i,INTEGER j,DOUBLE o8a[],DOUBLE b[]);
    DOUBLE o8sc3(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]);
    
    static INTEGER  i;
    static DOUBLE   term,term1;

    /* compute directional derivative of Zangwill function */
    
    L100:

    o8dirder = o8sc1(1,o8n,o8gradf,o8d)*o8scf;
    
    for (i = 1 ; i <= o8nres ; i++) 
    {
       if ( i <= o8n )
       {
       /*  bound constraints                                    */
          if ( o8low[i] == o8up[i] ) 
          /*    equality constraint type h(o8x)=o8x(i)-o8low(i)=0     */
          {
            if ( o8res[2*i-1] <= -o8tp3*o8epsmac )
            {
               o8dirder -=  o8d[i]*o8w[2*i-1];
            }
            else if ( o8res[2*i-1] >= o8tp3*o8epsmac )
            {
               o8dirder += o8d[i]*o8w[2*i-1];
            }
            else
            {
               o8dirder += fabs(o8d[i])*o8w[2*i-1];
            }
          }
          else
          {
            if ( o8bind[2*i-1] == 1 )
            /*  active lower bound */
            {
              term = o8w[2*i-1]*o8d[i] ;
              if ( fabs(o8res[2*i-1]) <= o8tp3*o8epsmac )
              {
                o8dirder -=  min(o8zero,term);
              }
              else
              { 
                if ( o8res[2*i-1] < -o8tp3*o8epsmac )
                {
                  term = min ( term , -o8res[2*i-1]*o8w[2*i-1] );
            /* penalty function can decrease at most by  -o8res[2*i-1]*o8w[2*i-1] */
                  o8dirder -= term ;
                }
              }
            }  
            if ( o8bind[2*i] == 1 ) 
            {
              /* active upper bound */
              term = -o8d[i]*o8w[2*i];
              if ( fabs(o8res[2*i]) <= o8tp3*o8epsmac )
              {
                o8dirder -= min(o8zero,term);
              }
              else
              {   
                if ( o8res[2*i] < -o8tp3*o8epsmac )
                {
                  term = min ( term , -o8res[2*i]*o8w[2*i] );
                  o8dirder -= term ;
                }
              }  
            }
          }
       }
       else
       {      
          term  = o8sc3(1,o8n,i-o8n,o8gres,o8d)*o8gres[0][i-o8n];
          if ( o8low[i] == o8up[i] ) 
          {
            if ( o8res[2*i-1] <= -o8tp3*o8epsmac )
            {
               o8dirder -=  term*o8w[2*i-1];
            }
            else if ( o8res[2*i-1] >= o8tp3*o8epsmac )
            {
               o8dirder += term*o8w[2*i-1];
            }
            else
            {
               o8dirder += fabs(term)*o8w[2*i-1];
            }
          }
          else
          {
            if ( o8bind[2*i-1] == 1 )
            {
              term = o8w[2*i-1]*term ;
              if ( fabs(o8res[2*i-1]) <= o8tp3*o8epsmac )
              {
                o8dirder -= min(o8zero,term);
              }
              else
              { 
                if ( o8res[2*i-1] < -o8tp3*o8epsmac )
                {
                  term = min ( term , -o8res[2*i-1]*o8w[2*i-1] );
                  o8dirder -= term ;
                }
              }
            }  
            if ( o8bind[2*i] == 1 ) 
            {
              term = term*o8w[2*i];
              if ( fabs(o8res[2*i]) <= o8tp3*o8epsmac )
              {
                o8dirder -= min(o8zero,term);
              }
              else
              {   
                if ( o8res[2*i] < -o8tp3*o8epsmac )
                {
                  term = min ( term , -o8res[2*i]*o8w[2*i] );
                  o8dirder -= term ;
                }
              }  
            }
          }
       }
    }
    return;
}

/* **************************************************************************** */
/*                      cut o8d if appropriate and rescale                        */
/* **************************************************************************** */
void o8cutd(void) 
{

    DOUBLE o8sc1 (INTEGER i,INTEGER j,DOUBLE o8a[],DOUBLE b[]);
    DOUBLE o8vecn(INTEGER nl,INTEGER nm,DOUBLE o8x[]);

    static INTEGER  i;
    static DOUBLE   term,term1;

    o8xnorm  = o8vecn(1,o8n,o8x);
    term   = o8beta*(o8xnorm+o8one);
    o8dnorm  = o8vecn(1,o8n,o8d);
    o8d0norm = o8vecn(1,o8n,o8d0);
    o8dscal  = o8one;
    if ( o8dnorm*o8d0norm != o8zero ) 
    {
        o8cosphi = o8sc1(1,o8n,o8d,o8d0)/(o8d0norm*o8dnorm);
    } 
    else 
    {
        o8cosphi = o8zero;
    }
    if ( o8dnorm > term ) 
    {
    
        /* o8d too long: rescale */
        
        term1 = term/o8dnorm;
        o8dnorm = term;
        o8dscal = term1;
        for (i = 1 ; i <= o8n ; i++) 
        {
            o8d[i]  = o8d[i]*term1;
            o8dd[i] = o8dd[i]*pow(term1,2);
        }
    }
    /* since we project the ray with respect to the bounds, be sure */
    /* to compute the directional derivative correctly              */
    /* therefore correct o8d and o8dd appropriately                     */
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        if ( o8llow[i] && o8x[i]+o8sigsm*o8d[i] <= o8ug[i] ) 
        {
            o8d[i]  = o8zero;
            o8dd[i] = max(o8zero,o8dd[i]);
        }
        if ( o8lup[i] && o8x[i]+o8sigsm*o8d[i] >= o8og[i] ) 
        {
            o8d[i]  = o8zero;
            o8dd[i] = min(o8zero,o8dd[i]);
        }
    }
    o8dnorm = o8vecn(1,o8n,o8d);
    
    return;
}

/* **************************************************************************** */
/* compute maximum stepsize o8stmaxl such that projection on the box of lower     */
/* and upper bounds changes for o8sig in [0,o8stmaxl],if such exists                */
/* **************************************************************************** */
void o8smax(void) {
    
    static INTEGER  i;
    static LOGICAL  exis;
    /* exis is true if there exists some sigmaxl, such that the projection */
    /* of o8x+sigma*o8d on the feasible box does not change for sigma>=sigmaxl */
    /* that means no coordinate changes any more                           */ 
    exis = TRUE;

    for (i = 1 ; i <= o8n ; i++) {
        exis = exis &&( ( o8d[i] == o8zero )
            || ( o8lup[i]  && o8d[i] > o8zero )
            || ( o8llow[i] && o8d[i] < o8zero ) );
    }
    if ( exis ) 
    {
        o8stmaxl = o8sigsm;
        for (i = 1 ; i <= o8n ; i++) {
            if ( o8llow[i] && o8d[i] < o8zero ) {
                if ( -o8d[i]*o8sigla >= o8x[i]-o8ug[i] ) {
                    o8stmaxl = max(o8stmaxl,(o8x[i]-o8ug[i])/(-o8d[i]));
                } else {
                    o8stmaxl = o8sigla;
                }
            }
            if ( o8lup[i] && o8d[i] > o8zero ) {
                if ( o8d[i]*o8sigla >= o8og[i]-o8x[i] ) {
                    o8stmaxl = max(o8stmaxl,(o8og[i]-o8x[i])/o8d[i]);
                } else {
                    o8stmaxl = o8sigla;
                }
            }
        }
    } else {
        o8stmaxl = o8sigla;
    }
    /* but never use stepsize larger than o8sigla */
    
    o8stmaxl = min(o8sigla,o8stmaxl);
    
    return;
}

/* **************************************************************************** */
/*        restore the best point found so far to be the current new point       */
/* **************************************************************************** */
void o8rest(void) {
    
    static INTEGER  j;

    o8phi1  = o8phimin;
    o8psi1  = o8psimin;
    o8upsi1 = o8upsim;
    o8sig   = o8sigmin;
    o8fx1   = o8fmin;
    for (j = 1 ; j <= o8n ; j++) {
        o8x1[j] = o8xmin[j];
    }
    for (j = 1 ; j <= 2*o8nres ; j++) {
        o8res1[j] = o8resmin[j];
    }
    return;
}

/* **************************************************************************** */
/*             save the best point found so far in ...min variables             */
/* **************************************************************************** */
void o8save(void) {
    
    static INTEGER  i;
    
    o8phimin = o8phi1;
    o8upsim  = o8upsi1;
    o8psimin = o8psi1;
    o8fmin   = o8fx1;
    /* Rprintf("Assigned o8fmin\n"); */
    o8sigmin = o8sig;
    for (i = 1 ; i <= o8n ; i++) {
        o8xmin[i] = o8x1[i];
    }
    for (i = 1 ; i <= 2*o8nres ; i++) {
        o8resmin[i] = o8res1[i];
    }

    return;
}

/* **************************************************************************** */
/*                    evaluate the functions at the new point                   */
/* **************************************************************************** */
void o8eval(DOUBLE sigact,DOUBLE *sigres,LOGICAL *reject,LOGICAL *error) 
{

    void    esf(DOUBLE o8x[],DOUBLE *o8fx);
    void    escon(INTEGER i,INTEGER liste[] , DOUBLE o8x[],
                                 DOUBLE con[],INTEGER err[]);
    DOUBLE  o8sc3(INTEGER i, INTEGER j , INTEGER k , 
        DOUBLE **mat, DOUBLE o8x[]);
    void    user_eval(DOUBLE xvar[],INTEGER mode);
    
    static INTEGER  i,j;
    static DOUBLE   term , denom ;
    /* liste as a formal placeholder, not used here since type of call to */
    /* escon is 1                                                         */
    static INTEGER   liste[2] ;
    static LOGICAL eval_err ;   
    
     
    liste[0] = 0 ;
    liste[1] = 0 ; 
    o8sig = sigact;
    for (i = 1 ; i <= o8n ; i ++) 
    {
        o8x1[i] = o8x[i]+o8sig*(o8d[i]+o8sig*o8dd[i]);
        
        /* project with respect to the box-constraints */
        
        o8x1[i] = max(o8x1[i],o8ug[i]);
        o8x1[i] = min(o8x1[i],o8og[i]);
    }
    *reject = FALSE;
    *error  = FALSE;
    *sigres = o8sig;
    o8upsi1   = o8zero;
    o8psi1    = o8zero;

/**********                                            *******************/
    if ( o8bloc ) 
    {   
      o8valid = FALSE ;
      /* user_eval must reset o8valid */
      user_eval(o8x1,-1);
    }
/********** only function values, from o8xtr = o8x1*o8xsc    *******************/
    
/********** compute o8res1 = restrictions at o8x1     *******************/
    for ( i = 1 ; i <= o8n ; i++ ) 
    {
       o8res1[2*i-1] = o8x1[i]-o8ug[i];
       o8res1[2*i  ] = o8og[i]-o8x1[i];
    }
    for ( i = 1 ; i <= o8nlin ; i++ )
    {
       term = o8sc3(1,o8n,i,o8gres,o8x1);
       o8cres[i] += 1 ;
       o8res1[2*(i+o8n)-1] = term - o8low[o8n+i];
       o8res1[2*(i+o8n)  ] = o8up[o8n+i]-term;
    }
    for ( i = 1 ; i <= o8nonlin ; i++ )
    {
      o8confuerr[i] = FALSE ;
    }
    escon(1,liste,o8x1,o8eval_con,o8confuerr);
    /* for call type 1, liste is dummy */
    eval_err = FALSE ;
    for (i = 1 ; i <= o8nonlin ; i++ )
    {
       o8cres[o8nlin+i] += 1 ;
       eval_err = eval_err || o8confuerr[i] ;
    }
    if ( eval_err ) 
    {
      *error = TRUE ;
      return ;
    }
    for ( i = 1 ; i <= o8nonlin ; i++ )
    {
       o8res1[2*(o8n+o8nlin+i)-1] = o8eval_con[i]-o8low[i+o8n+o8nlin];
       o8res1[2*(o8n+o8nlin+i)  ] = o8up[i+o8n+o8nlin]-o8eval_con[i];
    }

    for ( i = 1 ; i <= o8nres ; i++ )
    {
      if ( o8low[i] == o8up[i] ) 
      {
         term = fabs(o8res1[2*i-1]);
         o8upsi1 = o8upsi1+term;
         if ( o8upsi1 > o8tau0 && o8phase != -1 ) 
         {
            *reject = TRUE;
            return;
         }
 
         o8psi1 = o8psi1+term*o8w[2*i-1];
         denom = ( i<=o8n ) ? max(o8one,o8xsc[i]) : o8gresn[i-o8n];
         /*  only the odd numbered constraint is used for equalities */
         if ( o8res1[2*i-1]*o8res[2*i-1] < o8zero && o8sig <= o8one && 
              (fabs(o8res[2*i-1])/denom >= o8tp3*o8epsmac 
               || fabs(o8res1[2*i-1])/denom >= o8tp3*o8epsmac ) ) 
            *sigres = min(*sigres,o8sig*o8res[2*i-1]/(o8res[2*i-1]-o8res1[2*i-1]));
      }      
      else
      {
         term = -min(o8zero,o8res1[2*i-1]);
         if ( o8res1[2*i-1] < -o8delmin && o8bind[2*i-1] == 0 )
         {
           o8violis[0] +=1 ;
           o8violis[o8violis[0]] = 2*i-1 ;
         }

         o8upsi1 = o8upsi1+term;
         if ( o8upsi1 > o8tau0 && o8phase != -1 )
         {
            *reject = TRUE;
            return;
         }
         o8psi1 = o8psi1+term*o8w[2*i-1];
         denom = ( i <= o8n) ? max(o8one, o8xsc[i]) : o8gresn[i-o8n];
         if ( o8res1[2*i-1]*o8res[2*i-1] < o8zero && o8sig <= o8one && 
              ( o8bind[2*i-1] == 0 ||
              (o8bind[2*i-1] == 1 && (fabs(o8res[2*i-1])/denom >= o8tp3*o8epsmac
              || fabs(o8res1[2*i-1])/denom >= o8tp3*o8epsmac ) ) ) )
            *sigres = min(*sigres,o8sig*o8res[2*i-1]/(o8res[2*i-1]-o8res1[2*i-1]));

         term = -min(o8zero,o8res1[2*i]);
         o8upsi1 = o8upsi1+term;
         if ( o8res1[2*i] < -o8delmin && o8bind[2*i] == 0 )
         {
           o8violis[0] += 1;
           o8violis[o8violis[0]] = 2*i ;
         }

         if ( o8upsi1 > o8tau0 && o8phase != -1 )
         {
            *reject = TRUE;
            return;
         }
         o8psi1 = o8psi1+term*o8w[2*i];

         if ( o8res1[2*i]*o8res[2*i] < o8zero && o8sig <= o8one && ( o8bind[2*i] == 0 ||
            (o8bind[2*i] == 1 && (fabs(o8res[2*i])/denom >= o8tp3*o8epsmac
            || fabs(o8res1[2*i])/denom >= o8tp3*o8epsmac ) ) ) )
            *sigres = min(*sigres,o8sig*o8res[2*i]/(o8res[2*i]-o8res1[2*i]));

      }
    }
       /* o8violis is the list of constraints                            */
       /* not binding which have been hit during unidimensional search */



    if ( o8phase != -1 ) 
    {
        o8ffuerr = FALSE;
        
        esf(o8x1,&o8fx1);
        o8icf +=  1 ;
        if ( o8ffuerr ) 
        {
            *error = TRUE;
            
            return;
        }
    } 
    else 
    {
        o8fx1 = o8zero;
    }
    o8phi1 = o8scf*o8fx1+o8psi1;
    
    return;
}

/* **************************************************************************** */
/*          determination of stepsize by an Armijo-like test for descent        */
/* **************************************************************************** */
void o8unim(DOUBLE sig1th) {

    /* sig1th the first proposed stepsize for searching on the arc              */
    /* if o8sig = o8one did'nt o8work                                                 */

    /* o8n = number of variables                                                  */
    /* o8x = current point                                                        */
    /* o8d = direction of descent. o8dd = second order correction                   */
    /* o8x,o8d = input                                                              */
    /* o8x0,o8d0 etc. information from previous step                                */

    /* o8xnorm,o8dnorm = euclidean length of o8x and o8d                                */
    /* o8stptrm = 1 on success , = -1 or = -2 otherwise                           */
    /* o8sig    =  computed stepsize                                              */
    /* it is assumed that o8one is asymptotically optimal                         */
    /* o8sigsm = smallest acceptable stepsize                                     */
    /* o8sigla = largest  acceptable stepsize                                     */
    /* o8alpha = smallest feasible reduction factor for stepsize                  */
    /* o8delta = multiplier for derivative should be smaller than .25             */
    /* o8beta  = maximum feasible increase of o8x-norm for o8sig = 1                  */
    /* o8theta = bound for cos(angle(current direction, previous direction))      */
    /*         if overridden, stepsize larger than o8one is tried                 */

    /* ************************************************************************ */

    void    o8info(INTEGER icase);
    void    o8rest(void);
    void    o8save(void);
    void    o8eval(DOUBLE sigact,DOUBLE *sigres,LOGICAL *reject,LOGICAL *error);

    static INTEGER  i,l,j;
    static DOUBLE   term,maxphi;
    static DOUBLE   sigres ,diff;
    static LOGICAL  desc,descre,sminfe,lainfe,reject,error;
    

    o8unim_step[1] = o8p5;
    o8unim_step[2] = o8twom2;
    for (i = 3 ; i <= o8nstep ; i++) {
        o8unim_step[i] = o8tm1;
    }
    /* projection of o8d, rescaling and computing o8dirder has been done already */
    
    l         = 0;
    error     = FALSE;
    o8phi       = o8scf*o8fx+o8psi;
    o8sig       = sig1th;
    o8violis[0] = 0;
    if ( ! o8silent ) o8info(8);
    
    L100:

    l = l+1;
    if ( l > o8nstep ) {
        if ( error && ! o8silent ) o8msg(24);
        o8stptrm = -o8one;
        o8sig    = o8zero;
        
        return;
    }
    /* compute a new o8x and test for descent */
    
    o8eval(o8sig,&sigres,&reject,&error);
    
    if ( error ) {
        if ( o8sig > o8one ) {
        
            o8rest();
            
            goto L200;
            
        } else {
            if ( ! o8silent ) o8msg(25);
            o8sig = o8unim_step[l]*o8sig;
            
            goto L100;
        }
    }
    if ( reject ) {
        if ( o8sig > o8one ) {
        
            o8rest();
            
            goto L200;
            
        } else {
            o8sig = o8unim_step[l]*o8sig;
            
            goto L100;
        }
    }
    if ( ! o8silent ) o8info(9);
    
    /* new function value */
    
    if ( o8sig > o8one ) {
        if ( o8phi1 >= o8phimin ) {
        
            /* o8phi does'nt decrease further */

            o8rest();
            
            goto L200;
            
        } else {
            if ( o8sig < o8stmaxl ) {
            
                o8save();
                
                o8sig = min(o8stmaxl,o8sig+o8sig);
                
                goto L100;
                
            } else {
            
                goto L200;
            }
        }
    }
    if ( o8lastch >= o8itstep-3 || o8phase != 2 || o8singul ) {
    
        /* require monotonic behaviour */
    
        diff = o8phi-o8phi1;
    } else {
        maxphi = o8phi;
        for (j = 1 ; j <= 3 ; j++) {
            maxphi = max(o8scf*o8accinf[o8itstep-j][2]+o8accinf[o8itstep-j][4],maxphi);
        }
        diff = maxphi-o8phi1;
    }
    desc   = diff >= min(-o8sig*o8delta*o8dirder,o8level);
    descre = o8upsi - o8upsi1 >= o8sig*pow(o8delta,2)*o8upsi/o8tauqp;
    sminfe = o8upsi <= o8tau0*o8p5 && o8upsi1 <= o8tau0;
    lainfe = o8upsi > o8tau0*o8p5;
    
    if ( desc && ( sminfe || ( lainfe && descre ) ) ) {
    
        /* Goldstein-Armijo descent test satisfied */
        
        if ( o8sig == o8one && ( (o8cosphi >= o8theta && o8sig0 >= o8one 
             && (o8phase+1)*(o8phase-2) != 0
             && ! o8singul) || diff >= -o8sig*o8delta1*o8dirder )
             && o8stmaxl > o8one && o8upsi < o8tau0*o8p5 ) {
             
            /* 1 >= o8delta1 >> o8delta > 0 */
             
            /* try stepsize larger than o8one */
            /* save the current point as the best o8one */
            
            o8save();
            
            o8sig = min(o8stmaxl,o8sig+o8sig);
            
            goto L100;
        }
        if ( o8sig <= o8one && o8upsi > o8tau0*o8p5 && o8upsi1 > o8upsi ) goto L300;
        
        goto L200;
        
    } else {
    
        goto L300;
    }
    L200:

    /* accept new o8x, save old values */
    
    o8fx0    = o8fx;
    o8fx     = o8fx1;
    o8upsi0  = o8upsi;
    o8upsi   = o8upsi1;
    o8psi0   = o8psi;
    o8psi    = o8psi1;
    o8stptrm = o8one;
    o8sig0   = o8sig;
    for (i = 1 ; i <= o8n ; i++) {
        o8x0[i]   = o8x[i];
        o8d0[i]   = o8d[i];
        o8x[i]    = o8x1[i];
        o8difx[i] = o8x[i]-o8x0[i];
    }
    o8d0norm = o8dnorm;
    o8x0norm = o8xnorm;
    for (i = 1 ; i <= 2*o8nres ; i++) {
        o8res[i] = o8res1[i];
    }

    return;
    
    /* continue reducing o8sig */
    
    L300:

    if ( sigres < o8sig ) {
        o8sig = min(o8p5*o8sig,max(o8unim_step[l]*o8sig,sigres));
    } else {
        term = (diff-o8dirder*o8sig)*o8two;
        if ( term > o8epsmac*(o8scf*fabs(o8fx)+o8psi) ) {
            o8sig = min(o8p5*o8sig,max(o8unim_step[l]*o8sig,-o8dirder*pow(o8sig,2)/term));
        } else {
            o8sig = o8unim_step[l]*o8sig;
        }
    }

    
    if ( o8sig*max(o8one,o8dnorm) >= o8sigsm ) goto L100;
    o8stptrm = -o8one;
    o8sig    = o8zero;

    return;
}

/* **************************************************************************** */
/*                scalar product of o8two vectors or parts of vectors             */
/* **************************************************************************** */
DOUBLE o8sc1(INTEGER i,INTEGER j,DOUBLE o8a[],DOUBLE b[]) 
{

    /* multiply o8two vectors */
    
    static INTEGER  k;
    static DOUBLE   s;

    if ( i > j ) 
    {
        return (o8zero);
    } 
    else 
    {
        s = o8zero;
        for (k = i ; k <= j ; k++) 
        {
            s += o8a[k]*b[k];
        }
        return (s);
    }
}

/* **************************************************************************** */
/*                    multiply row j of matrix a with vector b                  */
/* **************************************************************************** */
DOUBLE o8sc2(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]) {
    
    static DOUBLE   s;
    static INTEGER  i;

    s = o8zero;
    for (i = o8n ; i <= m ; i++) 
    {
        s = s+o8a[j][i]*b[i];
    }
    return (s);
}

/* **************************************************************************** */
/*          multiply column j section (o8n to m) of matrix a with vector b        */
/* **************************************************************************** */
DOUBLE o8sc3(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]) 
{

    static DOUBLE   s;
    static INTEGER  i;

    s = o8zero;
    for (i = o8n ; i <= m ; i++) {
        s = s+o8a[i][j]*b[i];
    }
    return (s);
}
/*************************************************************************/
DOUBLE o8sc3b(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[])
{
 
    static DOUBLE   s;
    static INTEGER  i;
 
    s = o8zero;
    for (i = o8n ; i <= m ; i++) {
        s = s+o8a[i][j]*b[i];
    }
    return (s);
}

DOUBLE o8sc4(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a) 
{

    static DOUBLE   s;
    static INTEGER  i;

    s = o8zero;
    for (i = o8n ; i <= m ; i++) {
        s += pow(o8a[i][j],2);
    }
    return (s);
}

DOUBLE o8sc3_(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]) {

    static DOUBLE   s;
    static INTEGER  i;

    s = o8zero;
    for (i = o8n ; i <= m ; i++) {
        s = s+o8a[i][j]*b[i];
    }
    return (s);
}

/* **************************************************************************** */
/* subprogram for structured output of a submatrix a[ma][na]                    */
/* on channel lognum in fix or float format with heading "head".                */
/* uses a fixed format string with 70 print columns                             */
/* **************************************************************************** */
void o8mdru(DOUBLE **o8a,INTEGER ma,INTEGER na,char head[],
            FILE *lognum,LOGICAL fix) {

    static INTEGER  i,j,jo,ju;
    static INTEGER  anz;

    fprintf(lognum,"\n%40s\n", head);
    anz = 4;
    jo  = 0;
    while ( jo < na ) {
        ju = jo+1;
        jo = min(ju+anz-1,na);
        
        fprintf(lognum,"\nrow/column");
        for (j = ju ; j <= jo ; j++) {
            fprintf(lognum,"      %3i      ", j);
            if ( (j-ju+1) % 4 == 0 || j == jo ) fprintf(lognum,"\n");
        }
        for (i = 1 ; i <= ma ; i++) {
            if ( fix ) {
                fprintf(lognum,"   %4i   ",i);
                for (j = ju ; j <= jo ; j++) {
                    fprintf(lognum,"%14.7f ", o8a[i][j]);
                    if ( (j-ju+1) % 4 == 0 || j == jo ) fprintf(lognum,"\n");
                }
            } else {
                fprintf(lognum,"   %4i   ",i);
                for (j = ju ; j <= jo ; j++) {
                    fprintf(lognum," %13.6e ", o8a[i][j]);
                    if ( (j-ju+1) % 4 == 0 || j == jo ) fprintf(lognum,"\n");
                }
            }
        }
    }
    return;
}
void o8mdru_(DOUBLE **o8a,INTEGER ma,INTEGER na,char head[],
             FILE *lognum,LOGICAL fix) {

    static INTEGER  i,j,jo,ju;
    static INTEGER  anz;

    fprintf(lognum,"\n%40s\n", head);
    anz = 4;
    jo  = 0;
    while ( jo < na ) {
        ju = jo+1;
        jo = min(ju+anz-1,na);
        
        fprintf(lognum,"\nrow/column");
        for (j = ju ; j <= jo ; j++) {
            fprintf(lognum,"      %3i      ", j);
            if ( (j-ju+1) % 4 == 0 || j == jo ) fprintf(lognum,"\n");
        }
        for (i = 1 ; i <= ma ; i++) {
            if ( fix ) {
                fprintf(lognum,"   %4i   ",i);
                for (j = ju ; j <= jo ; j++) {
                    fprintf(lognum,"%14.7f ", o8a[i][j]);
                    if ( (j-ju+1) % 4 == 0 || j == jo ) fprintf(lognum,"\n");
                }
            } else {
                fprintf(lognum,"   %4i   ",i);
                for (j = ju ; j <= jo ; j++) {
                    fprintf(lognum," %13.6e ", o8a[i][j]);
                    if ( (j-ju+1) % 4 == 0 || j == jo ) fprintf(lognum,"\n");
                }
            }
        }
    }
    return;
}

/* **************************************************************************** */
/*                        compute gradient of lagrangian                        */
/* **************************************************************************** */
void o8egph(DOUBLE gphi[]) {

    static INTEGER  i,j,l,k;

    for (i = 1 ; i <= o8n ; i++) 
    {
         gphi[i] = o8gradf[i] * o8scf;
         for (j = 1 ; j <= o8aalist[0] ; j++) 
         {
             l = o8aalist[j];
             k = (l+1)/2;
             if( o8low[k] == o8up[k] )
             { 
               if ( k > o8n )
               {
                 gphi[i] -= o8gres[i][k-o8n]*o8u[l] ;
               }
               else
               {
                 gphi[i] -= o8xsc[k]*o8u[l]; 
                 /* since an equality constraint is always odd numbered */
               }
             }
             else
             {
                if( o8u[l] > o8zero ) 
                {
                  if ( k > o8n ) 
                  { 
                    gphi[i] = gphi[i]-o8gres[i][k-o8n]*o8gres[0][k-o8n]*o8u[l];
                  }
                  else
                  {
                    if ( l%2 == 0 )
                    {
                       gphi[i] += o8xsc[k]*o8u[l];
                    }
                    else
                    {
                       gphi[i] -= o8xsc[k]*o8u[l];
                    }
                  }
                }
             }

            /* include constraints, whose multipliers are of correct sign only */
         }
    }
    return;
}

/* detection and elimination of redundant equality constraints             */
/* this is done making them always inactive inequality constraints         */
/* by redefining o8low and o8up                                                */
void o8elim(void) {
    DOUBLE  o8sc3b 
      (INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]);
    DOUBLE  o8vecn(INTEGER nl,INTEGER nm,DOUBLE o8x[]);
    void o8msg(INTEGER num);
    static INTEGER  n1,n2;  
    static INTEGER  i,j,k,l,i1,icur,ipiv;
    static DOUBLE   sum,term,dalpha,dbeta,qrii;
    static DOUBLE   curle;
    
    o8optite = 0 ; /* indicates success */
    icur = 0 ;
    for ( i=1 ; i<= o8n+o8nlin ; i++ )
    {
      /* search for linear equality constraints */
      if ( o8low[i] == o8up[i] )
      {
        icur += 1;
        if ( i <= o8n )
        {
          for ( j = 1 ; j <= o8n ; j++ )
          {
            o8elim_qri[j] = o8zero ;
          }
          o8elim_qri[i] = o8xsc[i]  ;
        }
        else
        {   
          for ( j = 1 ; j <= o8n ; j++ )
          {
             o8elim_qri[j] = o8gres[j][i-o8n] ;
          }
        }  
        o8colle[icur] = o8vecn(1,o8n,o8elim_qri);
        o8colno[icur] = i; /* global constraint number */
        o8elim_col[icur] = icur ; /* local column number */   
        o8elim_rhsscal[icur] = o8one/o8colle[icur];
        for ( j = 1 ; j <= o8n ; j++ )
        {
           o8qr[j][icur] = o8elim_qri[j]/o8colle[icur] ;
        }
        o8colle[icur] = o8one;
        /* here o8colno points to the indices of the linear equality constraints */
      }
    }  
    if ( icur == 0 ) return ;
    /* there are no linear equality constraints */
    /* otherwise try to determine the o8rank of the current o8qr */
    n1  = 1 ;
    n2  = icur;
    o8rank = 0 ; 
    for ( i = n1 ; i <= n2 ; i++)
    {
       /* search for pivot column */
       ipiv  = i;
       curle = o8colle[i];
       for (j = i+1 ; j <= n2 ; j++)
       {
         if ( o8colle[j] > curle )
         {
                curle = o8colle[j];
                ipiv = j ;
         }
       }  
       /* interchange columns explicitly, if necessary   */
       if ( ipiv != i )
       {
          j           = o8elim_col[i];
          o8elim_col[i]    = o8elim_col[ipiv];
          o8elim_col[ipiv] = j;
          term        = o8colle[i];
          o8colle[i]    = o8colle[ipiv];
          o8colle[ipiv] = term;
          for (k = 1 ; k <= o8n ; k++)
          {
             term        = o8qr[k][i];
             o8qr[k][i]    = o8qr[k][ipiv];
             o8qr[k][ipiv] = term;
          }
       }   
       sum = o8zero;
       for (j = i ;  j <= o8n ;  j++)
       {
         term   = o8qr[j][i];
         o8elim_qri[j] = term;
         sum    = term*term+sum;
       }
        
       if ( sum <= pow(o8rho,2) )
       {
         /* set tiny values to o8zero */
         for (j = i ; j <= n2 ; j++)  
         {
           o8colle[j] = o8zero;
           for (k = i ; k <= o8n ; k++)
           {
              o8qr[k][j] = o8zero;
           }
         }  
         o8rank = i-1;
         break;
       }
       qrii   = o8elim_qri[i];
       dalpha = -sqrt(sum);
       if ( qrii < o8zero ) dalpha = -dalpha;
       dbeta    = o8one/(sum-qrii*dalpha);   
       o8diag[i]  = dalpha;
       o8betaq[i] = dbeta; 
       o8elim_qri[i]   = qrii-dalpha;
       o8qr[i][i] = o8elim_qri[i];
       o8rank     = i;
       i1       = i+1;
       for (j = i1 ; j <= n2 ; j++)
       {
         sum = dbeta*o8sc3b(i,o8n,j,o8qr,o8elim_qri);
         for (k = i ; k <= o8n ; k++)
         {
           o8qr[k][j] = o8qr[k][j]-sum*o8elim_qri[k];
         }
         o8colle[j] = o8colle[j]-pow(o8qr[i][j],2);
       }
    }   
    /* test */
    /* Rprintf("berechneter rang = %o8d \n",o8rank); */
    /* testende */
    if ( o8rank == n2 ) return ;
    if ( ! o8silent ) o8msg(28);
    /* the linear equality constraints are sufficiently independent*/
    /* otherwise, check first for compatibility of the (almost) dependent */
    /*  now
       R= [R11 , R12 ; O  O ]
       in matlab notation, where R11 is o8rank times o8rank
       R is stored in o8diag and above the diagonal in o8qr
    */
    /* constraints : solve transpose(R11)*y=P*b , */
    for ( i = 1 ; i <= o8rank ; i++ )
    {
       sum = o8low[o8colno[o8elim_col[i]]]*o8elim_rhsscal[o8elim_col[i]];
       /* the permuted right hand side, o8low[i]=o8up[i] */
       for ( j = i-1 ; j>=1 ; j-- )
       {
         sum -= o8qr[j][i]*o8elim_y[j] ;
       }
       o8elim_y[i] = sum/o8diag[i] ;
    }
    for ( i = o8rank +1 ; i <= n2 ; i++ )
    {
       sum  = o8low[o8colno[o8elim_col[i]]]*o8elim_rhsscal[o8elim_col[i]];
       for ( j = 1 ; j<= o8rank ; j++ )
       {
         term = o8qr[j][i]*o8elim_y[j] ;
         sum -= term ;
       }
       if ( fabs(sum) >= o8two*o8n*o8rho )
       {
         o8optite = 8 ;
         if ( ! o8silent ) o8msg(27);
         return ;
       }
    }   
        
    for ( i = o8rank +1 ; i <= n2 ; i++ )
    {
      o8low[o8colno[o8elim_col[i]]] = -o8big ;
      o8up[o8colno[o8elim_col[i]]]  =  o8big ;
    }
    /* inhibit use of dependent but compatible linear equality constraints */
    return;
}


/* **************************************************************************** */
/* O8QR-decomposition of matrix of gradients of binding constraints               */
/* this set may be expanded using multiple calls to o8dec.                      */
/* No exit on singular r-factor here. Information on                            */
/* the decompostion is stored in o8betaq and in and below the                     */
/* diagonal of o8qr. r-factor is stored in o8diag (diagonal ) and                   */
/* above the diagonal of o8qr. o8cscal is the column scaling of the                 */
/* original matrix. Column pivoting is done here and is stored in o8colno         */
/* **************************************************************************** */
void o8dec(INTEGER nlow,INTEGER nrl) {
    
    DOUBLE  o8sc3b (INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]);
    void    o8ht  (INTEGER id,INTEGER incr,INTEGER is1,INTEGER is2,INTEGER m,
                   DOUBLE **o8a,DOUBLE o8beta[],DOUBLE b[],DOUBLE c[]);    
    void    o8left(DOUBLE **o8a,DOUBLE b[],DOUBLE y[],DOUBLE *yl,INTEGER o8n);
    DOUBLE  o8vecn(INTEGER nl,INTEGER nm,DOUBLE o8x[]);
    
    static INTEGER  n1,n2;
    static INTEGER  i,j,k,l,i1,i2,ipiv;
    static DOUBLE   sum,term,dalpha,dbeta,qrii,dl,fac;
    static DOUBLE   curle;


    if ( nlow > nrl ) return;
    if ( nlow == 1 ) o8rank = 0;
    dl = o8one/(o8n+o8n+o8n);
    for ( i = nlow ; i <= nrl ; i++) {
        o8diag[i]  = o8zero;
        o8betaq[i] = o8zero;
        o8colno[i] = i;
        if ( o8aalist[i] <= 2*o8n )
        {
          /* o8a bound constraint */
          k = (o8aalist[i]+1)/2;
          /* component o8x[k]     */
          for ( j = 1 ; j <= o8n ; j++ )
          {
            o8dec_qri[j] = o8zero ;
          }
          o8dec_qri[k] = ( o8aalist[i] % 2 == 0 ) ? -o8xsc[k] : o8xsc[k] ;
        }
        else
        {
          k = (o8aalist[i]-2*o8n+1)/2 ;
          fac = o8gres[0][k];
          for (j = 1 ; j <= o8n ; j++) 
          {
            o8dec_qri[j] = o8gres[j][k]*fac;
          }
        }
        o8left(o8a,o8dec_qri,o8dec_qri0,&sum,o8n);
        
        if ( sum == o8zero ) {
            o8cscal[i] = o8one;
            o8colle[i] = o8zero;
            for (j = 1 ; j <= o8n ; j++) {
                o8qr[j][i] = o8zero;
            }
        } else {
            for (j = 1 ; j <= o8n ; j++) {
                o8dec_qri[j] = o8dec_qri0[o8perm[j]];
            }
            term     = o8one/sqrt(max(sum,pow(o8rho,2)));
            o8cscal[i] = term;
            if ( nlow > 1 ) {
            
                o8ht(1,0,1,o8rank,o8n,o8qr,o8betaq,o8dec_qri,o8dec_qri0);
                
                for (j = 1 ; j <= o8n ; j++) {
                    o8dec_qri[j] = o8dec_qri0[j];
                }
            }
            for (j = 1 ; j <= o8n ; j++) {
                o8qr[j][i] = o8dec_qri[j]*term;
            }
            /* o8colle : length of remaining column squared */
    
            o8colle[i] = pow(o8vecn(o8rank+1,o8n,o8dec_qri)*term,2);
        }
    }
    if ( nlow > 1 && o8rank < nlow-1 ) {
    
        /* shift o8zero block to the right */
        
        i1 = nlow-1-o8rank;
        i2 = nrl-nlow+1;
        for (i = 1 ; i <= min(i1,i2) ; i++) {
            ipiv        = o8rank+i;
            k           = nrl-i+1;
            term        = o8betaq[k];
            o8betaq[k]    = o8betaq[ipiv];
            o8betaq[ipiv] = term;
            j           = o8colno[k];
            o8colno[k]    = o8colno[ipiv];
            o8colno[ipiv] = j;
            term        = o8colle[k];
            o8colle[k]    = o8colle[ipiv];
            o8colle[ipiv] = term;
            for (j = 1 ; j <= o8n ; j++) {
                term        = o8qr[j][k];
                o8qr[j][k]    = o8qr[j][ipiv];
                o8qr[j][ipiv] = term;
             }
         }
    }
    if ( nlow > 1 ) {
        n1 = o8rank+1;
        n2 = n1+nrl-nlow;
    } else {
        n1 = nlow;
        n2 = nrl;
    }
    for ( i = n1 ; i <= n2 ; i++) {
    
        /* search for pivot column */
        
        ipiv  = i;
        curle = o8colle[i];
        for (j = i+1 ; j <= n2 ; j++) {
            if ( o8colle[j] > curle ) {
                curle = o8colle[j];
            }
        }
        for (j = n2 ; j >= i ; j--) {
            if ( o8colle[j] >= curle/o8three ) ipiv = j;
        }
        /* interchange columns explicitly, if necessary   */
        /* make interchanges continuous with respect to o8x */
        
        if ( ipiv != i ) {
            j           = o8colno[i];
            o8colno[i]    = o8colno[ipiv];
            o8colno[ipiv] = j;
            term        = o8colle[i];
            o8colle[i]    = o8colle[ipiv];
            o8colle[ipiv] = term;
            for (k = 1 ; k <= o8n ; k++) {
                term        = o8qr[k][i];
                o8qr[k][i]    = o8qr[k][ipiv];
                o8qr[k][ipiv] = term;
            }
        }
        sum = o8zero;
        for (j = i ;  j <= o8n ;  j++) {
            term   = o8qr[j][i];
            o8dec_qri[j] = term;
            sum    = term*term+sum;
        }
        if ( sum <= pow(o8rho,2) ) {
        
            /* set tiny values to o8zero */
            
            for (j = i ; j <= n2 ; j++) {
                o8colle[j] = o8zero;
                for (k = i ; k <= o8n ; k++) {
                    o8qr[k][j] = o8zero;
                }
            }
            o8rank = i-1;
            
            return;
        }
        qrii   = o8dec_qri[i];
        dalpha = -sqrt(sum);
        if ( fabs(qrii) <= -dalpha*dl ) {
            term = o8zero;
            for (j = i+1 ; j <= o8n ; j++) {
                if ( fabs(o8dec_qri[j]) > term ) {
                    term = fabs(o8dec_qri[j]);
                    l    = j;
                }
            }
            k        = o8perm1[i];
            o8perm1[i] = o8perm1[l];
            o8perm1[l] = k;
        }
        if ( qrii < o8zero ) dalpha = -dalpha;
        dbeta    = o8one/(sum-qrii*dalpha);
        o8diag[i]  = dalpha;
        o8betaq[i] = dbeta;
        o8dec_qri[i]   = qrii-dalpha;
        o8qr[i][i] = o8dec_qri[i];
        o8rank     = i;
        i1       = i+1;
        for (j = i1 ; j <= n2 ; j++) {
            sum = dbeta*o8sc3b(i,o8n,j,o8qr,o8dec_qri);
            for (k = i ; k <= o8n ; k++) {
                o8qr[k][j] = o8qr[k][j]-sum*o8dec_qri[k];
            }
            o8colle[j] = o8colle[j]-pow(o8qr[i][j],2);
        }
    }
    return;
}

/* **************************************************************************** */
/*                  application of Householder transformations                  */
/* **************************************************************************** */
void o8ht(INTEGER id,INTEGER incr,INTEGER is1,INTEGER is2,INTEGER m,
          DOUBLE **o8a,DOUBLE o8beta[],DOUBLE b[],DOUBLE c[]) {

    /* application of Householder transformations stored     */
    /* in the lower or strict lower (if incr = 0 or 1 resp.) */
    /* triangle of a and in the vector o8beta on b giving c.   */
    /* only columns is1 to is2 are used in forward manner    */
    /* if id > 0,backwards otherwise.                        */
    /* rows is1 to m of c are changed only                   */
    /* a stands here for the actual o8qr                       */
    DOUBLE o8sc3b(INTEGER o8n,INTEGER m,INTEGER j,DOUBLE **o8a,DOUBLE b[]);

    static DOUBLE   sum;
    static INTEGER  i,j,k,it;

    for (i = 1 ; i <= m ; i++) {
        c[i] = b[i];
    }
    if(is1 > m)   return;
    if(is2 < is1) return;
    for (i = is1 ; i <= is2 ; i++) {
        it = i;
        if(id < 0) it = is2-it+is1;
        
        /* it = index of transformation */
        
        j   = it+incr;
        sum = o8beta[it]*o8sc3b(j,m,it,o8a,c);
        for (k = j ; k <= m ; k++) {
            c[k] = c[k]-sum*o8a[k][it];
        }
    }
    return;
}
    
/* **************************************************************************** */
/* solve triangular system r*o8x = b, r defined by Householder-O8QR-                */
/* decomposition decomp (with column scaling)                                   */
/* **************************************************************************** */
void o8sol(INTEGER nlow,INTEGER nup,DOUBLE b[],DOUBLE o8x[]) {
    
    static DOUBLE   sum;
    static INTEGER  i,j;

    for (i = nup ; i >= nlow ; i--) {
        sum = o8zero;
        for (j = i+1 ; j <= nup ; j++) {
            sum = sum+o8qr[i][j]*o8sol_xl[j];
        }
        o8sol_xl[i] = (b[i]-sum)/o8diag[i];
    }
    for (i = nlow ; i <= nup ; i++) {
        o8x[i] = o8sol_xl[i]*o8cscal[o8colno[i]];
    }
    /* there must follow interchange of o8x as given by o8colno */
    /* e.g. xx(o8colno[i]) = o8x[i]                             */

    return;
}

/* **************************************************************************** */
/* solve triangular system r(transpose)*o8x = b, r defined by                     */
/* Householder-O8QR-decomposition decomp (with column scaling]                    */
/* **************************************************************************** */
void o8solt(INTEGER nlow,INTEGER nup,DOUBLE b[],DOUBLE o8x[]) {

    static INTEGER  i,j;
    static DOUBLE   sum;

    for (i = nlow ; i <= nup ; i++) {
    
        /* b has been permuted already ! */
        
        o8x[i] = b[i]*o8cscal[o8colno[i]];
    }
    for (i = nlow ; i <= nup ; i++) {
        sum = o8zero;
        for (j = nlow ; j <= i-1 ; j++) {
            sum = sum+o8qr[j][i]*o8x[j];
        }
        o8x[i] = (o8x[i]-sum)/o8diag[i];
    }
    return;
}
    
/* **************************************************************************** */
/* lenght of vector (a,b). numerically stable version with                      */
/* overflow / underflow saveguard                                               */
/* **************************************************************************** */
DOUBLE o8dsq1(DOUBLE o8a,DOUBLE b) {

    static DOUBLE   a1,b1;
    static DOUBLE   o8res;

    a1 = fabs(o8a);
    b1 = fabs(b);
    if ( a1 > b1 ) {
        o8res = a1*sqrt(o8one+pow(b1/a1,2));
    } else {
        if ( b1 > a1 ) {
            o8res = b1*sqrt(o8one+pow(a1/b1,2));
        } else {
            o8res = a1*sqrt(o8two);
        }
    }
    return (o8res);
}

/* **************************************************************************** */
/*                  computes the upper triangular Cholesky-factor               */
/* **************************************************************************** */
void o8upd(DOUBLE **o8r,DOUBLE z[],DOUBLE y[],INTEGER o8n,LOGICAL *fail) {
    
    DOUBLE  o8dsq1(DOUBLE o8a,DOUBLE b);
    void    o8left(DOUBLE **o8a,DOUBLE b[],DOUBLE y[],DOUBLE *yl,INTEGER o8n);
    static INTEGER  i,j,i1;
    static DOUBLE   yl,zl,wl,wn1,ai,bi,h;

    /* o8upd computes the upper triangular Cholesky-factor               */
    /* r1 of r(transpose)*r+z*z(transpose)-y*y(transpose)                */
    /* and restores it in r. The strict lower triangle of r              */
    /* remains unchanged.                                                */
    /* fail = TRUE if the decomposition does'nt exist, stop on dimension */
    /* error, fail = false on success.                                   */
    
    *fail = FALSE;
    
    /* save subdiagonal */
    
    for (i = 1 ; i <= o8n-1 ; i++) {
        o8upd_sdiag[i]  = o8r[i+1][i];
        o8r[i+1][i] = o8zero;
    }
    /* step o8one: include z*z(transpose) */
    
    zl = o8zero;
    for (i = 1 ; i <= o8n ; i++) {
        zl = zl + pow(z[i],2);
    }
    if ( zl != o8zero ) {
    
        /* solve o8w(transpose)*r = z(transpose) */
        
        o8left(o8r,z,o8upd_w,&wl,o8n);
        
        wl = sqrt(wl+o8one);
        
        /* o8u[2]*o8u[3]*...*o8u[o8n]*o8w = ( norm(o8w),0,..,0)(transpose) */
        /* o8u[i] rotations                                      */
        
        for (i = o8n ; i >= 2 ; i--) {
            if ( o8upd_w[i] != o8zero ) {
                i1        = i-1;
                ai        = o8upd_w[i1];
                bi        = o8upd_w[i];
                o8upd_w[i1]     = o8dsq1(ai,bi);
                ai        = ai/o8upd_w[i1];
                bi        = -bi/o8upd_w[i1];
                o8r[i][i1]  = bi*o8r[i1][i1];
                o8r[i1][i1] = ai*o8r[i1][i1];
                for (j = i ; j <= o8n ; j++) {
                    h        = ai*o8r[i1][j] - bi*o8r[i][j];
                    o8r[i][j]  = bi*o8r[i1][j] + ai*o8r[i][j];
                    o8r[i1][j] = h;
                }
            }
        }
        /* o8r = o8d*o8r, o8d = o8diag(wl,1,...,1), o8r now Hessenberg */
        
        for (i = 1 ; i <= o8n ; i++) {
            o8r[1][i] = o8r[1][i]*wl;
        }
        /* o8r = o8u[o8n-1]*...*o8u[1]*o8r now upper triangular, */
        /* o8u[i] givens-rotations                       */
        
        for (i = 1 ; i <= o8n-1 ; i++) {
            i1 = i+1;
            ai = o8r[i][i];
            bi = -o8r[i1][i];
            h  = o8dsq1(ai,bi);
            if ( h != o8zero ) {
                ai       = ai/h;
                bi       = bi/h;
                o8r[i][i]  = h;
                o8r[i1][i] = o8zero;
                for (j = i+1 ; j <= o8n ; j++) {
                    h        = ai*o8r[i][j] - bi*o8r[i1][j];
                    o8r[i1][j] = bi*o8r[i][j] + ai*o8r[i1][j];
                    o8r[i][j]  = h;
                }
            }
        }
    }
    /* step o8two : include -y*y(transpose) */
    
    yl = o8zero;
    for (i = 1 ; i <= o8n ; i++) {
        yl = yl + pow(y[i],2);
    }
    if ( yl != o8zero ) {
    
        o8left(o8r,y,o8upd_w,&wl,o8n);
        
        if ( wl >= o8one ) {
            *fail = TRUE;
        } else {
            wl  = sqrt(fabs(o8one-wl));
            wn1 = wl;
            
            /* ************************************************ */
            /*   ( o8r(new), 0 )                  (    o8r  , o8w )   */
            /*   (-----------)  =  o8u[1]*...o8u[o8n]*(-----------)   */
            /*   (y(transp),1)                  ((0,..,0),wl)   */
            /* ************************************************ */
            
            for (i = o8n ; i >= 1 ; i--) {
                ai  = wn1;
                bi  = o8upd_w[i];
                wn1 = o8dsq1(ai,bi);
                if ( wn1 != o8zero ) {
                    ai      = ai/wn1;
                    bi      = bi/wn1;
                    o8upd_rn1[i]  = bi*o8r[i][i];
                    o8r[i][i] = ai*o8r[i][i];
                    for (j = i+1 ; j <= o8n ; j++) {
                        h       = ai*o8r[i][j] - bi*o8upd_rn1[j];
                        o8upd_rn1[j]  = bi*o8r[i][j] + ai*o8upd_rn1[j];
                        o8r[i][j] = h;
                    }
                }
            }
        }
    }
    /* restore subdiagonal */
    
    for ( i = 1 ;  i <= o8n-1 ;  i++) {
        o8r[i+1][i] = o8upd_sdiag[i];
    }
    return;
}

/* **************************************************************************** */
/*                                    o8rght                                    */
/* **************************************************************************** */
void o8rght(DOUBLE **o8a,DOUBLE b[],DOUBLE y[],DOUBLE *yl,INTEGER o8n) {
    
    static INTEGER  i,j;
    static DOUBLE   h;

    /* o8rght assumes that the Cholesky-factor of o8a         */
    /* o8a = o8r(transpose)*o8r is stored in the upper half of o8a. */
    /* b is a right hand side. o8rght solves                */
    /* o8r*y = b                                              */
    /* yl = pow(norm(y),2)                                  */
    
    *yl = o8zero;
    for (i = o8n ; i >= 1 ; i--) {
        h = b[i];
        for (j = i+1 ; j <= o8n ; j++) {
            h = h - o8a[i][j]*y[j];
        }
        h    = h/o8a[i][i];
        y[i] = h;
        *yl  = pow(h,2) + *yl;
    }
    return;
}

/* **************************************************************************** */
/*                                  o8left                                      */
/* **************************************************************************** */
void o8left(DOUBLE **o8a,DOUBLE b[],DOUBLE y[],DOUBLE *yl,INTEGER o8n) {
    
    static INTEGER  i,j;
    static DOUBLE   h;

    /* o8left assumes that the Cholesky-factor of o8a         */
    /* o8a = o8r(transpose)*o8r is stored in the upper half of o8a. */
    /* b is a right hand side. o8left solves                */
    /* o8r(transpose)*y = b                                   */
    /* yl = pow(norm(y),2)                                  */
    
    *yl = o8zero;
    for (i = 1 ; i <= o8n ; i++) {
        h = b[i];
        for (j = 1 ; j <= i-1 ; j++) {
            h = h - o8a[j][i]*y[j];
        }
        h    = h/o8a[i][i];
        y[i] = h;
        *yl  = pow(h,2) + *yl;
    }
    return;
}

/* **************************************************************************** */
/*                      euclidean norm of o8x , avoid overflow                    */
/* **************************************************************************** */
DOUBLE o8vecn(INTEGER nl,INTEGER nm,DOUBLE o8x[]) {

    static INTEGER  i;
    static DOUBLE   xm,h;

    if ( nm < nl ) {
    
        return (o8zero);
    }
    xm = fabs(o8x[nl]);
    for (i = nl+1 ; i <= nm ; i++) {
        xm = max(xm,fabs(o8x[i]));
    }
    if ( xm == o8zero ) {
    
        return (o8zero);
        
    } else {
        h = o8zero;
        for (i = nl ; i <= nm ; i++) {
            h = h+pow(o8x[i]/xm,2);
        }
        return (xm*sqrt(h));
    }
}
    
/* ************************************************************************* */
/*  solution of extended quadratic program                                   */
/*                                                                           */
/*  o8scf*o8gradf(o8x)*dir+(1/2)*dir*o8a*dir+summe(o8tauqp*sl[i]+( my/2)*pow(sl[i],2) )*/
/*  minimal subject to                                                       */
/*  sl[i] >= 0, i = 1,...,o8nr  (the slacks)                                   */
/*  (o8gres[.][j]*dir+o8res[j])+vz*sl[j] = 0, j = 1,...,o8nh vz = -sign(o8res[j])    */
/*  (o8gres[.][o8aalist[j]])*dir+o8res[o8aalist[j]])+sl[j] >= 0, j = o8nh+1,....,o8nr    */
/*  the weight o8tauqp is adapted during solution                              */
/*  the quasi-Newton-matrix a is taken from o8comm.h                         */
/*  a is regularized if not sufficiently well conditioned                    */
/*  the resulting dir=xd[1+o8nr],...,xd[o8n+o8nr] is a direction of descent for    */
/*  the Zangwill function of the corresponding nonlinear                     */
/*  optimization problem                                                     */
/*  f(o8x) = min, o8res[j] = 0, j = 1,..o8nh, o8res[j] >= 0 , j = o8nh+1,o8nres          */
/*  at the currrent point o8x if the weight o8tauqp is chosen appropriately      */
/*  the quadratic programming problem is solved using the method             */
/*  of Goldfarb and Idnani                                                   */
/*  variables are stored in xd (solution) and o8ud (multipliers)               */
/*  in the following order xd = ( sl[j], j = 1,o8nr; dir = direct. of desc.)   */
/*  o8ud = ( multipliers for sl[i] >= 0 , i = 1,..,o8nr ;                        */
/*  multipliers for the equality constraints ,                               */
/*  multipliers for the general inequality constraints )                     */
/* ***************************************************************************/
void o8qpdu(void) 
{
    
    void    o8info(INTEGER icase);
    void    o8msg (INTEGER num);
    void    o8dird(void);
    void    o8cutd(void);
    DOUBLE  o8sc1 (INTEGER i,INTEGER j,DOUBLE o8a[],DOUBLE b[]);
    DOUBLE  o8vecn(INTEGER nl,INTEGER nm,DOUBLE o8x[]);
    void    o8zup (DOUBLE z[]);
    void    o8rup (DOUBLE rv[]);
    void    o8dlcd(INTEGER ai[],INTEGER l);
    void    o8adcd(void);
    void    o8rinv(INTEGER o8n,DOUBLE **o8a,INTEGER o8ndual,DOUBLE **o8x);
    
    static DOUBLE   infe1,s1,s2,tiny,
                    my,zz,ss,su,t,t1,t2,f,fmax,psid,c1,c2,cdiag,term,
                    su1,su2,condr,infiny,term1,term2,
                    diff0;
    static INTEGER  i,j,k,ip,l,incr,nosucc;
    static LOGICAL  wlow;

    o8ndual = o8n+o8nr;  /* o8nr=o8aalist[0] */
    
    /* number of equality constraints in QP-problem */
    incr = o8nr ;
    o8mi = 0 ;
    o8me = 0 ; 
    /* compute the index lists of the QP's equality constraints 
       eqlist and inequality constraints iqlist from o8aalist and
       o8low and o8up */
   for ( i = 1; i <= o8aalist[0] ; i++ )
   {
     l = (o8aalist[i]+1)/2;
     if ( o8low[l] == o8up[l] )
     {
       o8me += 1 ;
       o8qpdu_eqlist[o8me] = o8aalist[i] ;
     }
     else
     {
       o8mi += 1 ;
       o8qpdu_iqlist[o8mi+incr] = o8aalist[i]+o8nr ;
     }
   }
   for ( i = 1 ; i <= o8nr ; i++ ) 
   {
       o8qpdu_iqlist[i] = i ;   /* the slacks of the QP model */
   }
   o8mi += o8nr ;

   /* QP inequality constraints = active constraints - */
   /* equality-constraints + o8slack's                   */
    
   infiny = o8epsmac/o8tolmac;
   if ( o8analyt ) 
   {
        tiny = 2*o8nr*o8epsmac*o8tp3;
   } 
   else 
   {
       tiny = 2*o8nr*max(o8epsdif,o8epsmac*o8tp3);
   }

   o8qpterm = 0;
   for (i = 1 ; i <= o8nr ; i++) 
   {
    
        /* check gradients of active constraints against o8zero */
        o8qpdu_mult [i] = o8one ;   
        /* for a o8zero gradient the o8slack cannot be diminished:*/ 
        /*   set mult[i] = 0                                  */
        l = o8aalist[i] ;
        if ( l > 2*o8n ) 
        {  /* not a bound */
           for (j = 1 ; j <= o8n ; j++) 
           {
             o8qpdu_y[j] = o8gres[j][(l-2*o8n+1)/2];
           }
           if ( o8vecn(1,o8n,o8qpdu_y) == o8zero ) 
           {
             o8qpdu_mult[i] = o8zero;
             if ( ! o8silent ) o8msg(8);
           }
        }
   }  
   /* restart point in case of increase of o8tauqp */
    
    L10:
    nosucc = 0 ; 
    /* initialize matrices j and o8r */
    
    for (i = 1 ; i <= o8ndual ; i++) 
    {
        o8ddual[i] = o8zero;
        for (j = 1 ; j <= o8ndual ; j++) 
        {
            o8r[j][i]  = o8zero;
            o8xj[j][i] = o8zero;
        }
    }
    o8rnorm = o8one;
    o8rlow  = o8one;
    term1 = o8zero;
    for (i = 1 ; i <= 2*o8nres ; i++) 
    {
        o8u[i] = o8zero;
        if ( o8w[i] > term1 ) term1 = o8w[i];
    }
    o8accinf[o8itstep][19] = o8clow;
    o8accinf[o8itstep][20] = term1;
    o8accinf[o8itstep][31] = o8tauqp;
    for (i = 1 ; i <= 2*o8nr ; i++) 
    {   /* the multipliers in the QP */
        o8ud[i] = o8zero;
    }
    c1 = fabs(o8a[1][1]);
    for (i = 1 ; i <= o8n ; i++) 
    {
        c1 = max(c1,fabs(o8a[i][i]));
    }
    c1 = c1*o8tp1;
    
    /* we require much more regularity of o8a in the singular case */
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        if ( fabs(o8a[i][i]) < sqrt(o8rho1)*c1 ) o8a[i][i] = sqrt(o8rho1)*c1;
    }
    /* invert the Cholesky-factor and store in 
       the right lower block of dimension o8n of o8xj ( Idnanis j-matrix) */
    
    o8rinv(o8n,o8a,o8ndual,o8xj);
    
    c1   = fabs(o8a[1][1]);
    incr = o8nr;
    c2   = fabs(o8xj[1+incr][1+incr]);
    for (i = 1 ; i <= o8n ; i++) 
    {
        c1 = max(c1,fabs(o8a[i][i]));
        c2 = max(c2,fabs(o8xj[i+incr][i+incr]));
    }
    my = o8zero;
    for (i = 1 ; i <= o8n ; i++) 
    {
        for (j = i ; j <= o8n ; j++) 
        {
             my = my+pow(o8a[i][j],2);
        }
    }
    my    = my/o8n;
    cdiag = o8one/sqrt(my);
    for ( i = 1 ; i <= incr ; i++) 
    {
        o8xj[i][i] = cdiag;   
        /* the quadratic part for the slacks       */
        /* has same order of magnitude as the part */
        /* coming from the original Lagrangian     */ 
    }
    for (i = 1 ; i <= o8ndual ; i++) 
    {
        if ( i >  incr ) o8qpdu_g0[i] = o8gradf[i-incr]*o8scf;
        if ( i <= incr ) o8qpdu_g0[i] = o8tauqp; 
        /* the linear part for the slacks */
    }
    /* compute unconstrained solution */
    /* the Cholesky-factor of o8a is stored in the upper triangle */
    
    for (i = 1 ; i <= o8n ; i++) 
    {
        su = o8zero;
        for (j = 1 ; j <= i-1 ; j++) 
        {
            su = su+o8a[j][i]*o8qpdu_y[j+incr];
        }
        o8qpdu_y[i+incr] = (o8qpdu_g0[i+incr]-su)/o8a[i][i];
    }
    for (i = o8n ; i >= 1 ; i--) 
    {
        su = o8zero;
        for (j = i+1 ; j <= o8n ; j++) 
        {
            su = su+o8a[i][j]*o8qpdu_y[j+incr];
        }
        o8qpdu_y[i+incr] = (o8qpdu_y[i+incr]-su)/o8a[i][i];
    }
    for (i = 1 ; i <= incr ; i++) 
    {
    
        /* initially assume the slacks being o8zero */
        
        o8qpdu_y[i] = o8zero;
    }
    for (i = 1 ; i <= o8ndual ; i++) 
    {
        o8qpdu_xd[i] = -o8qpdu_y[i];
    }
    /* unconstrained minimizer of the QP: slacks come first */
    
    f = o8p5*o8sc1(1,o8ndual,o8qpdu_g0,o8qpdu_xd);
    fmax = f ;
    /* define the initial working set: all slacks are at their */
    /* lower bounds. o8iq = dimension of the QP's working set    */
    
    o8iq = o8nr;

    for (i = 1 ; i <= o8iq ; i++) 
    {
        o8qpdu_ai[i]   = i;
        o8r[i][i] = o8one;
        o8ud[i]   = o8tauqp;
    }
    /* slacks are at o8zero, multipliers for the slacks at o8tauqp */
    o8rnorm = o8one;
    o8rlow  = o8one;
    
    /* introduction of equality constraints                       */
    /* these are characterized by a negative index in ai and never*/ 
    /* considered for inactivation. they also play no role for the*/ 
    /* dual step size t1                                          */ 
 
    
    for (i = 1 ; i <= o8me ; i++) 
    {
        for ( j = 1 ; j <= o8iq ; j ++) 
        {
            o8ud1[j] = o8ud[j];
        }
        L20:

        o8ud1[o8iq+1] = o8zero;
        
        /* an equality constraint is indicated by the negative index */
    
        o8qpdu_ai[o8iq+1] = -o8qpdu_eqlist[i] ;
        l = (o8qpdu_eqlist[i]+1)/2 ;
        /* store the gradient of this constraint in o8np */
        for (j = 1 ; j <= o8n ; j++) 
        {
          if ( l > o8n )
          {  /* sign is 1 for equality constraints */       
            o8qpdu_cei[j+incr] = o8gres[j][l-o8n];
          }
          else
          {
            o8qpdu_cei[j+incr] = o8zero ;
          }
        }
        for (j = 1 ; j <= incr ; j++) 
        {
            o8qpdu_cei[j] = o8zero;
        }
        o8qpdu_cei[i] = o8one;
        if ( o8res[o8qpdu_eqlist[i]] > o8zero ) o8qpdu_cei[i] = -o8one;
        if ( l <= o8n ) 
        {
          o8qpdu_cei[l+incr] = o8xsc[l] ;   /* a fixed primal variable */
        }
        for (j = 1 ; j <= o8ndual ; j++) 
        {
            o8np[j] = o8qpdu_cei[j];
        }

        o8zup(o8qpdu_z);
        
        if ( o8iq != 0 ) o8rup(o8qpdu_vr);
        
        /* |z| = 0? */
        
        zz   = o8vecn(1,o8ndual,o8qpdu_z);
        term = o8sc1(1,o8ndual,o8qpdu_z,o8np);
        
        if ( zz >= tiny*o8rnorm && term > o8zero ) 
        {
            t2 = (-o8sc1(1,o8ndual,o8np,o8qpdu_xd)-o8res[o8qpdu_eqlist[i]])/term;
        } 
        else if ( (-o8sc1(1,o8ndual,o8np,o8qpdu_xd)-o8res[o8qpdu_eqlist[i]]) >= o8zero ) 
        {
            t2 = infiny;
        } 
        else 
        {
            t2 = -infiny;
        }
        /* for an equality constraint t2 may be positive or negative*/
        
        if ( o8iq != 0 ) o8rup(o8qpdu_vr);
        l = 0;
        if ( t2 > o8zero ) 
        {
            t1 = infiny;
            for (k = 1 ; k <= o8iq ; k++) 
            {
                if( o8qpdu_vr[k] > o8zero && o8qpdu_ai[k] > 0 ) 
                {
                    if( o8ud1[k]/o8qpdu_vr[k] < t1 ) 
                    {
                        t1 = o8ud1[k]/o8qpdu_vr[k];
                    }
                }
            }
            t = min(t1,t2);
        } 
        else 
        {
            t1 = infiny;
            for ( k = 1 ; k <= o8iq ; k++) 
            {
                if( o8qpdu_vr[k] < o8zero && o8qpdu_ai[k] > 0 ) 
                {
                    if( o8ud1[k]/fabs(o8qpdu_vr[k]) < t1 ) 
                    {
                        t1 = o8ud1[k]/fabs(o8qpdu_vr[k]);
                    }
                }
            }
            t1 = -t1;
            t  = max(t1,t2);
            
            /* t now negative */
        }
        /* add constraint , otherwise we must first delete some */
        /* inequality constraint with o8zero multiplier           */
        /* first delete then add!                               */
        
        if ( fabs(t)  >= infiny ) 
        {
          o8qpterm = -2 ;
          if ( ! o8silent ) o8msg(20);
          goto L2000;
        }        
        if ( fabs(t2) >= infiny ) 
        {
        
            /* purely dual step */
            
            for (k = 1 ; k <= o8iq ; k++) 
            {
                o8ud1[k] = o8ud1[k]+t*(-o8qpdu_vr[k]);
                if ( o8ud1[k] < o8zero && o8qpdu_ai[k] > 0 ) o8ud1[k] = o8zero;
                /*  o8ud1[k] < 0 is a roundoff effect */
            }
            o8ud1[o8iq+1] = o8ud1[o8iq+1]+t;
            o8qpdu_qpdel[0]  = 0;
            for (j = 1 ; j <= o8iq ; j++) 
            {
                if ( o8ud1[j] <= tiny && o8qpdu_ai[j] > 0 ) 
                {
                    o8qpdu_qpdel[0]        = o8qpdu_qpdel[0]+1;
                    o8qpdu_qpdel[o8qpdu_qpdel[0]] = o8qpdu_ai[j];
                }
            }
            for (k = 1 ; k <= o8qpdu_qpdel[0] ; k++) 
            {
                l      = o8qpdu_qpdel[k];
                o8qpdu_iai[l] = l;
                
                o8dlcd(o8qpdu_ai,l);
            }
            goto L20;
        }
        /* primal and dual step */
        for (k = 1 ; k <= o8ndual ; k++) 
        {
            o8qpdu_xd[k] = o8qpdu_xd[k]+t*o8qpdu_z[k];
        }
        for (k = 1 ; k <= o8iq ; k++) 
        {
            o8ud1[k] = o8ud1[k]+t*(-o8qpdu_vr[k]);
            if ( o8ud1[k] < o8zero && o8qpdu_ai[k] > 0 ) o8ud1[k] = o8zero;
        }
        o8ud1[o8iq+1] = o8ud1[o8iq+1]+t;
        
        f = f+t*o8sc1(1,o8ndual,o8qpdu_z,o8np)*(o8p5*t+o8ud1[o8iq+1]);
        if ( f <= fmax*(o8one+o8epsmac)+o8epsmac )
        {
          nosucc += 1 ;
          if ( nosucc > o8qpitma ) 
          {
            o8qpterm = - 3; 
            if ( ! o8silent ) o8msg(26);
            goto L2000;
          }
        }
        else
        {
          nosucc = 0 ; 
        }
        fmax = max( f , fmax ) ;

        if ( fabs(t2-t1) <= tiny ) 
        {
            o8qpdu_qpdel[0] = 0;
            for (j = 1 ; j <= o8iq ; j++) 
            {
                if ( o8ud1[j] <= tiny && o8qpdu_ai[j] > 0 ) 
                {
                    o8qpdu_qpdel[0]        = o8qpdu_qpdel[0]+1;
                    o8qpdu_qpdel[o8qpdu_qpdel[0]] = o8qpdu_ai[j];
                }
            }
            for (k = 1 ; k <= o8qpdu_qpdel[0] ; k++) 
            {
                l      = o8qpdu_qpdel[k];
                o8qpdu_iai[l] = l;
                
                o8dlcd(o8qpdu_ai,l);
            }
            o8qpdu_ai[o8iq+1] = -o8qpdu_eqlist[i] ;
            
            o8adcd();
            
        } 
        else if ( t == t2 ) 
        {
            o8qpdu_ai[o8iq+1] = -o8qpdu_eqlist[i] ;
            
            o8adcd();
            
        } 
        else 
        {
            o8qpdu_qpdel[0] = 0;
            for (j = 1 ; j <= o8iq ; j++) 
            {
                if ( o8ud1[j] <= tiny && o8qpdu_ai[j] > 0 ) 
                {
                    o8qpdu_qpdel[0]        = o8qpdu_qpdel[0]+1;
                    o8qpdu_qpdel[o8qpdu_qpdel[0]] = o8qpdu_ai[j];
                }
            }
            for (k = 1 ; k <= o8qpdu_qpdel[0] ; k++) 
            {
                l      = o8qpdu_qpdel[k];
                o8qpdu_iai[l] = l;
                o8dlcd(o8qpdu_ai,l);
            }
            goto L20;
        }
        for (j = 1 ; j <= o8iq ; j++) 
        {
            o8ud[j] = o8ud1[j];
        }  
    /* end of loop over equality constraints */
    }

    /* set iai = k\ai */
    
    for (i = 1 ; i <= o8mi ; i++) 
    {
        o8qpdu_iai[i] = i;
    }
    /* step 1 */
    
    L50:

    /* ai = QP - working set , iai[i] = 0 if i in ai */
    
    for (i = 1 ; i <= o8iq ; i++) 
    {
        ip = o8qpdu_ai[i];
        if ( ip > 0 ) o8qpdu_iai[ip] = 0;
    }
    /* s[o8qpdu_xd] = ci(trans)*o8qpdu_xd+ci0 >= 0 ? */
    
    psid = o8zero;
    
    /* psid : the measure of infeasibility */
    
    for (i = 1 ; i <= o8mi ; i++) 
    {
    
        /* iaexcl: if = 0, exclude from addition in this cycle */
        
        o8qpdu_iaexcl[i] = 1;
        su        = o8zero;
        k         =  0 ;
        /* numbers of inequality constraints:                           */
        /* i = 1,...,o8nr corresponds to the constraints v >= 0, o8u_a >= 0 */
        /* i = o8nr+1,....,o8mi to the regularized general inequalities     */
        
        if ( i > o8nr ) 
        {
            /* an original inequality constraint */
            k = (o8qpdu_iqlist[i]-o8nr+1)/2 ;
            if ( k > o8n )
            {  /* not a bound, gradient stored in o8gres, sign is in o8gres[0][] */
              for (j = 1 ; j <= o8n ; j++) 
              {
                o8qpdu_cii[j+incr] = o8gres[j][k-o8n]*o8gres[0][k-o8n];
              }
              for (j = 1 ; j <= incr ; j++) 
              {
                o8qpdu_cii[j] = o8zero;
              }
              o8qpdu_cii[o8me+i-incr] = o8one;
              o8qpdu_ci0[i]         = o8res[o8qpdu_iqlist[i]-o8nr];
            }
            else
            {
              for ( j = 1 ; j <= o8ndual ; j++ )
              {
                o8qpdu_cii[j] = o8zero ;
              }
              o8qpdu_cii[o8me+i-incr] = o8one ;
              if ( (o8qpdu_iqlist[i]-o8nr) % 2 == 0 ) 
              {
                o8qpdu_cii[k+incr] = -o8xsc[k];
                o8qpdu_ci0[i] = o8res[o8qpdu_iqlist[i]-o8nr];
              }
              else
              {
                o8qpdu_cii[k+incr] = o8xsc[k];
                o8qpdu_ci0[i]      = o8res[o8qpdu_iqlist[i]-o8nr];
              }
            }
          
        }
        else
        { /* a QP o8slack variable */
          for (j = 1 ; j <= o8ndual ; j++) 
          {
             o8qpdu_cii[j] = o8zero;
          }
          o8qpdu_ci0[i] = o8zero;
          o8qpdu_cii[i] = o8one;
        }
/* end computation of data of o8mi-th QP constraint data */

        su   = o8sc1(1,o8ndual,o8qpdu_cii,o8qpdu_xd)+o8qpdu_ci0[i];
        o8qpdu_s[i] = su;
        psid = psid+min(o8zero,su);
    }

    for (i = 1 ; i <= o8iq ; i++) {
        o8qpdu_udold[i] = o8ud[i];
        o8qpdu_aiold[i] = o8qpdu_ai[i];
    }
    for (i = 1 ; i <= o8ndual ; i++) {
        o8qpdu_xdold[i] = o8qpdu_xd[i];
    }
    L60:
    ss = o8zero;
    ip = 0;
    
    /* introduce most violated inequality constraint */
    
    for (i = 1 ; i <= o8mi ; i++) {
        if( o8qpdu_s[i] < ss && o8qpdu_iai[i] != 0 && o8qpdu_iaexcl[i] != 0 ) {
            ss = o8qpdu_s[i];
            ip = i;
        }
    }
    if ( o8iq > 1 ) {
        condr = o8rnorm/o8rlow;
    } else {
        condr = o8one;
    }

    if ( fabs(psid) <= tiny*(c1*c2+condr) || ip == 0) {
    
        /* successful termination of QP-solver for current o8tauqp */
        
        if ( fabs(psid) > tiny*(c1*c2+condr) && ! o8silent ) o8msg(10);
        o8qpterm = 1;
        o8accinf[o8itstep][30] = o8one;
        o8accinf[o8itstep][13] = condr;
        o8accinf[o8itstep][14] = c1*c2;
        for (i = 1 ; i <= o8n ; i++) {
            o8d[i] = o8qpdu_xd[i+incr];
        }
        /* new : o8dnorm added */
        
        o8dnorm  = o8vecn(1,o8n,o8d);
        o8infeas = o8zero;
        
        for (i = 1 ; i <= incr ; i++) {
            o8infeas = o8infeas+fabs(o8qpdu_xd[i]);
        }
        /* L1-norm of o8slack variables */
        
        o8accinf[o8itstep][31] = o8tauqp;
        o8accinf[o8itstep][32] = o8infeas;
        wlow = FALSE;
        su1  = o8zero;
        su2  = o8zero;
        for (i = 1 ; i <= o8iq ; i++) 
        {
            if ( o8qpdu_ai[i] < 0 ) 
            {
                o8u[-o8qpdu_ai[i]] = o8ud[i];
            } 
            else 
            {
                if ( o8qpdu_ai[i] > o8nr ) o8u[o8qpdu_iqlist[o8qpdu_ai[i]]-o8nr] = o8ud[i];
            }
        }

        term1 = o8zero;
        for (j = 1 ; j <= o8n ; j++) {
            o8np[j] = o8gradf[j]*o8scf;
        }
        for (i = 1 ; i <= o8nres ; i++) 
        {
          if ( i > o8n )
          { 
            for (j = 1 ; j <= o8n ; j++) 
            {
                o8np[j] = o8np[j]-o8gres[j][i-o8n]*(o8u[2*i-1]-o8u[2*i]);
            }
          }
          else
          {
            o8np[i] = o8np[i]-o8xsc[i]*(o8u[2*i-1]-o8u[2*i]);
          }
        }
        
        o8b2n = o8vecn(1,o8n,o8np);
        
        if ( o8scf != o8zero ) o8b2n = o8b2n/o8scf;
        
        /* correction in the original variables */
            
        infe1 = o8zero;
        for (i = 1 ; i <= o8nr ; i++) {
            infe1 = infe1+fabs(o8qpdu_xd[i])*o8qpdu_mult[i];
        }
        if ( o8upsi <= o8delmin*o8nres 
            && o8b2n <= (o8gfn+o8one)*o8epsx*o8tp2 && o8phase >= 0
            && o8infeas <= o8delmin*o8nres ) {
        /* dual feasibility is automatic here */ 
        /* since multipliers may be incorrect for o8infeas != o8zero be careful */
        /* we consider the problem as successfully solved with reduced      */
        /* requirements. we terminate here                                  */

            for (i = 1 ; i <= o8n ; i++) 
            {
                o8d[i] = o8zero;
            }
            o8dnorm  = o8zero;
            o8optite = o8three;
            o8dirder = o8zero ;
            o8qpterm = 1 ;
            return;
        }
        /* there may be an additional increase of o8tauqp necessary again */
        if ( infe1 > (o8one-o8delta1/o8tauqp)*o8upsi &&
            ( o8vecn(1,o8n,o8d) <= min(infe1,pow(infe1,2))*o8tp1
            || o8upsi > o8tau0*o8p5 ) ) 
        {
            
        /* further increase o8tauqp ! */
            
            for (i = 1 ; i <= 2*o8nres ; i++) 
            {
                o8u[i]     = o8zero;
                o8slack[i] = o8zero;
            }
            if ( ! o8silent ) o8msg(17);
            if ( o8tauqp*o8taufac > o8taumax ) {
                if ( ! o8silent ) o8msg(5);
                o8qpterm = -1;
                o8accinf[o8itstep][30] = o8qpterm;
                o8accinf[o8itstep][31] = o8tauqp;
                o8accinf[o8itstep][32] = o8infeas;
                for (i = 1 ; i <= o8n ; i++) {
                    o8d[i] = o8zero;
                }
                o8dnorm = o8zero;
                
                return;
                
            } else {
                o8tauqp = o8tauqp*o8taufac;
                
                goto L10;
            }
        }
        /* compute new weights for the penalty-function */
        
        L500:

        for (i = 1 ; i <= 2*o8nres ; i++) 
        {
            o8slack[i] = o8zero;
        }
        for (i = 1 ; i <= o8nr ; i++) 
        {
            o8slack[o8aalist[i]] = o8qpdu_xd[i];
        }
        wlow = FALSE;
        for (i = 1 ; i <= o8nres ; i++) 
        {
            o8w1[2*i-1] = o8w[2*i-1];
            o8w1[2*i] = o8w[2*i] ;
            if ( o8low[i] == o8up[i] ) 
            {
               if ( fabs(o8slack[2*i-1]) > fabs(o8res[2*i-1])+tiny ) 
               {
                    o8w1[2*i-1] = fabs(o8u[2*i-1]);
               } 
               else 
               {
                    o8w1[2*i-1] = o8ny*fabs(o8u[2*i-1])+o8tau;
               }
            } 
            else 
            {
               if ( o8bind[2*i-1] == 0 ) 
               {
                  o8w1[2*i-1] = max(o8w[2*i-1]*o8p8,o8tau);
               } 
               else 
               {
                  if ( o8res[2*i-1] >= o8zero && o8slack[2*i-1] <= tiny )
                    o8w1[2*i-1] = 
                    max(o8ny*fabs(o8u[2*i-1])+o8tau,(fabs(o8u[2*i-1])+o8w1[2*i-1])*o8p5);
                  if ( o8res[2*i-1] >= o8zero && o8slack[2*i-1] > tiny )
                    o8w1[2*i-1] = fabs(o8u[2*i-1]);
                  if ( o8res[2*i-1] < o8zero && o8slack[2*i-1] <= -o8res[2*i-1]+tiny )
                    o8w1[2*i-1] = 
                    max(o8ny*fabs(o8u[2*i-1])+o8tau,(fabs(o8u[2*i-1])+o8w1[2*i-1])*o8p5);
                  if ( o8res[2*i-1] < o8zero && o8slack[2*i-1] >  -o8res[2*i-1]+tiny )
                    o8w1[2*i-1] = fabs(o8u[2*i-1]);
               }
               if ( o8bind[2*i] == 0 ) 
               {
                  o8w1[2*i] = max(o8w[2*i]*o8p8,o8tau);
               } 
               else 
               {
                  if ( o8res[2*i] >= o8zero && o8slack[2*i] <= tiny )
                    o8w1[2*i] = 
                    max(o8ny*fabs(o8u[2*i])+o8tau,(fabs(o8u[2*i])+o8w1[2*i])*o8p5);
                  if ( o8res[2*i] >= o8zero && o8slack[2*i] > tiny )
                    o8w1[2*i] = fabs(o8u[2*i]);
                  if ( o8res[2*i] < o8zero && o8slack[2*i] <= -o8res[2*i]+tiny )
                    o8w1[2*i] = 
                    max(o8ny*fabs(o8u[2*i])+o8tau,(fabs(o8u[2*i])+o8w1[2*i])*o8p5);
                  if ( o8res[2*i] < o8zero && o8slack[2*i] >  -o8res[2*i]+tiny )
                    o8w1[2*i] = fabs(o8u[2*i]);
               }
            } 
            if ( o8w1[2*i-1] < o8w[2*i-1] || o8w1[2*i] < o8w[2*i] ) wlow = TRUE;
        }
        if ( wlow ) 
        {
            s1 = o8zero;
            s2 = o8zero;
            for (i = 1 ; i <= o8nres ; i++) 
            {
                if ( o8low[i] == o8up[i]  ) 
                {
                    s1 += o8w1[2*i-1]*fabs(o8resst[2*i-1]);
                    s2 += o8w1[2*i-1]*fabs(o8res[2*i-1]);
                } 
                else 
                {
                    s1 -= min(o8zero,o8resst[2*i-1])*o8w1[2*i-1];
                    s2 -= min(o8zero,o8res[2*i-1]  )*o8w1[2*i-1];
                    s1 -= min(o8zero,o8resst[2*i])*o8w1[2*i];
                    s2 -= min(o8zero,o8res[2*i]  )*o8w1[2*i];

                }
            }
            diff0 = (o8fxst-o8fx)*o8scf+(s1-s2);
            if ( diff0 >= o8eta*o8clow && o8itstep-o8lastdw >= max(5,min(o8n/10,20)) ) {
            
                /* accept new (diminished ) weights */
            
                o8lastdw = o8itstep;
                o8lastch = o8itstep;
                o8level  = diff0/o8iterma;
                o8psist  = s1;
                o8psi    = s2;
                for (i = 1 ; i <= 2*o8nres ; i++) {
                    if ( o8w1[i] != o8w[i] ) o8lastch = o8itstep;
                    o8w[i] = o8w1[i];
                }
                o8clow = o8clow+o8one;
                if ( o8clow > o8itstep/10 ) {
                
                    /* additional increase of o8eta */
                    
                    o8eta = o8eta*o8onep3;
                    if ( ! o8silent ) o8info(11);
                }
                if ( ! o8silent ) o8info(12);
                
                goto L1000;
            }
        }
        /* we cannot accept new weights */
        /* reset weights                */

        for (i = 1 ; i <= o8nres ; i++  ) 
        {
            o8w1[2*i-1] = o8w[2*i-1];
            o8w1[2*i  ] = o8w[2*i] ;
            if ( o8low[i] == o8up[i] ) 
            {
                if ( o8slack[2*i-1] > fabs(o8res[2*i-1]) ) 
                   o8w1[2*i-1] = fabs(o8u[2*i-1]);
                if ( o8slack[2*i-1] <= fabs(o8res[2*i-1]) ) 
                {
                  if ( o8w[2*i-1] <= fabs(o8u[2*i-1]) 
                       && fabs(o8u[2*i-1]) <= o8w[2*i-1]+o8tau ) 
                  {
                      o8w1[2*i-1] = o8w[2*i-1]+o8two*o8tau;
                  } 
                  else 
                  {
                      o8w1[2*i-1] = max(o8w[2*i-1],o8ny*fabs(o8u[2*i-1])+o8tau);
                  }
                }
            } 
            else 
            {
                if ( o8slack[2*i-1] > -min(-tiny,o8res[2*i-1]) && o8bind[2*i-1] == 1) 
                {
                   o8w1[2*i-1] = fabs(o8u[2*i-1]);
                } 
                else if(  o8bind[2*i-1] == 1 && 
                          o8slack[2*i-1] <= -min(-tiny,o8res[2*i-1])
                          && o8u[2*i-1] <= o8w[2*i-1]+o8tau && o8w[2*i-1] >= o8u[2*i-1] ) 
                {
                    
                    o8w1[2*i-1] = o8w[2*i-1]+o8two*o8tau;
                } 
                else if ( o8bind[2*i-1] == 1 ) 
                {
                    o8w1[2*i-1] = max(o8w[2*i-1],o8ny*fabs(o8u[2*i-1])+o8tau);
                }
                if ( o8slack[2*i] > -min(-tiny,o8res[2*i]) && o8bind[2*i] == 1 )
                {
                   o8w1[2*i] = fabs(o8u[2*i]);
                }
                else if ( o8bind[2*i] == 1 && o8slack[2*i] <= -min(-tiny,o8res[2*i])
                            && o8u[2*i] <= o8w[2*i]+o8tau && o8w[2*i] >= o8u[2*i] )
                {
                 
                    o8w1[2*i] = o8w[2*i]+o8two*o8tau;
                }
                else if ( o8bind[2*i] == 1 )
                {
                    o8w1[2*i] = max(o8w[2*i],o8ny*fabs(o8u[2*i])+o8tau);
                }

            }
        }
        term1 = o8zero;
        for (i = 1 ; i <= 2*o8nres ; i++) {
            if ( o8w1[i] > o8w[i] || o8w1[i] < o8w[i] ) o8lastch = o8itstep;
            if ( o8w1[i] > o8w[i] ) o8lastup = o8itstep;
            if ( o8w1[i] < o8w[i] ) o8lastdw = o8itstep;
            o8w[i]  = o8w1[i];
            term1 = max(term1,o8w[i]);
        }
        s1 = o8zero;
        s2 = o8zero;
        for (i = 1 ; i <= o8nres ; i++) {
            if ( o8low[i] == o8up[i] ) 
            {
                s1 += o8w[2*i-1]*fabs(o8resst[2*i-1]);
                s2 += o8w[2*i-1]*fabs(o8res[2*i-1]);
            } else {
                s1 -= o8w[2*i-1]*min(o8zero,o8resst[2*i-1]);
                s2 -= o8w[2*i-1]*min(o8zero,o8res[2*i-1]);
                s1 -= o8w[2*i]*min(o8zero,o8resst[2*i]);
                s2 -= o8w[2*i]*min(o8zero,o8res[2*i]);

            }
        }
        o8psist = s1;
        o8psi   = s2;
        if ( ! o8silent ) o8info(12);
        o8accinf[o8itstep][20] = term1;
        o8accinf[o8itstep][19] = o8clow;
        
        goto L1000;
    }
   /************************************************************/
   /**     continue QP solver                                 **/
   /************************************************************/

    if ( ip > o8nr ) 
    {
    /* compute gradient of current QP inequality constraint */
    /* this is a linearized original constraint */
        k = (o8qpdu_iqlist[ip]-o8nr+1)/2 ;
        /*  k is in {1,...,o8nres} */
        if ( k > o8n ) 
        {   /* a general constraint */
          for (j = 1 ; j <= o8n ; j++) 
          {
            o8qpdu_cii[j+incr] = o8gres[j][k-o8n]*o8gres[0][k-o8n];
          }
          for (j = 1 ; j <= incr ; j++) 
          {
            o8qpdu_cii[j] = o8zero;
          }
          o8qpdu_cii[o8me+ip-o8nr] = o8one;
          o8qpdu_ci0[ip]       = o8res[o8qpdu_iqlist[ip]-o8nr];
        }
        else
        {  /* an active bound constraint */
           for ( j = 1 ; j <= o8ndual ; j++ )
           {
             o8qpdu_cii[j] = o8zero ;
           }
           o8qpdu_cii[o8me+ip-o8nr] = o8one ;
           o8qpdu_ci0[ip      ] = o8res[o8qpdu_iqlist[ip]-o8nr];
           o8qpdu_cii[k+incr  ] = ( (o8qpdu_iqlist[ip]-o8nr) %2 ==0  ) ? -o8xsc[k] : o8xsc[k] ;
        }        

    } 
    else 
    {
        for (j = 1 ; j <= o8ndual ; j++) 
        {
            o8qpdu_cii[j] = o8zero;
        }
        o8qpdu_ci0[ip] = o8zero;
        o8qpdu_cii[ip] = o8one;
    }
    for (i = 1 ; i <= o8ndual ; i++) {
        o8np[i] = o8qpdu_cii[i];
    }

    for (i = 1 ; i <= o8iq ; i++) {
        o8ud1[i] = o8ud[i];
    }
    o8ud1[o8iq+1] = o8zero;
    o8qpdu_ai[o8iq+1]  = ip;     /* this is in {1,..,o8mi} */
    
    L100:

    /* step 2a */
    
    o8zup(o8qpdu_z);
    
    if ( o8iq != 0 ) o8rup(o8qpdu_vr);
    
    l  = 0;
    t1 = infiny;
    for (k = 1 ; k <= o8iq ; k++) {
        if(o8qpdu_ai[k] > 0 && o8qpdu_vr[k] > o8zero) {
            if ( o8ud1[k]/o8qpdu_vr[k] < t1 ) {
                t1 = o8ud1[k]/o8qpdu_vr[k];
            }
        }
    }
    /* |z| = 0? */
    
    /* old      zz = o8sc1(1,o8ndual,o8qpdu_z,o8qpdu_z) */
    
    zz   = o8vecn(1,o8ndual,o8qpdu_z);
    term = o8sc1(1,o8ndual,o8qpdu_z,o8np);
    
    /* old      if (zz != o8zero && term > o8zero ) { */
    
    if ( zz >= tiny*o8rnorm && term > o8zero ) {
        t2 = -o8qpdu_s[ip]/term;
    } else {
        t2 = infiny;
    }
    t = min(t1,t2);

    if ( t >= infiny ) 
    {
      o8qpterm = -2 ;
      if ( ! o8silent ) o8msg(20);
      goto L2000;
    }


    if( t2 >= infiny ) {
        for (k = 1 ; k <= o8iq ; k++) {
            o8ud1[k] = o8ud1[k]+t*(-o8qpdu_vr[k]);
            if ( o8ud1[k] < o8zero && o8qpdu_ai[k] > 0 ) o8ud1[k] = o8zero;
        }
        o8ud1[o8iq+1] = o8ud1[o8iq+1]+t;
        o8qpdu_qpdel[0]  = 0;
        for (i = 1 ; i <= o8iq ; i++) {
            if ( o8ud1[i] <= tiny && o8qpdu_ai[i] > 0 ) {
                o8qpdu_qpdel[0]        = o8qpdu_qpdel[0]+1;
                o8qpdu_qpdel[o8qpdu_qpdel[0]] = o8qpdu_ai[i];
            }
        }
        for (k = 1 ; k <= o8qpdu_qpdel[0] ; k++) {
            l      = o8qpdu_qpdel[k];
            o8qpdu_iai[l] = l;

            o8dlcd(o8qpdu_ai,l);
        }
        goto L100;
    }
    for (k = 1 ; k <= o8ndual ; k++) {
        o8qpdu_xd[k] = o8qpdu_xd[k]+t*o8qpdu_z[k];
    }
    for (k = 1 ; k <= o8iq ; k++) {
        o8ud1[k] = o8ud1[k]+t*(-o8qpdu_vr[k]);
        if ( o8ud1[k] < o8zero && o8qpdu_ai[k] > 0 ) o8ud1[k] = o8zero;
    }
    o8ud1[o8iq+1] = o8ud1[o8iq+1]+t;
    
    f = f+t*o8sc1(1,o8ndual,o8qpdu_z,o8np)*(o8p5*t+o8ud1[o8iq+1]);
    if ( f <= fmax*(o8one+o8epsmac)+o8epsmac )
    {
      nosucc += 1 ;
      if ( nosucc > o8qpitma )
      {
        o8qpterm = - 3;
        if ( ! o8silent ) o8msg(26);
        goto L2000;
      }
    }  
    else
    {   
      nosucc = 0 ;
    }
    fmax = max( f , fmax ) ;
    
    if ( t2 <= t1-tiny ) {
    
        /* o8ddual is computed by o8zup */
        
        if ( o8vecn(o8iq+1,o8ndual,o8ddual) < o8epsmac*o8rnorm ) {
        
            /* degeneracy: adding this constraint gives a singular working set */
            /* theoretically impossible, but due to roundoff this may occur.   */
            /* mark this constraint and try to add another o8one                 */
            
            o8iptr = ip;
            o8iqtr = o8iq;
            for (i = 1 ; i <= o8iq ; i++) {
                o8aitr[i] = o8qpdu_ai[i];
            }
            o8sstr  = ss;
            o8riitr = o8vecn(o8iq+1,o8ndual,o8ddual);
            
            if ( ! o8silent ) o8msg(19);
            o8qpdu_iaexcl[ip] = 0;
            for (i = 1 ; i <= o8mi ; i++) {
                o8qpdu_iai[i] = i;
            }
            for (i = 1 ; i <= o8iq ; i++) {
                o8qpdu_ai[i] = o8qpdu_aiold[i];
                if (o8qpdu_ai[i] > 0 ) o8qpdu_iai[o8qpdu_ai[i]] = 0;
                o8ud1[i] = o8qpdu_udold[i];
            }
            for (i = 1 ; i <= o8ndual ; i++) {
                o8qpdu_xd[i] = o8qpdu_xdold[i];
            }
            goto L60;
        }
        /* add constraint, l-pair */

        o8adcd();
        
        o8qpdu_iai[ip] = 0;
        for (i = 1 ; i <= o8iq ; i++) {
            o8ud[i] = o8ud1[i];
        }
        goto L50;
    }
    su = o8zero;
    if ( ip > o8nr ) 
    {
    
      /* an original linearized inequality constraint */
      k = (o8qpdu_iqlist[ip]-o8nr+1)/2 ;
      /*  k is in {1,...,o8nres} */
      if ( k > o8n ) 
      {   /* a general constraint */
         for (j = 1 ; j <= o8n ; j++) 
         {
           o8qpdu_cii[j+incr] = o8gres[j][k-o8n]*o8gres[0][k-o8n];
         }
         for (j = 1 ; j <= incr ; j++) 
         {
           o8qpdu_cii[j] = o8zero;
         }
         o8qpdu_cii[o8me+ip-o8nr] = o8one;
         o8qpdu_ci0[ip]       = o8res[o8qpdu_iqlist[ip]-o8nr];
      }
      else
      {  /* an active bound constraint */
         for ( j = 1 ; j <= o8ndual ; j++ )
         {
           o8qpdu_cii[j] = o8zero ;
         }
         o8qpdu_cii[o8me+ip-o8nr] = o8one ;
         o8qpdu_ci0[ip      ] = o8res[o8qpdu_iqlist[ip]-o8nr];
         o8qpdu_cii[k+incr  ] = ( (o8qpdu_iqlist[ip]-o8nr) %2 ==0  ) ? -o8xsc[k] : o8xsc[k] ;
      }        
      o8qpdu_s[ip] = o8sc1(1,o8ndual,o8qpdu_cii,o8qpdu_xd)+o8qpdu_ci0[ip];
    } 
    else 
    {
        /* a o8slack constraint */
        
        o8qpdu_s[ip] = o8qpdu_xd[ip];
    }
    /* now t = t1 */
    
    o8qpdu_qpdel[0] = 0;
    for (i = 1 ; i <= o8iq ; i++) {
        if ( o8ud1[i] <= tiny && o8qpdu_ai[i] > 0 ) {
            o8qpdu_qpdel[0]        = o8qpdu_qpdel[0]+1;
            o8qpdu_qpdel[o8qpdu_qpdel[0]] = o8qpdu_ai[i];
        }
    }
    for (k = 1 ; k <= o8qpdu_qpdel[0] ; k++) {
        l      = o8qpdu_qpdel[k];
        o8qpdu_iai[l] = l;
        
        o8dlcd(o8qpdu_ai,l);
    }
    if ( t2 <= t1+tiny ) {
        if ( o8vecn(o8iq+1,o8ndual,o8ddual) < o8epsmac*o8rnorm ) {
        
            /* degeneracy */
            
            o8iptr = ip;
            o8iqtr = o8iq;
            for (i = 1 ; i <= o8iq ; i++) {
                o8aitr[i] = o8qpdu_ai[i];
            }
            o8sstr  = ss;
            o8riitr = o8vecn(o8iq+1,o8ndual,o8ddual);
            
            if ( ! o8silent ) o8msg(19);
            o8qpdu_iaexcl[ip] = 0;
            for (i = 1 ; i <= o8mi ; i++) {
                o8qpdu_iai[i] = i;
            }
            for (i = 1 ; i <= o8iq ; i++) {
                o8qpdu_ai[i] = o8qpdu_aiold[i];
                if ( o8qpdu_ai[i] > 0 ) o8qpdu_iai[o8qpdu_ai[i]] = 0;
                o8ud1[i] = o8qpdu_udold[i];
            }
            for (i = 1 ; i <= o8ndual ; i++) {
                o8qpdu_xd[i] = o8qpdu_xdold[i];
            }
            goto L60;
        }
        /* add constraint, l-pair */
        
        o8adcd();
        
        o8qpdu_iai[ip] = 0;
        for (i = 1 ; i <= o8iq ; i++) {
            o8ud[i] = o8ud1[i];
        }
        goto L50;
        
    } else {
    
        goto L100;
    }
    /* this is the exit point of o8qpdu                                     */
    /* we either may have successful or unsuccessful termination here       */
    /* the latter with o8qpterm = -2 or -3, in which case it may nevertheless */
    /* be possible to use the computed o8d. -2 or -3 exit is theoretically    */
    /* impossible but may occur due to roundoff effects.                    */
    /* we check the directional derivative of the penalty-function now      */
    
    L1000:

    /* cut and rescale o8d if appropriate */
    
    o8cutd();
    
    /* compute the directional derivative o8dirder */
    
    o8dird();
    
    if ( o8dirder >= o8zero || ( -o8dirder <= o8epsmac*o8tp2*(o8scf*fabs(o8fx)+o8psi+o8one) &&
        o8infeas > max(o8upsi,o8nres*o8delmin) ) ) {
        if ( ! o8silent ) o8msg(18);
        if ( o8tauqp <= o8taumax/o8taufac ) {
            o8tauqp = o8tauqp*o8taufac;
            
            goto L10;
            
        } else {
            if ( ! o8silent ) o8msg(5);
            o8qpterm = -1;
            o8accinf[o8itstep][30] = o8qpterm;
            o8accinf[o8itstep][31] = o8tauqp;
            o8accinf[o8itstep][32] = o8infeas;
            for (i = 1 ; i <= o8n ; i++) {
                o8d[i] = o8zero;
            }
            o8dnorm = o8zero;
        }
    }
    return;
    
    L2000:

    /* QP infeasible ( in this application impossible , theoretically) */
    

    o8accinf[o8itstep][30] = -o8two;
    o8accinf[o8itstep][13] = condr;
    o8accinf[o8itstep][14] = c1*c2;
    for (i = 1 ; i <= o8n ; i++) {
        o8d[i] = o8qpdu_xd[i+incr];
    }
    o8dnorm = o8vecn(1,o8n,o8d);
    su1   = o8zero;
    for (i = 1 ; i <= incr ; i++) {
        su1 = su1+fabs(o8qpdu_xd[i]);
    }
    /* L1-norm of o8slack variables */
    
    o8accinf[o8itstep][32] = su1;
    wlow = FALSE;
    su1  = o8zero;
    su2  = o8zero;
    for (i = 1 ; i <= o8iq ; i++) {
        if ( o8qpdu_ai[i] < 0 ) {
            o8u[-o8qpdu_ai[i]] = o8ud[i];
        } else {
            if ( o8qpdu_ai[i] > o8nr ) o8u[o8qpdu_iqlist[o8qpdu_ai[i]]-o8nr] = o8ud[i];
            /*  o8qpdu_iqlist[ai[i]] = o8aalist[k]+o8nr for some k */
        }
    }
    /* compute Lagrangian violation */
    term1 = o8zero;
    for (j = 1 ; j <= o8n ; j++) {
        o8np[j] = o8gradf[j]*o8scf;
    }
    for (i = 1 ; i <= o8nres ; i++) 
    {
      if ( i > o8n )
      {  
        for (j = 1 ; j <= o8n ; j++) 
        {
          o8np[j] = o8np[j]-o8gres[j][i-o8n]*(o8u[2*i-1]-o8u[2*i]);
        }
      }
      else
      {
          o8np[i] -= o8xsc[i]*(o8u[2*i-1]-o8u[2*i]) ;
      }
      o8w1[2*i-1] = o8w[2*i-1];
      o8w1[2*i  ] = o8w[2*i] ;
      if ( o8low[i] == o8up[i]  ) 
      {
        if ( o8slack[2*i-1] > fabs(o8res[2*i-1]) ) o8w1[2*i-1] = fabs(o8u[2*i-1]);
        if ( o8slack[2*i-1] <= fabs(o8res[2*i-1]) ) 
        {
           if ( o8w[2*i-1] <= fabs(o8u[2*i-1]) 
                && fabs(o8u[2*i-1]) <= o8w[2*i-1]+o8tau ) 
           {
               o8w1[2*i-1] = o8w[2*i-1]+o8two*o8tau;
           } 
           else 
           {
               o8w1[2*i-1] = max(o8w[2*i-1],o8ny*fabs(o8u[2*i-1])+o8tau);
           }
        }
        su1 += fabs(o8res[2*i-1]  )*o8w1[2*i-1];
        su2 += fabs(o8resst[2*i-1])*o8w1[2*i-1];
      } 
      else 
      {
        if ( o8slack[2*i-1] > -min(-tiny,o8res[2*i-1]) && o8bind[2*i-1] == 1 ) 
        {
           o8w1[2*i-1] = fabs(o8u[2*i-1]);
        } 
        else if ( o8bind[2*i-1] == 1 && o8slack[2*i-1] <= -min(-tiny,o8res[2*i-1])
                  && o8u[2*i-1] <= o8w[2*i-1]+o8tau && o8w[2*i-1] >= o8u[2*i-1] ) 
        {
           o8w1[2*i-1] = o8w[2*i-1]+o8two*o8tau;
        } 
        else if ( o8bind[2*i-1] == 1 ) 
        {
           o8w1[2*i-1] = max(o8w[2*i-1],o8ny*fabs(o8u[2*i-1])+o8tau);
        }
        su1 -= o8w1[2*i-1]*min(o8zero,o8res[2*i-1]);
        su2 -= o8w1[2*i-1]*min(o8zero,o8resst[2*i-1]);
     
        if ( o8slack[2*i] > -min(-tiny,o8res[2*i]) && o8bind[2*i] == 1 ) 
        {
           o8w1[2*i] = fabs(o8u[2*i]);
        } 
        else if ( o8bind[2*i] == 1 && o8slack[2*i] <= -min(-tiny,o8res[2*i])
                  && o8u[2*i] <= o8w[2*i]+o8tau && o8w[2*i] >= o8u[2*i] ) 
        {
           o8w1[2*i] = o8w[2*i]+o8two*o8tau;   
        } 
        else if ( o8bind[2*i] == 1 ) 
        {
           o8w1[2*i] = max(o8w[2*i],o8ny*fabs(o8u[2*i])+o8tau);
        }
        su1 -= o8w1[2*i]*min(o8zero,o8res[2*i]);  
        su2 -= o8w1[2*i]*min(o8zero,o8resst[2*i]);

      }
      if ( o8w[i] != o8w1[i] ) o8lastch = o8itstep;
      o8w[i]  = o8w1[i];
      term1 = max(term1,o8w[i]);
    }
    o8psist = su2;
    o8psi   = su1;
    
    o8b2n   = sqrt(o8sc1(1,o8n,o8np,o8np));
    
    if ( o8scf != o8zero ) o8b2n = o8b2n/o8scf;
    if ( wlow ) {
        o8clow   = o8clow+o8one;
        o8lastch = o8itstep;
        o8lastdw = o8itstep;
    }
    if ( ! o8silent ) o8info(12);
    o8accinf[o8itstep][19] = o8clow;
    o8accinf[o8itstep][20] = term1;
    o8accinf[o8itstep][31] = o8tauqp;
        if ( o8upsi <= o8delmin*o8nres   
            && o8b2n <= (o8gfn+o8one)*o8epsx*o8tp2 && o8phase >= 0
            && o8infeas <= o8delmin*o8nres )
    {
     
      /* since multipliers may be incorrect for o8infeas != o8zero be careful */
      /* we consider the problem as successfully solved with reduced      */
      /* requirements                                                     */
     
            for (i = 1 ; i <= o8n ; i++) {
                o8d[i] = o8zero;
            }
            o8dnorm  = o8zero;
            o8optite = o8three;
            o8qpterm = 1 ;   /* new */
            return;
    }

    goto L1000;
}

/* **************************************************************************** */
/*                  compute updated projected gradient (primal)                 */
/* **************************************************************************** */
void o8zup(DOUBLE z[]) {

    static INTEGER  i,j;
    static DOUBLE   su;

    /* o8d = j(trans) *o8np */
    
    for (i = 1 ; i <= o8ndual ; i++) {
        su = o8zero;
        for (j = 1 ; j <= o8ndual ; j++) {
            su = su+o8xj[j][i]*o8np[j];
        }
        o8ddual[i] = su;
    }
    /* computation of z */
    
    for (i = 1 ; i <= o8ndual ; i++) {
        z[i] = o8zero;
        for (j = o8iq+1 ; j <= o8ndual ; j++) {
            z[i] = z[i]+o8xj[i][j]*o8ddual[j];
        }
    }
    return;
}

/* **************************************************************************** */
/*                    compute correction of dual multipliers                    */
/* **************************************************************************** */
void o8rup(DOUBLE rv[]) {

    static DOUBLE   s;
    static INTEGER  i,j;

    for (i = o8iq ; i >= 1 ; i--) {
        s = o8zero;
        for (j = i+1 ; j <= o8iq ; j++) {
            s = s+o8r[i][j]*rv[j];
        }
        rv[i] = (o8ddual[i]-s)/o8r[i][i];
    }
    return;
}

/* **************************************************************************** */
/*                          delete constraint o8nr. l                             */
/* **************************************************************************** */
void o8dlcd(INTEGER ai[],INTEGER l) {
    
    DOUBLE o8dsq1(DOUBLE o8a,DOUBLE b);
    
    static INTEGER  qq,i,j,k;
    static DOUBLE   t1,t2,cc,ss,h,c1,s1,xny;

    for (i = 1 ; i <= o8iq ; i++) {
        if ( ai[i] == l ) {
            qq = i;
            
            goto L10;
         }
    }
    L10:

    for (i = qq ; i <= o8iq-1 ; i++) {
        ai[i]  = ai[i+1];
        o8ud1[i] = o8ud1[i+1];
        for (j = 1 ; j <= o8ndual ; j++) {
            o8r[j][i] = o8r[j][i+1];
        }
    }
    L20:

    ai[o8iq]    = ai[o8iq+1];
    o8ud1[o8iq]   = o8ud1[o8iq+1];
    ai[o8iq+1]  = 0;
    o8ud1[o8iq+1] = o8zero;
    for (j = 1 ; j <= o8iq ; j++) {
        o8r[j][o8iq] = o8zero;
    }
    o8iq = o8iq-1;
    
    if ( o8iq == 0 ) goto L100;
    
    for (j = qq ; j <= o8iq ; j++) {
        cc = o8r[j][j];
        ss = o8r[j+1][j];
        h  = o8dsq1(cc,ss);
        
        if ( h == o8zero ) goto L90;
        
        c1 = cc/h;
        s1 = ss/h;
        o8r[j+1][j] = o8zero;
        if ( c1 < o8zero ) {
            o8r[j][j] = -h;
            c1      = -c1;
            s1      = -s1;
        } else {
            o8r[j][j] = h;
        }
        xny = s1/(o8one+c1);
        for (k = j+1 ; k <= o8iq ; k++) {
            t1        = o8r[j][k];
            t2        = o8r[j+1][k];
            o8r[j][k]   = t1*c1+t2*s1;
            o8r[j+1][k] = xny*(t1+o8r[j][k])-t2;
        }
        for (k = 1 ; k <= o8ndual ; k++) {
            t1         = o8xj[k][j];
            t2         = o8xj[k][j+1];
            o8xj[k][j]   = t1*c1+t2*s1;
            o8xj[k][j+1] = xny*(o8xj[k][j]+t1)-t2;
        }
        L90:;
    }
    L100:

    o8rnorm = o8one;
    o8rlow  = o8one;
    
    /* in order to avoid a compiler error of hp in +op3 mode */
    
    if ( o8iq >= 1 ) {
        o8rnorm = fabs(o8r[1][1]);
        o8rlow  = fabs(o8r[1][1]);
        i     = 1;
        while ( i < o8iq ) {
            i     = i+1;
            o8rnorm = max(o8rnorm,fabs(o8r[i][i]));
            o8rlow  = min(o8rlow, fabs(o8r[i][i]));
        }
    }
    return;
}

/* **************************************************************************** */
/*                  add constraint whose gradient is given by o8np                */
/* **************************************************************************** */
void o8adcd(void) {
    
    DOUBLE o8dsq1(DOUBLE o8a,DOUBLE b);
                    
    static INTEGER  i,j,k;
    static DOUBLE   cc,ss,h,s1,c1,t1,t2,xny;

    for (j = o8ndual ; j >= o8iq+2 ; j--) {
        cc = o8ddual[j-1];
        ss = o8ddual[j];
        h  = o8dsq1(cc,ss);
        
        if ( h == o8zero ) goto L20;
        
        o8ddual[j] = o8zero;
        s1       = ss/h;
        c1       = cc/h;
        if ( c1 < o8zero ) {
            c1         = -c1;
            s1         = -s1;
            o8ddual[j-1] = -h;
        } else {
            o8ddual[j-1] = h;
        }
        xny = s1/(o8one+c1);
        for (k = 1 ; k <= o8ndual ; k++) {
            t1         = o8xj[k][j-1];
            t2         = o8xj[k][j];
            o8xj[k][j-1] = t1*c1+t2*s1;
            o8xj[k][j]   = xny*(t1+o8xj[k][j-1])-t2;
        }
        L20:;
    }
    o8iq = o8iq+1;
    for (i = 1 ; i <= o8iq ; i++) {
        o8r[i][o8iq] = o8ddual[i];
    }
    o8rnorm = o8one;
    o8rlow  = o8one;
    
    /* in order to avoid a compiler error of hp in +op3 mode */
    
    if ( o8iq >= 1 ) {
        o8rnorm = fabs(o8r[1][1]);
        o8rlow  = fabs(o8r[1][1]);
        i     = 1;
        while ( i < o8iq ) {
            i     = i+1;
            o8rnorm = max(o8rnorm,fabs(o8r[i][i]));
            o8rlow  = min(o8rlow, fabs(o8r[i][i]));
        }
    }
    return;
}

/* **************************************************************************** */
/*          computes the inverse of the upper triangular matrix part            */
/* **************************************************************************** */
void o8rinv(INTEGER o8n,DOUBLE **o8a,INTEGER o8ndual,DOUBLE **o8x) {
            
    /* computes the inverse of the upper triangular matrix part */
    /* of a and stores it in the upper triangle of the          */
    /* right lower minor of o8x                                   */
    /* actual dimension of a is o8n and of o8x o8ndual                */
    
    static INTEGER  l,j,k,incr;
    static DOUBLE   su;

    incr = o8ndual-o8n;
    
    /* incr = o8nr */
    
    for (j = o8n ; j >= 1 ; j--) {
    
        /* we assume a being sufficiently regular here. given in this */
        /* application. See top of o8qpdu                             */
        
        o8x[j+incr][j+incr] = o8one/o8a[j][j];
        for (k = j-1 ; k >= 1 ; k--) {
            su = o8zero;
            for (l = k+1 ; l <= j ; l++) {
                su = su+o8a[k][l]*o8x[l+incr][j+incr];
            }
            o8x[k+incr][j+incr] = -su/o8a[k][k];
        }
    }
    return;
}

/* **************************************************************************** */
/* suite of function interfaces, for performing scaling and                     */
/* external evaluations                                                         */
/* if o8bloc = TRUE then it is assumed that prior to calls to es<xyz>             */
/* o8valid function information is computed (via a call of user_eval)             */
/* and stored in o8fu and o8fugrad , setting o8valid = TRUE afterwards                */
/* **************************************************************************** */

/* **************************************************************************** */
/*                              objective function                              */
/* **************************************************************************** */
void esf(DOUBLE o8x[],DOUBLE *o8fx) 
{

    void ef(DOUBLE o8x[],DOUBLE *o8fx);
    
    static INTEGER  i;
    
    if ( o8bloc ) 
    {
        if ( o8valid ) 
        {
            *o8fx = o8fu[0];
        } 
        else 
        {
            Rprintf("donlp2: o8bloc-call, function info invalid\n");
            exit(1);
        }
    } 
    else 
    {
        for (i = 1 ; i <= o8n ; i++) 
        {
            o8xtr[i] = o8x[i]*o8xsc[i];
        }
        ef(o8xtr,o8fx);
       
    }
    return;
}

/* **************************************************************************** */
/*                        gradient of objective function                        */
/* **************************************************************************** */
void esgradf(DOUBLE o8x[],DOUBLE o8gradf[]) 
{

    void    ef    (DOUBLE o8x[],DOUBLE *o8fx);
    void    egradf(DOUBLE o8x[],DOUBLE o8gradf[]);
    
    static INTEGER  j;
    static DOUBLE   d1,d2,d3,sd1,sd2,sd3,fhelp,fhelp1,fhelp2,
                    fhelp3,fhelp4,fhelp5,fhelp6,xincr,xhelp,floc;

    if ( o8bloc ) 
    {
        if ( o8valid ) 
        {
            for (j = 1 ; j <= o8n ; j++) 
            {
                o8gradf[j] = o8xsc[j]*o8fugrad[j][0];
            }
            return;
        }
        else 
        {
            Rprintf("donlp2: o8bloc call with function info invalid\n");
            exit(1);
        }
    } 
    else 
    {
        for (j = 1 ; j <= o8n ; j++) 
        {
            o8xtr[j] = o8xsc[j]*o8x[j];
        }
        if ( o8analyt ) 
        {
        
             egradf(o8xtr,o8gradf);
             
        } 
        else 
        {
            if ( o8difftype == 1 ) 
            {
                o8deldif = min(o8tm1*sqrt(o8epsfcn),o8tm2);

                ef(o8xtr,&floc);
                
                for (j = 1 ; j <= o8n ; j++) 
                {
                    xhelp = o8xtr[j];
                    xincr = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd);
                    if ( xhelp >= o8zero ) 
                    {
                        o8xtr[j] = xhelp+xincr;
                    }
                    else 
                    {
                        o8xtr[j] = xhelp-xincr;
                    }
                    ef(o8xtr,&fhelp);
                    
                    o8gradf[j] = (fhelp-floc)/(o8xtr[j]-xhelp);
                    o8xtr[j]   = xhelp;
                }
            } 
            else if ( o8difftype == 2 ) 
            {
                o8deldif = min(o8tm1*pow(o8epsfcn,o8one/o8three),o8tm2);
                for (j = 1 ; j <= o8n ; j++) 
                {
                    xhelp  = o8xtr[j];
                    xincr  = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd);
                    o8xtr[j] = xhelp+xincr;
                    
                    ef(o8xtr,&fhelp1);
                    
                    o8xtr[j] = xhelp-xincr;
                    
                    ef(o8xtr,&fhelp2);
                    
                    o8gradf[j] = (fhelp1-fhelp2)/(xincr+xincr);
                    o8xtr[j]   = xhelp;
                }
            } 
            else   /* o8difftype ==3 */ 
            {
                o8deldif = min(o8tm1*pow(o8epsfcn,o8one/o8seven),o8tm2);
                for (j = 1 ; j <= o8n ; j++) 
                {
                   xhelp  = o8xtr[j];
                   xincr  = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd/o8four);
                   o8xtr[j] = xhelp-xincr;
                    
                   ef(o8xtr,&fhelp1);
                    
                   o8xtr[j] = xhelp+xincr;
                    
                   ef(o8xtr,&fhelp2);
                    
                   xincr  = xincr+xincr;
                   d1     = xincr;
                   o8xtr[j] = xhelp-xincr;
                    
                   ef(o8xtr,&fhelp3);
                    
                   o8xtr[j] = xhelp+xincr;
                    
                   ef(o8xtr,&fhelp4);
                    
                   xincr  = xincr+xincr;
                   d2     = xincr;
                   o8xtr[j] = xhelp-xincr;
                   
                   ef(o8xtr,&fhelp5);
                    
                   o8xtr[j] = xhelp+xincr;
                    
                   ef(o8xtr,&fhelp6);
                    
                   o8xtr[j]   = xhelp;
                   d3       = xincr+xincr;
                   sd1      = (fhelp2-fhelp1)/d1;
                   sd2      = (fhelp4-fhelp3)/d2;
                   sd3      = (fhelp6-fhelp5)/d3;
                   sd3      = sd2-sd3;
                   sd2      = sd1-sd2;
                   sd3      = sd2-sd3;
                   o8gradf[j] = sd1+o8p4*sd2+sd3/o8c45;
                }
            }
        }
        for (j = 1 ; j <= o8n ; j++) 
        {
            o8gradf[j] = o8xsc[j]*o8gradf[j];
        }
    }
    return;
}

/********************************************************************/

void escon( INTEGER call_type , INTEGER liste[], DOUBLE o8x[], DOUBLE constr[],
            LOGICAL errliste[])
{
void  econ(INTEGER call_type,INTEGER liste[],DOUBLE  o8xtr[],DOUBLE constr[],
            LOGICAL errliste[]);
INTEGER j    ;

     if ( o8bloc ) 
     {
        if ( o8valid ) 
        {
            if ( call_type == 1 )
            {
               for ( j = 1 ; j <= o8nonlin ; j++ )
               {
                 constr[j] = o8fu [j] ;
                 errliste[j] = o8confuerr[j] ;
               }
            }
            else
            {
               for ( j = 1 ; j <= liste[0] ; j++ )
               {
                 constr[liste[j]] = o8fu[liste[j]] ;
                 errliste[liste[j]] = o8confuerr[liste[j]] ;
               }
            }
        }
        else
        {
            Rprintf("donlp2: o8bloc call with function info invalid\n");
            exit(1);
        }   
     } 
     else 
     {
        for (j = 1 ; j <= o8n ; j++) 
        {
            o8xtr[j] = o8x[j]*o8xsc[j];
        }
        econ(call_type,liste,o8xtr,constr,errliste);
     } 
     return;  
}     
void  escongrad(INTEGER liste[], INTEGER shift , DOUBLE o8x[],
 DOUBLE **grad_constr)
/********************************************************************/
/*  evaluate gradients of nonlinear constraints from liste for the  */
/*  internally scaled o8x                                             */     
/********************************************************************/
{
void  econ(INTEGER call_type,INTEGER liste[],DOUBLE  o8xtr[],DOUBLE constr[],
            LOGICAL errliste[]);
void  econgrad(INTEGER liste[], INTEGER shift , DOUBLE o8x[],
            DOUBLE **grad_constr);
	    
/* econgrad : the gradient for the external unscaled o8x */
static INTEGER i,j   ;
static DOUBLE   d1,d2,d3,sd1,sd2,sd3,
                xincr,xhelp;

/*  after rescaling o8x , evaluate the gradients of the nonlinear constraints */
/*  given by liste                                                          */

    if ( o8bloc ) 
    {
        if ( o8valid ) 
        {
           for ( i = 1 ; i <= liste[0] ; i++ )
           {
             for (j = 1 ; j <= o8n ; j++) 
             {
                grad_constr[j][shift+liste[i]] = o8xsc[j]*o8fugrad[j][liste[i]];
             }
           }
           return;
        } 
        else 
        {
            Rprintf("donlp2: o8bloc call with function info invalid\n");
            exit(1);
        }
    } 
    else 
    {
       for (j = 1 ; j <= o8n ; j++) 
       {
          o8xtr[j] = o8x[j]*o8xsc[j];   
       }
       if ( o8analyt ) 
       {
        
            econgrad(liste,shift,o8xtr,grad_constr);
            for ( j = 1 ; j <= o8n ; j++ )
            {
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                grad_constr[j][shift+liste[i]] *= o8xsc[j] ;
              }
            }        
       } 
       else 
       {
          if ( o8difftype == 1 ) 
          {
            o8deldif = min(o8tm1*sqrt(o8epsfcn),o8tm2);
            econ(2,liste,o8xtr,escongrad_fhelp1,escongrad_errloc);
            for ( i = 1 ; i <= liste[0] ; i++ )
            {
              if ( escongrad_errloc[liste[i]] )
              {
                Rprintf("donlp2: error in evaluating \n");
                Rprintf("nonlinear user function %i \n",liste[i]);
                Rprintf("during numerical differentiation \n");
                exit(1);
              }
            }            
            for (j = 1 ; j <= o8n ; j++) 
            {
              xhelp = o8xtr[j];
              xincr = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd);
              if ( xhelp >= o8zero ) 
              {   
                 o8xtr[j] = xhelp+xincr;
              } 
              else 
              {
                 o8xtr[j] = xhelp-xincr;
              }

              econ(2,liste,o8xtr,escongrad_fhelp2,escongrad_errloc);  
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");
                  exit(1);
                }
              }
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                grad_constr[j][shift+liste[i]] = 
                   o8xsc[j]*(escongrad_fhelp2[liste[i]]-escongrad_fhelp1[liste[i]])/(o8xtr[j]-xhelp);
              }
              o8xtr[j] = xhelp ;
            }
          }
          else if ( o8difftype == 2 )
          {
            o8deldif = min(o8tm1*pow(o8epsfcn,o8one/o8three),o8tm2);
            for (j = 1 ; j <= o8n ; j++) 
            {
              xhelp  = o8xtr[j];
              xincr  = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd);
              o8xtr[j] = xhelp+xincr;
                   
              econ(2,liste,o8xtr,escongrad_fhelp1,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");
                  exit(1);
                }
              }
                    
              o8xtr[j] = xhelp-xincr;
                    
              econ(2,liste,o8xtr,escongrad_fhelp2,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");
                  exit(1);
                }
              }

              for ( i = 1 ; i <= liste[0] ; i++)
              {      
                grad_constr[j][shift+liste[i]] = 
                    o8xsc[j]*(escongrad_fhelp1[liste[i]]-escongrad_fhelp2[liste[i]])/(xincr+xincr);
              }
              o8xtr[j]    = xhelp;
            }
          }
          else
          {
            o8deldif = min(o8tm1*pow(o8epsfcn,o8one/o8seven),o8tm2);
            for (j = 1 ; j <= o8n ; j++) 
            {
              xhelp  = o8xtr[j];
              xincr  = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd/o8four);
              o8xtr[j] = xhelp-xincr;
                    
              econ(2,liste,o8xtr,escongrad_fhelp1,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");   
                  exit(1);
                }
              }  

              o8xtr[j] = xhelp+xincr;
              econ(2,liste,o8xtr,escongrad_fhelp2,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");   
                  exit(1);
                }
              }  
                    
              xincr  = xincr+xincr;
              d1     = xincr;
              o8xtr[j] = xhelp-xincr;
              econ(2,liste,o8xtr,escongrad_fhelp3,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");   
                  exit(1);
                }
              }  

              o8xtr[j] = xhelp+xincr;
              econ(2,liste,o8xtr,escongrad_fhelp4,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");   
                  exit(1);
                }
              }  
                                        
              xincr  = xincr+xincr;
              d2     = xincr;
              o8xtr[j] = xhelp-xincr;
                    
              econ(2,liste,o8xtr,escongrad_fhelp5,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");   
                  exit(1);
                }
              }  
                                        
              o8xtr[j] = xhelp+xincr;
              econ(2,liste,o8xtr,escongrad_fhelp6,escongrad_errloc);
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                if ( escongrad_errloc[liste[i]] )
                {
                  Rprintf("donlp2: error in evaluating \n");
                  Rprintf("nonlinear user function %i \n",liste[i]);
                  Rprintf("during numerical differentiation \n");   
                  exit(1);
                }
              }  

              o8xtr[j]    = xhelp;
              d3        = xincr+xincr;
              for ( i = 1 ; i <= liste[0] ; i++ )
              {
                 sd1       = (escongrad_fhelp2[liste[i]]-escongrad_fhelp1[liste[i]])/d1;
                 sd2       = (escongrad_fhelp4[liste[i]]-escongrad_fhelp3[liste[i]])/d2;
                 sd3       = (escongrad_fhelp6[liste[i]]-escongrad_fhelp5[liste[i]])/d3;
                 sd3       = sd2-sd3;
                 sd2       = sd1-sd2;
                 sd3       = sd2-sd3; 
                 grad_constr[j][shift+liste[i]] =
                 o8xsc[j]*(sd1+o8p4*sd2+sd3/o8c45);
              }  /* end for i */
            } /* end for j */
          }
       } /* end if o8analyt else */
    } /* end if o8bloc else */
    return;
}
/* **************************************************************************** */
/* suite of functions to allocate and free memory for dynamic memory allocation */
/* **************************************************************************** */

#ifdef ARRAY_BORDER_CHECK
/**********************************************************************/
/***           check 1D array borders for any violations        *******/
/**********************************************************************/
void checkArrayBorders(char *str) 
{
    INTEGER i, violations;
    violations = 0;

    /* Check integer borders */
    for (i=0; i<o8i_borders_count; i++) {
        if (**(o8i_borders+2*i) != o8i_unique) {
            Rprintf("checkArrayBorders: Integer 1D array index %4d:  start border violated\n", i);  
            fflush(stdout);
            violations++;
        }
        if (**(o8i_borders+2*i+1) != o8i_unique) {
            Rprintf(" checkArrayBorders: Integer 1D array index %4d:  end border violated\n", i);  
            fflush(stdout);
            violations++;
        }
    }
    
    /* Check double borders */
    for (i=0; i<o8d_border_count; i++) {
        if (**(o8d_border+2*i) != o8d_unique) {
            Rprintf("checkArrayBorders: Double  1D array index %4d:  start border violated\n", i);  
            fflush(stdout);
            violations++;
        }
        if (**(o8d_border+2*i+1) != o8d_unique) {
            Rprintf("checkArrayBorders: Double  1D array index %4d:  end border violated\n", i);  
            fflush(stdout);
            violations++;
        }
    }
    
    /* Check logical borders */
    for (i=0; i<o8l_borders_count; i++) {
        if (**(o8l_borders+2*i) != o8l_unique) {
            Rprintf(" checkArrayBorders: Logical 1D array index %4d:  start border violated\n", i);  
            fflush(stdout);
            violations++;
        }
        if (**(o8l_borders+2*i+1) != o8l_unique) {
            Rprintf(" checkArrayBorders: Logical 1D array index %4d:  end border violated\n", i);  
            fflush(stdout);
            violations++;
        }
    }

    if (violations == 0) 
        Rprintf("checkArrayBorders: no violations: %s\n", str);
    else
        Rprintf("checkArrayBorders: ************** violations %4d:  %s\n", violations, str);
    fflush(stdout);
}
#endif


/**********************************************************************/
/***              allocate memory for integer 1D array          *******/
/**********************************************************************/

INTEGER* i1_malloc(INTEGER size1, INTEGER init)
{

    /* Allocate INTEGER precision array with length "size1"          */
    /* assuming o8zero origin.  If initialize is non o8zero             */
    /* then initize the allocated array to o8zero.                    */   

    INTEGER* array;
    INTEGER i,j;
    
    array = (INTEGER*) S_alloc((size_t) ((size1+2*ABC)),(sizeof(INTEGER)));
    if (!array) {
        printf( "ERROR: i1_malloc: memory error: malloc failed");  
        exit(-1);
    }

#ifdef ARRAY_BORDER_CHECK
    /* setup array borders */
    if (o8i_borders_count > 2*ABC_NUM_1DARRAYS-4) {
        Rprintf("ERROR: ARRAYBORDERS: ABC_NUM_1DARRAYS is too small\n");
	exit(-1);
    }	
    array[0] = o8i_unique;
    o8i_borders[2*o8i_borders_count] = &(array[0]);
    array[size1+1] = o8i_unique;
    o8i_borders[2*o8i_borders_count+1] = &(array[size1+1]);
    o8i_borders_count++;
    array++;
#endif    

    /* initialize the array to 0 */
    if (init) {
        for (i=0; i<size1; i++) 
            array[i] = 0;
    }

    return array;
}


/**********************************************************************/
/***                free memory for INTEGER 1D array             *******/
/**********************************************************************/
void i1_free(INTEGER* array) 
{
    INTEGER i;

    /* Check for null pointer */
    if (!array) {
        printf( "ERROR: i1_free: memory error: pointer is null");  
        exit(-1);
    }

    /* Free the memory for the INTEGER 1D array */
    free(array-ABC);
}

/**********************************************************************/
/***              allocate memory for INTEGER 2D array           *******/
/**********************************************************************/
INTEGER** i2_malloc(INTEGER size1, INTEGER size2, INTEGER init) 
{

    /* Allocate INTEGER precision array with lengths "size1" and     */
    /* size2 assuming o8zero origin.  If initialize is non o8zero       */
    /* then initize the allocated array to o8zero.                    */   

    INTEGER** array;
    INTEGER* arraytemp;
    INTEGER i,j;
    
    array = (INTEGER**) S_alloc((size_t) (size1),(sizeof(INTEGER*)));
    if (!array) {
        printf( "ERROR: d2_malloc: memory error: malloc failed");  
        exit(-1);
    }
    for (i=0; i<size1; i++) {
        arraytemp = (INTEGER*) S_alloc((size_t) ((size2+2*ABC)),(sizeof(INTEGER)));
        if (!arraytemp) {  
            printf( "ERROR: d2_malloc: memory error: malloc failed");  
            exit(-1);
        }
	
#ifdef ARRAY_BORDER_CHECK
        /* setup array borders */
        if (o8i_borders_count > 2*ABC_NUM_1DARRAYS-4) {
            Rprintf("ERROR: ARRAY_BORDERS_CHECK: ABC_NUM_1DARRAYS is too small\n");
	    exit(-1);
        }	
        arraytemp[0] = o8i_unique;
        o8i_borders[2*o8i_borders_count] = &(arraytemp[0]);
        arraytemp[size2+1] = o8i_unique;
        o8i_borders[2*o8i_borders_count+1] = &(arraytemp[size2+1]);
        o8i_borders_count++;
#endif
        array[i] = arraytemp + ABC;
    }

    if (init) {
       for (i=0; i<size1; i++) 
           for (j=0; j<size2; j++) 
               array[i][j] = 0.0;
    }

    return array;
}

/**********************************************************************/
/***                free memory for INTEGER 2D array             *******/
/**********************************************************************/
void i2_free(INTEGER** array, INTEGER size1) 
{
    INTEGER i;

    /* Check for null pointer */
    if (!array) {
        printf( "ERROR: d2_free: memory error: pointer is null");  
        exit(-1);
    }

    /* Free the memory for the INTEGER 2D array piece by piece */
    for (i=0; i<size1; i++) 
        free(array[i]-ABC);
    free(array);
}


/**********************************************************************/
/***              allocate memory for double 1D array           *******/
/**********************************************************************/
DOUBLE* d1_malloc(INTEGER size1, INTEGER init) 
{

    /* Allocate DOUBLE precision array with length "size1"          */
    /* assuming o8zero origin.  If initialize is non o8zero             */
    /* then initize the allocated array to o8zero.                    */   

    DOUBLE* array;
    INTEGER i,j;
    
    array = (DOUBLE*) S_alloc((size_t) ((size1+2*ABC)),(sizeof(DOUBLE)));
    if (!array) {
        printf( "ERROR: d1_malloc: memory error: malloc failed");  
        exit(-1);
    }

#ifdef ARRAY_BORDER_CHECK
    /* setup array borders */
    if (o8d_border_count > 2*ABC_NUM_1DARRAYS-4) {
        Rprintf("ERROR: ARRAYBORDERS: ABC_NUM_1DARRAYS is too small\n");
	exit(-1);
    }	
    array[0] = o8d_unique;
    o8d_border[2*o8d_border_count] = &(array[0]);
    array[size1+1] = o8d_unique;
    o8d_border[2*o8d_border_count+1] = &(array[size1+1]);
    o8d_border_count++;
    array++;
#endif

    if (init) {
        for (i=0; i<size1; i++) 
            array[i] = 0.0;
    }

    return array;
}

/**********************************************************************/
/***                free memory for DOUBLE 1D array             *******/
/**********************************************************************/
void d1_free(DOUBLE* array) 
{
    INTEGER i;

    /* Check for null pointer */
    if (!array) {
        printf( "ERROR: d1_free: memory error: pointer is null");  
        exit(-1);
    }

    /* Free the memory for the DOUBLE 1D array */
    free(array-ABC); 
}

/**********************************************************************/
/***              allocate memory for DOUBLE 2D array           *******/
/**********************************************************************/
DOUBLE** d2_malloc(INTEGER size1, INTEGER size2, INTEGER init) 
{

    /* Allocate DOUBLE precision array with lengths "size1" and     */
    /* size2 assuming o8zero origin.  If initialize is non o8zero       */
    /* then initize the allocated array to o8zero.                    */   

    DOUBLE** array;
    DOUBLE* arraytemp;
    INTEGER i,j;
    
    array = (DOUBLE**) S_alloc((size_t) (size1),(sizeof(DOUBLE*)));
    if (!array) {
        printf( "ERROR: d2_malloc: memory error: malloc failed");  
        exit(-1);
    }
    for (i=0; i<size1; i++) {

        /* Allocate memory */
        arraytemp = (DOUBLE*) S_alloc((size_t) ((size2+2*ABC)),(sizeof(DOUBLE)));
        if (!arraytemp) {  
            printf( "ERROR: d2_malloc: memory error: malloc failed");  
            exit(-1);
        }

#ifdef ARRAY_BORDER_CHECK
        /* setup array borders */
        if (o8d_border_count > 2*ABC_NUM_1DARRAYS-4) {
            Rprintf("ERROR: ARRAY_BORDERS_CHECK: ABC_NUM_1DARRAYS is too small\n");
	    exit(-1);
        }	
        arraytemp[0] = o8d_unique;
        o8d_border[2*o8d_border_count] = &(arraytemp[0]);
        arraytemp[size2+1] = o8d_unique;
        o8d_border[2*o8d_border_count+1] = &(arraytemp[size2+1]);
        o8d_border_count++;
#endif
        array[i] = arraytemp + ABC;
    }

    if (init) {
       for (i=0; i<size1; i++) 
           for (j=0; j<size2; j++) 
               array[i][j] = 0.0;
    }

    return array;
}

/**********************************************************************/
/***                free memory for DOUBLE 2D array             *******/
/**********************************************************************/
void d2_free(DOUBLE** array, INTEGER size1) 
{
    INTEGER i;

    /* Check for null pointer */
    if (!array) {
        printf( "ERROR: d2_free: memory error: pointer is null");  
        exit(-1);
    }

    /* Free the memory for the DOUBLE 2D array piece by piece */
    for (i=0; i<size1; i++) 
        free(array[i]-ABC);
    free(array);
}

/**********************************************************************/
/***              allocate memory for LOGICAL 1D array           *******/
/**********************************************************************/
LOGICAL* l1_malloc(INTEGER size1, INTEGER init) 
{

    /* Allocate LOGICAL precision array with length "size1"          */
    /* assuming o8zero origin.  If initialize is non o8zero             */
    /* then initize the allocated array to o8zero.                    */   

    LOGICAL* array;
    INTEGER i,j;
    
    array = (LOGICAL*) S_alloc((size_t) ((size1+2*ABC)),(sizeof(LOGICAL)));
    if (!array) {
        printf( "ERROR: l1_malloc: memory error: malloc failed");  
        exit(-1);
    }

#ifdef ARRAY_BORDER_CHECK
    /* setup array borders */
    if (o8i_borders_count > 2*ABC_NUM_1DARRAYS-4) {
        Rprintf("ERROR: ARRAYBORDERS: ABC_NUM_1DARRAYS is too small\n");
	exit(-1);
    }	
    array[0] = o8i_unique;
    o8i_borders[2*o8i_borders_count] = &(array[0]);
    array[size1+1] = o8i_unique;
    o8i_borders[2*o8i_borders_count+1] = &(array[size1+1]);
    o8i_borders_count++;
    array++;
#endif

    if (init) {
        for (i=0; i<size1; i++) 
            array[i] = 0.0;
    }

    return array;
}

/**********************************************************************/
/***                free memory for LOGICAL 1D array             *******/
/**********************************************************************/
void l1_free(LOGICAL* array) 
{
    INTEGER i;

    /* Check for null pointer */
    if (!array) {
        printf( "ERROR: l1_free: memory error: pointer is null");  
        exit(-1);
    }

    /* Free the memory for the LOGICAL 1D array */
    free(array-ABC);
}

/**********************************************************************/
/***              allocate memory for LOGICAL 2D array           *******/
/**********************************************************************/
LOGICAL** l2_malloc(INTEGER size1, INTEGER size2, INTEGER init)
{

    /* Allocate LOGICAL precision array with lengths "size1" and     */
    /* size2 assuming o8zero origin.  If initialize is non o8zero       */
    /* then initize the allocated array to o8zero.                    */   

    LOGICAL** array;
    LOGICAL* arraytemp;
    INTEGER i,j;
    
    array = (LOGICAL**) S_alloc((size_t) (size1),(sizeof(LOGICAL*)));
    if (!array) {
        printf( "ERROR: l2_malloc: memory error: malloc failed");  
        exit(-1);
    }
    for (i=0; i<size1; i++) {
        arraytemp = (LOGICAL*) S_alloc((size_t) ((size2+2*ABC)),(sizeof(LOGICAL)));
        if (!arraytemp) {  
            printf( "ERROR: l2_malloc: memory error: malloc failed");  
            exit(-1);
        }
	
#ifdef ARRAY_BORDER_CHECK
        /* setup array borders */
        if (o8i_borders_count > 2*ABC_NUM_1DARRAYS-4) {
            Rprintf("ERROR: ARRAY_BORDERS_CHECK: ABC_NUM_1DARRAYS is too small\n");
	    exit(-1);
        }	
        arraytemp[0] = o8i_unique;
        o8i_borders[2*o8i_borders_count] = &(arraytemp[0]);
        arraytemp[size2+1] = o8i_unique;
        o8i_borders[2*o8i_borders_count+1] = &(arraytemp[size2+1]);
        o8i_borders_count++;
#endif
        array[i] = arraytemp + ABC;
    }

    if (init) {
       for (i=0; i<size1; i++) 
           for (j=0; j<size2; j++) 
               array[i][j] = 0.0;
    }

    return array;
}

/**********************************************************************/
/***                free memory for LOGICAL 2D array             *******/
/**********************************************************************/
void l2_free(LOGICAL** array, INTEGER size1) 
{
    INTEGER i;

    /* Check for null pointer */
    if (!array) {
        printf( "ERROR: l2_free: memory error: pointer is null");  
        exit(-1);
    }

    /* Free the memory for the LOGICAL 2D array piece by piece */
    for (i=0; i<size1; i++) 
        free(array[i]-ABC);
    free(array);
}


/**********************************************************************/
/***              allocate memory for global arrays            *******/
/**********************************************************************/
void global_mem_malloc(
		       long n, 
		       long nlin, 
		       long nonlin,
		       long iterma, 
		       long nstep, 
		       long ndualm, 
		       long mdualm
		       ) 
{


    /* o8comm.h */
    o8accinf     = d2_malloc(iterma+1, 33, 1);
    o8x          = d1_malloc(n+1, 1);
    o8x0         = d1_malloc(n+1, 1);
    o8x1         = d1_malloc(n+1, 1);
    o8xmin       = d1_malloc(n+1, 1);
    o8d          = d1_malloc(n+1, 1);
    o8d0         = d1_malloc(n+1, 1);
    o8dd         = d1_malloc(n+1, 1);
    o8difx       = d1_malloc(n+1, 1);
    o8resmin     = d1_malloc(2*(n+nlin+nonlin+1), 1);
    o8gradf      = d1_malloc(n+1, 1);
    o8qgf        = d1_malloc(n+1, 1); 
    o8gphi0      = d1_malloc(n+1, 1);
    o8gphi1      = d1_malloc(n+1, 1);
    o8gres       = d2_malloc(n+1, nlin+nonlin+1, 1);
    o8gresn      = d1_malloc(nlin+nonlin+1, 1);
    o8perm       = i1_malloc(n+1, 1);
    o8perm1      = i1_malloc(n+1, 1);
    o8colno      = i1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8qr         = d2_malloc(n+1, n+nlin+nonlin+1, 1);    
    o8betaq      = d1_malloc(n+nlin+nonlin+1, 1); 
    o8diag       = d1_malloc(n+nlin+nonlin+1, 1);
    o8cscal      = d1_malloc(n+nlin+nonlin+1, 1);
    o8colle      = d1_malloc(n+nlin+nonlin+1, 1);
    
    o8a          = d2_malloc(n+1, n+1, 1);
    o8diag0      = d1_malloc(n+1, 1);
    o8violis     = i1_malloc(2*nstep*(n+nlin+nonlin)+1, 1);
    o8alist      = i1_malloc(n+nlin+nonlin+1, 1);
    o8bind     = i1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8bind0    = i1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8aalist     = i1_malloc(n+nlin+nonlin+1, 1);
    o8clist      = i1_malloc(nlin+nonlin+1, 1);  
    o8u          = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8u0         = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8w          = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8w1         = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8res        = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8res0       = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8res1       = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8resst      = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8yu         = d1_malloc(n+nlin+nonlin+1, 1);
    o8slack      = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8work       = d1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8ug         = d1_malloc(n+1, 1);
    o8og         = d1_malloc(n+1, 1);
    o8low        = d1_malloc(n+nlin+nonlin+1, 1);
    o8up         = d1_malloc(n+nlin+nonlin+1, 1);
    o8xst        = d1_malloc(n+1, 1);
        

    /* o8fint.h */
    o8xtr        = d1_malloc(n+1, 1);
    o8xsc        = d1_malloc(n+1, 1);
    o8fu         = d1_malloc(nlin+nonlin+1, 1);
    o8fugrad     = d2_malloc(n+1, nlin+nonlin+1, 1);
    o8fud        = d2_malloc(nlin+nonlin+1, 7, 1);
    
    /* o8fuco.h */
    o8val        = l1_malloc(nlin+nonlin+1, 1);
    o8llow       = l1_malloc(n+1, 1);
    o8lup        = l1_malloc(n+1, 1);
    o8cres       = i1_malloc(nlin+nonlin+1, 1);
    o8cgres      = i1_malloc(nlin+nonlin+1, 1);
    o8confuerr   = l1_malloc(nlin+nonlin+1, 1);
    o8nonlinlist = i1_malloc(nlin+nonlin+1, 1);
    o8gunit      = i2_malloc(4, nlin+nonlin+1, 1);
    o8gconst     = l1_malloc(nlin+nonlin+1, 1);
    o8cfuerr     = l1_malloc(nlin+nonlin+1, 1);

    /* o8gene.h */
    o8np         = d1_malloc(ndualm+1, 1);
    o8xj         = d2_malloc(ndualm+1, ndualm+1, 1);
    o8ddual      = d1_malloc(ndualm+1, 1);
    o8r          = d2_malloc(ndualm+1, ndualm+1, 1);
    o8ud         = d1_malloc(mdualm+1, 1);
    o8ud1        = d1_malloc(mdualm+1, 1);
    o8aitr       = i1_malloc(mdualm+1, 1);
 
    
    /* GLOBAL VARIABLES USED LOCALLY */
    
    donlp2_yy     = d1_malloc(n+1, 1);
    o8bfgs_dg     = d1_malloc(n+1, 1); 
    o8bfgs_adx    = d1_malloc(n+1, 1);
    o8bfgs_ltdx   = d1_malloc(n+1, 1); 
    o8bfgs_gtdx   = d1_malloc(n+nlin+nonlin+1, 1);  
    o8bfgs_updx   = d1_malloc(n+1, 1);
    o8bfgs_updz   = d1_malloc(n+1, 1);
    o8opti_qtx    = d1_malloc(n+1, 1);
    o8opti_yy     = d1_malloc(n+1, 1);
    o8opti_yx     = d1_malloc(n+1, 1);
    o8opti_trvec  = d1_malloc(n+1, 1);
    o8opti_con    = d1_malloc(nlin+nonlin+1, 1);
    o8opti_delist = i1_malloc(n+nlin+nonlin+1, 1);
    o8opti_bindba = i1_malloc(2*(n+nlin+nonlin)+1, 1);
    o8eval_con    = d1_malloc(nlin+nonlin+1, 1);
    o8unim_step   = d1_malloc(nstep+1, 1);
    o8dec_qri     = d1_malloc(n+1, 1);
    o8dec_qri0    = d1_malloc(n+1, 1);
    o8elim_qri    = d1_malloc(n+1, 1);
    o8elim_y      = d1_malloc(n+1, 1);
    o8elim_rhsscal= d1_malloc(n+nlin+1, 1);
    o8elim_col    = i1_malloc(n+nlin+1, 1); 
    o8sol_xl      = d1_malloc(n+1, 1);
    o8upd_sdiag   = d1_malloc(n+1, 1);    
    o8upd_rn1     = d1_malloc(n+1, 1);
    o8upd_w       = d1_malloc(n+1, 1); 
    o8qpdu_g0     = d1_malloc(ndualm+1, 1); 
    o8qpdu_ci0    = d1_malloc(mdualm+1, 1); 
    o8qpdu_cii    = d1_malloc(mdualm+1, 1); 
    o8qpdu_cei    = d1_malloc(ndualm+1, 1); 
    o8qpdu_xd     = d1_malloc(ndualm+1, 1); 
    o8qpdu_s      = d1_malloc(mdualm+1, 1); 
    o8qpdu_z      = d1_malloc(ndualm+1, 1); 
    o8qpdu_vr     = d1_malloc(mdualm+1, 1); 
    o8qpdu_xdold  = d1_malloc(ndualm+1, 1); 
    o8qpdu_udold  = d1_malloc(mdualm+1, 1); 
    o8qpdu_ai     = i1_malloc(mdualm+1, 1);
    o8qpdu_iai    = i1_malloc(mdualm+1, 1);  
    o8qpdu_iaexcl = i1_malloc(mdualm+1, 1);
    o8qpdu_aiold  = i1_malloc(mdualm+1, 1);
    o8qpdu_eqlist = i1_malloc(n+1, 1);
    o8qpdu_iqlist = i1_malloc(mdualm+1, 1);
    o8qpdu_y      = d1_malloc(ndualm+1, 1);
    o8qpdu_mult   = d1_malloc(n+nlin+nonlin+1, 1);
    o8qpdu_qpdel  = i1_malloc(mdualm+1, 1);
    escongrad_errloc = i1_malloc(nlin+nonlin+1, 1);
    escongrad_fhelp1 = d1_malloc(nlin+nonlin+1, 1);
    escongrad_fhelp2 = d1_malloc(nlin+nonlin+1, 1);
    escongrad_fhelp3 = d1_malloc(nlin+nonlin+1, 1);
    escongrad_fhelp4 = d1_malloc(nlin+nonlin+1, 1);
    escongrad_fhelp5 = d1_malloc(nlin+nonlin+1, 1);
    escongrad_fhelp6 = d1_malloc(nlin+nonlin+1, 1);

}

/**********************************************************************/
/***                free memory from global arrays              *******/
/**********************************************************************/
void global_mem_free(
		     long n, 
		     long nlin, 
		     long nonlin,
		     long iterma, 
		     long ndualm, 
		     long mdualm
		     ) 
{
     
    /* GLOBAL VARIABLES USED GLOBALLY */

    /* o8comm.h */
    d2_free(o8accinf, iterma+1);
    d1_free(o8x);
    d1_free(o8x0);
    d1_free(o8x1);
    d1_free(o8xmin);
    d1_free(o8d);
    d1_free(o8d0);
    d1_free(o8dd);
    d1_free(o8difx);
    d1_free(o8resmin);
    d1_free(o8gradf);
    d1_free(o8qgf);
    d1_free(o8gphi0);
    d1_free(o8gphi1); 
    d2_free(o8gres, n+1);
    d1_free(o8gresn); 
    i1_free(o8perm);  
    i1_free(o8perm1); 
    i1_free(o8colno); 
    d2_free(o8qr, n+1);
    d1_free(o8betaq);
    d1_free(o8diag); 
    d1_free(o8cscal);
    d1_free(o8colle);
    
    d2_free(o8a, n+1); 
    d1_free(o8diag0);  
    i1_free(o8violis); 
    i1_free(o8alist);  
    i1_free(o8bind);   
    i1_free(o8bind0);  
    i1_free(o8aalist); 
    i1_free(o8clist);  
    d1_free(o8u);      
    d1_free(o8u0);     
    d1_free(o8w);      
    d1_free(o8w1);     
    d1_free(o8res);    
    d1_free(o8res0);   
    d1_free(o8res1);   
    d1_free(o8resst);  
    d1_free(o8yu);     
    d1_free(o8slack); 
    d1_free(o8work);  
    d1_free(o8ug);    
    d1_free(o8og);    
    d1_free(o8low);   
    d1_free(o8up);    
    d1_free(o8xst);   
        

    /* o8fint.h */
    d1_free(o8xtr);
    d1_free(o8xsc);
    d1_free(o8fu);
    d2_free(o8fugrad, n+1);
    d2_free(o8fud, nlin+nonlin+1);
    
    /* o8fuco.h */
    l1_free(o8val); 
    l1_free(o8llow);
    l1_free(o8lup); 
    i1_free(o8cres);
    i1_free(o8cgres);
    l1_free(o8confuerr);
    i1_free(o8nonlinlist);
    i2_free(o8gunit, 4);
    l1_free(o8gconst);
    l1_free(o8cfuerr);

    /* o8gene.h */
    d1_free(o8np);
    d2_free(o8xj, ndualm+1);
    d1_free(o8ddual);
    d2_free(o8r, ndualm+1);
    d1_free(o8ud);
    d1_free(o8ud1);
    i1_free(o8aitr); 
    
    /* GLOBAL VARIABLES USED LOCALLY */
    d1_free(donlp2_yy); 
    d1_free(o8bfgs_dg); 
    d1_free(o8bfgs_adx);
    d1_free(o8bfgs_ltdx);
    d1_free(o8bfgs_gtdx); 
    d1_free(o8bfgs_updx); 
    d1_free(o8bfgs_updz); 
    d1_free(o8opti_qtx); 
    d1_free(o8opti_yy);  
    d1_free(o8opti_yx);  
    d1_free(o8opti_trvec); 
    d1_free(o8opti_con);  
    i1_free(o8opti_delist); 
    i1_free(o8opti_bindba); 
    d1_free(o8eval_con);   
    d1_free(o8unim_step);  
    d1_free(o8dec_qri);    
    d1_free(o8dec_qri0);
    d1_free(o8elim_qri);
    d1_free(o8elim_y);
    d1_free(o8elim_rhsscal);
    i1_free(o8elim_col);   
    d1_free(o8sol_xl);  
    d1_free(o8upd_sdiag);
    d1_free(o8upd_rn1);  
    d1_free(o8upd_w);  
    d1_free(o8qpdu_g0);  
    d1_free(o8qpdu_ci0); 
    d1_free(o8qpdu_cii); 
    d1_free(o8qpdu_cei); 
    d1_free(o8qpdu_xd);  
    d1_free(o8qpdu_s); 
    d1_free(o8qpdu_z); 
    d1_free(o8qpdu_vr);  
    d1_free(o8qpdu_xdold); 
    d1_free(o8qpdu_udold); 
    i1_free(o8qpdu_ai);    
    i1_free(o8qpdu_iai);   
    i1_free(o8qpdu_iaexcl);
    i1_free(o8qpdu_aiold); 
    i1_free(o8qpdu_eqlist);
    i1_free(o8qpdu_iqlist);
    d1_free(o8qpdu_y); 
    d1_free(o8qpdu_mult); 
    i1_free(o8qpdu_qpdel); 
    i1_free(escongrad_errloc);
    d1_free(escongrad_fhelp1);
    d1_free(escongrad_fhelp2);
    d1_free(escongrad_fhelp3);
    d1_free(escongrad_fhelp4);
    d1_free(escongrad_fhelp5);
    d1_free(escongrad_fhelp6);
}


   
