##################
##
## cosmo function
##
##################
cosmo <- function(seqs="browse", constraints="None", minW=6, maxW=15,
                  models = "ZOOPS", revComp=TRUE, minSites=NULL,
                  maxSites=NULL, starts=5, approx="over", cutFac=5,
                  wCrit="bic", wFold=5, wTrunc=100, modCrit="lik",
                  modFold=5, modTrunc=100, conCrit="likCV", conFold=5,
                  conTrunc=90, intCrit="lik", intFold=5,
                  intTrunc=100, maxIntensity=FALSE, lstarts=FALSE,
                  backSeqs=NULL, backFold=5, bfile=NULL, transMat=NULL,
                  order=NULL, maxOrder=6, silent=FALSE) { 


    ## get sequences in matrix format
    seqIn   <- getSequences(seqs)
    seqMat  <- seqIn$seqMat
    desc    <- seqIn$descList
    seqList <- seqIn$seqList
    
    if (!is.matrix(seqMat) || !is.numeric(seqMat)){
      stop("seqMat must be a numeric matrix")}
    if (!all(seqMat %in% 0:5)){
      stop("All elements of seqMat must be in {0, 1, 2, 3, 4, 5}")}

    numSeqs          <- ncol(seqMat)
    maxSeqLength     <- nrow(seqMat)
    
    ## get constraints in constraintGroup format
    empty <-  FALSE
    if (is.character(constraints)){
      if (constraints == "GUI") {

        if (!require(cosmoGUI))
          stop("To use the 'GUI' option, the cosmoGUI package needs to be installed.")
        
        ## invoke constraint GUI
        cons <- constraintBuilder()

        if (is.null(cons)){
          paste("No valid constraint set was constructed\n")
          return(NULL)
        }

      }else if (constraints == "None"){
        ## create empty constaint set
        cons <- makeConSet(1,"V",NA)
        empty <- TRUE

      }else{        
        ## read in constraints from file        
        cons <- readConFile(constraints)
      }

    }else if (class(constraints) == "list"){
      if (!all(unlist(lapply(constraints, function(x) class(x)[1])) ==
               "constraintSet")){
        stop("If constraints is a list, all elements must be of class constraintSet.")
      }
      cons <- constraints
    }else if (class(constraints) == "constraintSet"){
      cons <- constraints
    }else{
      ## wrong class
      stop("Wrong class for constraints")
    }    

    if (!is.list(cons)) cons <- list(cons)
    
    if (empty){
      numConSets <- 0
      constraints <- NULL
    }else{
      numConSets <- length(cons)
        
      ## now output constraintGroup as string
      if (!is.null(cons)) constraints <- writeConFile(cons, stringOutput=TRUE) 
    }
    
    ## get back sequences in matrix format
    if (!is.null(backSeqs)){
      backIn <- getSequences(backSeqs)
      backSeqs <- backIn$seqMat
      
      if (!is.matrix(backSeqs) ||
          !is.numeric(backSeqs)) stop("backSeqs must be a numeric matrix")
      if (!all(backSeqs %in% 0:5))
          stop("All elements of backSeqs must be in {0, 1, 2, 3, 4, 5}")
    }
    numBackSeqs      <- ifelse(is.null(backSeqs), 0, ncol(backSeqs))
    maxBackSeqLength <- ifelse(is.null(backSeqs), 0, nrow(backSeqs))
    backFold         <- ifelse(!is.null(backSeqs) &&
                               backFold>numBackSeqs,
                               numBackSeqs, backFold)

    ## check remaining arguments
    if (maxSeqLength < maxW){
      stop("maxW cannot be greater than the length of the longest sequence")}
    
    if (minW < 2 || minW > 30){
      stop("minW must be at least 2 and no greater than 30")}
    if (maxW < 2 || maxW > 30){
      stop("maxW must be at least 2 and no greater than 30")}
    if (minW > maxW){
      stop("minW cannot be greater than maxW")}

    if (!approx %in% c("over", "cut", "exact")){
      stop("approx must be one of: over, cut, exact")}
    if (!wCrit %in% c("lik", "aic", "bic", "eval", "likCV")){
      stop("wCrit must be one of: lik, aic, bic, eval, likCV")
    }
    if (!modCrit %in% c("lik", "aic", "bic", "eval", "likCV")){
      stop("modCrit must be one of: lik, aic, bic, eval, likCV")
    }
    if (!conCrit %in% c("lik", "eval", "likCV", "pwmCV")){
      stop("conCrit must be one of: lik, eval, likCV, pwmCV")
    }
    if (conCrit == "pwmCV") conFold <- 2
    if (!intCrit %in% c("lik", "eval", "likCV")){
      stop("intCrit must be one of: lik, eval, likCV")
    }

    
    if (wTrunc < 1   || wTrunc>100)   stop("wTrunc must be between 1 and 100")
    if (modTrunc < 1 || modTrunc>100) stop("modTrunc must be between 1 and 100")
    if (conTrunc < 1 || conTrunc>100) stop("conTrunc must be between 1 and 100")
    if (intTrunc < 1 || intTrunc>100) stop("intTrunc must be between 1 and 100")

    
    if (wFold < 2)   stop("wFold must be at least 2")
    if (modFold < 2) stop("modFold must be at least 2")
    if (conFold < 2) stop("conFold must be at least 2")
    if (intFold < 2) stop("intFold must be at least 2")
    if (backFold < 2) stop("backFold must be at least 2")

    if (!is.logical(maxIntensity)) stop("maxIntensity must be a logical")
    
    if (!is.null(order) && (order<0 || order>6)) stop("order must be between 0 and 6")
    if (starts<1 || starts>100) stop("starts must be between 1 and 100")
    if (!is.logical(lstarts)) stop("lstarts must be a logical")

    if (length(intersect(models, c("OOPS", "ZOOPS", "TCM"))) == 0)
      stop("models must contain at least one of: OOPS, ZOOPS, TCM")
    if (any(!models%in% c("OOPS", "ZOOPS", "TCM")))
      stop("models may only contain: OOPS, ZOOPS, TCM")

    oops  <- is.element("OOPS", models)
    zoops <- is.element("ZOOPS", models)
    tcm <- is.element("TCM", models)

    if (maxOrder<=0 || maxOrder >6)
      stop("maxOrder must be between 0 and 6")

    if ((!is.null(transMat)) && (!is.null(bfile)))
      stop("Only one of transMat and bfile may be given")

    if ((!is.null(order)) && (!is.null(transMat))){
      warning(paste("Specified order for background model of", order, "will be ignored since transition matrix is given."))
    }

    if ((!is.null(order)) && (!is.null(bfile))){
      warning(paste("Specified order for background model of", order, "will be ignored since background file is given."))
    }
        
    if (!is.logical(silent)) stop("silent must be a logical")
    msgLevel <- ifelse(silent, 0, 2)

    ## extract transition matrices out of list and concatenate
    ## them in one big matrix, then take the transpose
    ## for compatibility with C
    tM <- NULL
    if (!is.null(bfile)) transMat <- bfile2tmat(bfile)
    if (!is.null(transMat)){

      if (!is.list(transMat)) stop("transMat must be a list")
      
      for (i in 1:length(transMat)){
        if (!is.matrix(transMat[[i]]) || !is.numeric(transMat[[i]]) || (ncol(transMat[[i]])!=4))
          stop("The entries of the list transMat must be numeric matrices with four columns")
      }
      
      for (i in 1:length(transMat)) tM <- rbind(tM, transMat[[i]])
      
      if (!is.matrix(tM) || !is.numeric(tM)) stop("transMat must be a numeric matrix")
      totals <- apply(tM, 1, sum)
      if (!all(abs(totals-1)<1e-8)) stop("All rows of transMat must add up to 1.0")
      correctRows <- FALSE
      numRows <- 0
      for (i in 0:6){
        numRows <- numRows + 4^i
        if (nrow(tM) == numRows){
          correctRows <- TRUE
          order <- i
        }
      }
      if (!correctRows){
        stop("transMat must have number of rows equal to sum_{j=0}^k 4^j for k=0, 1, 2,..., or 6")}
      tM <- t(tM)

    }


    
    ## initialize output variables
    selOrder      <- 0
    selWidth      <- 0
    selModel      <- 0
    selConSet     <- 0
    selNumSites   <- 0
    numModels     <- 0
    eval          <- 0
    selIntCrit    <- 0 
    success       <- 1

    klDivs        <- rep(0, maxOrder+1)
    
    ncells <- 0
    for (i in 0:maxOrder) ncells <- ncells + 4^(i+1)
    estTransMat   <- rep(0, ncells)
    
    ## add one constraint set for the empty constraint set
    numW          <- maxW - minW + 1
    numMods       <- oops + zoops + tcm
    wCrits        <- rep(0, (numConSets+1)*numW*numMods)
    modCrits      <- rep(0, (numConSets+1)*numW*numMods)
    conCrits      <- rep(0, (numConSets+1)*numW*numMods)  
    wInit         <- rep(0, (numConSets+1)*numW*numMods)
    modInit       <- rep(0, (numConSets+1)*numW*numMods)
    conInit       <- rep(0, (numConSets+1)*numW*numMods)  
    
    pwm           <- rep(0, 4*maxW)
    motProbs      <- rep(0, numSeqs*maxSeqLength)
    motRevComp    <- rep(0, numSeqs*maxSeqLength)
    alignSeqs     <- rep(0, numSeqs*maxSeqLength)
    alignStarts   <- rep(0, numSeqs*maxSeqLength)
    
    ## call C function
    com <- .C("ccosmo",
      as.integer(seqMat),
      as.integer(numSeqs),
      as.integer(maxSeqLength),
      as.character(constraints),
      as.integer(minW),
      as.integer(maxW),
      as.integer(revComp),
      as.integer(minSites),
      as.integer(maxSites),
      as.integer(oops),
      as.integer(zoops),
      as.integer(tcm),
      as.character(approx),
      as.integer(cutFac),
      as.character(wCrit),
      as.integer(wFold),
      as.integer(wTrunc),
      as.character(modCrit),
      as.integer(modFold),
      as.integer(modTrunc),
      as.character(conCrit),
      as.integer(conFold),
      as.integer(conTrunc),
      as.character(intCrit),
      as.integer(intFold),
      as.integer(intTrunc),
      as.integer(maxIntensity),
      as.integer(starts),
      as.integer(lstarts),
      as.integer(backSeqs),
      as.integer(numBackSeqs),
      as.integer(maxBackSeqLength),
      as.integer(backFold),
      as.double(tM),
      as.integer(order),
      as.integer(maxOrder),
      selOrder=as.integer(selOrder),
      klDivs=as.double(klDivs),
      estTransMat=as.double(estTransMat),
      selIntCrit=as.double(selIntCrit),
      wCrits=as.double(wCrits),
      modCrits=as.double(modCrits),
      conCrits=as.double(conCrits),    
      wInit=as.integer(wInit),
      modInit=as.integer(modInit),
      conInit=as.integer(conInit),    
      selWidth=as.integer(selWidth),
      selModel=as.integer(selModel),
      selConSet=as.integer(selConSet),
      selNumSites=as.integer(selNumSites),
      numModels=as.integer(numModels),
      pwm=as.double(pwm),
      motProbs=as.double(motProbs),
      motRevComp=as.integer(motRevComp),
      alignSeqs=as.integer(alignSeqs),
      alignStarts=as.integer(alignStarts),
      eval=as.double(eval),
      success=as.integer(success),
      as.integer(msgLevel)
      )
    
    if (!com$success) return(NULL)
    
    ## prepare output
    out.klDivs      <- com$klDivs
    out.klDivs[out.klDivs == 0] <- Inf
    out.klDivs[out.klDivs == -1] <- NA
    out.numModels   <- com$numModels
    out.wCrits      <- com$wCrits[1:out.numModels]
    out.modCrits    <- com$modCrits[1:out.numModels]
    out.conCrits    <- com$conCrits[1:out.numModels]
    out.wInit       <- as.logical(com$wInit[1:out.numModels])
    out.modInit     <- as.logical(com$modInit[1:out.numModels])
    out.conInit     <- as.logical(com$conInit[1:out.numModels])
    out.wCrits[!out.wInit] <- NA
    out.modCrits[!out.modInit] <- NA
    out.conCrits[!out.conInit] <- NA
    out.conSetNum   <- com$selConSet
    out.pwm         <- matrix(com$pwm[1:(4*com$selWidth)], 4, com$selWidth)
    out.motProbs    <- matrix(com$motProbs, maxSeqLength, numSeqs)
    out.motOnRC     <- matrix(as.logical(com$motRevComp), maxSeqLength, numSeqs)
    out.alignSeqs   <- com$alignSeqs[1:com$selNumSites]
    out.alignStarts <- com$alignStarts[1:com$selNumSites]
    out.modelType   <- c("OOPS", "ZOOPS", "TCM")[com$selModel]

    ## extract estimated transition matrices
    ## and create list from them
    first <- 1
    tMs <- list()
    transMat <- com$estTransMat
    dnaLetters <- c("A", "C", "G", "T")
    for (suborder in 0:com$selOrder){
      tM <- matrix(transMat[first:(first+4*4^suborder-1)], 4^suborder, 4, byrow=TRUE)
      colnames(tM) <- dnaLetters
      if (suborder == 0){
        rnames <- "--"
      }else{
        rnames <- dnaLetters
        if (suborder>1){
          for (i in 2:suborder){
            nrnames <- NULL
            for (j in 1:length(rnames)){
              for (k in 1:4) nrnames <- c(nrnames, paste(rnames[j],dnaLetters[k],sep=""))
            }
            rnames <- nrnames
          }
        }
      }
      rownames(tM) <- rnames
            
      first <- first + 4*4^suborder
      tMs <- c(tMs, mat=list(tM)) 
    }
    names(tMs) <- paste("order", 0:com$selOrder, sep="")

    ## create PWM object
    pwm <- makePWM(out.pwm)
    
    ## create alignment of predicted sites
    seqNames <- unlist(desc[com$alignSeqs])
    sites <- NULL
    for (i in 1:com$selNumSites){
      site <- seqMat[com$alignStarts[i]:(com$alignStarts[i]+com$selWidth-1),
                     com$alignSeqs[i]]
      sites <- c(sites, toString(site))
    }
    strands <- out.motOnRC[cbind(out.alignStarts,out.alignSeqs)]
    strands <- ifelse(strands, -1, 1)    
    probs <- out.motProbs[cbind(out.alignStarts,out.alignSeqs)]

    motifs <- new("align",seq=seqNames, pos=out.alignStarts, orient=strands,
                  motif=sites, prob=probs,eval=com$eval)

    ## create list of sequences
    seqs <- list()
    seqLengths <- NULL
    for (i in 1:length(seqList)){
      seqs[[i]] <- list(seq=seqList[[i]], desc=desc[[i]]) 
      seqLengths[i] <- nchar(seqList[[i]])
    }

    
    ## create data.frame of candidate models
    if (empty) out.conSetNum <- 1
    conSet <- cons[[out.conSetNum]]

    if (numSeqs == 1){
      modTypes <- "TCM"
    }else{
      modTypes <- models
    }
    numModTypes <- length(modTypes)
    
    consets <- NULL
    models <- NULL
    widths <- NULL
    upper <- ifelse(empty, 1, numConSets)
    for (c in 1:upper){
      for (j in 1:numModTypes){
        modType <- modTypes[j]
        for (w in minW:maxW){
          consets <- c(consets, c)
          models <- c(models, modType)
          widths <- c(widths, w)
        }
      }
    }
    cand <- data.frame(conSet=consets, model=models, width=widths,
                              wCrit=out.wCrits, modCrit=out.modCrits, conCrit=out.conCrits)

    ## create date.frame of selected model
    choices <- c(out.conSetNum, out.modelType, com$selWidth, com$selNumSites, com$selOrder)
    crits <- c(conCrit, modCrit, wCrit, intCrit, "likCV")

    cand.row <- (cand$conSet == out.conSetNum) &
                (cand$model  == out.modelType) &
                (cand$width  == com$selWidth)
    if (all(cand.row==FALSE)) stop("Couldn't find selected model in candidates data.frame")
    
    cand.row <- (1:nrow(cand))[cand.row]
    critVals <- c(rev(unlist(cand[cand.row, 4:6])), com$selIntCrit, out.klDivs[com$selOrder+1])    

    sel <- data.frame(choice=choices, crit=crits, critVal=critVals,
                            row.names=c("Constraint", "Model", "Width", "NumSites", "Markov Order"))
    sel[,1] <- as.character(sel[,1])
    

    ## create data.frame of candidate orders
    orders <- 0:(length(out.klDivs)-1)
    back=data.frame(order=orders,klDiv=out.klDivs)

    ## create postProbs
    motProbs <- ifelse(out.motOnRC, -out.motProbs, out.motProbs)
    probs <- list()
    for (i in 1:ncol(motProbs)){
      probs[[i]] <- motProbs[1:seqLengths[i],i]
    }
    
    out <- new("cosmo", seqs=seqs, pwm=pwm, back=back,
               cand=cand, cons=conSet, sel=sel, motifs=motifs,
               probs=probs, tmat=tMs, objectCall=match.call())   
    

    out
    
  }


bgModel <- function(seqs, order=NULL, fold=5, maxOrder=6) {   
  
  seqIn <- getSequences(seqs)
  mySeqs <- seqIn$seqMat
      
  if (!is.matrix(mySeqs) ||
      !is.numeric(mySeqs)) stop("mySeqs must be a numeric matrix")
  if (!all(mySeqs %in% 0:5))
    stop("All elements of mySeqs must be in {0, 1, 2, 3, 4, 5}")

  ## Error checking
  if (fold < 2) stop("fold must be at least 2")   
  if (maxOrder<0)
    stop("maxOrder must be between 0 and 10")
  if (!is.null(order) && (order<0))
    stop("order must be between 0 and 10")  

  
  numBackSeqs      <- ifelse(is.null(mySeqs), 0, ncol(mySeqs))
  maxBackSeqLength <- ifelse(is.null(mySeqs), 0, nrow(mySeqs))

  if (fold>numBackSeqs){
    fold <- numBackSeqs
    warning(paste("fold cannot exceed number of sequences. fold is set to ", fold,".", sep=""))
  }
  if (maxOrder>6){
    maxOrder <- 6
    warning("maxOrder cannot exceed 6. maxOrder is set to 6.")
  }
  if ((!is.null(order)) && (order>6)){
    order <- 6
    warning("order cannot exceed 6. order is set to 6.")
  }
  
  
  ## initialize output variables
  success       <- 1
  selOrder      <- 0
  
  klDivs        <- rep(0, maxOrder+1)

  ncells <- 0
  for (i in 0:maxOrder) ncells <- ncells + 4^(i+1)
  estTransMat   <- rep(0, ncells)
  
  ## call C function
  com <- .C("bgm",
      as.integer(mySeqs),
      as.integer(numBackSeqs),
      as.integer(maxBackSeqLength),
      as.integer(fold),
      as.integer(order),
      as.integer(maxOrder),
      estTransMat=as.double(estTransMat),
      klDivs=as.double(klDivs),
      selOrder=as.integer(selOrder),
      success=as.integer(success)
      )
    
  if (!com$success) return(NULL)
  
  ## prepare output
  out.klDivs      <- com$klDivs
  out.klDivs[out.klDivs == 0] <- Inf
  out.klDivs[out.klDivs == -1] <- NA
  
  ## extract estimated transition matrices
    ## and create list from them
  first <- 1
  tMs <- list()
  transMat <- com$estTransMat
  dnaLetters <- c("A", "C", "G", "T")
  for (suborder in 0:com$selOrder){
    tM <- matrix(transMat[first:(first+4*4^suborder-1)], 4^suborder, 4, byrow=TRUE)
    colnames(tM) <- dnaLetters
    if (suborder == 0){
      rnames <- "--"
    }else{
      rnames <- dnaLetters
      if (suborder>1){
        for (i in 2:suborder){
          nrnames <- NULL
          for (j in 1:length(rnames)){
            for (k in 1:4) nrnames <- c(nrnames, paste(rnames[j],dnaLetters[k],sep=""))
          }
          rnames <- nrnames
        }
      }
    }
    rownames(tM) <- rnames
    
    first <- first + 4*4^suborder
    tMs <- c(tMs, mat=list(tM)) 
    }
  names(tMs) <- paste("order", 0:com$selOrder, sep="")
  
  ## create data.frame of candidate orders
  orders <- 0:(length(out.klDivs)-1)
  back=data.frame(order=orders,klDiv=out.klDivs)

  list(transMat=tMs, order=com$selOrder, klDivs=back)
  
  }


getSequences <-
  function(seqs) {


    ## first get a list of sequences
    ## in string format
    if (is.character(seqs)){

      ## select FASTA file through browser      
      if (seqs == "browse") {
        seqs <-
          fileBrowser(textToShow="Sequence File (in FASTA format)" )
      }
      if (!file.exists(seqs)) stop(paste("Sequence file", seqs,"not found."))
      
      seqs <- readFASTA(seqs)

    }
      
    if (class(seqs) != "list")
      stop("seqs must be browse, the path to a sequence file, or a list of sequences")   

      
    seqList <- list()
    descList <- list()
    for (i in 1:length(seqs)){
      seqList[[i]] <- seqs[[i]]$seq
      descList[[i]] <- seqs[[i]]$desc
    }


    
    ## now convert to matrix
    seqMat <- seqList2seqMat(seqList)

    list(seqMat=seqMat, seqList=seqList, descList=descList)
}


seqList2seqMat <- function(seqList){
  
    max<-max(unlist(lapply(seqList, nchar)))
    seqMat<-matrix(0, nrow=max, ncol=length(seqList))
    for (j in 1:length(seqList)){
      seq <- seqList[[j]]
      seq <- gsub("[A,a]","1", seq)
      seq <- gsub("[C,c]","2", seq)
      seq <- gsub("[G,g]","3", seq)
      seq <- gsub("[T,t]","4", seq)  
      seq <- gsub("[A-Z,a-z]",5, seq)
      seq <- as.numeric(strsplit(seq,"")[[1]])

      ## fill in with zeros
      if (length(seq) != max) seq <- c(seq, rep(0, max - length(seq)))
      seqMat[,j] <- seq
    }

    seqMat
  }

writeFASTA <- function (seqs, file){

  seqList <- list()
  descList <- list()
  for (i in 1:length(seqs)){
    seqList[[i]] <- seqs[[i]]$seq    
    descList[[i]] <- seqs[[i]]$desc
  }

  
  names(seqList)<-paste(">", descList, sep="")
  write.table(unlist(seqList), file=file, quote=FALSE, sep="\n", col.names=FALSE)
  
}

readFASTA <- function (file, maxNameLength=24, checkComments = TRUE){

    if (is.character(file)) {
      if (!file.exists(file))
        stop(paste("Sequence file", file," not found.")) 

      file <- file(file, "r")
        on.exit(close(file))
    }
    else {
        if (!inherits(file, "connection")) 
            stop("'file' must be a character string or connection")
        if (!isOpen(file)) {
            open(file, "r")
            on.exit(close(file))
        }
    }
  
    s1 <- scan(file = file, what = "", sep = "\n", quote = "", 
        allowEscapes = FALSE, quiet = TRUE)
    
    if (checkComments) {
        comments = grep("^;", s1)
        if (length(comments) > 0) 
            s1 = s1[-comments]
    }
    descriptions <- grep("^>", s1)
    numF <- length(descriptions)
    if (numF == 0) 
        stop("no FASTA sequences found")
    dp <- descriptions + 1
    dm <- descriptions - 1
    end = c(dm[-1], length(s1))
    ans = NULL
    for (i in 1:numF){
      desc <- s1[descriptions[[i]]]

      ## get rid of ">"
      desc <- gsub(">","",desc)

      ## keep first word
      desc <- strsplit(desc," ")[[1]][1]

      ## truncate if necessary
      if (nchar(desc) > maxNameLength) desc <- substr(desc, 1, maxNameLength)

      seq <- paste(s1[dp[i]:end[i]], collapse = "")

      ## check for non-letter characters
      nonAlph <- grep("[^A-Za-z]", seq)
      if (length(nonAlph) > 0){
        chars <- strsplit(seq,"")[[1]]
        nonAlph <- grep("[^A-Za-z]", chars)
        nonAlphs <- paste(chars[nonAlph],collapse="")
        stop(paste("Sequence ", i, " contains the characters ", nonAlphs, ". Only letters are allowed.", sep=""))
      }
      
      ans[[i]] <- list(desc = desc, seq = seq)
    }
    ans
}

license.cosmo <- function(){

  license.file <- system.file("COPYING", package="cosmo")

  rl <- readLines(license.file)
  rl <- rl[-c(1,length(rl))]
  rl <- gsub("*", "", rl, fixed=TRUE)
  for (i in 1:length(rl)) cat(paste(rl[i],"\n"))

}


bfile2tmat <- function (file){

    if (is.character(file)) {
      if (!file.exists(file))
        stop(paste("Background file", file," not found.")) 

      file <- file(file, "r")
        on.exit(close(file))
    }else {
        if (!inherits(file, "connection")) 
            stop("'file' must be a character string or connection")
        if (!isOpen(file)) {
            open(file, "r")
            on.exit(close(file))
        }
    }
  
    s1 <- scan(file = file, what = "", sep = "\n", quote = "", 
        allowEscapes = FALSE, quiet = TRUE)

    s1 <- sub('^\\s+', '', s1, perl = TRUE)
    
    comments = grep("^#", s1)
    if (length(comments) > 0) s1 = s1[-comments]
    tup <- unlist(lapply(strsplit(s1,'\\s+'),function(x) x[[1]]))
 
    tup <- gsub("[A,a]","1", tup)
    tup <- gsub("[C,c]","2", tup)
    tup <- gsub("[G,g]","3", tup)
    tup <- gsub("[T,t]","4", tup)  

    order <- max(nchar(tup)) - 1

    freqs <- as.numeric(unlist(lapply(strsplit(s1,'\\s+'),function(x) x[[2]])))

    dnaLetters <- c("A", "C", "G", "T")    
    transMat <- list()
    for (subOrder in 0:order){
      ntups <- 4^subOrder
      tM <- matrix(0, ntups, 4)
      colnames(tM) <- dnaLetters

      if (subOrder == 0){
        rnames <- "--"
      }else{
        rnames <- dnaLetters
        if (subOrder>1){
          for (i in 2:subOrder){
            nrnames <- NULL
            for (j in 1:length(rnames)){
              for (k in 1:4) nrnames <- c(nrnames, paste(rnames[j],dnaLetters[k],sep=""))
            }
            rnames <- nrnames
          }
        }
      }
      rownames(tM) <- rnames
 
      for (i in 1:length(tup)){
        if (nchar(tup[i]) == subOrder + 1){
          c <- as.numeric(substr(tup[i],nchar(tup[i]), nchar(tup[i])))
          if (subOrder == 0){
            r <- 1
          }else{
            ind <- as.numeric(unlist(strsplit(substr(tup[i], 1, nchar(tup[i])-1),"")))
            r <- 1
            fac <- 1
            for (k in length(ind):1){            
              r <- r + fac*(ind[k]-1)
              fac <- fac*4
            }
          }
          tM[r,c] <- freqs[i]
        }
      }
      
      if (abs(sum(tM)-1.0)>0.01)
        stop("Frequencies of tuples must add up to 1.0")

      ## normalize
      tM <- t(apply(tM, 1, function(x) x/sum(x)))

      transMat <- c(transMat, list(tM))
    }

    names(transMat) <- paste("order", 0:order, sep="")

    transMat
  }
