######################
##
## simulate sequences
##
######################
rseq <-
  function(numSeqs, seqLength, rate, pwm, transMats, model="ZOOPS", posOnly=FALSE){

    if (class(pwm) == "pwm") pwm <- pwm@pwm
    if (any(abs(apply(pwm,2,sum)-1)>0.001)) stop("PWM columns must add to one")

    if (!length(seqLength)%in%c(1,numSeqs)) stop("seqLength must be either of length 1 or numSeqs")

    if (length(seqLength) == 1) seqLength <- rep(seqLength, numSeqs)
    
    order <- length(transMats) - 1
    
    for (i in 0:order){
      numRows <- 4^i
      if (nrow(transMats[[i+1]]) != numRows){
        stop(paste("transition matrix", i+1, "should have", numRows, "rows, but has", nrow(transMats[[i+1]]), "rows"))
      }
    }
     
    if (!model%in%c("ZOOPS","TCM")) stop("model must be either ZOOPS or TCM")

    if ((model == "ZOOPS") &&((rate<0.0 || rate >1.0)))
      stop("For the ZOOPS model, the rate parameter must be between 0.0 and 1.0") 

    if ((model == "TCM") &&(rate<=0.0))
      stop("For the TCM model, the rate parameter must be greater than 0.0") 
    
    
    W <- ncol(pwm)
    
    sequences <- NULL

    strands <- NULL
    starts <- NULL
    motifs <- NULL
    seqNums <- NULL
    numMotifs <- NULL
    
    if (model == "ZOOPS"){

      myStarts <- NULL
      for (i in 1:numSeqs) myStarts[i] <- sample(1:(seqLength[i]-W+1), 1, replace=TRUE)
      
      L1 <- myStarts-1
      L2 <- seqLength-(myStarts+W-1)
      
      for (i in 1:numSeqs){
        if (runif(1)<rate){
          thisMotif <- simMotif(pwm)
          numMotifs <- rbind(numMotifs, thisMotif)
          motifs <- c(motifs, toString(thisMotif))
          
          if (!posOnly && runif(1)>0.5){
            thisMotif <- revComp(thisMotif)
            strands <- c(strands, -1)
          }else{
            strands <- c(strands, 1)
          }
          starts <- c(starts, myStarts[i])
          seqNums <- c(seqNums, i) 
          
          newSeq <- c(simBackground(L1[i], transMats),
                      thisMotif, simBackground(L2[i], transMats))
        }else{
          newSeq <- simBackground(seqLength[i], transMats)
          
        }
        
        sequences[i] <- toString(newSeq)
      }

    }else if (model == "TCM"){
      
      for (i in 1:numSeqs){
        pos <- 1
        newSeq <- NULL
        newStarts <- NULL
        newStrands <- NULL
        
        while(pos <= seqLength[i]){
          len<-round(rexp(1, rate))
          
          if (pos+len <= seqLength[i]-W+1){
            thisMotif <- simMotif(pwm)
            numMotifs <- rbind(numMotifs, thisMotif)
            motifs <- c(motifs, toString(thisMotif))

            if (!posOnly & runif(1)>0.5){
              thisMotif <- revComp(thisMotif)
              strands <- c(strands,-1)
            }else{
              strands <- c(strands,1)
            }                           
            
            starts <- c(starts, pos+len)
            seqNums <- c(seqNums, i) 
            
            newSeq <- c(newSeq,simBackground(len, transMats), thisMotif)            
            pos <- pos+len+W
          }else{
            len <- seqLength[i]-pos+1
            newSeq <- c(newSeq, simBackground(len, transMats))
            pos <- seqLength[i]+1
          }
        }

        
        sequences[i]<- toString(newSeq)
      }    
    }



    
    seqs <- list()
    seqNames <- NULL
    for (i in 1:numSeqs){
      seqName <- paste("Seq",i,sep="")
      seqNames <- c(seqNames, seqName)
      seqs[[i]] <- list(seq=sequences[i],desc=seqName)
    }

    
    if (length(starts)>0){
      motifs <- new("align",seq=seqNames[seqNums], pos=starts, orient=strands,motif=motifs,prob=rep(1,length(seqNums)),eval=0)
      empPWM <- apply(numMotifs, 2, function(x) c(sum(x==1),sum(x==2), sum(x==3), sum(x==4))/length(x))
      empPWM <- makePWM(empPWM)
    }else{
      motifs <- NULL
      empPWM <- NULL
    }

      
    list(seqs=seqs, motifs=motifs, empPWM=empPWM)
  }

toString <-
  function(seq){
    paste(c("A","C","G","T","N")[seq],sep="",collapse="")
  }

simBackground <-
  function(len, transMats){
    if (len == 0) return(NULL)

    order <- length(transMats) -1 

    oligo <- NULL
    seq <- NULL
    for (i in 1:len){
      if (i<=order){
        tM <- transMats[[i]]
      }else{
        tM <- transMats[[order+1]]
      }
      probs <- as.numeric(tM[findRow(oligo),])
      mat <- rmultinom(1,1,probs)
      index <- (1:4)[1:4*mat]
      seq <- c(seq, index)
      oligo <- c(oligo, index)
      if (length(oligo) > order) oligo <- oligo[-1]
    }
    seq
  }

 
simMotif <-
  function(pwm){
    seqmat <- apply(pwm, 2, function(x){rmultinom(1, 1, x)})
    (1:4)[1:4*seqmat]
  }


revComp <- function(seq){
  rc <- 5 - rev(seq)
  rc
}


findRow <-
  function(sequence){

    row <- 1    
    if (!is.null(sequence)){
      row <- 1
      for (i in 1:length(sequence)){
        row <- row + (4^(length(sequence)-i))*(sequence[i]-1)
      }      
    }
    return(row)
  }

## score simulation results
simScore <-
  function(truth, cosmoOut, minOverlap=0.25){


    
    if ((!is.null(truth) && class(truth) != "align")){
      stop("truth must be NULL or an object of class align")
    }

    if ((!is.null(cosmoOut)) && (class(cosmoOut) != "cosmo")){
      stop("cosmoOut must be NULL or an object of class cosmo")
    }

    if (is.null(truth)){
      sens <- NA
      spec <- NA
      roc <- NA
    }else if (is.null(cosmoOut)){
      sens <- 0
      spec <- 0
      roc <- 0
    }else{

      found <-  cosmoOut@motifs
      
      trueW <- nchar(truth@motif[1])
      foundW <- nchar(found@motif[1])

      nTrue  <- length(truth@pos)
      nFound <- length(found@pos)
      
      trueFound <- 0
      for (i in 1:nTrue){
        for (j in 1:nFound){
          if (truth@seq[i] == found@seq[j]){
            truePos <- truth@pos[i]:(truth@pos[i]+trueW-1)
            foundPos <- found@pos[j]:(found@pos[j]+foundW-1)
            overlap <- intersect(truePos,foundPos)
            if (length(overlap) >= ceiling(minOverlap*trueW)) trueFound <- trueFound + 1
          }
        }
      }
      sens <- trueFound/nTrue      
      spec <- trueFound/nFound

      
      ## first order posterior probabilites
      probs <- cosmoOut@probs
      numSeqs <- length(probs)
      seqLengths <- unlist(lapply(probs, length))
      maxLength <- max(seqLengths)
      probMat <- matrix(0,maxLength, numSeqs)
      for (i in 1:numSeqs) probMat[,i] <- abs(probs[[i]])        
      probs <- as.vector(probMat)
      numMax <- sum(probs>0)
      
      seqNames <- NULL
      for (i in 1:length(cosmoOut@seqs)) seqNames[i] <- cosmoOut@seqs[[i]]$desc
      
      trueMat <- matrix(0,maxLength, numSeqs)
      
      for (i in 1:nTrue){
        seqNum <- (1:numSeqs)[truth@seq[i]==seqNames]
        offset <- floor((1-minOverlap)*trueW)
        first <- max(1, truth@pos[i] - offset)
        last <- min(seqLengths[seqNum], truth@pos[i] + offset)
        
        maxProb <- 0
        best <- first
        for (pos in first:last){
          if (probMat[pos, seqNum] > maxProb){
            maxProb <- probMat[pos, seqNum]
            best <- pos
          }
        }
        trueMat[best,seqNum] <- 1
      }
        
      trueVec <- as.vector(trueMat)
      index <- rev(order(probs))
      hits <- trueVec[index][1:numMax]
      
      misses <- 1 - hits
      hits <- cumsum(hits)
      misses <- cumsum(misses)
      
        
      tpp <- hits/nTrue
      fpp <- misses/(numMax-nTrue)
      
      tpp1 <- tpp
      tpp0 <- c(0,tpp[-length(tpp)])
      mid  <- 0.5*(tpp0+tpp1)
      dx   <- diff(c(0,fpp))
      roc <- sum(dx*mid)
      
    }
    
    list(sens=sens, spec=spec, roc=roc)

  }




  
