/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/
  

/*****************************************************************************
*								       	     *
*      This module contains functions for requesting memory for various      *
*      structures                                                            *
*                                                                            *
*****************************************************************************/
   

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "matLib.h"
#include "cosmo.h"
#include "alloc.h" 
  
/**********************************************************************/
/*
  distrMalloc

  Allocate memory for n DISTR structs.

*/
/**********************************************************************/
DISTR *distrMalloc(
		   long n       /* number of structs to allocate memory for */
		   )
{
  DISTR *distrs;                 /* pointer to struct that will be returned */
  int i;


  if (n <= 0){
    Rprintf("distrMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }


  distrs = (DISTR *)S_alloc((unsigned)(n),(sizeof(DISTR)));
  if (distrs==NULL){
    Rprintf("distrMalloc: Couldn't get %ld DISTRs. Exiting ...\n", n);
    exit(1);
  }

  for (i=0; i<n; i++) {
    distrs[i].w = 0;
    distrs[i].offset = NULL;
    distrs[i].range = NULL;
    distrs[i].d = NULL;
    distrs[i].cdf = NULL;
    distrs[i].mean = 0;
  }


  return distrs;


}



/**********************************************************************/
/*
  critMalloc

  Allocate memory for n CRIT structs.

*/
/**********************************************************************/
CRIT *critMalloc(
		 long n       /* number of structs to allocate memory for */
		 )
{
  CRIT *crit;                 /* pointer to struct that will be returned */

  if (n <= 0){
    Rprintf("critMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }


  crit = (CRIT *)S_alloc((unsigned)(n),(sizeof(CRIT)));
  if (crit==NULL){
    Rprintf("critMalloc: Couldn't get %ld CRITs. Exiting ...\n", n);
    exit(1);
  }

  return crit;

}

/**********************************************************************/
/*
  intMotProbMalloc

  Allocate memory for n INTMOTPROB structs.

*/
/**********************************************************************/
INTMOTPROB *intMotProbMalloc(
			     long n       /* number of structs to allocate memory for */
			     )
{
  int num;                    /* index for iterating over structs */
  INTMOTPROB *intMotProb;     /* pointer to struct that will be returned */

  if (n <= 0){
    Rprintf("intMotProbMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }


  intMotProb = (INTMOTPROB *)S_alloc((unsigned)(n),(sizeof(INTMOTPROB)));
  if (intMotProb==NULL){
    Rprintf("intMotProbMalloc: Couldn't get %ld INTMOTPROBs. Exiting ...\n", n);
    exit(1);
  }

  for (num=0; num<n; num++){
    intMotProb[num].seq   = NULL;
    intMotProb[num].prob  = 0;
  }

  return intMotProb;

}



/**********************************************************************/
/*
  svSpaceMalloc

  Allocate memory for n SVSPACE structs.

*/
/**********************************************************************/
SVSPACE *svSpaceMalloc(
		       long n   /* number of structs to allocate memory for */
		       )
{
  SVSPACE *svSpace;     /* pointer to struct that will be returned */
  int num;              /* index for iterating over structs */

  if (n <= 0){
    Rprintf("svSpaceMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  svSpace = (SVSPACE *)S_alloc((unsigned)(n),(sizeof(SVSPACE)));
  if (svSpace==NULL){
    Rprintf("svSpaceMalloc: Couldn't get %ld SVSPACE's. Exiting ...\n", n);
    exit(1);
  }

  for(num=0; num<n;num++){
    svSpace[num].mType       = 0;
    svSpace[num].nnSites     = 0;
    svSpace[num].numSites    = NULL;
    svSpace[num].numSv       = 0;
    svSpace[num].svTheta     = NULL;
    svSpace[num].svLogLR     = NULL;
  }

  return svSpace;
}

/**********************************************************************/
/*
  mTypeMalloc

  Allocate memory for n MTYPEs.

*/
/**********************************************************************/
MTYPE *mTypeMalloc(
		   long n    /* number of structs to allocate memory for */
		   ) 
{
  MTYPE *mType;      /* pointer to model type that will be returned */
  int num;           /* index for iterating over structs */

  if (n <= 0){
    Rprintf("mTypeMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }


  mType = (MTYPE *)S_alloc((unsigned)(n),(sizeof(MTYPE)));
  if (mType==NULL){
    Rprintf("mTypeMalloc: Couldn't get %ld MTYPEs. Exiting ...\n", n);
    exit(1);
  }

  for (num=0; num<n;num++) mType[num] = 0;

  return mType;
}

/**********************************************************************/
/*
  postProbMalloc

  Allocate memory for n POSTPROB structs.

*/
/**********************************************************************/
POSTPROB *postProbMalloc(
			 long n    /* number of structs to allocate memory for */
			 )
{
  POSTPROB *postProb;    /* pointer to struct that will be returned */
  int num;               /* index for iterating over structs */

  if (n <= 0){
    Rprintf("postProbMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  postProb = (POSTPROB *)S_alloc((unsigned)(n),(sizeof(POSTPROB)));
  if (postProb==NULL){
    Rprintf("postProbMalloc: Couldn't get %ld POSTPROBs. Exiting ...\n", n);
    exit(1);
  }

  for (num=0; num<n; num++){
    postProb[num].seqNum     = 0;
    postProb[num].start      = 0;
    postProb[num].prob       = 0;
    postProb[num].revComp    = 0;
  }

  return postProb;
}


/**********************************************************************/
/*
  thetaMalloc

  Allocate memory for n THETA structs.

*/
/**********************************************************************/
THETA *thetaMalloc(
		   long n,            /* number of structs to allocate memory for */
		   int maxWidth,      /* maximum motid width */
		   int numProfParms   /* maximum number of IC profile parameters */
		   )
{
  THETA *theta;      /* pointer to struct that will be returned */
  int num;           /* index for iterating over structs */


  if (n <= 0){
    Rprintf("thetaMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  theta = (THETA *)S_alloc((unsigned)(n),(sizeof(THETA)));
  if (theta==NULL){
    Rprintf("thetaMalloc: Couldn't get %ld THETAs. Exiting ...\n", n);
    exit(1);
  }

  for (num=0; num<n; num++){
    theta[num].mType         = 0;
    theta[num].width         = maxWidth;
    theta[num].pwm           = DoubleMalloc(4*maxWidth);    
    theta[num].intensity     = 0;
    theta[num].numSites      = 0;
    theta[num].numProfParms  = numProfParms;
    theta[num].profParms     = numProfParms ? DoubleMalloc(numProfParms) : NULL;
    theta[num].logLR         = 0;
    theta[num].KL            = 0;    
    theta[num].posKL         = DoubleMalloc(maxWidth);
    theta[num].IC            = 0;
    theta[num].posIC         = DoubleMalloc(maxWidth);
    theta[num].logEval       = 0;
    theta[num].aic           = BIG;
    theta[num].bic           = BIG;
    theta[num].score         = LITTLE;
    theta[num].conSetNum     = 0;
    theta[num].conMod        = NULL;
  }

  return theta;
}


/**********************************************************************/
/*
  sampleMalloc

  Allocate memory for n SAMPLE structs.

*/
/**********************************************************************/
SAMPLE *sampleMalloc(
		     long n        /* number of structs to allocate memory for */
		     )
{

  SAMPLE *samples;       /* pointer to struct that will be returned */
  int num;               /* index for iterating over structs */


  if (n <= 0){
    Rprintf("sampleMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  samples = (SAMPLE *)S_alloc((unsigned)(n),(sizeof(SAMPLE)));
  if (samples==NULL){
    Rprintf("sampleMalloc: Unable to get %ld samples. Exiting ...\n", n);
    exit(1);
  }

  for (num=0; num<n; num++){

    samples[num].name          = NULL;
    samples[num].nameLength    = 0;

    samples[num].length        = 0;
    samples[num].seq           = NULL;
    samples[num].rcSeq         = NULL;
  
    samples[num].cvGroup       = 0;
    samples[num].exclude       = FALSE;    

    samples[num].logCumBack    = NULL;
    samples[num].backLR        = NULL;
    samples[num].cumProbBack   = NULL;

    samples[num].mProbs1       = NULL;
    samples[num].rcmProbs1     = NULL;
    samples[num].mProbs2       = NULL;
    samples[num].rcmProbs2     = NULL;

    samples[num].seqProbs      = NULL;
    samples[num].rcSeqProbs    = NULL;

    samples[num].seqLR         = 0;
    samples[num].cutLRs        = NULL;
    
  }

  return samples;
}



/**********************************************************************/
/*
  conSetMalloc

  Allocate memory for n CONSET structs.

*/
/**********************************************************************/
CONSET * conSetMalloc(
		      long n,         /* number of structs to allocate memory for */
		      int widthLow    /* minimum motif width */
		      )
{
    
  int cSetNum;             /* index for iterating over structs */
  CONSET *conSets;         /* pointer to struct that will be returned */


  if (n <= 0){
    Rprintf("conSetMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  conSets = (CONSET*)S_alloc((unsigned)(n),(sizeof(CONSET)));    
  if (conSets==NULL){
    Rprintf("conSetMalloc: Couldn't get %ld CONSETs. Exiting...\n", n);
    exit(1);
  }	

  for (cSetNum=0; cSetNum<n;cSetNum++){
    conSets[cSetNum].numInts             = 1;
    conSets[cSetNum].numShapeCons        = 0;	
    conSets[cSetNum].numBoundCons        = 0;	
    conSets[cSetNum].numIntFreqCons      = 0;
    conSets[cSetNum].intCons             = NULL;

    conSets[cSetNum].hasPalCon           = 0;
    conSets[cSetNum].palCon.int1         = 0;
    conSets[cSetNum].palCon.int2         = 0;
    conSets[cSetNum].palCon.error        = 0;

    conSets[cSetNum].hasSubCon           = 0;
    conSets[cSetNum].subCon.conType      = sub;
    conSets[cSetNum].subCon.length       = 0;
    conSets[cSetNum].subCon.motif        = NULL;
    conSets[cSetNum].subCon.cutOff       = 0;

    conSets[cSetNum].numLinCons          = 0;
    conSets[cSetNum].linCons             = NULL;

    conSets[cSetNum].sharpestBound       = 0.25;
    conSets[cSetNum].widthLow            = widthLow;
    
    conSets[cSetNum].isEmpty             = TRUE;
  }

  return conSets;
}


/**********************************************************************/
/*
  posFreqConMalloc

  Allocate memory for n POSFREQCON structs.

*/
/**********************************************************************/
POSFREQCON *posFreqConMalloc(
			     long n    /* number of structs to allocate memory for */
			     )
{
  int numPosFreqCon;             /* index for iterating over structs */
  POSFREQCON * posFreqCon;       /* pointer to struct that will be returned */


  if (n <= 0){
    Rprintf("posFreqConMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  posFreqCon = (POSFREQCON *)S_alloc((unsigned) (n),(sizeof(POSFREQCON)));
  if (posFreqCon==NULL){
    Rprintf("posFreqConMalloc: Couldn't get %ld POSFREQCONs. Exiting...\n", n);
    exit(1);
  }	
    
  for (numPosFreqCon=0; numPosFreqCon<n;numPosFreqCon++){
    posFreqCon[numPosFreqCon].low    = 0;
    posFreqCon[numPosFreqCon].pos    = 0;
    posFreqCon[numPosFreqCon].nuc    = A;
    posFreqCon[numPosFreqCon].next   = (numPosFreqCon < n-1) ? (posFreqCon + numPosFreqCon + 1) : NULL;
  }

  return posFreqCon;
}


/**********************************************************************/
/*
  inFreqConMalloc

  Allocate memory for n INTFREQCON structs.

*/
/**********************************************************************/
INTFREQCON *intFreqConMalloc(
			     long n    /* number of structs to allocate memory for */
			     )
{
  int numIntFreqCon;            /* index for iterating over structs */
  INTFREQCON * intFreqCon;      /* pointer to struct that will be returned */


  if (n <= 0){
    Rprintf("intFreqConMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  intFreqCon = (INTFREQCON *) S_alloc((unsigned)(1),(sizeof(INTFREQCON)));
  if (intFreqCon==NULL){
    Rprintf("intFreqConMalloc: Couldn't get %ld INTFREQCONs. Exiting...\n", n);
    exit(1);
  }	
    
  for (numIntFreqCon=0; numIntFreqCon<n;numIntFreqCon++){
    intFreqCon[numIntFreqCon].low    = 0;
    intFreqCon[numIntFreqCon].nuc    = A;
    intFreqCon[numIntFreqCon].next   = (numIntFreqCon < n-1) ? (intFreqCon + numIntFreqCon + 1) : NULL;
  }

  return intFreqCon;
}


/**********************************************************************/
/*
  linConMalloc

  Allocate memory for n LINCON structs.

*/
/**********************************************************************/
LINCON *linConMalloc(
		     long n      /* number of structs to allocate memory for */
		     )
{
  int linConNum;         /* index for iterating over structs */
  LINCON * linCon;       /* pointer to struct that will be returned */


  if (n <= 0){
    Rprintf("linConMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  linCon = (LINCON *) S_alloc((unsigned) (n),(sizeof(LINCON)));
  if (linCon==NULL){
    Rprintf("linConMalloc: Couldn't get %ld LINCONs. Exiting...\n", n);
    exit(1);
  }

  for (linConNum=0; linConNum<n;linConNum++){
    linCon[linConNum].pos1    = start;
    linCon[linConNum].pos2    = start;
    linCon[linConNum].int1    = 0;
    linCon[linConNum].int2    = 0;
    linCon[linConNum].index1  = 0;
    linCon[linConNum].index2  = 0;
    linCon[linConNum].low     = 0;
    linCon[linConNum].up      = 0;
    linCon[linConNum].next    = (linConNum < n-1) ? (linCon + linConNum + 1) : NULL;
  }

  return linCon;
}

 
/**********************************************************************/
/*
  intConMalloc

  Allocate memory for n INTCON structs.

*/
/**********************************************************************/
INTCON * intConMalloc(
		      long n  /* number of structs to allocate memory for */
		      )
{
    
  int intConNum;        /* index for iterating over structs */
  INTCON *intCons;      /* pointer to struct that will be returned */
    

  if (n <= 0){
    Rprintf("intConMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  intCons = (INTCON*)S_alloc((unsigned)(n),(sizeof(INTCON)));
  if (intCons == NULL){  
    Rprintf("intConMalloc: Couldn't get %ld INTCONs. Exiting...\n", n);
    exit(1);
  }

  /* initialize */
  for (intConNum=0;intConNum<n;intConNum++){
    intCons[intConNum].lenType           = bp;
    intCons[intConNum].prop              = 0;
    intCons[intConNum].numBasePairs      = 0;

    intCons[intConNum].hasBoundCon       = 0;
    intCons[intConNum].boundCon.low      = 0;
    intCons[intConNum].boundCon.up       = 0;

    intCons[intConNum].hasShapeCon       = 0;
    intCons[intConNum].shapeCon.shape    = lin;
    intCons[intConNum].shapeCon.startLow = 0;
    intCons[intConNum].shapeCon.startUp  = 2;
    intCons[intConNum].shapeCon.endLow   = 0;
    intCons[intConNum].shapeCon.endUp    = 2;
    intCons[intConNum].shapeCon.error    = 0;
    intCons[intConNum].shapeCon.number   = 0;

    intCons[intConNum].numIntFreqCons    = 0;
    intCons[intConNum].intFreqCons       = NULL;

    intCons[intConNum].numPosFreqCons    = 0;
    intCons[intConNum].posFreqCons       = NULL;
  }


  return intCons;
}


/**********************************************************************/
/*
  modMalloc

  Allocate memory for n MODEL structs.

*/
/**********************************************************************/
MODEL * modMalloc(
		  long n   /* number of structs to allocate memory for */
		  )
{
    
  int modNum;          /* index for iterating over structs */
  MODEL *mod;          /* pointer to struct that will be returned */


  if (n <= 0){
    Rprintf("modMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  mod = (MODEL *) S_alloc((unsigned)(n),(sizeof(MODEL)));
  if (mod==NULL){
    Rprintf("modMalloc: Couldn't get %ld MODELs. Exiting ...\n", n);
    exit(1);
  }

  for (modNum = 0; modNum < n; modNum++){
    mod[modNum].mType               = ZOOPS;
    mod[modNum].width               = 0;
    mod[modNum].n                   = 0;
    mod[modNum].nlin                = 0;
    mod[modNum].nonlin              = 0;

    mod[modNum].hasPosFreqCons      = 0;
    mod[modNum].numLinFreqCons      = 0;
    mod[modNum].hasSubCon           = 0;
    mod[modNum].subConIndex         = 0;

    mod[modNum].numProfParms        = 0;
    mod[modNum].profParmIndex       = 0;
    mod[modNum].sharpestBound       = 0;


    mod[modNum].linConMatrix        = NULL;
    mod[modNum].low                 = NULL;
    mod[modNum].up                  = NULL;
    mod[modNum].nlCons              = NULL;
  }
    
  return mod;
}



/**********************************************************************/
/*
  nlConMalloc

  Allocate memory for n NLCON structs.

*/
/**********************************************************************/
NLCON *nlConMalloc(
		   long n   /* number of structs to allocate memory for */
		   )
{
  NLCON *nlCon;         /* pointer to struct that will be returned */
  int numNlCon;         /* index for iterating over structs */


  if (n <= 0){
    Rprintf("nlConMalloc: Don't allocate n = %ld objects!\n", n);
    exit(1);
  }

  nlCon =  (NLCON*)S_alloc((unsigned)(n),(sizeof(NLCON)));
  if (nlCon == NULL){  
    Rprintf("nlConMalloc: Couldn't get %ld NLCONs. Exiting...\n", n);
    exit(1);
  }

  /* Initialize as ProfCon */
  for (numNlCon=0;numNlCon<n;numNlCon++){
    nlCon[numNlCon].pCon.conType     = lin;
    nlCon[numNlCon].pCon.ref         = prevPos;
    nlCon[numNlCon].pCon.interval    = 0;
    nlCon[numNlCon].pCon.xPos        = 0;
    nlCon[numNlCon].pCon.parm1       = 0;
    nlCon[numNlCon].pCon.parm2       = 0;
  }

  return nlCon;
}



/**********************************************************************/
/*
  allocMod

  Allocate memory for arrays mod->low, mod->up, and mod->linConMatrix
  inside a MODEL struct mod.

*/
/**********************************************************************/
void allocMod(
	      MODEL *mod    /* model for whose components to allocate memory */
	      )
{
      
  int cellNum;    /* index for iterating over cells in matrix */
  int numCells;   /* number of such cells */
  int boundNum;   /* index for iterating over bounds */
  int numBounds;  /* number of such bounds */

  numBounds = mod->n + mod->nlin + mod->nonlin;
  numCells  = mod->n* mod->nlin;

  if (numBounds<=0){
    Rprintf("allocMod: Don't allocate memory for %d bounds. Exiting...\n", numBounds);
    exit(1);
  }

  if (numCells<=0){
    Rprintf("allocMod: Don't allocate memory for %d cells. Exiting...\n", numCells);
    exit(1);
  }

  mod->low = (double*) realloc(mod->low, (unsigned)numBounds * sizeof(double));
  if (mod->low==NULL){
    Rprintf("allocMod: Couldn't get %d doubles for mod->low. Exiting ...\n", numBounds);
    exit(1);
  }

  mod->up  = (double*) realloc(mod->up, (unsigned)numBounds * sizeof(double));
  if (mod->up==NULL){
    Rprintf("allocMod: Couldn't get %d doubles for mod->up. Exiting ...\n", numBounds);
    exit(1);
  }

  mod->linConMatrix = (int*) realloc(mod->linConMatrix, (unsigned)numCells * sizeof(int));
  if (mod->linConMatrix==NULL){
    Rprintf("allocMod: Couldn't get %d ints for mod->linConMatrix. Exiting ...\n", numCells);
    exit(1);
  }
  
  /* initialize */
  for (boundNum=0; boundNum<numBounds;boundNum++){
    mod->low[boundNum] = 0;
    mod->up[boundNum]  = 1;
  }

  for (cellNum=0; cellNum<numCells;cellNum++){
    mod->linConMatrix[cellNum] = 0;
  }

}


