/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains the implemtation of the framework for            *
*      specifying constraints                                                * 
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include "matLib.h"
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "constraints.h"
#include "estarts.h"
#include "ctype.h"          /* for toupper() */
  

/**********************************************************************/
/*
  deriveAllModels

  Interpret constraint sets for all model types and candidate motif
  widths and store interpreted constraint sets.
  
	
*/
/**********************************************************************/
int deriveAllModels(
		     DATASET *dataSet   /* the dataset */
		     )
{

  int cSetNum;     /* number of current constraint set */
  int modTypeNum;  /* number of current model type */
  int width;       /* current candidate motif width */

  

  for (cSetNum=0; cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width <= dataSet->maxWidth; width++){
	if (!deriveModel(dataSet, width, modTypeNum, cSetNum)) return(0);
      }
    }
  }

  return(1);
}
   
/**********************************************************************/
/*
  deriveModel

  Interprets a given constraint set for a particular motif
  width, storing all the information needed by the optimization
  routine in a MODEL.
	
  Returns 1 if successful, 0 otherwise

*/
/**********************************************************************/
BOOLEAN deriveModel(
		    DATASET *dataSet,  /* the dataset */
		    int width,         /* motif width */
		    int modTypeNum,    /* model type number */
		    int conSetNum      /* constraint set number */
		    )
{

  MODEL *mod;                      /* the model that is to be derived */
  CONSET cSet;                     /* the constraint set that will be interpreted */
    

  int *intNumBasePairs;            /* the number of bps in each interval */
  int *intFirst;                   /* the first position in each interval */
  int *intLast;                    /* the last position in each interval */    

  int intNum, j, k, index, pos;    /* various temporary variables */
  int nuc, conNum, numBasePairs;   
  int maxNumSites;
  int extraParm;                   /* extra parameter beyond PWM: pi for ZOOPS, lambda for TCM */
  MTYPE mType, modType;
  
  POSFREQCON *posFreqCon;          /* various temporary pointers */
  INTFREQCON *intFreqCon;
  LINCON *linCon;


  mType               = dataSet->mTypes[modTypeNum];
  mod                 = dataSet->mod + MOD(width, modTypeNum, conSetNum);

  extraParm           = mType != OOPS && dataSet->maxIntensity ? 1 : 0;

  cSet                = dataSet->cSets[conSetNum];
  mod->mType          = mType;
  mod->width          = width;
  mod->numProfParms   = 2*cSet.numShapeCons;
  mod->profParmIndex  = 4*width + extraParm;
  mod->sharpestBound  = cSet.sharpestBound;
  intNumBasePairs     = dataSet->intNumBasePairs;
  intFirst            = dataSet->intFirst;
  intLast             = dataSet->intLast;


  /* divide motif up into intervals */
  numBasePairs = 0;
  for (intNum=0;intNum<cSet.numInts;intNum++){
    if (cSet.intCons[intNum].lenType==prop){
      intNumBasePairs[intNum] = NINT(cSet.intCons[intNum].prop*width); 
      numBasePairs += intNumBasePairs[intNum];
    }else if (cSet.intCons[intNum].lenType==bp){
      intNumBasePairs[intNum] = cSet.intCons[intNum].numBasePairs;
      numBasePairs += intNumBasePairs[intNum];
    }
  }

  if (numBasePairs>width){
    Rprintf("deriveModel: Fixed and proportional length intervals require at least %d bps\n", numBasePairs);
    return FALSE;
  }

  for (intNum=0;intNum<cSet.numInts;intNum++){
    if (cSet.intCons[intNum].lenType==var){
      intNumBasePairs[intNum] = width - numBasePairs;
      if (intNumBasePairs[intNum]<=0){
	Rprintf("deriveModel: Variable length interval has length zero at motif width %d\n", width);
	return FALSE;
      }
      break;
    }
  }

  intFirst[0] = 0;
  intLast[0]  = intNumBasePairs[0] - 1;
  for (intNum=1;intNum<cSet.numInts;intNum++){
    intFirst[intNum]  = intLast[intNum-1] + 1;
    intLast[intNum]   = intFirst[intNum] + intNumBasePairs[intNum] - 1;
  }


  /********************************* 
   * Calculate n, nlin, and nonlin *
   *********************************/

  mod->n      = 4*mod->width + extraParm;                /* the PWM parameters plus, possibly, pi or lambda*/
  mod->nlin   = mod->width;                              /* constrain all position frequencies to sum to 1 */
  mod->nonlin = 0;

  /* 2 extra parameters for each interval shape constraint */
  mod->n += 2*cSet.numShapeCons;

  /* PosFreqCons for nucleotides 'AT' and 'GC' are linear */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    for(posFreqCon = cSet.intCons[intNum].posFreqCons;posFreqCon!=NULL;posFreqCon=posFreqCon->next){
      if (posFreqCon->nuc==AT || posFreqCon->nuc==GC){
	(mod->nlin)++;
	(mod->numLinFreqCons)++;
      }
    }
  }

    
  /* IntervalFreqCons are linear */
  mod->nlin           += cSet.numIntFreqCons;  
  mod->numLinFreqCons += cSet.numIntFreqCons;

  /* palindromic constraints are linear */
  if (cSet.hasPalCon){
    mod->nlin += 4*intNumBasePairs[cSet.palCon.int1];
  }

  /* linear constraints on profile parameters */
  mod->nlin += cSet.numLinCons;
    
  /* Find out how many nonlinear profile constraint there are */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    /* bound constraints */
    if (cSet.intCons[intNum].hasBoundCon){
      mod->nonlin += intNumBasePairs[intNum];
    }
    else if (cSet.intCons[intNum].hasShapeCon){
      /* have one extra constraint for monotone shape */
      if (cSet.intCons[intNum].shapeCon.shape == monInc || cSet.intCons[intNum].shapeCon.shape == monDec){
	mod->nonlin += intNumBasePairs[intNum] + 1;
		
	/* linear shape constraints */
      }else if (cSet.intCons[intNum].shapeCon.shape == lin){ 
	mod->nonlin += intNumBasePairs[intNum];
      }
    }
  }

  mod->nonlin += cSet.hasSubCon;

  if (COSMO_MSG_LEVEL>=3) Rprintf("n = %d and nlin = %d and nonlin = %d\n", mod->n, mod->nlin, mod->nonlin);
    
  /* Allocate memory for nonlinear constraints */
  if (mod->nonlin > 0) mod->nlCons = nlConMalloc(mod->nonlin);
    
 
  /* Now interpret the shape constraints */
  conNum = 0;
  for (intNum=0;intNum<cSet.numInts;intNum++){

    /* bound constraints first */
    if (cSet.intCons[intNum].hasBoundCon){
      for (pos=intFirst[intNum];pos<=intLast[intNum];pos++){
	mod->nlCons[conNum].pCon.conType = bdd;
	mod->nlCons[conNum].pCon.pos = pos;
	mod->nlCons[conNum].pCon.interval = intNum;
	mod->nlCons[conNum].pCon.parm1 = -1;
	mod->nlCons[conNum].pCon.parm2 = -1;
	mod->nlCons[conNum].pCon.xPos  = ((double)pos-intFirst[intNum])/(intNumBasePairs[intNum]-1);
	conNum++;
      }

    }else if(cSet.intCons[intNum].hasShapeCon){
	    
      /* have one extra constraint for monotone shape */
      if (cSet.intCons[intNum].shapeCon.shape == monInc || cSet.intCons[intNum].shapeCon.shape == monDec){
		
	mod->nlCons[conNum].pCon.conType = cSet.intCons[intNum].shapeCon.shape;
	mod->nlCons[conNum].pCon.pos = intFirst[intNum];
	mod->nlCons[conNum].pCon.interval = intNum;
	mod->nlCons[conNum].pCon.parm1 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number;
	mod->nlCons[conNum].pCon.parm2 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number + 1;
	mod->nlCons[conNum].pCon.ref = prevBrk;
	mod->nlCons[conNum].pCon.xPos = 0.0;
	conNum++;
		
	for (pos=intFirst[intNum]+1;pos<=intLast[intNum];pos++){
	  mod->nlCons[conNum].pCon.conType = cSet.intCons[intNum].shapeCon.shape;
	  mod->nlCons[conNum].pCon.pos = pos;
	  mod->nlCons[conNum].pCon.interval = intNum;
	  mod->nlCons[conNum].pCon.parm1 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number;
	  mod->nlCons[conNum].pCon.parm2 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number + 1;
	  mod->nlCons[conNum].pCon.ref = prevPos;
	  mod->nlCons[conNum].pCon.xPos = ((double)pos-intFirst[intNum])/(intNumBasePairs[intNum]-1);
	  conNum++;
	}
		
	mod->nlCons[conNum].pCon.conType = cSet.intCons[intNum].shapeCon.shape;
	mod->nlCons[conNum].pCon.pos = intLast[intNum];
	mod->nlCons[conNum].pCon.interval = intNum;
	mod->nlCons[conNum].pCon.parm1 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number;
	mod->nlCons[conNum].pCon.parm2 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number + 1;
	mod->nlCons[conNum].pCon.ref = nextBrk;
	mod->nlCons[conNum].pCon.xPos = 1.0;
	conNum++;
	      
      }else{
	
	/* linear shape constraint */
	for (pos=intFirst[intNum];pos<=intLast[intNum];pos++){
	  mod->nlCons[conNum].pCon.conType = lin;
	  mod->nlCons[conNum].pCon.pos = pos;
	  mod->nlCons[conNum].pCon.interval = intNum;
	  mod->nlCons[conNum].pCon.parm1 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number;
	  mod->nlCons[conNum].pCon.parm2 = 4*width + extraParm + 2*cSet.intCons[intNum].shapeCon.number + 1;
	  mod->nlCons[conNum].pCon.xPos = ((double)pos-intFirst[intNum])/(intNumBasePairs[intNum]-1);
	  conNum++;
	}
      }
    }
  }



  if (cSet.hasSubCon){
    if (COSMO_MSG_LEVEL>=3) Rprintf("deriveModel: Has subcon at conNum = %d\n", conNum);
    mod->hasSubCon                     = 1;
    mod->subConIndex                   = conNum;
    mod->nlCons[conNum].subCon.conType = sub;
    mod->nlCons[conNum].subCon.length  = cSet.subCon.length;	
    mod->nlCons[conNum].subCon.motif   = cSet.subCon.motif;	
    mod->nlCons[conNum].subCon.cutOff  = cSet.subCon.cutOff;	
  }
  if (COSMO_MSG_LEVEL>=3){
    for (intNum=0;intNum < mod->nonlin - cSet.hasSubCon;intNum++){
      Rprintf("\nnlCON %d:\n",intNum);
      Rprintf("type = %d\n",mod->nlCons[intNum].pCon.conType);
      Rprintf("position = %d\n",mod->nlCons[intNum].pCon.pos);
      Rprintf("interval = %d\n",mod->nlCons[intNum].pCon.interval);
      Rprintf("parm1 = %d\n",mod->nlCons[intNum].pCon.parm1);
      Rprintf("parm2 = %d\n",mod->nlCons[intNum].pCon.parm2);
      Rprintf("xPosition = %lf\n",mod->nlCons[intNum].pCon.xPos);
    }
    if (cSet.hasSubCon){
      Rprintf("\nnlCON %d:\n", intNum);
      Rprintf("type = %d\n", mod->nlCons[conNum].subCon.conType);
      Rprintf("length = %d\n",mod->nlCons[conNum].subCon.length);
      PrintIntMatrix2File(stderr,(int *)mod->nlCons[conNum].subCon.motif,1, mod->nlCons[conNum].subCon.length);
      Rprintf("cut-off = %lf\n",mod->nlCons[conNum].subCon.cutOff);
    }
  }



  /* allocate memory for mod->low, mod->up, and mod->linConMatrix */
  allocMod(mod);


  /***********************************
   * Creating lower and upper bounds *
   ***********************************/

  /* constrain all PWM paramters to 0<p<1 */
  for (index=0;index<4*width;index++){
    /* log-likelihood function is not differentiable for p=0 */
    mod->low[index] = 1e-16;
    mod->up[index]  = 1.0;
  }

  if (extraParm){
    mod->low[index] = 1e-16;
    modType         = mType == TCM ? dataSet->approx : mType;
    maxNumSites     = mType == TCM ? dataSet->tcmMaxNumSites : dataSet->zoopsMaxNumSites;
    mod->up[index]  = getIntensity(modType, width, maxNumSites, dataSet) - 1e-16;
    index++;
  }



  /* Bounds on PWM parameters */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    for(posFreqCon = cSet.intCons[intNum].posFreqCons;posFreqCon!=NULL;posFreqCon=posFreqCon->next){
      /* convert A->0, C->1, G->2, T->3, AT->4, GC->5 */
      nuc = (int) posFreqCon->nuc;
      if (nuc<4){  
	mod->low[4*(intFirst[intNum] + posFreqCon->pos) +nuc] = posFreqCon->low;
	mod->up[4*(intFirst[intNum] + posFreqCon->pos)+nuc]  = 1.0;
	mod->hasPosFreqCons = 1;
      }
    }
  }


  /* Bounds on extra shape parameters */
  index = 4*width + extraParm;
  for (intNum=0;intNum<cSet.numInts;intNum++){
    if (cSet.intCons[intNum].hasShapeCon){
      mod->low[index] = cSet.intCons[intNum].shapeCon.startLow;
      mod->up[index] = cSet.intCons[intNum].shapeCon.startUp;
      index++;
      mod->low[index] = cSet.intCons[intNum].shapeCon.endLow;
      mod->up[index] = cSet.intCons[intNum].shapeCon.endUp;
      index++;
    }
  }

  /* Constraints to make all pos frequencies add up to 1 */
  for (pos=0;pos<width;pos++){
    mod->low[index] = 1;
    mod->up[index] = 1;
    index++;
  }


  /* PosFreqCons for nucleotides 'AT' and 'GC' are linear */
  /* bounds on linear frequency constraints start at index = mod->n + width */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    for(posFreqCon = cSet.intCons[intNum].posFreqCons;posFreqCon!=NULL;posFreqCon=posFreqCon->next){
      if (posFreqCon->nuc==AT || posFreqCon->nuc==GC){
	mod->low[index] = posFreqCon->low;
	mod->up[index]  = 1.0;
	index++;
      }
    }
  }


  /* Bounds on IntervalFreqCons */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    for (intFreqCon = cSet.intCons[intNum].intFreqCons;intFreqCon!=NULL;intFreqCon=intFreqCon->next){
      for (j=0;j<cSet.intCons[intNum].numIntFreqCons;j++){
	mod->low[index] = intNumBasePairs[intNum] * intFreqCon->low;
	mod->up[index]  = intNumBasePairs[intNum] * 1;
	index++;
      }
    }
  }


  /* Error bounds on palindromic constraints */ 
  if (cSet.hasPalCon){
    for (intNum=0;intNum<4*intNumBasePairs[cSet.palCon.int1];intNum++){
      mod->low[index] = -cSet.palCon.error;
      mod->up[index]  =  cSet.palCon.error;
      index++;
    }
  }


  /* linear constraints on profile parameters */
  for (linCon = cSet.linCons; linCon!=NULL;linCon=linCon->next){
    mod->low[index] = linCon->low;
    mod->up[index]  = linCon->up;
    index++;
  }


  /* Nonlinear profile constraints */
  if (mod->nonlin>0){
    for (intNum=0;intNum<mod->nonlin;intNum++){
      if (mod->nlCons[intNum].pCon.conType == bdd){
	mod->low[index] = cSet.intCons[mod->nlCons[intNum].pCon.interval].boundCon.low;
	mod->up[index]  = cSet.intCons[mod->nlCons[intNum].pCon.interval].boundCon.up;
	index++;
      }else if(mod->nlCons[intNum].pCon.conType == monInc){
	if (mod->nlCons[intNum].pCon.ref==prevBrk ||mod->nlCons[intNum].pCon.ref==nextBrk){
	  mod->low[index] = 0;
	  mod->up[index]  = 0;
	}else{
	  mod->low[index] = -cSet.intCons[mod->nlCons[intNum].pCon.interval].shapeCon.error;
	  mod->up[index]  =  2;
	}
	index++;
      }else if(mod->nlCons[intNum].pCon.conType == monDec){
	if (mod->nlCons[intNum].pCon.ref==prevBrk ||mod->nlCons[intNum].pCon.ref==nextBrk){
	  mod->low[index] = 0;
	  mod->up[index]  = 0;
	}else{
	  mod->low[index] = -2;
	  mod->up[index]  =  cSet.intCons[mod->nlCons[intNum].pCon.interval].shapeCon.error;
	}
	index++;
      }else if(mod->nlCons[intNum].pCon.conType == lin){
	mod->low[index] = -cSet.intCons[mod->nlCons[intNum].pCon.interval].shapeCon.error;
	mod->up[index]  =  cSet.intCons[mod->nlCons[intNum].pCon.interval].shapeCon.error;
	index++;
      }else if (mod->nlCons[intNum].pCon.conType == sub){
	mod->low[index] = -1;
	/* upper limit is set so that penalty on average can be as large as p=exp(-5*p_cutoff) */ 
	mod->up[index] =   exp(-5*cSet.subCon.cutOff)*cSet.subCon.length;
	index++;
      }
    }
  }
  if (COSMO_MSG_LEVEL>=4){  
    for (intNum=0;intNum<index;intNum++){
      Rprintf("%d: lower = %lf and upper = %lf\n", intNum, mod->low[intNum], mod->up[intNum]);
    }
  }   

  /*****************************************
   * Creating the linear constraint matrix *
   *****************************************/

  /* Making all position frequencies sum up to 1 */
  for (index=0;index<width;index++){
    for (k=0;k<mod->n;k++){
      mod->linConMatrix[LCM(index,k)] = 0;
    }
    for (k=0;k<4;k++){
      mod->linConMatrix[LCM(index,index*4 + k)] = 1;
    }
  }

  /* PosFreqCons for nucleotides 'AT' and 'GC' are linear */
  /* linFreqCons start at column width in linConMatrix */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    for (posFreqCon = cSet.intCons[intNum].posFreqCons;posFreqCon!=NULL;posFreqCon=posFreqCon->next){
      if (posFreqCon->nuc==AT || posFreqCon->nuc==GC){
	for (k=0;k<mod->n;k++){
	  mod->linConMatrix[LCM(index,k)] = 0;
	}
	if (posFreqCon->nuc==AT){
	  mod->linConMatrix[LCM(index,4*posFreqCon->pos + 0)] = 1;
	  mod->linConMatrix[LCM(index,4*posFreqCon->pos + 3)] = 1;			
	}else{
	  mod->linConMatrix[LCM(index,4*posFreqCon->pos + 1)] = 1;
	  mod->linConMatrix[LCM(index,4*posFreqCon->pos + 2)] = 1;			
	}
	index++;
      }
    }
  }


  /* IntervalFreqCons */
  for (intNum=0;intNum<cSet.numInts;intNum++){
    for (intFreqCon = cSet.intCons[intNum].intFreqCons;intFreqCon!=NULL;intFreqCon=intFreqCon->next){
      for (k=0;k<mod->n;k++){
	mod->linConMatrix[LCM(index,k)] = 0;
      }
      switch (intFreqCon->nuc){
      case AT:
	for (pos=intFirst[intNum];pos<=intLast[intNum];pos++){
	  mod->linConMatrix[LCM(index,4*pos + 0)] = 1;
	  mod->linConMatrix[LCM(index,4*pos + 3)] = 1;
	}
	break;
      case GC:
	for (pos=intFirst[intNum];pos<=intLast[intNum];pos++){
	  mod->linConMatrix[LCM(index,4*pos + 1)] = 1;
	  mod->linConMatrix[LCM(index,4*pos + 2)] = 1;
	}
	break;
      default:
	/* convert A->0, C->1, G->2, T->3*/
	nuc = (int) intFreqCon->nuc;
	for (pos=intFirst[intNum];pos<=intLast[intNum];pos++){
	  mod->linConMatrix[LCM(index,4*pos + nuc)] = 1;
	}
      }	    
      index++;
    }
  }

  /* Palindromic constraint */
  if (cSet.hasPalCon){
    for (j=0;j<4*intNumBasePairs[cSet.palCon.int1];j++){
      for (k=0;k<mod->n;k++){
	mod->linConMatrix[LCM(index,k)] = 0;
      }
      mod->linConMatrix[LCM(index,intFirst[cSet.palCon.int1] + j)] =  1;
      mod->linConMatrix[LCM(index,4*width - intFirst[cSet.palCon.int1] - j - 1)] = -1;
      index++;
    }
  }


  /* linear constraints on extra shape parameters */
  if (cSet.numLinCons>0){
    for (linCon = cSet.linCons; linCon!=NULL;linCon=linCon->next){
      for (k=0;k<mod->n;k++){
	mod->linConMatrix[LCM(index,k)] = 0;
      }
      mod->linConMatrix[LCM(index,4*width + extraParm + linCon->index1)] =  1;
      mod->linConMatrix[LCM(index,4*width + extraParm + linCon->index2)] = -1;
      index++;
    }
  }

  if (COSMO_MSG_LEVEL>=3){
    Rprintf("\nLinCon matrix:\n");
    PrintIntMatrix2File(stderr,mod->linConMatrix,mod->n,mod->nlin);
  }


  return TRUE;
}



/**********************************************************************/
/*
  findLine

  Find the next line in file 'conFile' that starts with the string
  'param'. This line must be found before the next line starting
  with the character 'before', e.g. before='>' indicating the beginning 
  of a new command, or before='@' indicating the beginning of a new
  constraint set.

*/
/**********************************************************************/
int findLine(
	     CONSTRING *conString,    /* the constraint file CONSTRING */
	     char *line,              /* character array for line-by-line input */
	     char *first,             /* character array for line-by-line input */
	     char *param,             /* the parameter to find */
	     char *command,           /* the command for which to find the parameter */             
	     char before,             /* character that must not be found before parameter */
	     int chop                 /* should the last character be chopped off? */
	     )
{

  while (getLine(conString,line)!=NULL){

    if (!strcmp(line,"\n"))  continue;
    if (!strcmp(line,"\r\n"))  continue; /* for constraints from HTML form */
    if (sscanf(line,"%s", first)==EOF){
      Rprintf("%s :Couldn't find parameter %s\n", command, param);
      return(0);
    }

    /* chop off ':' from the end */
    if (chop) first[strlen(first)-1] = '\0';
    if (!strcmp(first,param)){
      return(1);
    }else if(first[0]==before){
      Rprintf("%s: Couldn't find parameter %s\n", command, param);
      Rprintf("    Encountered new command %s\n", line);
      return(0);
    }
	
  }
    
  Rprintf("%s: Couldn't find parameter %s\n", command, param);
  return(0);

}
/**********************************************************************/
/*
  findCommand

  Find the next line in file 'conFile' that starts with 'symbol', e.g. 
  symbol='>'  indicating a command, or symbol='@' indicating a new
  constraint set. If the character 'before' is found before 'symbol',
  the function returns -1.

  Returns 1 if successful, 0 otherwise.

*/
/**********************************************************************/
int findCommand(
		CONSTRING *conString,    /* the constraint file */
		char *line,              /* character array for line-by-line input */
		char symbol,             /* the symbol at the beginning of the line to look for */
		char before              /* the character that must not be found before the symbol */
		)
{

  while (getLine(conString,line)!=NULL){
    if (line[0]==symbol){
      /* first character of "line" is ">", so we found a potential new command */
      return 1;
    }else if (line[0]==before){
      return -1;
    }
  }

  return 0;
}

    

/**********************************************************************/
/*
  addShapeCon

  Add a shape constraint to a given interval of a given constraint
  set. The following parameters are read in from the constraint file:

  1. The shape of the IC profile across this interval
  2. A lower bound on the IC at the beginning of the interval
  3. An upper bound on the IC at the beginning of the interval
  4. A lower bound on the IC at the end of the interval
  5. An upper bound on the IC at the end of the interval
  6. A bound on how much this constraint may be violated

  The format in which these parameters must be specified is in 
  the constraint file is:

  Interval: {1,...,numInts}
  Shape: {Linear,MonotoneIncreasing,MonotoneDecreasing}          
  LeftBounds: [0,2] to [0,2]
  RightBounds: [0,2] to [0,2]
  ErrorTol: [0-1]

  Commenting lines may be present between these lines.


*/
/**********************************************************************/
int addShapeCon(
		 CONSET *cSet,            /* array of constraint sets */
		 CONSTRING *conString,    /* constraint file in CONSTRING */
		 int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
		 int intNum,              /* the interval to which the shape constraint is added */
		 char *line,              /* character arrays for line-by-line input */
		 char *word,
		 char *first
		 )
{
  int number, i;


  /* read in parameter "Shape" */
  if (!findLine(conString, line, first, "Shape", "addShapeCon",'>',1)) return(0);
  if (sscanf(line,"%s %s", first, word)!=2){
    Rprintf("addShapeCon: Error reading in parameter Shape\n");
    return(0);
  }
  if (!strcmp(word,"Linear")){
    cSet[cSetNum].intCons[intNum].shapeCon.shape = lin;
  }else if (!strcmp(word,"MonotoneIncreasing")){
    cSet[cSetNum].intCons[intNum].shapeCon.shape = monInc;
  }else if (!strcmp(word,"MonotoneDecreasing")){
    cSet[cSetNum].intCons[intNum].shapeCon.shape = monDec;
  }else{
    Rprintf("Couldn't recognize profile constraints %s. Exiting...\n", word);
    return(0);
  }


  /* read in parameter "LeftBounds" */
  if (!findLine(conString, line, first, "LeftBounds", "addShapeCon",'>',1)) return(0);
  if (sscanf(line,"%s %lf %s %lf", first, &(cSet[cSetNum].intCons[intNum].shapeCon.startLow), word, &(cSet[cSetNum].intCons[intNum].shapeCon.startUp))!=4){
    Rprintf("addShapeCon: Error reading in parameter LeftBounds\n");
    return(0);
  }

  /* read in parameter "RightBounds" */
  if (!findLine(conString, line, first, "RightBounds", "addShapeCon",'>',1)) return(0);
  if (sscanf(line,"%s %lf %s %lf", first, &(cSet[cSetNum].intCons[intNum].shapeCon.endLow), word, &(cSet[cSetNum].intCons[intNum].shapeCon.endUp))!=4){
    Rprintf("addShapeCon: Error reading in parameter RightBounds\n");
    return(0);
  }

  /* read in parameter "ErrorTol" */
  if (!findLine(conString, line, first, "ErrorTol", "addShapeCon",'>',1)) return(0);
  if (sscanf(line,"%s %lf", first, &(cSet[cSetNum].intCons[intNum].shapeCon.error))!=2){
    Rprintf("addShapeCon: Error reading in parameter ErrorTol\n");
    return(0);
  }

  if (cSet[cSetNum].intCons[intNum].shapeCon.startUp<cSet[cSetNum].intCons[intNum].shapeCon.startLow){
    Rprintf("addShapeCon: upper bound on first shape parameter is lower than lower bound\n");
    return(0);
  }

  if (cSet[cSetNum].intCons[intNum].shapeCon.endUp<cSet[cSetNum].intCons[intNum].shapeCon.endLow){
    Rprintf("addShapeCon: upper bound on second shape parameter is lower than lower bound\n");
    return(0);
  }


  cSet[cSetNum].intCons[intNum].hasShapeCon = 1;
  cSet[cSetNum].numShapeCons++;

  /* updating the shape constraint numbers */
  number = 0;

  for (i=0;i<cSet[cSetNum].numInts;i++){
    if (cSet[cSetNum].intCons[i].hasShapeCon){
      cSet[cSetNum].intCons[i].shapeCon.number = number;
      number++;
    }
  }
  if (COSMO_MSG_LEVEL>=2) Rprintf("addShapeCon: Added shape constraint to interval %d of conSet %d\n", intNum+1, cSetNum);

  return(1);
}


/**********************************************************************/
/*
  addBoundCon

  Add a bound constraint on the information content in a given 
  interval of a given constraint set. The following parameters
  are read in from the constraint file:

  1. A lower bound on the IC across this interval
  2. An upper bound on the IC across this interval

  The format in which these parameters must be specified is in 
  the constraint file is:

  Interval: {1,...,numInts}
  Bounds: [0,2] to [0,2]


  Commenting lines may be present between these lines.

  Returns 1 if input is ok, 0 otherwise.

*/
/**********************************************************************/
int addBoundCon(
		 CONSET *cSet,            /* array of constraint sets */
		 CONSTRING *conString,    /* the constraint file */
		 int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
		 int intNum,              /* the interval to which the shape constraint is added */
		 char *line,              /* character arrays for line-by-line input */
		 char *word,
		 char *first
		 )
{

  /* read in parameter "Lower" */
  if (!findLine(conString, line, first, "Bounds", "addBoundCon",'>',1)) return(0);
  if (sscanf(line,"%s %lf %s %lf", first, &(cSet[cSetNum].intCons[intNum].boundCon.low), word, &(cSet[cSetNum].intCons[intNum].boundCon.up))!=4){
    Rprintf("addBoundCon: Error reading in parameter Bounds\n");
    return(0);
  }
    
    
  if (cSet[cSetNum].intCons[intNum].boundCon.up<cSet[cSetNum].intCons[intNum].boundCon.low){
    Rprintf("addBoundCon: Upper bound is lower than lower bound\n");
    return(0);
  }
  cSet[cSetNum].intCons[intNum].hasBoundCon = 1;
  cSet[cSetNum].numBoundCons++;

  if (COSMO_MSG_LEVEL>=2) Rprintf("addBoundCon: Added bound constraint to conSet %d interval %d\n", cSetNum, intNum+1);    
  return(1);
}

/**********************************************************************/
/*
  addIntFreqCon

  Add a constraint on the nucleotide frequencies across a given 
  interval in a given constraint set. The following parameters 
  are read in from the constraint file:

  1. The nucleotide or pair of nucleotides
  2. A lower bound on the average frequency of this nucleotide
  or pair of nucleotides across the interval

  The format in which these parameters must be specified is in 
  the constraint file is:

  Nuc: {a,A,c,C,g,G,t,T,at,AT,gc,GC}
  LowerBound: [0-1]                                             

  Commenting lines may be present between these lines.

  Returns 1 if input is ok, 0 otherwise.

*/
/**********************************************************************/
int addIntFreqCon(
		   CONSET *cSet,            /* array of constraint sets */
		   CONSTRING *conString,    /* the constraint file */
		   int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
		   int intNum,              /* the interval to which the shape constraint is added */
		   char *line,              /* character arrays for line-by-line input */
		   char *word,
		   char *first
		   )
{
  INTFREQCON *lastIntFreqCon;  /* pointer to the last IntFreq constraint */
  INTFREQCON *newIntFreqCon;   /* pointer to the IntFreq cosntraint to be added */

  newIntFreqCon = intFreqConMalloc(1); 
  if (cSet[cSetNum].intCons[intNum].intFreqCons==NULL){
    cSet[cSetNum].intCons[intNum].intFreqCons = newIntFreqCon;
  }else{
    lastIntFreqCon = cSet[cSetNum].intCons[intNum].intFreqCons;
    while (lastIntFreqCon->next != NULL) lastIntFreqCon = lastIntFreqCon->next;
    lastIntFreqCon->next = newIntFreqCon;
  }


  /* read in parameter "Nuc" */
  if (!findLine(conString, line, first, "Nuc", "addIntFreqCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %s", first, word)!=2){
    Rprintf("addIntFreqCon: Error reading in nucleotide for interval frequency constraint\n");
    return(0);
  }
  if ((!strcmp(word,"A"))||(!strcmp(word,"a"))){
    newIntFreqCon->nuc = A;
  }else if((!strcmp(word,"C"))||(!strcmp(word,"c"))){
    newIntFreqCon->nuc = C;
  }else if((!strcmp(word,"G"))||(!strcmp(word,"g"))){
    newIntFreqCon->nuc = G;
  }else if((!strcmp(word,"T"))||(!strcmp(word,"t"))){
    newIntFreqCon->nuc = T;
  }else if((!strcmp(word,"AT"))||(!strcmp(word,"at"))){
    newIntFreqCon->nuc = AT;
  }else if((!strcmp(word,"GC"))||(!strcmp(word,"gc"))){
    newIntFreqCon->nuc = GC;
  }else{
    Rprintf("addIntFreqCon: Couldn't recognize nucleotide %s. Exiting...\n",word);
    return(0);
  }


  /* read in parameter "LowerBound" */
  if (!findLine(conString, line, first, "LowerBound", "addIntFreqCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %lf", first, &(newIntFreqCon->low))!=2){
    Rprintf("addIntFreqCon: Error reading in lower bound for interval frequency constraint\n");
    return(0);
  }

  if (newIntFreqCon->low<0 || newIntFreqCon->low>1){
    Rprintf("addIntFreqCon: Lower bound must be between 0 and 1\n");
    return(0);
  }
    
  if (cSet[cSetNum].sharpestBound<newIntFreqCon->low) cSet[cSetNum].sharpestBound = newIntFreqCon->low;
  cSet[cSetNum].intCons[intNum].numIntFreqCons++;
  cSet[cSetNum].numIntFreqCons++;

  if (COSMO_MSG_LEVEL>=2) Rprintf("addIntFreqCon: Added interval frequency constraint to conSet %d interval %d\n", cSetNum, intNum+1);    
  return(1);
}

/**********************************************************************/
/*
  addPosFreqCon

  Add a constraint on the nucleotide frequency of a given nucleotide
  in a given position of a given fixed interval in a given constraint
  set. The following parameters are read in from the constraint file:

  1. The position of this nucleotide within in the interval
  2. The nucleotide whose frequency is to be bounded
  3. A lower bound for the nucleotide's frequency

  The format in which these parameters must be specified is in 
  the constraint file is:

  Pos: {1,...,numBasePairs}
  Nuc: {a,A,c,C,g,G,t,T,at,AT,gc,GC}
  LowerBound: [0-2]                                             

  Commenting lines may be present between these lines.

  Returns 1 if input is ok, 0 otherwise.

*/
/**********************************************************************/
int addPosFreqCon(
		   CONSET *cSet,            /* array of constraint sets */
		   CONSTRING *conString,    /* the constraint file */
		   int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
		   int intNum,              /* the interval to which the shape constraint is added */
		   int pos,                 /* position in interval for which we give a lower bound */
		   char *line,              /* character arrays for line-by-line input */
		   char *word,
		   char *first
		   )
{
  POSFREQCON *lastPosFreqCon;  /* pointer to the last PosFreq constraint */
  POSFREQCON *newPosFreqCon;   /* pointer to the PosFreq constraint to be added */

  newPosFreqCon = posFreqConMalloc(1);
  if (cSet[cSetNum].intCons[intNum].posFreqCons==NULL){
    cSet[cSetNum].intCons[intNum].posFreqCons = newPosFreqCon;    
  }else{
    lastPosFreqCon = cSet[cSetNum].intCons[intNum].posFreqCons;
    while (lastPosFreqCon->next != NULL) lastPosFreqCon = lastPosFreqCon->next;
    lastPosFreqCon->next = newPosFreqCon;
  }

  newPosFreqCon->pos = pos-1; /*for zero-based arrays */

  /* read in parameter "Nucleotide" */
  if (!findLine(conString, line, first, "Nuc", "addPosFreqCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %s", first, word)!=2){
    Rprintf("addPosFreqCon: Error reading in nucleotide for position frequency constraint\n");
    return(0);
  }
  if ((!strcmp(word,"A"))||(!strcmp(word,"a"))){
    newPosFreqCon->nuc = A;
  }else if((!strcmp(word,"C"))||(!strcmp(word,"c"))){
    newPosFreqCon->nuc = C;
  }else if((!strcmp(word,"G"))||(!strcmp(word,"g"))){
    newPosFreqCon->nuc = G;
  }else if((!strcmp(word,"T"))||(!strcmp(word,"t"))){
    newPosFreqCon->nuc = T;
  }else if((!strcmp(word,"AT"))||(!strcmp(word,"at"))){
    newPosFreqCon->nuc = AT;
  }else if((!strcmp(word,"GC"))||(!strcmp(word,"gc"))){
    newPosFreqCon->nuc = GC;
  }else{
    Rprintf("addPosFreqCon: Couldn't recognize nucleotide %s. Exiting...\n",word);
    return(0);
  }

  /* read in parameter "Lower" */
  if (!findLine(conString, line, first, "LowerBound", "addPosFreqCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %lf", first, &(newPosFreqCon->low))!=2){
    Rprintf("addPosFreqCon: Error reading in lower bound for position frequency constraint\n");
    return(0);
  }


  if (newPosFreqCon->low<0 || newPosFreqCon->low>1){
    Rprintf("addPosFreqCon: lower bound must be between 0 and 1\n");
    return(0);
  }

  cSet[cSetNum].intCons[intNum].numPosFreqCons++;

  if (COSMO_MSG_LEVEL>=2) Rprintf("addPosFreqCon: Added position frequency constraint to conSet %d interval %d\n", cSetNum, intNum+1);
  return(1);
}

/**********************************************************************/
/*
  addLinCon

  Impose bounds on the difference of two parameters of the 
  information content profile. The following parameters are read 
  in from the constraint file:

  1. The interval that the first parameter belongs to
  2. The position of the paramter in that interval (start==a or end==b)
  3. The interval that the second parameter belongs to
  4. The position of the paramter in that interval (start==a or end==b)
  5. A lower bound on (parameter1 - parameter2)
  6. An upper bound on (parameter1 - parameter2)

  The format in which these parameters must be specified is in 
  the constraint file is:

  Parameters: {1,...,numInts}{a,b} - {1,...,numInts}{a,b} 
  Bounds: [-2,2] to [-2,2]


  Commenting lines may be present between these lines.

  Returns 1 if input is ok, 0 otherwise.

*/
/**********************************************************************/
int addLinCon(
	       CONSET *cSet,            /* array of constraint sets */
	       CONSTRING *conString,    /* the constraint file */
	       int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
	       char *line,              /* character arrays for line-by-line input */
	       char *word,
	       char *first
	       )
{

  LINCON *lastLinCon;     /* pointer to the last linear constraint */
  LINCON *newLinCon;      /* pointer to the linear constraint to be added */

  char *firstParm, *secParm, letter;

  firstParm   = CharMalloc(MAXNUMCHARS+1);
  secParm     = CharMalloc(MAXNUMCHARS+1);


  /* get pointer to the linear constraint to be added */
  newLinCon = linConMalloc(1);
  if (cSet[cSetNum].linCons==NULL){
    cSet[cSetNum].linCons = newLinCon;
  }else{
    lastLinCon = cSet[cSetNum].linCons;
    while (lastLinCon->next!= NULL) lastLinCon = lastLinCon->next;
    lastLinCon->next = newLinCon;
  }

  /* read in parameter "Parameters" */
  if (!findLine(conString, line, first, "Parameters", "addLinCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %s %s %s", first, firstParm,word,secParm)!=4){
    Rprintf("addLinCon: Error reading in parameters for parameter difference constraint\n");
    return(0);
  }

  letter = firstParm[1];
  firstParm[strlen(firstParm)-1] = '\0';
  if (sscanf(firstParm,"%d", &(newLinCon->int1))!=1){
    Rprintf("addLinCon: Error reading in interval 1 for parameter difference constraint\n");
    return(0);
  }
  newLinCon->int1--; /* for zero-based arrays */


  if (letter=='a'){
    if (COSMO_MSG_LEVEL>=4) Rprintf("parm1 = interval %d start\n", newLinCon->int1 + 1);
    newLinCon->pos1 = start;
  }else if (letter=='b'){
    if (COSMO_MSG_LEVEL>=4) Rprintf("parm1 = interval %d end\n", newLinCon->int1 + 1);
    newLinCon->pos1 = end;
  }else{
    Rprintf("addLinCon: Error reading in edge 1 for parameter difference constraint\n");
    return(0);
  }

  letter = secParm[1];
  secParm[strlen(secParm)-1] = '\0';
  if (sscanf(secParm,"%d", &(newLinCon->int2))!=1){
    Rprintf("addLinCon: Error reading in interval 2 for parameter difference constraint\n");
    return(0);
  }
  newLinCon->int2--; /* for zero-based arrays */

  if (letter=='a'){
    if (COSMO_MSG_LEVEL>=4) Rprintf("parm2 = interval %d start\n", newLinCon->int2 + 1);
    newLinCon->pos2 = start;
  }else if (letter=='b'){
    if (COSMO_MSG_LEVEL>=4)Rprintf("parm2 = interval %d end\n", newLinCon->int2 + 1);
    newLinCon->pos2 = end;
  }else{
    Rprintf("addLinCon: Error reading in edge 2 for parameter difference constraint\n");
    return(0);
  }
  
	
  /* read in parameter "Bounds" */
  if (!findLine(conString, line, first, "Bounds", "addLinCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %lf %s %lf", first, &(newLinCon->low),word,&(newLinCon->up))!=4){
    Rprintf("addLinCon: Error reading in bounds for parameter difference constraint\n");
    return(0);
  }
  if (COSMO_MSG_LEVEL>=4)  Rprintf("Bounds are from %lf to %lf\n",newLinCon->low, newLinCon->up); 

  if (!cSet[cSetNum].intCons[newLinCon->int1].hasShapeCon){
    Rprintf("addLinCon: Interval %d has no shape constraint\n", newLinCon->int1+1);
    return(0);
  }

  if (!cSet[cSetNum].intCons[newLinCon->int2].hasShapeCon){
    Rprintf("addLinCon: Interval %d has no shape constraint\n", newLinCon->int2+1);
    return(0);
  }

  if (newLinCon->up<newLinCon->low){
    Rprintf("addLinCon: Upper bound is lower than lower bound\n");
    return(0);
  }

  /* give the proper indices to the shape parameters involved */
  newLinCon->index1 = 2*cSet[cSetNum].intCons[newLinCon->int1].shapeCon.number + (newLinCon->pos1==end);
  newLinCon->index2 = 2*cSet[cSetNum].intCons[newLinCon->int2].shapeCon.number + (newLinCon->pos2==end);

  cSet[cSetNum].numLinCons++;
  if (COSMO_MSG_LEVEL>=2) Rprintf("addLinCon: Added parameter difference constraint to conSet %d\n", cSetNum);
    
  return(1);
}

/**********************************************************************/
/*
  addPalCon

  Add a palindromicity constraint to a given constraint set. The
  following parameters are read in from the constraint file:

  1. A first interval
  2. A second interval that is palindromic wrt the first one
  3. An error bound for how different a given pair of supposedly
  palindromic nucleotide frequencies can be.

  The format in which these parameters must be specified is in 
  the constraint file is:

  Intervals: {1,...,numInts} and {1,...,numInts}
  ErrorTol: [0,1]

  Commenting lines may be present between these lines.

  Returns 1 if input is ok, 0 otherwise.

*/
/**********************************************************************/
int addPalCon(
	       CONSET *cSet,            /* array of constraint sets */
	       CONSTRING *conString,    /* the constraint file */
	       int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
	       char *line,              /* character arrays for line-by-line input */
	       char *word,
	       char *first
	       )
{
  int int1, int2;

  /* read in parameter "Interval1" */
  if (!findLine(conString, line, first, "Intervals", "addPalCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %d %s %d", first, &(cSet[cSetNum].palCon.int1), word,&(cSet[cSetNum].palCon.int2))!=4){
    Rprintf("addPalCon: Error reading in intervals for palindromic constraint\n");
    return(0);
  }
  cSet[cSetNum].palCon.int1--; /* for zero-based arrays */
  cSet[cSetNum].palCon.int2--; /* for zero-based arrays */
	
  /* read in parameter "ErrorBound" */
  if (!findLine(conString, line, first, "ErrorTol", "addPalCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %lf", first, &(cSet[cSetNum].palCon.error))!=2){
    Rprintf("addPalCon: Error reading in lower bound for palindromic constraint\n");
    return(0);
  }
	
  if ((cSet[cSetNum].palCon.error<0) || (cSet[cSetNum].palCon.error>1)){
    Rprintf("addPalCon: Palindromic error bound must between zero and one\n");
    return(0);
  }

  if ((cSet[cSetNum].palCon.int1<0) || (cSet[cSetNum].palCon.int1>cSet[cSetNum].numInts)){
    Rprintf("addPalCon: Invalid interval\n");
    return(0);
  }

  int1 = cSet[cSetNum].palCon.int1;
  int2 = cSet[cSetNum].palCon.int2;

  if ((cSet[cSetNum].intCons[int1].lenType != cSet[cSetNum].intCons[int2].lenType)){
    Rprintf("addPalCon: The two intervals are not of the same type\n");
    return(0);
  }

  if (cSet[cSetNum].intCons[int1].lenType==bp){
    if ((cSet[cSetNum].intCons[int1].numBasePairs != cSet[cSetNum].intCons[int2].numBasePairs)){
      Rprintf("addPalCon: The two intervals are not of the same length\n");
      return(0);
    }
  }else if (cSet[cSetNum].intCons[int1].lenType==prop){
    if ((cSet[cSetNum].intCons[int1].prop != cSet[cSetNum].intCons[int2].prop)){
      Rprintf("addPalCon: The two intervals are not of the same length\n");
      return(0);
    }
  }else{
    Rprintf("addPalCon: The two intervals can't be of type 'variable'\n");
    return(0);
  }

    
  cSet[cSetNum].hasPalCon = 1;
  if (COSMO_MSG_LEVEL>=2) Rprintf("addPalCon: Added palindromic constraint to conSet %d\n", cSetNum);        
  return(1);
}

/**********************************************************************/
/*
  addSubCon

  Add a submotif constraint to a given constraint set. The
  following parameters are read in from the constraint file:

  1. The submotif, e.g. "AAT"
  2. A cutoff probability that the given nucleotide frequencies
  must exceed

  The format in which these parameters must be specified is in 
  the constraint file is:

  Motif: e.g. AAT or aat
  MinFreq: [0,1]

  Commenting lines may be present between these lines.

  Returns 1 if input is ok, 0 otherwise.

*/
/**********************************************************************/
int addSubCon(
	       CONSET *cSet,            /* array of constraint sets */
	       CONSTRING *conString,    /* the constraint file */
	       int cSetNum,             /* the index of the constraint set to which the shape constraint is added */
	       char *line,              /* character arrays for line-by-line input */
	       char *word,
	       char *first
	       )
{
  int numChar;
  char oneLetter;

  /* read in parameter "Motif" */
  if (!findLine(conString, line, first, "Motif", "addSubCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %s", first, word)!=2){
    Rprintf("addSubCon: Error reading in motif for submotif constraint\n");
    return(0);
  }

  cSet[cSetNum].subCon.conType = sub;
  cSet[cSetNum].subCon.length = strlen(word);
  cSet[cSetNum].subCon.motif = (NUC*)S_alloc((unsigned)((cSet[cSetNum].subCon.length+1)),(sizeof(NUC)));
  if (cSet[cSetNum].subCon.motif == NULL){  
    Rprintf("addSubCon: Couldn't get %d NUCs. Exiting...\n", cSet[cSetNum].subCon.length);
    return(0);
  }

  for (numChar=0;numChar<cSet[cSetNum].subCon.length;numChar++){
    oneLetter = toupper(word[numChar]);
   
    if (oneLetter=='A' || oneLetter=='a'){
      cSet[cSetNum].subCon.motif[numChar] = A;
    }else if (oneLetter=='C' || oneLetter=='c'){
      cSet[cSetNum].subCon.motif[numChar] = C;
    }else if (oneLetter=='G' || oneLetter=='g'){
      cSet[cSetNum].subCon.motif[numChar] = G;
    }else if (oneLetter=='T' || oneLetter=='t'){
      cSet[cSetNum].subCon.motif[numChar] = T;
    }else{
      Rprintf("addSubCon: Invalid character in submotif: %c\n", oneLetter);
      return(0);
    }	       
  }

  /* read in parameter "MinFreq" */
  if (!findLine(conString, line, first, "MinFreq", "addSubCon",'>',1)) return(0);	    
  if (sscanf(line,"%s %lf", first, &(cSet[cSetNum].subCon.cutOff))!=2){
    Rprintf("addSubCon: Error reading in minimum frequency for submotif constraint\n");
    return(0);
  }
	
  if ((cSet[cSetNum].subCon.cutOff<0) || (cSet[cSetNum].subCon.cutOff>1)){
    Rprintf("addSubCon: Cut-off must between zero and one\n");
    return(0);
  }

  if (cSet[cSetNum].subCon.length>cSet[cSetNum].widthLow){
    Rprintf("addSubCon: Length of submotif (%d) is greater than smallest overall motif width (%d)\n", cSet[cSetNum].subCon.length,cSet[cSetNum].widthLow);
    return(0);
  }

    

  cSet[cSetNum].hasSubCon = 1;
  if (COSMO_MSG_LEVEL>=2) Rprintf("addSubCon: Added submotif constraint to conSet %d\n", cSetNum);        
  return(1);
}

/**********************************************************************/
/*
  checkInput

  Checks whether an interval number and a constraint set number
  are within the necessary bounds.


*/
/**********************************************************************/
int checkInput(
		CONSET *conSets,      /* array of constraint sets */
		int intNum,           /* the interval number to check */
		int cSetNum,          /* the constraint set number to check */
		int numConSets        /* the number of constraint sets */
		)
{
  if (cSetNum<0 || cSetNum>=numConSets){
    Rprintf("readConFile: invalid conSet number\n");
    return(0);
  }
  if (intNum<0 || intNum>=conSets[cSetNum].numInts){
    Rprintf("readConFile: invalid interval number\n");
    return(0);
  }
    
  return(1);
}

/**********************************************************************/
/*
  checkConSetNum

  Checks whether constraint set number is within the necessary 
  bounds.


*/
/**********************************************************************/
int checkConSetNum(
		    int conSetNum,         /* the constraint set number to check*/
		    int numConSets         /* the number of constraint sets */
		    )
{
  if (conSetNum<0 || conSetNum>=numConSets){
    Rprintf("readConFile: invalid conSet number\n");
    return(0);
  }
  return(1);
}


/**********************************************************************/
/*
  readConFile

  Read constraint file to add constraints to constraint set that
  has already been initialized.


*/
/**********************************************************************/
int readConFile(
		DATASET *dataSet
		)
{

  int numFree;                   /* number of empty constraint sets */
  int cSetNum;                   /* index for iterating over constraint sets */
  char *line;
  line = dataSet->line;


  if (dataSet->hasConstraints){

    if (!initConSets(dataSet)) return(0);
    if (!addConstraints(dataSet)) return(0);

    /* add unconstrained set if desired */
    if (dataSet->addFree){

      /* check if there is more than one free constraint set */
      cSetNum = dataSet->hasConstraints ? 1 : 0;
      numFree = 0;
      for (;cSetNum<dataSet->numConSets; cSetNum++) numFree += dataSet->cSets[cSetNum].isEmpty;
      if (numFree>0){
	Rprintf( "\n*********************************** WARNING ****************************************\n");
	Rprintf( "readConFile: Constraint file contains empty constraint set. Option -addfree ignored.\n");
	Rprintf( "************************************************************************************\n");
      }else{
	dataSet->cSets[dataSet->numConSets].numInts            = 1;
	dataSet->cSets[dataSet->numConSets].intCons            = intConMalloc(1);
	dataSet->cSets[dataSet->numConSets].intCons[0].lenType = var;
	dataSet->numConSets++;
      }
    }

    /* subtract the first empty constraint set to get the real number of candidate models */
    dataSet->effNumModels    = (dataSet->numConSets-1)*dataSet->numModTypes*dataSet->numWidths;

  }else{
    /* set up empty constraint set */

    dataSet->numConSets                   = 1;
    dataSet->cSets                        = conSetMalloc(dataSet->numConSets, dataSet->minWidth);
    dataSet->cSets[0].numInts             = 1;
    dataSet->cSets[0].intCons             = intConMalloc(dataSet->cSets[0].numInts);
    dataSet->cSets[0].intCons[0].lenType  = var;

    dataSet->effNumModels       = dataSet->numConSets*dataSet->numModTypes*dataSet->numWidths;
  }

    


  if (!getHiProb(dataSet)) return(0);
  if (!getMaxNumProfParms(dataSet)) return(0);
  if (!getMaxNumInts(dataSet)) return(0);


  dataSet->numModels       = dataSet->numConSets*dataSet->numModTypes*(dataSet->maxWidth - dataSet->minWidth + 1);
  dataSet->mod             = modMalloc(dataSet->numModels);
  dataSet->allCrits        = critMalloc(dataSet->numModels);
  dataSet->svTheta         = thetaMalloc(1,dataSet->maxWidth,dataSet->maxNumProfParms);
  dataSet->theta           = thetaMalloc(1,dataSet->maxWidth,dataSet->maxNumProfParms);
  dataSet->theta1          = thetaMalloc(1,dataSet->maxWidth,dataSet->maxNumProfParms);
  dataSet->theta2          = thetaMalloc(1,dataSet->maxWidth,dataSet->maxNumProfParms);
  dataSet->intNumBasePairs = IntMalloc(dataSet->maxNumInts);
  dataSet->intFirst        = IntMalloc(dataSet->maxNumInts);
  dataSet->intLast         = IntMalloc(dataSet->maxNumInts);

  dataSet->refThetas       = thetaMalloc(dataSet->conCrit.fold, dataSet->maxWidth, dataSet->maxNumProfParms);
  dataSet->refThetaConv    = IntMalloc(dataSet->conCrit.fold);
  dataSet->nRefThetas      = dataSet->conCrit.fold;
    
  return(1);
}
/**********************************************************************/
/*
  addConstraints

  Read constraint file to add constraints to constraint set that
  has already been initialized.


*/
/**********************************************************************/
int addConstraints(
		   DATASET *dataSet
		   )
{

  int  intNum, cSetNum, pos;
  CONSTRING *conString;   /* the constraint file */
  CONSET *cSet;           /* the constraint sets */
  int numConSets;
  char *line;
  char *first;
  char *word;

  conString  = &(dataSet->conString);
  numConSets = dataSet->numConSets;
  line       = dataSet->line;
  first      = dataSet->first;
  word       = dataSet->word;
  cSet       = dataSet->cSets;
  
  if (!findCommand(conString,line,'@',EOF)){
    Rprintf("Couldn't find beginning of constraint set 1\n");
    return(0);
  }


  if (COSMO_MSG_LEVEL>=2) Rprintf("\n");
  for (cSetNum=1; cSetNum<numConSets;cSetNum++){

    if (COSMO_MSG_LEVEL>=2) Rprintf("addConstraints: constraint set %d\n",cSetNum);

    while(findCommand(conString,line,'>','@')==1){
      sscanf(line,"%s", first);
      if (COSMO_MSG_LEVEL>=3) Rprintf("addConstraints: Got line -- %s\n", first);
      
      if ((!strcmp(first,">ICShape"))|(!strcmp(first,">IcShape"))){
	
	/* Read in interval number */
	if (!findLine(conString, line, first, "Interval", "addConstraints",'>',1)) return(0);	
	if (sscanf(line,"%s %d", first, &intNum)!=2){
	  Rprintf("addConstraints: Error reading in interval number for shape constraint\n");
	  return(0);
	}
	intNum--;

	if (!checkInput(cSet,intNum, cSetNum,numConSets)) return(0);
	if (cSet[cSetNum].intCons[intNum].hasShapeCon){
	  Rprintf("addConstraints: Can't add shape constraint to interval %d of constraint set %d:\n",intNum, cSetNum);
	  Rprintf("             Interval already has shape constraint\n");
	  return(0);
	}
	if (cSet[cSetNum].intCons[intNum].hasBoundCon){
	  Rprintf("addConstraints: Can't add shape constraint to interval %d of constraint set %d:\n",intNum, cSetNum);
	  Rprintf("             Interval already has bound constraint\n");
	  return(0);
	}
	if (!addShapeCon(cSet, conString, cSetNum, intNum, line, word,first)) return(0);
	
      }else if((!strcmp(first,">ICBounds"))|(!strcmp(first,">IcBounds"))|(!strcmp(first,">ICBound"))|(!strcmp(first,">IcBound"))){

	/* Read in interval number */
	if (!findLine(conString, line, first, "Interval", "addConstraints",'>',1)) return(0);	
	if (sscanf(line,"%s %d", first, &intNum)!=2){
	  Rprintf("addConstraints: Error reading in interval number for bound constraint\n");
	  return(0);
	}
	intNum--;

	if (!checkInput(cSet,intNum, cSetNum,numConSets)) return(0);
	if (cSet[cSetNum].intCons[intNum].hasBoundCon){
	  Rprintf("addConstraints: Can't add bound constraint to interval %d of constraint set %d:\n",intNum, cSetNum);
	  Rprintf("             Interval already has bound constraint\n");
	  return(0);
	}
	if (cSet[cSetNum].intCons[intNum].hasShapeCon){
	  Rprintf("addConstraints: Can't add bound constraint to interval %d of constraint set %d:\n",intNum, cSetNum);
	  Rprintf("             Interval already has shape constraint\n");
	  return(0);
	}
	if (!addBoundCon(cSet, conString, cSetNum, intNum, line, word,first)) return(0);
	
      }else if((!strcmp(first,">NucFreq"))|(!strcmp(first,">NucProb"))|(!strcmp(first,">NucFrequency"))){

	/* Read in interval number */
	if (!findLine(conString, line, first, "Interval", "addConstraints",'>',1)) return(0);	
	if (sscanf(line,"%s %d", first, &intNum)!=2){
	  Rprintf("addConstraints: Error reading in interval number for bound constraint\n");
	  return(0);
	}
	intNum--;
	if (!checkInput(cSet,intNum, cSetNum,numConSets)) return(0);

	/* Read in nucleotide position */
	if (!findLine(conString, line, first, "Pos", "addConstraints",'>',1)) return(0);	
	if (sscanf(line,"%s %s", first, word)!=2){
	  Rprintf("addConstraints: Error reading in interval number for bound constraint\n");
	  return(0);
	}

	if ((!strcmp(word,"all"))|(!strcmp(word,"avg"))){
	  if (!addIntFreqCon(cSet, conString, cSetNum, intNum, line, word,first)) return(0);

	}else{
	  if (cSet[cSetNum].intCons[intNum].lenType!=bp){
	    Rprintf("addConstraints: position frequency constraints only apply to intervals of fixed length\n");
	    return(0);
	  }

	  if (sscanf(word,"%d", &pos)!=1){
	    Rprintf("addConstraints: Error reading in position for position frequency constraint\n");
	    return(0);
	  }
	  
	  if ((pos<1) || (pos>cSet[cSetNum].intCons[intNum].numBasePairs)){
	    Rprintf("addConstraints: Position for position frequency constraint lies outside the interval\n");
	    return(0);
	  }
	  

	  if (!addPosFreqCon(cSet, conString, cSetNum, intNum, pos, line, word,first)) return(0);
	}
      }else if((!strcmp(first,">Pal"))|(!strcmp(first,">Palindrome"))|(!strcmp(first,">Palindromes"))){

	if (!checkConSetNum(cSetNum, numConSets)) return(0);
	if (cSet[cSetNum].hasPalCon){
	  Rprintf("addConstraints: Can't add palindromic constraint to constraint set %d:\n", cSetNum);
	  Rprintf("             Constraint set already has palindromic constraint\n");
	  return(0);
	}
	if (!addPalCon(cSet, conString, cSetNum, line, word,first)) return(0);
	
      }else if((!strcmp(first,">Submotif"))|(!strcmp(first,">SubMotif"))|(!strcmp(first,">Sub"))){

	if (!checkConSetNum(cSetNum, numConSets)) return(0);
	if (cSet[cSetNum].hasSubCon){
	  Rprintf("addConstraints: Can't add submotif constraint to constraint set %d:\n", cSetNum);
	  Rprintf("             Constraint set already has submotif constraint\n");
	  return(0);
	}
	if (!addSubCon(cSet, conString, cSetNum, line, word,first)) return(0);
	
      }else if((!strcmp(first,">ParameterDiff"))|(!strcmp(first,">ParmDiff"))|(!strcmp(first,">ParameterDifference"))){

	if (!checkConSetNum(cSetNum, numConSets)) return(0);
	if (!addLinCon(cSet, conString, cSetNum, line, word,first)) return(0);
      }else if ((!strcmp(first,">IntervalSetup"))){

      }else{
 
	Rprintf("addConstraints: Encountered unknown command %s\n", first);
	return(0);
      }
    }
    
    /* see if constraint set is empty or not */
    cSet[cSetNum].isEmpty = !((cSet[cSetNum].numShapeCons>0) ||(cSet[cSetNum].numBoundCons>0) || (cSet[cSetNum].numIntFreqCons>0) || cSet[cSetNum].hasPalCon || cSet[cSetNum].hasSubCon || (cSet[cSetNum].numLinCons>0));
    
    if (COSMO_MSG_LEVEL>=3) Rprintf("addConstraints: constraint set %d is empty? %d\n", cSetNum, cSet[cSetNum].isEmpty);
  }

  return(1);
}


/**********************************************************************/
/*
  initConSets

  Initialize constraint sets from the constraint file. Interval 
  lengths are read in and checked for errors. Memory is allocated
  for interval constraints.
	

*/
/**********************************************************************/
int initConSets(
		DATASET *dataSet
		)
{
  char *word, *line, *first;
  int cSetNum;                  /* the constraint set number */
  int intNum;                   /* the interval number */
  int numVarInt;                /* number of variable length intervals */
  int numFixedInt;              /* number of fixed length intervals */
  int numBasePairs;             /* for error checking */
  double sumProps;              /* for error checking */
  int i;
  CONSTRING *conString;         /* the constraint file */  
  int widthLow;                 /* the lowest motif width with which the interval breakdown must be compatible */  
  BOOLEAN addFree;              /* add unconstrained set to the end? */

  line      = dataSet->line;
  word      = dataSet->word;
  first     = dataSet->first;
  conString = &(dataSet->conString);
  widthLow  = dataSet->minWidth;
  addFree   = dataSet->addFree;


  /* get Number of constraint sets */
  if (!getNumConSets(dataSet)){
    Rprintf("initConSets: constraint file contains no constraint sets (indicated by starting character @)\n");
    return(0);
  }
  
  if (COSMO_MSG_LEVEL>=2) Rprintf("\n");
  
  /* first constraint set is always empty*/
  dataSet->numConSets++;

  /* allocate memory for additional unconstrained set if necessary */
  dataSet->cSets = conSetMalloc(dataSet->numConSets + addFree, widthLow);

  dataSet->cSets[0].numInts             = 1;
  dataSet->cSets[0].intCons             = intConMalloc(dataSet->cSets[0].numInts);
  dataSet->cSets[0].intCons[0].lenType  = var;
  

  for (cSetNum=1; cSetNum<dataSet->numConSets;cSetNum++){

    if (COSMO_MSG_LEVEL>=2) Rprintf("initConSets: constraint set %d\n",cSetNum);


    if (!findCommand(conString,line,'@',EOF)){
      Rprintf("Couldn't find beginning of constraint set %d\n",cSetNum);
      return(0);
    }

    /* read in interval setup */
    if (!findLine(conString, line, first, ">IntervalSetup", "initConSets",'@',0)) return(0);

    if (!getNumInts(dataSet, &(dataSet->cSets[cSetNum].numInts))) return(0);


    if (dataSet->cSets[cSetNum].numInts==0){
      Rprintf("initConSets: found 0 intervals for constraint set %d\n", cSetNum);
      return(0);
    }

    dataSet->cSets[cSetNum].intCons = intConMalloc(dataSet->cSets[cSetNum].numInts);
    if (COSMO_MSG_LEVEL>=3)  Rprintf("initConSets: NumberOfIntervals for ConSet %d: %d\n", cSetNum, dataSet->cSets[cSetNum].numInts);

    numVarInt = 0;
    numFixedInt = 0;
    sumProps = 0;
    /* get length type for each interval */
    for (intNum=0;intNum<dataSet->cSets[cSetNum].numInts;intNum++){
      

      /* read in parameter "Length" */
      if (!findLine(conString, line, first, "Length", "initConSets",'>',1)) return(0);

      if (!sscanf(line,"%s %s", first, word)){
	Rprintf("initConSets: Error reading length type for conSet %d interval %d\n", cSetNum, intNum);
	return(0);
      }
      

      if((!strcmp(word,"variable"))|(!strcmp(word,"Variable"))){
	dataSet->cSets[cSetNum].intCons[intNum].lenType = var;
	numVarInt++;
	if (COSMO_MSG_LEVEL>=2) Rprintf("initConSets: Interval %d is variable length\n", intNum+1);
      }else{
	if (strchr(word,'%')!=NULL){
	  dataSet->cSets[cSetNum].intCons[intNum].lenType = prop;

	  /* chop of '%' at end */
	  word[strlen(word)-1] = '\0';
	  if (!sscanf(word,"%lf", &(dataSet->cSets[cSetNum].intCons[intNum].prop))){
	    Rprintf("initConSets: Error reading proportion for conSet %d interval %d\n", cSetNum, intNum);
	    return(0);
	  }
	  dataSet->cSets[cSetNum].intCons[intNum].prop /= 100;

	  sumProps += dataSet->cSets[cSetNum].intCons[intNum].prop;
	  if (dataSet->cSets[cSetNum].intCons[intNum].prop <=0 || dataSet->cSets[cSetNum].intCons[intNum].prop >1 ){
	    Rprintf("initConSets: Interval Proportions must be between 0 and 1\n");
	    return(0); 
	  }
	  if (COSMO_MSG_LEVEL>=2) Rprintf("initConSets: Proportion for interval %d: %lf\n", intNum+1, dataSet->cSets[cSetNum].intCons[intNum].prop);    
	}else{

	  dataSet->cSets[cSetNum].intCons[intNum].lenType = bp;
	  if (!sscanf(word,"%d", &(dataSet->cSets[cSetNum].intCons[intNum].numBasePairs))){
	    Rprintf("initConSets: Error reading numBasePairs for conSet %d interval %d\n", cSetNum, intNum);
	    return(0);
	  }
	  numFixedInt++;
	  if (COSMO_MSG_LEVEL>=2) Rprintf("initConSets: NumBasePairs for interval %d: %d\n",intNum+1, dataSet->cSets[cSetNum].intCons[intNum].numBasePairs);
	  
	}
      }

    }

    /* various error checking of constraint set input */
    if (numVarInt>1){
      Rprintf("initConSets: At most one variable length interval may be given\n");
      return(0);
    }

    if (numFixedInt>0 &&numVarInt==0){
      Rprintf("initConSets: If fixed length intervals are present, there must also be a variable length interval\n");
      return(0);
    }

    if (sumProps>1){
      Rprintf("initConSets: The interval proportions sum up to %lf\n", sumProps);
      return(0);
    }

    numBasePairs = 0;
    for (i=0;i<dataSet->cSets[cSetNum].numInts;i++){
      if (dataSet->cSets[cSetNum].intCons[i].lenType==prop){
	numBasePairs += NINT(dataSet->cSets[cSetNum].intCons[i].prop*widthLow);
      }else if (dataSet->cSets[cSetNum].intCons[i].lenType==bp){
	numBasePairs += dataSet->cSets[cSetNum].intCons[i].numBasePairs;
      }
    }
    if (numBasePairs>widthLow){
      Rprintf("initConSets: Fixed and proportional length intervals require at least %d bps\n", numBasePairs);
      return(0);
    }

    for (i=0;i<dataSet->cSets[cSetNum].numInts;i++){
      if (dataSet->cSets[cSetNum].intCons[i].lenType==var){
	numBasePairs = widthLow - numBasePairs;
	break;
      }
    }
    if (numBasePairs<=0){
      Rprintf("initConSets: Variable length interval has length zero at lowest motif width\n");
      return(0);
    }
  }
  
  

  /* rewind file */
  rewindConString(conString);

  return 1;
}





/**********************************************************************/
/*
  getMaxNumProfParms

  Extract maximum number of information content profile parameters
  from all constraint sets so that we can allocate appropriate
  memory.

	
*/
/**********************************************************************/
int getMaxNumProfParms(
			DATASET *dataSet
			)
{
  int cSetNum;               /* current constraint set number */
  int max;                   /* current maximum number of parameters */
  int numConSets;            /* number of constraint sets */
			 
  numConSets = dataSet->numConSets;


  /* There are two parameters for each interval with a shape constraint */
  max = 2*dataSet->cSets[0].numShapeCons;
  for (cSetNum=1;cSetNum<numConSets;cSetNum++){
    if (2*dataSet->cSets[cSetNum].numShapeCons > max) max = 2*dataSet->cSets[cSetNum].numShapeCons;
  }

  dataSet->maxNumProfParms = max;

  return(1);
}

/**********************************************************************/
/*
  getMaxNumInts

  Extract maximum number of intervals from all constraint sets so that 
  we can allocate appropriate memory.

	
*/
/**********************************************************************/
int getMaxNumInts(
		   DATASET *dataSet
		   )
{
  int cSetNum;               /* current constraint set number */
  int max;                   /* current maximum number of parameters */
  int numConSets;            /* number of constraint sets */
			 
  numConSets = dataSet->numConSets;


  max = 1;
  for (cSetNum=1;cSetNum<numConSets;cSetNum++){
    if (dataSet->cSets[cSetNum].numInts > max) max = dataSet->cSets[cSetNum].numInts;
  }

  dataSet->maxNumInts = max;

  return(1);
}
 

/**********************************************************************/
/*
  getHiProb

  Calculate the probability to assign to a nucleotide that actually
  occurs in a given position in the oligo from which we are deriving
  a staring value for the position weight matrix. This probability
  must be set >= to any lower bounds that may be imposed on nucleotide
  frequencies in order to ensure that valid starting values will 
  be found. We set it to the midpoint between the greatest lower 
  bound on nucleotide frequencies and 1.

  Returns the probability.
	
*/
/**********************************************************************/
int getHiProb(
	       DATASET *dataSet
	       )
{
  double sharpestBound;      /* greatest lower bound on nucleotide frequencies */
  int cSetNum;               /* number of current constraint set */
  int numConSets;            /* number of constraint sets */

  numConSets = dataSet->numConSets;
      
  sharpestBound = dataSet->cSets[0].sharpestBound;
  for (cSetNum=1;cSetNum<numConSets;cSetNum++){
    if (sharpestBound<dataSet->cSets[cSetNum].sharpestBound) sharpestBound = dataSet->cSets[cSetNum].sharpestBound;
  }

  dataSet->hiProb   =   0.5*(sharpestBound+1);
  dataSet->lowProb  =   (1 - dataSet->hiProb)/3;
  if (COSMO_MSG_LEVEL>=3) Rprintf("hiProb = %lf and lowProb = %lf\n", dataSet->hiProb, dataSet->lowProb);

  return(1);
}


/**********************************************************************/
/*
  getNumConSets

  Extract the number of constraint sets from the file. Each new 
  constraint set starts with a line that has as first character "@".
  The file is rewound at the end.

  Returns the number of constraint sets.
	
*/
/**********************************************************************/
int getNumConSets(
		  DATASET *dataSet
		  )
{
  char start;
  CONSTRING *conString;        /* the constraint file */ 
  char *line;
  char *first;

  line      = dataSet->line;
  first     = dataSet->first;
  conString = &(dataSet->conString);

  start = '@';

  dataSet->numConSets = 0;

  while (getLine(conString,line)!=NULL){
    if (!strcmp(line,"\n"))  continue;
    if (!strcmp(line,"\r\n"))  continue; /* for constraints from HTML form */
    if (sscanf(line,"%s", first)==EOF) break;
    if (first[0]==start) dataSet->numConSets++;
  }

  
  /*rewind the file*/
  rewindConString(conString);

  return(dataSet->numConSets);
}

/**********************************************************************/
/*
  getNumInts

  Get the number of intervals in a given constraint set. The intervals
  are specified in the form

  >IntervalSetup
  Length: 3 bp
  Length: variable
  Length: 30%

  The file is rewound to the line ">IntervalSetup"

  Returns the number of intervals.
	
*/
/**********************************************************************/
int getNumInts(
	       DATASET *dataSet,
	       int *numInts
	       )
{
  int streamPos;                /* position in stream */
  char nextConSet = '@';
  char *line = dataSet->line;
  char *first = dataSet->first;
  CONSTRING *conString;
  
  conString   = &(dataSet->conString);
  streamPos   = conString->index; 

  *numInts = 0;

  while (getLine(conString,line)!=NULL){
    if (!strcmp(line,"\n"))  continue;
    if (!strcmp(line,"\r\n"))  continue; /* for constraints from HTML form */
    if (sscanf(line,"%s", first)==EOF) break;
    if (first[0]==nextConSet) break;

    /* chop off ':' from the end */
    first[strlen(first)-1] = '\0';
    if (!strcmp(first,"Length")) (*numInts)++;
  }

  
  /*rewind the constraint set file*/
  conString->index = streamPos;

  return(1);
}




/**********************************************************************/
/*
  conFile2conString
  
  Read constraint file into CONSTRING structure.

  Returns TRUE if successful, FALSE otherwise
	
*/
/**********************************************************************/
int conFile2conString(
		      DATASET *dataSet
		      )
{
  char start;
  char *line;
  char *first;
  int nlines;
  int i,j;
  char oneLetter;
  FILE *conFile;

  line    = dataSet->line;
  first   = dataSet->first;
  conFile = dataSet->conFile;

  start = '@';

  nlines  = 0;
  while (fgets(line,MAXNUMCHARS,conFile)!=NULL) nlines++;
  dataSet->conString.nlines = nlines;
  dataSet->conString.index = 0;

  /*rewind the file*/
  fseek(conFile,(long)0,0);

  /* allocate memory */
  dataSet->conString.con = (char *)S_alloc((unsigned)(nlines*MAXNUMCHARS),(sizeof(char *)));
  if (dataSet->conString.con==NULL){
    Rprintf("ERROR: Couldn't get %d characters. Exiting...\n", nlines*MAXNUMCHARS);
    exit(1);
  }


  /* read lines into character array */
  i = 0;
  while (fgets(line,MAXNUMCHARS,conFile)!=NULL){
    oneLetter = line[0];
    j = 0;
    while (oneLetter != '\0'){
      dataSet->conString.con[i*MAXNUMCHARS + j] = oneLetter;
      j++;
      oneLetter = line[j];
    }
    dataSet->conString.con[i*MAXNUMCHARS + j] = oneLetter;
    i++;
  }

  /*rewind the file*/
  fseek(conFile,(long)0,0);

  return TRUE;
}


/**********************************************************************/
/*
  getLine

  Reads next line in CONSTRING struct into character buffer pointed to
  by line.

  Returns NULL if CONSTRING struct is already at end.

*/
/**********************************************************************/
char* getLine(
	      CONSTRING *conString,
	      char *line
	      )
{

  int i;

  if (conString->index>=conString->nlines) return NULL;

  for (i=0;i<MAXNUMCHARS;i++) line[i] = conString->con[conString->index*MAXNUMCHARS + i];
  conString->index++;

  return line;

}

/**********************************************************************/
/*
  rewindConString
  
  Reset index of CONSTRING struct to zero.

*/
/**********************************************************************/
int rewindConString(
		    CONSTRING *conString
		    )
{
  conString->index = 0;
  return TRUE;
}


/**********************************************************************/
/*
  string2conString
  
  Read constraints from string into CONSTRING structure.
  Needed for R package only.

  Returns TRUE if successful, FALSE otherwise
	
*/
/**********************************************************************/
int string2conString(
		     DATASET *dataSet,
		     char *constraints
		     )
{
  int nlines;
  int i;
  int length;          /* length of string constraints */
  int row, col;
     
  length = strlen(constraints);

  /* find number of lines */
  nlines  = 0;
  for (i=0; i<length;i++) if (constraints[i]=='\n') nlines++;
  dataSet->conString.nlines = nlines;
  dataSet->conString.index = 0;


  /* allocate memory */
  dataSet->conString.con = (char *)S_alloc((unsigned)(nlines*MAXNUMCHARS),(sizeof(char *)));
  if (dataSet->conString.con==NULL){
    Rprintf("ERROR: Couldn't get %d characters. Exiting...\n", nlines*MAXNUMCHARS);
    exit(1);
  }
  

  /* read lines into character array */
  i = 0;
  row = 0;
  col = 0;
  for (i=0;i<length;i++){
    dataSet->conString.con[row*MAXNUMCHARS + col] = constraints[i];
    col++;

    /* insert string terminator and advance j */
    if (constraints[i]=='\n'){
      dataSet->conString.con[col] = '\0'; 
      col = 0;
      row++;
    }
  }

  return TRUE;
}

/**********************************************************************/
/*
  getIntensity
  
  For a given expected number of motif occurrences and  model type, 
  find the corresponding intensity parameter.

  Sequences with exclude==FALSE are excluded if desired.

  Returns that intensity parameter.
  
 */
/**********************************************************************/
double getIntensity(
		    MTYPE mType,          /* model type */
 		    int width,            /* width of motif */
		    int numSites,         /* expected number of motif occurrences */
		    DATASET *dataSet      /* the dataset */
		    )
{


  double intensity;      /* intensity parameter */
  int minWidth;          /* minimum motif width to consider */
  int seqNum;            /* sequence number */
  double numSeqs;        /* number of sequences */
  double totalNucs;      /* total number of nucleotides */
  double numCuts;        /* number of cuts */

  minWidth = dataSet->minWidth;

  /* 
     get number of sequences, total nucleotides,
     and number of cuts 
  */  
  numSeqs = 0;
  totalNucs = 0;
  numCuts = 0;
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    if (!dataSet->samples[seqNum].exclude){
      numSeqs++;
      totalNucs += dataSet->samples[seqNum].length;
      numCuts += dataSet->samples[seqNum].numCuts[width-minWidth];
    }
  }
  
  switch(mType){
  case OOPS:
    intensity = 1;
    break;

  case ZOOPS:
    intensity = (double) numSites/numSeqs;
    break;
    
  case TCM:
    intensity = (double) numSites /totalNucs;
    break;

  case CUT:
  case OVER:
    intensity = 1 - pow(1-1/numCuts, numSites);
    break;
  default:
    Rprintf("getIntensity: mType %d not recognized. Exiting ...\n", mType);
    return(0);
    break;
  }
  
  return intensity;
}
