/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/



/*****************************************************************************
*								       	     *
*      This module contains the function to be called by R.                  *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include "general.h"
#include "macros.h"
#include "rcosmo.h"
#include "cosmo.h"
#include "alloc.h"
#include "back.h"
#include "constraints.h"
#include "estarts.h"
#include "fit.h"
#include "free.h"
#include "init.h"
#include "io.h"
#include "rio.h"
#include "matLib.h"
#include "optimize.h"
#include "precalc.h"

int COSMO_MSG_LEVEL;

void ccosmo(
	   int *seqs,
	   int *numSeqs,
	   int *maxSeqLength,
	   char **constraints,
	   int *minWidth,
	   int *maxWidth,
	   int *revComp,
	   int *minSites,
	   int *maxSites,
	   int *oops,
	   int *zoops,
	   int *tcm,
	   char **approx,
	   int *cutFac,
	   char **wCrit,
	   int *wFold,
	   int *wTrunc,
	   char **modCrit,
	   int *modFold,
	   int *modTrunc,
	   char **conCrit,
	   int *conFold,
	   int *conTrunc,
	   char **intCrit,
	   int *intFold,
	   int *intTrunc,
	   int *maxIntensity,
	   int *starts,
	   int *lstarts,
	   int *backSeqs,
	   int *numBackSeqs,
	   int *maxBackSeqLength,
	   int *backFold,
	   double *transMat,
	   int *order,
	   int *maxOrder,
	   int *selOrder,
	   double *klDivs,
	   double *estTransMat,
	   double *selIntCrit,
	   double *wCrits,
	   double *modCrits,
	   double *conCrits,
	   int *wInit,
	   int *modInit,
	   int *conInit,
	   int *selWidth,	   
	   int *selModel,
	   int *selConSet,
	   int *selNumSites,
	   int *numModels,
	   double *pwm,
	   double *motProbs,
	   int *motRevComp,
	   int *alignSeqs,
	   int *alignStarts,
	   double *eVal,
	   int *success,
	   int *msgLevel
	   )
{
  


  DATASET dataSet;     /* the dataset storing all relevant information */

  *success = 1;
  COSMO_MSG_LEVEL = *msgLevel;

  

  /* set default values for dataset */
  if (!setDefaults(&dataSet)){
    (*success) = 0;
    return;
  }

  dataSet.R                                             = TRUE;
  dataSet.numSeqs                                       = *numSeqs;
  dataSet.maxSeqLength                                  = *maxSeqLength;
  dataSet.minWidth                                      = *minWidth;                 /* minimum motif width to consider */
  dataSet.maxWidth                                      = *maxWidth;                 /* maximum motif width to consider */
  dataSet.oops                                          = *oops;                     /* BOOLEAN: consider OOPS model? */
  dataSet.zoops                                         = *zoops;                    /* BOOLEAN: consider ZOOPS model? */
  dataSet.tcm                                           = *tcm;                      /* BOOLEAN: consider TCM model? */
  dataSet.revComp                                       = *revComp;                  /* BOOLEAN: consider reverse complement? */

  if ((minSites)!=NULL)       dataSet.minNumSites       = *minSites;              /* minimum number of motif sites */
  if ((maxSites)!=NULL){
                          dataSet.zoopsMaxNumSites      = *maxSites;              /* maximum number of motif sites */
                          dataSet.tcmMaxNumSites        = *maxSites;              /* maximum number of motif sites */
  }
  dataSet.cutFactor                                     = *cutFac;
  dataSet.wCrit.fold                                    = *wFold;
  dataSet.wCrit.quant                                   = ((double) *wTrunc)/100;
  dataSet.wCrit.trunc                                   = dataSet.wCrit.quant!=1.0;
  dataSet.modCrit.fold                                  = *modFold;
  dataSet.modCrit.quant                                 = ((double) *modTrunc)/100;
  dataSet.modCrit.trunc                                 = dataSet.modCrit.quant!=1.0;
  dataSet.conCrit.fold                                  = *conFold;
  dataSet.conCrit.quant                                 = ((double) *conTrunc)/100;
  dataSet.conCrit.trunc                                 = dataSet.conCrit.quant!=1.0;
  dataSet.intCrit.fold                                  = *intFold;
  dataSet.intCrit.quant                                 = ((double)*intTrunc)/100;
  dataSet.intCrit.trunc                                 = dataSet.intCrit.quant!=1.0;
  dataSet.maxIntensity                                  = *maxIntensity;
  dataSet.numStarts                                     = *starts;
  dataSet.likStarts                                     = *lstarts;
  dataSet.numBackSeqs                                   = *numBackSeqs;
  dataSet.maxBackSeqLength                              = *maxBackSeqLength;
  dataSet.backFold                                      = *backFold;
  dataSet.maxOrder                                      = *maxOrder;
  
  if (order!=NULL){
                              dataSet.order             = *order;
                              dataSet.orderByCV         = FALSE;
  }

  if (!strcmp(*approx,"exact")){
    dataSet.approx = TCM;
  }else if (!strcmp(*approx,"over")){
    dataSet.approx = OVER;
  }else if (!strcmp(*approx,"cut")){
    dataSet.approx = CUT;
  }else{
    Rprintf("approx must be one of: exact, over, cut. Got: %s\n", *approx);
    (*success) = 0;
    return;
  }

  if (!strcmp(*wCrit,"lik")){
    dataSet.wCrit.crit = lik;
  }else if (!strcmp(*wCrit,"aic")){
    dataSet.wCrit.crit = aic;
  }else if (!strcmp(*wCrit,"bic")){
    dataSet.wCrit.crit = bic;
  }else if (!strcmp(*wCrit,"eval")){
    dataSet.wCrit.crit = eval;
  }else if (!strcmp(*wCrit,"likCV")){
    dataSet.wCrit.crit = likCV;
  }else{
    Rprintf("wCrit must be one of: lik, aic, bic, eval, likCV. Got: %s\n", *wCrit);
    (*success) = 0;
    return;
  }

  if (!strcmp(*modCrit,"lik")){
    dataSet.modCrit.crit = lik;
  }else if (!strcmp(*modCrit,"aic")){
    dataSet.modCrit.crit = aic;
  }else if (!strcmp(*modCrit,"bic")){
    dataSet.modCrit.crit = bic;
  }else if (!strcmp(*modCrit,"eval")){
    dataSet.modCrit.crit = eval;
  }else if (!strcmp(*modCrit,"likCV")){
    dataSet.modCrit.crit = likCV;
  }else{
    Rprintf("modCrit must be one of: lik, aic, bic, eval, likCV. Got: %s\n", *modCrit);
    (*success) = 0;
    return;
  }

  if (!strcmp(*conCrit,"lik")){
    dataSet.conCrit.crit = lik;
  }else if (!strcmp(*conCrit,"aic")){
    dataSet.conCrit.crit = aic;
  }else if (!strcmp(*conCrit,"bic")){
    dataSet.conCrit.crit = bic;
  }else if (!strcmp(*conCrit,"eval")){
    dataSet.conCrit.crit = eval;
  }else if (!strcmp(*conCrit,"likCV")){
    dataSet.conCrit.crit = likCV;
  }else if (!strcmp(*conCrit,"pwmCV")){
    dataSet.conCrit.crit = pwmCV;
  }else{
    Rprintf("conCrit must be one of: lik, aic, bic, eval, likCV, pwmCV. Got: %s\n", *conCrit);
    (*success) = 0;
    return;
  }

  if (!strcmp(*intCrit,"lik")){
    dataSet.intCrit.crit = lik;
  }else if (!strcmp(*intCrit,"eval")){
    dataSet.intCrit.crit = eval;
  }else if (!strcmp(*intCrit,"likCV")){
    dataSet.intCrit.crit = likCV;
  }else{
    Rprintf("intCrit must be one of: lik, eval, likCV. Got: %s\n", *intCrit);
    (*success) = 0;
    return;
  }
   
    
  /* process command line arguments */
  if (!processArgs(&dataSet)){
    (*success) = 0;
    return;
  }

  /* read constraints into CONSTRING struct */
  if (constraints!=NULL){
    if (!string2conString(&dataSet, *constraints)){
      (*success) = 0;
      return;
    }
    dataSet.hasConstraints = TRUE;
  }

 
  /* read in sequences */
  if (!readSeqsFromVec(seqs, &dataSet)){
    (*success) = 0;
    return;
  }

  /* read in background sequences if necessary */
  if (backSeqs!=NULL){
    if (!readBackSeqsFromVec(backSeqs, 1, &dataSet)){
      (*success) = 0;
      return;
    }
  }

  /* get transition matrix */
  if (transMat!=NULL){
    if (!readTransMatFromVec(transMat,&dataSet)){
      (*success) = 0;
      return;
    }
  }else{
    if (!getTransMat(dataSet.orderByCV, &dataSet)){
      (*success) = 0;
      return;
    }
  }

  /* read in constraints */
  if (!readConFile(&dataSet)){
    (*success) = 0;
    return;
  }

  /* interpret constraints for different widths etc. */
  if (!deriveAllModels(&dataSet)){
    (*success) = 0;
    return;
  }

  /* allocate memory for donlp */
  if (!donlpMalloc(&dataSet)){
    (*success) = 0;
    return;
  }

  /* create cut sequences if necessary */ 
  if (dataSet.approx!=TCM){
    if (!cutSequences(&dataSet)){
      (*success) = 0;
      return;
    }
  }

  /* print first status message if desired */
  if (dataSet.printStatus) Rprintf("0/%d   0/%d\n", dataSet.numSeqs, dataSet.effNumModels);


  /* precalculate arrays for background model */
  if (!precalcBack(&dataSet)){
    (*success) = 0;
    return;
  }

  /* get starting values */
  if (!getStarts(&dataSet)){
    (*success) = 0;
    return;
  }

  /* get pointer to dataSet to NLP namespace */
  initNLP(&dataSet);
  
  /* select model parameters */
  if (!selectModel(&dataSet)){
    Rprintf( "No model selected.\n");
    (*success) = 0;
    return;
  }


  /* fit final model */
  if (!finalModel(&dataSet)){
    Rprintf( "Final model did not converge. Are constraints feasible?\n");
    (*success) = 0;
    return;
  }

  /* generate output */
  output2vec(&dataSet, selOrder, klDivs, estTransMat, selIntCrit, wCrits, modCrits, conCrits, wInit, modInit, conInit, selWidth, selModel, selConSet, selNumSites, numModels, pwm, motProbs, motRevComp, alignSeqs, alignStarts, eVal);

}

void bgm(
	 int *backSeqs,
	 int *numBackSeqs,
	 int *maxBackSeqLength,
	 int *backFold,
	 int *order,
	 int *maxOrder, 
	 double *estTransMat,
	 double *klDivs,
	 int *selOrder,
	 int *success
	 )
{
  
  int subOrder;            /* a particular order */
  double *transMat;        /* transition matrix for given order */
  int addRows;             /* number of rows to add for previous transition matrices */
  int numRows;             /* number of rows in this transition matrix */
  int row;                 /* current row */
  int col;                 /* current column */
  int i;                   /* counter */

  DATASET dataSet;     /* the dataset storing all relevant information */

  *success = 1;

  /* set default values for dataset */
  if (!setDefaults(&dataSet)){
    (*success) = 0;
    return;
  }
  dataSet.numBackSeqs       = *numBackSeqs;
  dataSet.maxBackSeqLength  = *maxBackSeqLength;
  dataSet.backFold          = *backFold;
  dataSet.maxOrder          = *maxOrder;
  
  if (order!=NULL){
    dataSet.order             = *order;
    dataSet.orderByCV         = FALSE;
  }
    
 
  /* read in background sequences */
  if (!readBackSeqsFromVec(backSeqs, 0, &dataSet)){
    (*success) = 0;
    return;
  }

  /* get transition matrix */
  if (!getTransMat(dataSet.orderByCV, &dataSet)){
    (*success) = 0;
    return;
  }

  /* output selected order and estimated transition matrices */
  *selOrder     = dataSet.order;

  for (subOrder=0, addRows=0; subOrder<=dataSet.order; subOrder++, addRows+=numRows){

    transMat = dataSet.transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      for (col=0; col<4; col++){
	estTransMat[4*(addRows+row) + col] = exp(transMat[4*row+col]);
      }
    }
  }

  /* background Kullback-Leibler divergences */
  for (i=0; i<=dataSet.maxOrder; i++) klDivs[i] = dataSet.backKL[i];


}


