/************************************************************************
*  								       	*
*	MEME							       	*
*	Author: Timothy L. Bailey				       	*
*									*
*	Copyright							*
*	(1994 - 2000) The Regents of the University of California.	*
*	All Rights Reserved.						*
*									*
*	Permission to use, copy, modify, and distribute any part of 	*
*	this software for educational, research and non-profit purposes,*
*	without fee, and without a written agreement is hereby granted, *
*	provided that the above copyright notice, this paragraph and 	*
*	the following three paragraphs appear in all copies.		*
*									*
*	Those desiring to incorporate this software into commercial 	*
*	products or use for commercial purposes should contact the 	*
*	Technology Transfer Office, University of California, San Diego,*
*	9500 Gilman Drive, La Jolla, California, 92093-0910, 		*
*	Ph: (619) 534 5815.						*
*									*
*	IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO 	*
*	ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 	*
*	CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF 	*
*	THE USE OF THIS SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA 	*
*	HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 		*
*									*
*	THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE *
*	UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE 		*
*	MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  *
*	THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND 	*
*	EXTENDS NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, *
*	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 	*
*	MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT 	*
*	THE USE OF THE MATERIAL WILL NOT INFRINGE ANY PATENT, 		*
*	TRADEMARK OR OTHER RIGHTS.  					*
*								       	*
************************************************************************/


/************************************************************************************************
 *								       	                        *
 *      This module contains functions for calculating the E-value of a multiple alignment.     *
 *                                                                                              *
 ************************************************************************************************/
  
#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "eval.h"
#include "matLib.h"
#include "alloc.h"


/**********************************************************************/
/*
        int_compare

        Compare two integers.  Return <0 0 >0
        if the second int is <, =, > the first int.
	For sorting with qsort in decreasing order.
*/
/**********************************************************************/
extern int int_compare(
  const void *v1,
  const void *v2
)
{
  const int * s1 = (const int *) v1;
  const int * s2 = (const int *) v2;

  return(*s2 - *s1);
} /* int_compare */


/************************************************************************/
/*
	init_llr_pv_tables

	Initialize the single-column log likelihood ratio p-value tables.


*/
/************************************************************************/
extern void init_llr_pv_tables(
			       DATASET *dataSet,  /* the dataset */
			       int min,		  /* minimum number of sites */
			       int max,		  /* maximum number of sites */
			       double *back       /* background frequencies */
)
{
  int nsites;				/* number of sites */
  int alength;
  int n;
  DISTR *distrs;

  alength = 4;
  distrs  = dataSet->distrs;

  if (COSMO_MSG_LEVEL>=3) Rprintf( "init_llr_pv_tables: Initializing look-up tables for log-likelihood ratio p-values\n\n");


  /* make sure the distr table gets initialized on all nodes */
  (void) get_llr_pv(dataSet, 0, 1, 1, LLR_RANGE, 1.0, alength, back);

  for (nsites=min; nsites<=max; nsites++) {   /* nsites */

    /* allocate space for table */
    (void) get_llr_pv(dataSet, 0, nsites, 0, LLR_RANGE, 1.0, alength, back);

    /* create table */
    if (COSMO_MSG_LEVEL>=3) Rprintf( "nsites = %d\r", nsites);
    (void) get_llr_pv(dataSet, 0, nsites, 1, LLR_RANGE, 1.0, alength, back);

  } /* nsites */


} /* init_llr_pv_tables */

/******************************************************************************/
/*
	llr_distr

	Compute the probability distribution of the log-likelihood
	ratio (llr) statistic for a discrete distribution.
	Because of roundoff, LLR may be as small as -1 (instead of 0).

	Returns an array p where
		p[i] = log(Pr(LLR=i))

*/
/******************************************************************************/
double *llr_distr(
		  DATASET *dataSet,     /* the dataset */
		  int A,		/* dimension of discrete distribution */
		  double *dd,		/* discrete distribution */
		  int N,		/* number of samples */
		  int desired_range,	/* desired range for scaled LLR */
		  double frac,		/* fraction of scores to use */
		  double *alpha,	/* scale factor for scaled LLR */
		  int *offset,		/* prob[0] = prob(offset) */
		  int *range		/* range for scaled LLR */
)
{
  int i; 				/* index over alphabet */
  int n;				/* index over samples */
  int k;				/* other index */
  int I;				/* LLR */
  double dd_sum;			/* sum of dd */
  int *IP;				/* I'_i[n] */
  int *minI;	             		/* minimum intermediate value of I */
  int *maxI;			        /* maximum intermediate value of I */
  int Irange;				/* maxI-minI+1 */
  double logNfact;			/* log N! */
  double *logP; 			/* log P_i[n] */
  double *logSP;			/* log script_P[# samples][LLR] */
  double *prob=NULL;			/* final probability distribution */
  double min, max, min_dd;

  IP = dataSet->IP;
  logP = dataSet->logP;
  minI = dataSet->minI; 
  maxI = dataSet->maxI;
  logSP = dataSet->logSP;

  /* make sure distribution sums to 1.0 and has no 0's */
  for (i=dd_sum=0; i<A; i++) dd_sum += dd[i] + EPSILON;
  for (i=0; i<A; i++) dd[i] = (dd[i]+EPSILON)/dd_sum;

  /* compute N! */
  logNfact = 0;
  for (i=2; i<=N; i++) logNfact += log(i); 

  /* get estimates of minimum and miximum values of llr */
  for (i=0, min_dd=1; i<A; i++) min_dd = MIN(min_dd, dd[i]);
  max = NINT(-N * log(min_dd));
  for (i=min=0; i<A; i++) min += dd[i]*N*(log(dd[i]) - log(dd[i]));
  min = NINT(min);
  if (COSMO_MSG_LEVEL>=3) Rprintf("min = %f max = %f\n", min, max);

  /* set alpha to achieve the desired range */
  *alpha = desired_range/((max-min));

  /* compute I', P, minI and maxI */ 
  for (n=0; n<=N; n++) minI[n] = maxI[n] = 0;
  for (i=0; i<A; i++) {				/* index over alphabet */
    double logdd = LOG(dd[i]);			/* log(dd[i]) */
    IP[II(i,0)] = 0; 				/* I'_i(0) */
    logP[LL(i,0)] = 0;				/* log P_i(0) */
    for (n=1; n<=N; n++) {			/* index over samples */
      IP[II(i,n)] = NINT(*alpha*n*log(n/(N*dd[i]))); 	   /* I'_i(n) */
      logP[LL(i,n)] = logP[LL(i,n-1)] + logdd - log(n);	   /* log P_i(n) */
      for (k=1; k<=n; k++) {			/* index over samples of new */
	minI[n] = MIN(minI[n], minI[n-k] + IP[II(i,k)]);
	maxI[n] = MAX(maxI[n], maxI[n-k] + IP[II(i,k)]);
      }
    }
  }

  /* get overall minI and maxI */
  for (n=1; n<=N; n++) {
    /*Rprintf("minI[%d] %d maxI[%d] %d\n", n, minI[n], n, maxI[n]);*/
    minI[0] = MIN(minI[0], minI[n]);		/* min for intermediates */
    maxI[0] = MAX(maxI[0], maxI[n]);		/* max for intermediates */
    minI[n] = LOGZEROI;
    maxI[n] = 0;
  }
  Irange = maxI[0] - minI[0] + 2;
  *offset = minI[0] - 1;			/* I offset: I=-1 is array 0 */
  /*Rprintf("minI %d maxI %d Irange %d\n", minI[0], maxI[0], Irange);*/
  minI[0] = LOGZEROI;
  maxI[0] = 0;

  /* create script_P arrays with enough space for intermediate calculations */
  if (Irange > dataSet->Irange){
    Rprintf( "new Irange %d is greater than pre-calculated Irange %ld. numSites = %d. Exiting.\n", Irange, dataSet->Irange, N);
    exit(1);
  }

  
  /* clear intermediate probability array */
  for (n=0; n<=N; n++) for(I=0; I<Irange; I++) logSP[SS(n,I)] = LOGZERO;

  /* init probability array for first letter in alphabet */
  for (n=0; n<=N; n++) {
    I = IP[II(0,n)] - *offset;			/* offset I */
    logSP[SS(n,I)] = logNfact + logP[LL(0,n)];	/* init */
    minI[n] = maxI[n] = I;
  }

  /* compute probabilities recursively */
  for (i=1; i<A; i++) {			/* index over (rest of) alphabet */
    for (n=N; n>=0; n--) {		/* index over samples */
      for (k=1; k<=n; k++) {		/* index over samples of new letter */
        int min = minI[n-k];
        int max = MAX(min, maxI[n-k] - (1-frac)*(maxI[n-k]-minI[n-k]+1));
        /*Rprintf("min %d maxI %d max %d\n", min, maxI[n-k], max);*/
        for (I=min; I<=max; I++) {	/* index over I */
          if (logSP[SS(n-k,I)] > LOGZERO) {
	    /*Rprintf("i %d old: %d %d new: %d %d\n", i, n-k, I, n,I+IP[i][k]);*/
            logSP[SS(n,I+IP[II(i,k)])] = 
              LOG_SUM(logSP[SS(n,I+IP[II(i,k)])], logP[LL(i,k)] + logSP[SS(n-k,I)]);
          }
	}
	/* get current minimum and maximum I in intermediate arrays */
	minI[n] = MIN(minI[n], minI[n-k]+IP[II(i,k)]);
	maxI[n] = MAX(maxI[n], maxI[n-k]+IP[II(i,k)]);
      }
      if (n==N && i==A-1) break;	/* all done */
    }
  }

  /* compute range */
  /*Rprintf("minI[N] %d maxI[N] %d\n", minI[N], maxI[N]);*/
  *range = maxI[N] - minI[N]; 

  /* move to probability array with prob(offset) in position 0 */
  *offset += minI[N];			/* prob[0] = prob(offset) */
  prob = DoubleMalloc(*range+2);

  for (I=minI[N]; I<=maxI[N]; I++) prob[I-minI[N]] = logSP[SS(N,I)];
  /*Rprintf( "N= %d range= %d offset= %d alpha= %f\n", N, *range, 
    *offset, *alpha);*/
      

  return prob;
} /* llr_distr */

/******************************************************************************/
/*
	sum_distr

	Compute the distribution of the sum of two integer-valued
	random variables whose ranges are [0..r1] and [0..r2], 
	respectively, given the log of their distributions.
*/
/******************************************************************************/
double *sum_distr(
  double *d1,				/* (log) distribution of RV1 */
  int r1,				/* range of RV1 */
  double *d2,				/* (log) distribution of RV2 */
  int r2, 				/* range of RV2 */
  int *r_sum				/* range of sum of RV1 and RV2 */
)
{
  int i, j, k;
  int range = r1 + r2;			/* potential range of sum */
  double *d_sum = NULL;			/* distribution of sum */

  d_sum = DoubleMalloc(range+1);	/* space for distribution */

  for (i=0; i<=range; i++) {		/* value of sum */
    d_sum[i] = LOGZERO;
  }

  for (i=0; i<=r1; i++) {		/* range of RV1 */
    if (d1[i]==LOGZERO) continue;
    for (j=0, k=i; j<=r2; j++, k++) {	/* range of RV2 */
      if (d2[j]==LOGZERO) continue;
      d_sum[k] = LOG_SUM(d_sum[k], d1[i]+d2[j]);
    } /* RV2 */
  } /* RV1 */

  for (i=range; i>=0; i--) {		/* value of sum */
    if (d_sum[i] > LOGZERO) break;
  }
  *r_sum = i;				/* non-zero range */

  return d_sum;
} /* sum_distr */

/******************************************************************************/
/*
  	cdf

	Compute (log) 1-CDF of an integer-valued (log) distribution.
	Smooths the CDF by linear interpolation so that adjacent positions
	in the table will have different values if possible.
*/
/******************************************************************************/
double *cdf(
  double *d,				/* integer valued distribution */
  int r					/* range [0..r] */
)
{
  double *cdf=NULL, slope=0;
  int I, i, j, k;

  cdf = DoubleMalloc(r+1);

  cdf[r] = d[r];
  for (I=r-1; I>=0; I--) {
    cdf[I] = LOG_SUM(cdf[I+1], d[I]);
  }

  /* smooth cdf by linear interpolation in logs */
  for (i=r; i>0; i=j) {
    for (j=i-1; j>0 && d[j]==LOGZERO; j--) ;	/* find next non-zero p */ 
    if (i!=j) slope = (cdf[i]-cdf[j])/(i-j);	/* slope */
    for (k=j+1; k<i; k++) cdf[k] = cdf[j] + (k-j)*slope;
  }

  return cdf;
} /* cdf */

/******************************************************************************/
/*
	get_scaled_llr

	Given a set of observed counts (o_i) and expected counts (e_i)
	and a scale factor alpha, return the scaled information content: 
		Sum_i=1..A NINT(alpha * o_i log(o_i/e_i))
*/
/******************************************************************************/
int get_scaled_llr(
  double alpha,			/* scale factor */
  int A,			/* size of alphabet */
  double *o,			/* observed counts */
  double *e 			/* expected counts */
) {
  int i;
  int llr = 0;

  for (i=0; i<A; i++) {
    if (o[i]) {
      llr += NINT(alpha * o[i] * log(o[i]/e[i]));
    }
  }
  return llr >= 0 ? llr : 0;
} /* get_scaled_llr */


/******************************************************************************/
/*
	get_llr_pv

	Get the log p-value of a weighted log-likelihood ratio given:
		llr	log-likelihood ratio
		n	(weighted) number of sequences in the alignment
		w	width of the alignment
		range 	desired number of values in range
		frac	speedup factor
		alength	number of letters in alphabet
		dd[]	alphabet frequency distribution

	Computes the distributions on the fly only as needed.
	Range should be the same each time get_llr_pv is called.
	Call with integral n and w=0 to allocate table in parallel mode.

	Returns the p-value.
*/
/******************************************************************************/
extern double get_llr_pv(
			 DATASET *dataSet,   /* the dataset */
			 double llr,         /* log likelihood ratio */
			 double n,           /* wgtd number sequences in alignment */
			 int w,		     /* width of alignment */
			 int range,	     /* desired range for resolution */
			 double frac,	     /* speedup factor */
			 int alength,	     /* length of alphabet */
			 double *dd 	     /* alphabet frequency distribution */
			 )
{
  int i, N;
  double I;				/* weighted log likelihood ratio */
  int I0, I1;				/* position of llr in table */
  double logpv;				/* log pvalue */
  double n0, n1;			/* floor and ceil of n */
  double alpha;				/* scale factor used */
  DISTR *distrs;

  if (n<=1) return 0.0;			/* only one site p-value = 1.0 */

  /* return geometric mean if N is not integral */
  if ( (n0=floor(n)) != (n1=ceil(n)) )
    return ( 
      (n1-n)*get_llr_pv(dataSet, llr, n0, w, range, frac, alength, dd) +
      (n-n0)*get_llr_pv(dataSet, llr, n1, w, range, frac, alength, dd) 
    );

  N = (int) n;				/* make n an integer */

  
  /* N larger than any previous N? */
  if (dataSet->ndistrs < N) {			/* first call */
    dataSet->distrs = distrMalloc(N+1);
    dataSet->ndistrs = N;				/* set maximum N */
  }

  distrs = dataSet->distrs;

  /* done if w == 0 */
  if (w == 0) return 0.0;

  /* w larger than any previous w for this N? */
  if (distrs[N].w < w) {			/* larger w */
    distrs[N].d      = DoubleMalloc(w+1);
    distrs[N].cdf    = DoubleMalloc(w+1);
    distrs[N].offset = IntMalloc(w+1);
    distrs[N].range  = IntMalloc(w+1);


    /* first time? */ 
    if (distrs[N].w == 0) {			/* get the w=1 distribution */
      distrs[N].d = llr_distr(dataSet, alength, dd, N, range, frac,
        &distrs[N].alpha, &distrs[N].offset[1], &distrs[N].range[1]);
      /* get mean of LLR for w = 1 */
      for (i=0; i<=distrs[N].range[1]; i++) {
        double llr = (i + distrs[N].offset[1]) / distrs[N].alpha;
        distrs[N].mean += exp(distrs[N].d[i])*llr;
      }
      distrs[N].cdf = cdf(distrs[N].d, distrs[N].range[1]);
      distrs[N].w = 1;
    } /* first time */

    distrs[N].w = w; 				/* set maximum w */
  } /* new w */
  
  /* get position in table */
  alpha = distrs[N].alpha;
  I = (alpha * llr) - distrs[N].offset[w];		/* position in table */
  I0 = (int) I;						/* floor of position */
  I1 = I0 + 1;						/* ceil. of position */
  if (I < 0) {						/* lower bound I */
    logpv = distrs[N].cdf[0];
  } else if (I0 >= distrs[N].range[w]) {		/* upper bound I */
    logpv = distrs[N].cdf[distrs[N].range[w]];
  } else {						/* lin. interpolate */
    logpv =
      distrs[N].cdf[I0] + (I-I0)*(distrs[N].cdf[I1]-distrs[N].cdf[I0]);
  }

  /* return log p-value */
  return logpv;
} /* get_llr_pv */




/**********************************************************************/
/*
	get_log_nalign

	Get an upper bound on the number of independent alignments
	of segments of length w.
*/
/**********************************************************************/
extern double get_log_nalign(
			     MTYPE mtype,				/* type of model */
			     int w,	 				/* width of motif */
			     int N,					/* number of occurrences */
			     DATASET *dataSet  				/* the dataset */
			     )
{
  int i, j, t;
  int nseqs;	                         	/* number of sequences */
  double log_nalign;    			/* log number alignments */
  int icfactor;	                                /* double the possible sites */
  int seqNum;                                   /* sequence number */
  int numSeqs;                                  /* number of sequence in entire dataset */
  int *len;                                     /* sequence lengths in decreasing order */

  numSeqs = dataSet->numSeqs;
  len     = dataSet->len;

  nseqs = 0;
  for (seqNum=0; seqNum<numSeqs; seqNum++) if (!dataSet->samples[seqNum].exclude) nseqs++;

  log_nalign = 0;			
  icfactor = dataSet->revComp ? 2 : 1;	

  /* 
    sort the sequence lengths in decreasing order in array len[] first time thru
  */
  for (i=0,j=0; i< numSeqs; i++){
    if (!dataSet->samples[i].exclude){
      len[j] = dataSet->samples[i].length;
      j++;
    }
  }  
  qsort((void *) len, nseqs, sizeof(int), int_compare);

  /*
    get upper bound on number of alignments 
  */
  switch (mtype) {
  case OOPS:
  case ZOOPS:
      if (w > len[N-1]) {				/* impossible w */
	log_nalign = BIG;
      } else {
	for (i=0; i<N; i++) log_nalign += log(icfactor*(len[i]-w+1.0));
        if (N < nseqs) log_nalign += log_comb(nseqs, N);
      }
      break;
  case TCM:
  case CUT:
  case OVER:
    for (i=t=0; i<nseqs; i++) t += len[i] - w + 1;	/* # starts */
    if (N > t) {					/* impossible w & N */
      log_nalign = BIG;
    } else {					/* remove 1 site per site */
      for (i=0; i<N; i++) log_nalign += log((t-i)*icfactor/(i+1));
    }
    break;
  default:
    Rprintf("Model type %d not recognized\n", mtype);
    exit(1);
  } /* mtype */

  return log_nalign;
} /* double get_log_nalign */

/**********************************************************************/
/*
        log_comb
 
        Compute logarithm of m choose n.

*/
/**********************************************************************/
 
extern double log_comb(
		       int m,
		       int n
		       ) 
{
  int i;
  double x = 0;
  int big, little;
  if (m-n > n) { big = m-n; little = n; } else { big = n; little = m-n; }
  for (i=m; i>big; i--) x += log((double) i);
  for (i=2; i<=little; i++) x -= log((double) i);
  return x;
}

/**********************************************************************/
/*
	log_qfast
	
	Calculate the log p-value of the log of the 
	product of uniform [0,1] random variables.

*/
/**********************************************************************/
extern double log_qfast(
  int n,			/* number of random variables in product */
  double logk			/* product of random variables */
)
{
  int i;
  double term, phi;
 
  if (n == 0) return 0;			/* worst possible log p-value */

  phi = term = 1;
  for (i=1; i<n; i++) {
    term *= -logk/i;
    phi += term;
  }

  if (COSMO_MSG_LEVEL>=4) Rprintf("log_qfast: for w = %d, log(phi) = %lf\n", n, log(phi));

  return(logk + log(phi));
} /* qfast */
