/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*		                               	                             *
*      This module contains functions for fitting and evaluating candidate   *
*      models.                                                               *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "estarts.h"
#include "fit.h"
#include "io.h"
#include "matLib.h"
#include "optimize.h"
#include "theta.h"
#include "donlp.h"


/**********************************************************************/
/*
  fit

  Estimate parameters for a given constraint set, model type, motif
  width, number of motif occurrences, and starting value. 
  Model type TCM may be approximated by OVER, CUT model type or not.

  Final parameter estimates are stored in dataSet->svTheta.
  Returns TRUE if algorithm converged, FALSE otherwise

*/
/**********************************************************************/
BOOLEAN fit(
	    DATASET *dataSet,     /* the dataset */
	    THETA *theta,         /* parameter estimates on exit */
	    int startNum,         /* index of starting value */
	    int nSitesNum,        /* index of candidate value for expected number of motif occurrences */
	    int width,            /* motif width */
	    int modTypeNum,       /* index of model type */
	    int conSetNum,        /* constraint set number */
	    SVSPACE *baseSvSpace  /* starting value space to use */
	    )
{


  MTYPE mType;            /* model type */
  int numSv;              /* number of available starting values for a given expected number of motif occurrences */
  int nnSites;            /* effective number of candidate values for expected number of motif occurrences */
  THETA *svTheta;         /* starting value */

  mType     = dataSet->mTypes[modTypeNum];
  nnSites   = baseSvSpace[modTypeNum].nnSites;
  numSv     = baseSvSpace[modTypeNum].numSv;
  svTheta   = dataSet->svTheta;

  /* create theta from starting value */
  copyTheta(baseSvSpace[modTypeNum].svTheta + SV(width, conSetNum, nSitesNum, startNum), svTheta);


  /* use overlap or cutting approximations if desired */
  if (mType==TCM) switchMod(svTheta, dataSet->approx, dataSet);

  if ((COSMO_MSG_LEVEL>=2 && !dataSet->printStatus) || (COSMO_MSG_LEVEL>=2 && dataSet->R)){
    Rprintf( "\rfit: mType = %s conSet = %d width = %d nSitesNum = %d/%d starting value = %d/%d", 
	    getModName(dataSet,mType,1), conSetNum, width, nSitesNum + 1, nnSites, startNum+1, numSv);
  }

  if (COSMO_MSG_LEVEL>=3){
    Rprintf("\nfit: Starting value %d for mType %d  with intensity = %lf (numSites = %d) and score %lf:\n", 
	    startNum+1, svTheta->mType, svTheta->intensity, svTheta->numSites, svTheta->score);
    PrintDoubleMatrix2File(stderr,svTheta->pwm, 4, width);
    Rprintf("\n");
  }
	
  if (svTheta->score == LITTLE){
    if (COSMO_MSG_LEVEL>=3) Rprintf( "This starting value is uninitialized.\n");
    return(FALSE);
  }

      
  /* run optimization */
  donlp2();


  /* copy parameter estimates into output theta */
  copyTheta(svTheta, theta);

  /* set model type back to original one (if approximations CUT or OVER were used) */
  switchMod(theta, mType, dataSet);

  /* return FALSE if optimization did not converge */
  if (!theta->converged){
    if (COSMO_MSG_LEVEL>=3) Rprintf("\nfit: Model didn't converge\n");
    return (FALSE);
  }
	
  /* if desired, find best expected number of sites for these parameter estimates */
  if (dataSet->findBestNumSites) getBestNSites(theta, dataSet);
	
  /* get various criteria values */
  getPwmKlDiv(theta, dataSet->nucFreq);
  getIcProfile(theta);
  getLogLik(theta, dataSet);
  getLLR(theta, dataSet);
  getAIC(theta, dataSet);
  getBIC(theta, dataSet);
  getLogEval(theta, dataSet);

  if (COSMO_MSG_LEVEL>=3){
    Rprintf("fit:  intensity = %lf (numSites = %d): logEval = %e\n",  theta->intensity, theta->numSites, theta->logEval);
    PrintDoubleMatrix2File(stderr,theta->pwm, 4, theta->width);
    Rprintf("\n");
  }


  return(TRUE);
}


/**********************************************************************/
/*
  initCrit

  Initialize a CRIT struct.

*/
/**********************************************************************/
void initCrit(
	      CRIT *crit   /* pointer to struct to be initialized */
	      )
{
  crit->intInit     = FALSE;
  crit->startNum    = 0;
  crit->nSitesNum   = 0;
  crit->intCrit     = BIG;

  crit->wInit       = FALSE;
  crit->width       = 0;
  crit->wCrit       = BIG;

  crit->modInit     = FALSE;
  crit->modTypeNum  = 0;
  crit->mType       = 0;
  crit->modCrit     = BIG;

  crit->conInit     = FALSE;
  crit->conSetNum   = 0;
  crit->conCrit     = BIG;

}


/**********************************************************************/
/*
  copyCrit

  Copy one CRIT struct to another.

*/
/**********************************************************************/
void copyCrit(
	      CRIT *from,   /* copy from this struct */
	      CRIT *to      /* to this struct */
	      )
{
  to->intInit     = from->intInit;
  to->startNum    = from->startNum;
  to->nSitesNum   = from->nSitesNum;
  to->intCrit     = from->intCrit;

  to->wInit       = from->wInit;
  to->width       = from->width;
  to->wCrit       = from->wCrit;

  to->modInit     = from->modInit;
  to->modTypeNum  = from->modTypeNum;
  to->mType       = from->mType;
  to->modCrit     = from->modCrit;

  to->conInit     = from->conInit;
  to->conSetNum   = from->conSetNum;
  to->conCrit     = from->conCrit;

}


/**********************************************************************/
/*
  selectModel

  Chooses optimal fine-tuning parameters by evaluating model selection
  criteria as necessary.
  It does so by choosing the best width for given model type and 
  constraint set, then the best model type for given constraint set,
  and finally the best constraint set.

*/
/**********************************************************************/
int selectModel(
		DATASET *dataSet   /* the dataset */
		)
{
  int width;             /* motif width */
  int minWidth;          /* minimum motif width */
  int maxWidth;          /* maximum motif width */
  int modTypeNum;        /* model type index */
  int numModTypes;       /* number of model types */
  int cSetNum;           /* constraint set index */
  int numConSets;        /* number of constraint sets */
  int startNum;          /* index of current starting value */
  int numSv;             /* number of available starting values for a given expected number of motif occurrences */
  int nSitesNum;         /* index of current expected number of motif occurrences */
  int nnSites;           /* effective number of candidate values for expected number of motif occurrences */
  CRIT new;              /* criteria for given starting value */
  CRIT bestInt;          /* criteria for given constraint set, model type and width */
  CRIT bestW;            /* criteria for given constraint set and model type */
  CRIT bestMod;          /* criteria for given constraint set */
  CRIT bestCon;          /* criteria for best model over all */
  SVSPACE *svSpace;      /* starting value space */
  int candNum;           /* number of current candidate model */
  int seqNum;            /* sequence number */

  candNum      = 0;
  minWidth     = dataSet->minWidth;
  maxWidth     = dataSet->maxWidth;
  numModTypes  = dataSet->numModTypes;
  numConSets   = dataSet->numConSets;
  cSetNum      = dataSet->hasConstraints ? 1 : 0;

  /* initialize CRIT structs */
  initCrit(&new);
  initCrit(&bestInt);
  initCrit(&bestW);
  initCrit(&bestMod);
  initCrit(&bestCon);

  bestCon.conInit = FALSE;


  /* find best constraint set, model type and motif width */
  for (; cSetNum < dataSet->numConSets; cSetNum++){

    bestMod.modInit = FALSE;

    /* fit model on entire dataset by default*/											    
    for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++) dataSet->samples[seqNum].exclude = FALSE;

    /* for given constraint set, find best model type and width */
    for (modTypeNum=0; modTypeNum<numModTypes; modTypeNum++){

      svSpace  = dataSet->svSpace + modTypeNum;
      nnSites  = svSpace->nnSites;
      numSv    = svSpace->numSv;
      
      bestW.wInit = FALSE;

      /* for given constraint set and model type, find best width */
      for (width = minWidth; width <= maxWidth; width++){

	bestInt.intInit = FALSE;


	/* loop over candidate numbers of motif occurrences */
	for (nSitesNum=0; nSitesNum<nnSites; nSitesNum++){
	  
	  /* loop over different starting values */
	  for (startNum=0; startNum<numSv; startNum++){

	    new.startNum   = startNum;
	    new.nSitesNum  = nSitesNum;
	    new.width      = width;
	    new.modTypeNum = modTypeNum;
	    new.mType      = dataSet->mTypes[modTypeNum];
	    new.conSetNum  = cSetNum;
	    
	    switch(dataSet->intCrit.crit){
	    case lik:
	    case eval:
	      getSimpleCriteria(dataSet, &new, dataSet->svSpace);
	      break;
	    case likCV:
	      getCvLikRisk(dataSet, &new, dataSet->intCrit.fold, dataSet->svSpace);
	      break;
	    default: 
	      Rprintf("selectModel: criterion %d for intensity parameter not valid. Exiting...\n", dataSet->intCrit.crit);
	      return(FALSE);
	    }
 	    

	    /* new best value for intensity criterion ? */
	    if (new.intInit && (!bestInt.intInit || (new.intCrit < bestInt.intCrit))){
	      copyCrit(&new, &bestInt);   
	      if (COSMO_MSG_LEVEL>=3 && bestInt.intInit){
		Rprintf( "\nselectModel: new best intensity parameter for\n");
		Rprintf( "             cSet %d mType %s width %d nSitesNum  %d startNum %d\n", 
			bestInt.conSetNum, getModName(dataSet, bestInt.mType,0), bestInt.width, bestInt.nSitesNum, bestInt.startNum);
		Rprintf( "             with criterion %lf\n", bestInt.intCrit);
	      }
	    }
	  }
	}


	/* print first status message if desired */
	candNum++;
	if (dataSet->printStatus) Rprintf("%d/%d   %d/%d\n",dataSet->numSeqs,dataSet->numSeqs, candNum, dataSet->effNumModels);

	if (!bestInt.intInit){
	  Rprintf( "\nNo models converged for width %d model %d and constraint set %d\n", width, modTypeNum, cSetNum);
	  continue;
	}
	

	if (!bestInt.wInit && (dataSet->numWidths>1)){

	  switch(dataSet->wCrit.crit){
	  case lik:
	  case aic:
	  case bic:
	  case eval:
	    getSimpleCriteria(dataSet, &bestInt, dataSet->svSpace);
	    break;
	  case likCV:
	    getCvLikRisk(dataSet, &bestInt, dataSet->wCrit.fold, dataSet->svSpace);
	    break;
	  default: 
	    Rprintf("selectModel: criterion %d for width parameter not valid. Exiting...\n", dataSet->wCrit.crit);
	    return(FALSE);
	  }
	}


	/* new best value for width criterion ? */
	if ((bestInt.wInit || dataSet->numWidths == 1) && (!bestW.wInit || (bestInt.wCrit < bestW.wCrit))){
	  copyCrit(&bestInt, &bestW);   
	  
	  if (COSMO_MSG_LEVEL>=3 && bestW.wInit){
	    Rprintf( "\nselectModel: new best motif width for\n");
	    Rprintf( "             cSet %d mType %s width %d nSitesNum  %d startNum %d\n", 
		    bestW.conSetNum, getModName(dataSet, bestW.mType,0), bestW.width, bestW.nSitesNum, bestW.startNum);
		Rprintf( "             with criterion %lf\n", bestW.wCrit);
	  }


	}

	/* store results in array dataSet->allCrits */
	copyCrit(&bestInt, dataSet->allCrits + MOD(width, modTypeNum, cSetNum));

	/* set criteria for model type and constraint to unintialized */
	if (dataSet->numWidths==1) dataSet->allCrits[MOD(width,modTypeNum, cSetNum)].wInit = FALSE;
	dataSet->allCrits[MOD(width,modTypeNum, cSetNum)].modInit = FALSE;
	dataSet->allCrits[MOD(width,modTypeNum, cSetNum)].conInit = FALSE;

      }

      if (!bestW.intInit) continue;


      if (!bestW.modInit && (dataSet->numModTypes>1)){
	
	switch(dataSet->modCrit.crit){
	case lik:
	case aic:
	case bic:
	case eval:
	  getSimpleCriteria(dataSet, &bestW, dataSet->svSpace);
	  break;
	case likCV:
	  getCvLikRisk(dataSet, &bestW, dataSet->modCrit.fold, dataSet->svSpace);
	  break;
	default: 
	  Rprintf("selectModel: criterion %d for model type not valid. Exiting...\n", dataSet->modCrit.crit);
	  return(FALSE);
	}
      }

      /* store results in array dataSet->allCrits */    
      copyCrit(&bestW, dataSet->allCrits + MOD(bestW.width,modTypeNum, cSetNum));
      if (dataSet->numWidths==1)    dataSet->allCrits[MOD(bestW.width,modTypeNum, cSetNum)].wInit = FALSE;
      if (dataSet->numModTypes==1)  dataSet->allCrits[MOD(bestW.width,modTypeNum, cSetNum)].modInit = FALSE;
      dataSet->allCrits[MOD(bestW.width,modTypeNum, cSetNum)].conInit = FALSE;




      /* new best value for width criterion ? */
      if ((bestW.modInit || dataSet->numModTypes==1) && (!bestMod.modInit || (bestW.modCrit < bestMod.modCrit))){

	copyCrit(&bestW, &bestMod);   

	if (COSMO_MSG_LEVEL>=3 && bestMod.modInit){
	  Rprintf( "\nselectModel: new best model type for\n");
	  Rprintf( "             cSet %d mType %s width %d nSitesNum  %d startNum %d\n", 
              bestMod.conSetNum, getModName(dataSet, bestMod.mType,0), bestMod.width, bestMod.nSitesNum, bestMod.startNum);
	  Rprintf( "             with criterion %lf\n", bestMod.modCrit);
	}
      }
	
    }

    if (!bestMod.intInit) continue;

    if (!bestMod.conInit && (dataSet->numConSets>2)){
      
      switch(dataSet->conCrit.crit){
      case lik:
      case aic:
      case bic:
      case eval:
	getSimpleCriteria(dataSet, &bestMod, dataSet->svSpace);
	break;
      case likCV:
	getCvLikRisk(dataSet, &bestMod, dataSet->conCrit.fold, dataSet->svSpace);
	break;
      case pwmCV:
	getCvPwmRisk(dataSet, &bestMod);
	break;
      default: 
	Rprintf("selectModel: criterion %d for constraint set not valid. Exiting...\n", dataSet->conCrit.crit);
	return(FALSE);
      }
    }

    /* store results in array dataSet->allCrits */    
    copyCrit(&bestMod, dataSet->allCrits + MOD(bestMod.width,bestMod.modTypeNum, cSetNum));
    if (dataSet->numWidths==1)   dataSet->allCrits[MOD(bestMod.width,bestMod.modTypeNum, cSetNum)].wInit = FALSE;
    if (dataSet->numModTypes==1) dataSet->allCrits[MOD(bestMod.width,bestMod.modTypeNum, cSetNum)].modInit=FALSE;
    if (dataSet->numConSets<=2) dataSet->allCrits[MOD(bestMod.width,bestMod.modTypeNum, cSetNum)].conInit = FALSE;

    /* new best value for width criterion ? */
    if ((bestMod.conInit || dataSet->numConSets<=2) && (!bestCon.conInit || (bestMod.conCrit < bestCon.conCrit))){
      copyCrit(&bestMod, &bestCon);   

      if (COSMO_MSG_LEVEL>=3 && bestCon.conInit){
	Rprintf( "\nselectModel: new best constraint set for\n");
	Rprintf( "             cSet %d mType %s width %d nSitesNum  %d startNum %d\n", 
		bestCon.conSetNum, getModName(dataSet, bestCon.mType,0), bestCon.width, bestCon.nSitesNum, bestCon.startNum);
	Rprintf( "             with criterion %lf\n", bestCon.conCrit);
      }
    }

  }
  
  if (!bestCon.intInit) return(FALSE);

  copyCrit(&bestCon, &(dataSet->crits));

  /* get string descriptors of the choices that have been made */
  getCritDesc(dataSet); 


  if (COSMO_MSG_LEVEL>=3){
    Rprintf("\nselectModel: Final model parameter\n");
    Rprintf("%s\n", dataSet->wDesc);
    Rprintf("%s\n", dataSet->modDesc);
    Rprintf("%s\n", dataSet->conDesc);
  }

  return(TRUE);
}


/**********************************************************************/
/*
  getSimpleCriteria

  For a given candidate model indexed by constraint set, model type, 
  motif width, number of motif occurrences, and starting value, 
  evaluate the simple criteria likelihood, AIC, BIC, and E-value on 
  the whole dataset.

  Criteria are stored as needed in the struct 'crit'.

*/
/**********************************************************************/
void getSimpleCriteria(
		       DATASET *dataSet,    /* the dataset */
		       CRIT *crit,          /* CRIT struct in which to enter the criterion values */
		       SVSPACE *backSvSpace /* starting value space to use */
		       )
{

  BOOLEAN success;     /* did algorithm converge? */
  THETA *theta;        /* parameter estimates */

  theta            = dataSet->theta;

  success = fit(dataSet, theta, crit->startNum, crit->nSitesNum, crit->width, crit->modTypeNum, crit->conSetNum, backSvSpace);

  /* enter intensity criterion if necessary */
  switch(dataSet->intCrit.crit){
  case lik: 
   crit->intCrit = success ? -theta->logLik : BIG;
    crit->intInit = success;
    break;
  case eval:
    crit->intCrit = success ? theta->logEval : BIG;
    crit->intInit = success;
    break;
  default:
    break;
  }

  /* enter width criterion if necessary */
  switch(dataSet->wCrit.crit){
  case lik:
    crit->wCrit = success ? -theta->logLik : BIG;
    crit->wInit = success;
    break;
  case aic:
    crit->wCrit = success ? theta->aic : BIG;
    crit->wInit = success;
    break;
  case bic:
    crit->wCrit = success ? theta->bic : BIG;
    crit->wInit = success;
    break;
  case eval:
    crit->wCrit = success ? theta->logEval : BIG;
    crit->wInit = success;
    break;
  default:
    break;
  }

  /* enter model type criterion if necessary */
  switch(dataSet->modCrit.crit){
  case lik:
    crit->modCrit = success ? -theta->logLik : BIG;
    crit->modInit = success;
    break;
  case aic:
    crit->modCrit = success ? theta->aic : BIG;
    crit->modInit = success;
    break;
  case bic:
    crit->modCrit = success ? theta->bic : BIG;
    crit->modInit = success;
    break;
  case eval:
    crit->modCrit = success ? theta->logEval : BIG;
    crit->modInit = success;
    break;
  default:
    break;
  }

  /* enter constraint set criterion if necessary */
  switch(dataSet->conCrit.crit){
  case lik:
    crit->conCrit = success ? -theta->logLik : BIG;
    crit->conInit = success;
    break;
  case aic:
    crit->conCrit = success ? theta->aic : BIG;
    crit->conInit = success;
    break;
  case bic:
    crit->conCrit = success ? theta->bic : BIG;
    crit->conInit = success;
    break;
  case eval:
    crit->conCrit = success ? theta->logEval : BIG;
    crit->conInit = success;
    break;
  default:
    break;
  }
}


/**********************************************************************/
/*
  getCvLikRisk

  Get likelihood-based cross-validation criteria for a given starting
  value, number of motif occurrences, motif width, model type, and 
  constraint set. 

  Criteria are stored as needed in the struct 'crit'.

*/
/**********************************************************************/
void getCvLikRisk(
		  DATASET *dataSet,    /* the dataset */
		  CRIT *crit,          /* criteria */
		  int fold,            /* fold for CV */
		  SVSPACE *backSvSpace /* starting value space to use */
		  )
{

  double intKlDiv;      /* KL divergence for choosing intensity parameter */
  double wKlDiv;        /* KL divergence for choosing width */
  double modKlDiv;      /* KL divergence for choosing model type */
  double conKlDiv;      /* KL divergence for choosing constraint set */
  int valGroup;         /* validation group */
  BOOLEAN conv;         /* did algorithm converge? */
  BOOLEAN cvFail;       /* did the entire cross-validation fail? */
  int effFold;          /* effective fold, which may be less than fold if some models do not converge */
  THETA *theta;         /* parameter estimates */
  int seqNum;           /* sequence number */
  int numSeqs;          /* number of sequences */
  int numSites0;        /* original number of sites for starting value */
  int numSites1;        /* adjusted number of sites */
  int effNumSeqs;       /* effective number of sequences */
  int numSv;            /* number of available starting values for a given expected number of motif occurrences */
  int nnSites;          /* effective number of candidate values for expected number of motif occurrences */

  nnSites    = backSvSpace[crit->modTypeNum].nnSites;
  numSv      = backSvSpace[crit->modTypeNum].numSv;
  theta      = dataSet->theta;
  numSeqs    = dataSet->numSeqs;

  /* create CV groups */
  createCvGroups(dataSet->samples, dataSet->numSeqs, fold);
  
  intKlDiv = 0;
  wKlDiv   = 0;
  modKlDiv = 0;
  conKlDiv = 0;
  effFold  = 0;
  cvFail   = TRUE;
  
  /* loop over different validation groups */
  for (valGroup=0; valGroup<fold; valGroup++){

    /* exclude validation sample for fitting model */
    effNumSeqs = 0;
    for (seqNum=0; seqNum<numSeqs; seqNum++){
      dataSet->samples[seqNum].exclude = dataSet->samples[seqNum].cvGroup == valGroup;
      if (!dataSet->samples[seqNum].exclude) effNumSeqs++;
    }

    /* adjust number of sites since starting values were found on whole data set */
    numSites0 = backSvSpace[crit->modTypeNum].svTheta[SV(crit->width, crit->conSetNum, crit->nSitesNum, crit->startNum)].numSites;
    numSites1 = numSites0 *((double)effNumSeqs/(double)numSeqs);
    backSvSpace[crit->modTypeNum].svTheta[SV(crit->width, crit->conSetNum, crit->nSitesNum, crit->startNum)].numSites = numSites1;
    
    /* fit model */
    conv = fit(dataSet, theta, crit->startNum, crit->nSitesNum, crit->width, crit->modTypeNum, crit->conSetNum, backSvSpace);
    
    backSvSpace[crit->modTypeNum].svTheta[SV(crit->width, crit->conSetNum, crit->nSitesNum, crit->startNum)].numSites = numSites0;

    /* cvFails only if all sample splits fail to converge */
    cvFail = cvFail && (!conv);
    if (!conv) continue;      
    
    effFold++;

    /* exclude training sample for computing KL divergence */
    for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = dataSet->samples[seqNum].cvGroup != valGroup;

    /* do we need a CV for the intensity criterion with the same fold? */
    if (dataSet->intCrit.crit==likCV && dataSet->intCrit.fold==fold){
      intKlDiv += getKlDiv(dataSet,theta,dataSet->intCrit.trunc,dataSet->intCrit.quant);
    }
    
    /* do we need a CV for the width criterion with the same fold? */
    if (dataSet->wCrit.crit==likCV && dataSet->wCrit.fold==fold){
      wKlDiv += getKlDiv(dataSet, theta,dataSet->wCrit.trunc,dataSet->wCrit.quant);
    }
    
    /* do we need a CV for the model type criterion with the same fold? */
    if (dataSet->modCrit.crit==likCV && dataSet->modCrit.fold==fold){
      modKlDiv += getKlDiv(dataSet, theta,dataSet->modCrit.trunc,dataSet->modCrit.quant);
    }
    
    /* do we need a CV for the constraint set  criterion with the same fold? */
    if (dataSet->conCrit.crit==likCV && dataSet->conCrit.fold==fold){
      conKlDiv += getKlDiv(dataSet, theta,dataSet->conCrit.trunc,dataSet->conCrit.quant);
    }
  } 

  /* reset exclude to FALSE for all sequences */
  for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = FALSE;


  /* do we need a CV for the width criterion with the same fold? */
  if (dataSet->intCrit.crit==likCV && dataSet->intCrit.fold==fold){
    crit->intCrit = cvFail ? BIG : intKlDiv/effFold;
    crit->intInit = !cvFail;
  }

  /* do we need a CV for the width criterion with the same fold? */
  if (dataSet->wCrit.crit==likCV && dataSet->wCrit.fold==fold){
    crit->wCrit = cvFail ? BIG : wKlDiv/effFold;
    crit->wInit = !cvFail;
  }

  /* do we need a CV for the model type criterion with the same fold? */
  if (dataSet->modCrit.crit==likCV && dataSet->modCrit.fold==fold){
    crit->modCrit = cvFail ? BIG : modKlDiv/effFold;
    crit->modInit = !cvFail;
  }

  /* do we need a CV for the model type criterion with the same fold? */
  if (dataSet->conCrit.crit==likCV && dataSet->conCrit.fold==fold){
    crit->conCrit = cvFail ? BIG : conKlDiv/effFold;
    crit->conInit = !cvFail;
  }

}

/**********************************************************************/
/*
  getCvPwmRisk

  Get Euclidean distance-based cross-validation criteria for a given 
  starting value, number of motif occurrences, motif width, model type, 
  and constraint set. 

  Criterion is stored in struct 'crit'.

*/
/**********************************************************************/
void getCvPwmRisk(
		  DATASET *dataSet,   /* the dataset */
		  CRIT *crit          /* criteria */
		  )
{

  double loss;          /* value of loss function */
  int valGroup;         /* validation group */
  MTYPE mType;          /* model type */
  BOOLEAN conv;         /* did model converge? */
  BOOLEAN cvFail;       /* did all models for current CV fail to converge? */
  BOOLEAN groupFail;    /* did one of the two models (constraint & unconstrained) for the current sample split fail to converge? */ 
  int fold;             /* current fold */
  int effFold;          /* effective fold, which may be less than fold if some models do not converge */
  int startNum;         /* index of current starting value */
  int nSitesNum;        /* index of current expected number of motif occurrences */
  int width;            /* motif width */
  int modTypeNum;       /* index of model type */
  int cSetNum;          /* constraint set number */
  THETA *theta;         /* constrained parameter estimates */
  THETA *refTheta;      /* unconstrained parameter estimates */
  SVSPACE *svSpace;     /* starting value space */
  int nnSites;          /* effective number of candidate values for number of sites */
  int numSv;            /* number of starting values to try */
  int seqNum;           /* sequence number */
  int numSeqs;          /* number of sequences */
  int bestStartNum;     /* best starting value */
  int bestNSitesNum;    /* best number of sites index */
  double bestEValue;    /* best E-value */
  BOOLEAN success;      /* did fit converge? */
  double eval;          /* E-value criterion */
  double dist;          /* Euclidean distance between 2 PWM estimates */
  int numSites0;        /* original number of sites for starting value */
  int numSites1;        /* adjusted number of sites */
  int effNumSeqs;       /* effective number of sequences */


  
  fold          = dataSet->conCrit.fold;
  width         = crit->width;
  modTypeNum    = crit->modTypeNum;
  cSetNum       = crit->conSetNum;
  theta         = dataSet->theta1;
  refTheta      = dataSet->theta2;
  mType         = dataSet->mTypes[modTypeNum];
  numSeqs       = dataSet->numSeqs;

  /* create CV groups */
  createCvGroups(dataSet->samples, dataSet->numSeqs, fold);

  cvFail  = TRUE;
  loss    = 0;
  effFold = 0;


  /* get unconstrained reference thetas */
  if (COSMO_MSG_LEVEL>=3) Rprintf( "Fitting unconstrained models\n");

  for (valGroup=0; valGroup<fold;valGroup++){

    if (dataSet->cvStarts){
      /* exclude valGroup v */      
      for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = dataSet->samples[seqNum].cvGroup == valGroup;
      svSpace = valGroup==0 ? dataSet->cvSvSpace1 : dataSet->cvSvSpace2;
    }else{
      for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = FALSE;    
      svSpace = dataSet->svSpace;
    }

    nnSites  = svSpace[modTypeNum].nnSites;
    numSv    = svSpace[modTypeNum].numSv;

    /* select starting value and number of sites by E-value only */
    bestStartNum  = 0;
    bestNSitesNum = 0;
    bestEValue    = BIG;


    /* loop over candidate numbers of motif occurrences */
    for (nSitesNum=0; nSitesNum<nnSites; nSitesNum++){
    
      /* loop over different starting values */
      for (startNum=0; startNum<numSv; startNum++){

	if (COSMO_MSG_LEVEL>=4) Rprintf( "Unconstrained valgroup %d nSitesNum %d startNum %d\n", 
					valGroup, nSitesNum, startNum);

	success = fit(dataSet, theta, startNum, nSitesNum, width, modTypeNum, 0, svSpace);

	eval = success ? theta->logEval : BIG;

	if (COSMO_MSG_LEVEL>=3) Rprintf( "Unconstrained valgroup %d nSitesNum %d startNum %d: conv = %d and logEval = %e\n", 
					valGroup, nSitesNum, startNum, success, eval);



	/* new best starting value and number of sites? */
	if ((nSitesNum==0 && startNum==0) || (eval<bestEValue)){
	  bestStartNum  = startNum;
	  bestNSitesNum = nSitesNum;
	  bestEValue = eval;
	}
      }
    }

    effNumSeqs = 0;
    for (seqNum=0; seqNum<numSeqs; seqNum++){
      dataSet->samples[seqNum].exclude = dataSet->samples[seqNum].cvGroup == valGroup;
      if (!dataSet->samples[seqNum].exclude) effNumSeqs++;
    }

    /* adjust number of sites if necessary */
    numSites0 = svSpace[modTypeNum].svTheta[SV(width, 0, bestNSitesNum, bestStartNum)].numSites;
    if (!dataSet->cvStarts){
      numSites1 = numSites0 *((double)effNumSeqs/(double)numSeqs);
      svSpace[modTypeNum].svTheta[SV(width, 0, bestNSitesNum, bestStartNum)].numSites = numSites1;
    }

    refTheta = dataSet->refThetas + valGroup;      
    dataSet->refThetaConv[valGroup] = fit(dataSet, refTheta, bestStartNum, bestNSitesNum, width, modTypeNum, 0, svSpace);

    /* change back number of sites if necessary */
    if (!dataSet->cvStarts) svSpace[modTypeNum].svTheta[SV(width, 0, bestNSitesNum, bestStartNum)].numSites = numSites0;

    if (COSMO_MSG_LEVEL>=3){
      Rprintf("\nUnconstrained model for width %d and model type number %d and valGroup %d:\n", 
	      width, modTypeNum, valGroup);
      PrintDoubleMatrix2File(stderr, refTheta->pwm, 4, width);
    }      
    
  }      


  /* loop over validation groups */
  for (valGroup=0; valGroup<fold; valGroup++){
    
    groupFail = FALSE;
    if (COSMO_MSG_LEVEL>=3) Rprintf( "cSetNum %d is empty: %d\n", cSetNum, dataSet->cSets[cSetNum].isEmpty);

    /* Fit constrained model first */
    if (dataSet->cSets[cSetNum].isEmpty && fold==2){
      
      /* Can retrieve theta from one of the pre-calculated thetas */
      theta = dataSet->refThetas + valGroup;

      conv = dataSet->refThetaConv[valGroup];

      /* group fails if any one model fails to converge */
      groupFail = groupFail || (!conv);

    }else{

      
      /* fit model */
      if(COSMO_MSG_LEVEL>=3){
	Rprintf("getCvPwmRisk: Fitting constraint set %d for valGroup %d\n", cSetNum, valGroup);
      }

      if (dataSet->cvStarts){
	/* exclude validation sample */
	for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = dataSet->samples[seqNum].cvGroup == valGroup;
	svSpace = valGroup==0 ? dataSet->cvSvSpace1 : dataSet->cvSvSpace2;

	nnSites  = svSpace[modTypeNum].nnSites;
	numSv    = svSpace[modTypeNum].numSv;
	
	/* select starting value and number of sites by E-value only */
	bestStartNum  = 0;
	bestNSitesNum = 0;
	bestEValue    = BIG;
	
	/* loop over candidate numbers of motif occurrences */
	for (nSitesNum=0; nSitesNum<nnSites; nSitesNum++){
	  
	  /* loop over different starting values */
	  for (startNum=0; startNum<numSv; startNum++){
	    
	    if (COSMO_MSG_LEVEL>=3) Rprintf( "Constrained valgroup %d nSitesNum %d startNum %d\n", 
					  valGroup, nSitesNum, startNum);
	    success = fit(dataSet, theta, startNum, nSitesNum, width, modTypeNum, cSetNum, svSpace);
	    
	    eval = success ? theta->logEval : BIG;
	    
	    if (COSMO_MSG_LEVEL>=3) Rprintf( "Constrained valgroup %d nSitesNum %d startNum %d: conv = %d and logEval = %e\n", 
					    valGroup, nSitesNum, startNum, success, eval);
	    
	    
	    /* new best starting value and number of sites? */
	    if ((nSitesNum==0 && startNum==0) || (eval<bestEValue)){
	      bestStartNum  = startNum;
	      bestNSitesNum = nSitesNum;
	      bestEValue = eval;
	    }
	    
	  }
	}
	
      }else{
	svSpace = dataSet->svSpace;    

	bestStartNum  = crit->startNum;
	bestNSitesNum = crit->nSitesNum;

      }

      effNumSeqs = 0;
      for (seqNum=0; seqNum<numSeqs; seqNum++){
	dataSet->samples[seqNum].exclude = dataSet->samples[seqNum].cvGroup == valGroup;
	if (!dataSet->samples[seqNum].exclude) effNumSeqs++;
      }
      
      /* adjust number of sites if necessary */
      if (!dataSet->cvStarts){
	numSites0 = svSpace[modTypeNum].svTheta[SV(width, 0, bestNSitesNum, bestStartNum)].numSites;
	numSites1 = numSites0 *((double)effNumSeqs/(double)numSeqs);
	svSpace[modTypeNum].svTheta[SV(width, 0, bestNSitesNum, bestStartNum)].numSites = numSites1;
      }
	
      conv = fit(dataSet, theta, bestStartNum, bestNSitesNum, width, modTypeNum, cSetNum,  svSpace);

      /* change back number of sites if necessary */
      if (!dataSet->cvStarts) svSpace[modTypeNum].svTheta[SV(width, 0, bestNSitesNum, bestStartNum)].numSites = numSites0;
      

      /* group fails if any one model fails to converge */
      groupFail = groupFail || (!conv);
    }

    if (COSMO_MSG_LEVEL>=3){
      Rprintf("\n\nConstrained model for constraint set %d and model type number %d and valGroup %d:\n", 
	      cSetNum, modTypeNum, valGroup);
      PrintDoubleMatrix2File(stderr, theta->pwm, 4, width);
    }      
    
      


    /* retrieve pre-calculated reference theta for no constraints */
    refTheta = dataSet->refThetas + fold - valGroup - 1;
    
    conv = dataSet->refThetaConv[fold - valGroup - 1];
    
    /* group fails if any one model fails to converge */
    groupFail = groupFail || (!conv);
            
    if (COSMO_MSG_LEVEL>=3){
      Rprintf( "\n\ngetCvPwmRisk: unconstrained model for valgroup %d\n", valGroup);
      PrintDoubleMatrix2File(stderr,refTheta->pwm, 4, refTheta->width);
    }

    /* CV only fails if all groups fail */
    cvFail = cvFail && groupFail;
    if (groupFail) continue;

    dist  = euclDist(theta->pwm, refTheta->pwm,theta->width);

    loss += dist;

    if (COSMO_MSG_LEVEL>=3) Rprintf( "\ndistance = %lf\n", dist);

    effFold++;
  } 
  
  if (COSMO_MSG_LEVEL>=3) Rprintf( "loss = %lf, effFold = %d\n", loss, effFold);

  for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = FALSE;

  /* record criterion */
  crit->conCrit = cvFail ? BIG : loss/effFold;
  crit->conInit = !cvFail;

}

/**********************************************************************/
/*
  finalModel

  Fit final model for given choice of fine-tuning parameters stored
  in dataSet->pChoice.

  Returns TRUE if model is fit successfully, FALSE otherwise.

*/
/**********************************************************************/
BOOLEAN finalModel(
		   DATASET *dataSet  /* the dataset */
		   )
{

  int startNum;      /* number of starting value */
  int nSitesNum;     /* index of candidate value for expected number of motif occurrences */
  int width;         /* chosen motif width */
  MTYPE mType;       /* chosen model type */
  int cSetNum;       /* chosen constraint set number */
  int modTypeNum;    /* chosen model type number */
  BOOLEAN success;   /* did the algorithm converge for this choice of parameters? */
  int seqNum;        /* sequence number */
  int numSeqs;       /* number of sequences */

  /* these are the choices for the fine-tuning parameters */
  startNum   = dataSet->crits.startNum;
  nSitesNum  = dataSet->crits.nSitesNum;
  width      = dataSet->crits.width;
  mType      = dataSet->crits.mType;
  cSetNum    = dataSet->crits.conSetNum;
  modTypeNum = dataSet->crits.modTypeNum;
  numSeqs    = dataSet->numSeqs;

  /* fit model on whole dataset */  
  for (seqNum=0; seqNum<numSeqs; seqNum++) dataSet->samples[seqNum].exclude = FALSE;


  if (COSMO_MSG_LEVEL>=2) Rprintf( "\nfinalModel: fitting model for width %d modType %s and conSet %d\n", 
			width, getModName(dataSet,mType,0), cSetNum);

  if (COSMO_MSG_LEVEL>=2) Rprintf( "\nfinalModel: startNum %d and nSitesNum %d\n", startNum, nSitesNum); 

  /* fit model */
  success = fit(dataSet, dataSet->theta, startNum, nSitesNum, width, modTypeNum, cSetNum, dataSet->svSpace);


  if (COSMO_MSG_LEVEL>=2) Rprintf("\n\n");

  return(success);
}

  
/**********************************************************************/
/*
  createCvGroups

  Divides samples into cross-validation groups
  Currently, this is done deterministically rather than at random

*/
/**********************************************************************/
void createCvGroups(
		    SAMPLE* samples,          /* the samples that are to be divided up */
		    int numSeqs,              /* the number of samples */
		    int fold                  /* the number of groups that are to be created */
		    )  
{

  int v;           /* current validation sample group */
  int seqNum;      /* current sequence number */


  /* split up sequences into 'fold' groups */
  for (v=0;v<fold;v++){
    for (seqNum=0;seqNum<numSeqs;seqNum++){
      if (seqNum>=((double)numSeqs)*v/fold && seqNum<((double)numSeqs)*(v+1)/fold){
	samples[seqNum].cvGroup = v;
      }
    }
  }

  if (COSMO_MSG_LEVEL>=3){
    for (seqNum=0;seqNum<numSeqs;seqNum++){
      Rprintf( "Seq %d is in valGroup %d\n", seqNum, samples[seqNum].cvGroup);
    }
  }

}

