/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains functions for freeing memory for various         * 
*      structures                                                            *
*                                                                            *
*****************************************************************************/


#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "free.h" 
#include "donlp.h"  

/**********************************************************************/
/*
  dataSetFree

  Free memory allocated for this dataset.
	
*/
/**********************************************************************/
void dataSetFree(
		 DATASET *dataSet     /* the dataset */
		 )
{

  if (dataSet==NULL) return;

  /* these objects can be freed the normal way */
  myfree(dataSet->nucFreq);
  myfree(dataSet->mTypes);
  myfree(dataSet->modTypeName);
  myfree(dataSet->backProb);
  myfree(dataSet->logLiks);
  myfree(dataSet->refThetaConv);
  myfree(dataSet->pwm1);
  myfree(dataSet->pwm2);
  myfree(dataSet->lPwm1);
  myfree(dataSet->lPwm2);
  myfree(dataSet->cumProb);
  myfree(dataSet->motifCounts);  
  myfree(dataSet->intMotProbs1);
  myfree(dataSet->intMotProbs2);
  myfree(dataSet->maxMotProbs);
  myfree(dataSet->postProbs);
  myfree(dataSet->maxPostProbs); 
  myfree(dataSet->allCrits);
  myfree(dataSet->intNumBasePairs);
  myfree(dataSet->intFirst);
  myfree(dataSet->intLast);
  myfree(dataSet->line);
  myfree(dataSet->word);
  myfree(dataSet->first);
  myfree(dataSet->len);
  myfree(dataSet->minI);
  myfree(dataSet->maxI);
  myfree(dataSet->IP);
  myfree(dataSet->logP);
  myfree(dataSet->logSP);
  myfree(dataSet->fields[0]);
  myfree(dataSet->fields[1]);
  myfree(dataSet->fields);
  myfree(dataSet->backKL);
  
  if (dataSet->hasConstraints) myfree(dataSet->conString.con);

  
  /* these structure need to memory free inside of them */
  donlpFree(dataSet);
  tmFree(dataSet);
  svSpaceFree(dataSet->svSpace, dataSet->numModTypes); 
  if ((dataSet->numConSets>2) & (dataSet->conCrit.crit==pwmCV)){
    svSpaceFree(dataSet->cvSvSpace1, dataSet->numModTypes); 
    svSpaceFree(dataSet->cvSvSpace2, dataSet->numModTypes); 
  }  
  sampleFree(dataSet->samples, dataSet->numSeqs);
  sampleFree(dataSet->backSamples, dataSet->numBackSeqs);
  conSetFree(dataSet->cSets, dataSet->numConSets);
  modFree(dataSet->mod, dataSet->numModels);
  thetaFree(dataSet->refThetas, dataSet->nRefThetas);
  thetaFree(dataSet->svTheta,1);
  thetaFree(dataSet->theta,1);
  thetaFree(dataSet->theta1,1);
  thetaFree(dataSet->theta2,1);
  distrFree(dataSet->distrs, MIN(2, dataSet->minNumSites), dataSet->ndistrs);


}

/**********************************************************************/
/*
  svSpaceFree

  Free memory allocated inside n SVSPACE structs

*/
/**********************************************************************/
void svSpaceFree(
		 SVSPACE *svSpace, /* pointer to the first SVSPACE struct */
		 long n            /* number of structs to free memory for */
		 )
{
  int num;              /* index for iterating over structs */

  if (svSpace==NULL) return;

  for(num=0; num<n;num++){

    myfree(svSpace[num].svLogLR);
    myfree(svSpace[num].numSites);
    thetaFree(svSpace[num].svTheta, svSpace[num].totSv);  
  }

  myfree(svSpace);
}

/**********************************************************************/
/*
  thetaFree

  Free memeory allocated inside n THETA structs

*/
/**********************************************************************/
void  thetaFree(
		THETA *theta,      /* pointer to first THETA struct */
		long n             /* number of structs to free memory for */
		)
{
  int num;           /* index for iterating over structs */

  if (theta==NULL) return;

  for (num=0; num<n; num++){

    myfree(theta[num].pwm);
    myfree(theta[num].profParms);
    myfree(theta[num].posKL);
    myfree(theta[num].posIC);

  }

  myfree(theta);
}

/**********************************************************************/
/*
  sampleFree

  Free memory inside n SAMPLE structs.

*/
/**********************************************************************/
void sampleFree(
		SAMPLE *sample,  /* pointer to first sample */
		long n           /* number of structs to free memory for */
		)
{

  int num;               /* index for iterating over structs */

  if (sample==NULL) return;

  for (num=0; num<n; num++){

    myfree(sample[num].name);
    myfree(sample[num].seq);
    myfree(sample[num].rcSeq);
    myfree(sample[num].numCuts);
    myfree(sample[num].logCumBack);
    myfree(sample[num].backLR);;
    myfree(sample[num].cumProbBack);
    myfree(sample[num].mProbs1);
    myfree(sample[num].rcmProbs1);
    myfree(sample[num].mProbs2);
    myfree(sample[num].rcmProbs2);
    myfree(sample[num].seqProbs);
    myfree(sample[num].rcSeqProbs);
    myfree(sample[num].cutLRs); 

  }

  myfree(sample);
}

/**********************************************************************/
/*
  conSetFree

  Free memory inside n CONSET structs.

*/
/**********************************************************************/
void conSetFree(
		CONSET *conSets,   /* pointer to first CONSET struct */
		long n            /* number of structs to allocate memory for */
		)
{
    
  int cSetNum;             /* index for iterating over structs */

  if (conSets==NULL) return;

  for (cSetNum=0; cSetNum<n;cSetNum++){
    
    intConFree(conSets[cSetNum].intCons, conSets[cSetNum].numIntFreqCons);
    linConFree(conSets[cSetNum].linCons);
  }

  myfree(conSets);

}

/**********************************************************************/
/*
  posFreqConFree

  Free memory for linked POSFREQCON structs.

*/
/**********************************************************************/
void posFreqConFree(
		    POSFREQCON * posFreqCon       /* pointer to first struct */
		    )
{

  if (posFreqCon==NULL) return;

  if (posFreqCon->next) posFreqConFree(posFreqCon->next);
  myfree(posFreqCon);

}

/**********************************************************************/
/*
  intFreqConFree

  Free memory for linked INTFREQCON structs.

*/
/**********************************************************************/
void intFreqConFree(
		    INTFREQCON * intFreqCon       /* pointer to first struct */
		    )
{

  if (intFreqCon==NULL) return;

  if (intFreqCon->next) intFreqConFree(intFreqCon->next);
  myfree(intFreqCon);

}


/**********************************************************************/
/*
  linConFree

  Free memory for linked LINCONCON structs.

*/
/**********************************************************************/
void linConFree(
		LINCON * linCon       /* pointer to first struct */
		)
{
  
  if (linCon==NULL) return;

  if (linCon->next) linConFree(linCon->next);
  myfree(linCon);

}


/**********************************************************************/
/*
  intConFree

  Free memory inside n INTCON structs.

*/
/**********************************************************************/
void intConFree(
		INTCON *intCons,  /* pointer to first struct */
		long n            /* number of structs to allocate memory for */
		)
{
    
  int intConNum;        /* index for iterating over structs */

  if (intCons==NULL) return;

  for (intConNum=0;intConNum<n;intConNum++){

    intFreqConFree(intCons[intConNum].intFreqCons);
    posFreqConFree(intCons[intConNum].posFreqCons);
  }

  myfree(intCons);
}


/**********************************************************************/
/*
  modFree

  Free memory inside n MODEL structs.

*/
/**********************************************************************/
void modFree(
	     MODEL *mod,   /* pointer to first struct */
	     long n        /* number of structs to allocate memory for */
	     )
{
    
  int modNum;          /* index for iterating over structs */

  if (mod==NULL) return;

  for (modNum = 0; modNum < n; modNum++){

    myfree(mod[modNum].linConMatrix);
    myfree(mod[modNum].low);
    myfree(mod[modNum].up);
    myfree(mod[modNum].nlCons);
  }
    
  myfree(mod);

}


/**********************************************************************/
/*
  distrFree

  Free memory inside n DISTR structs.

*/
/**********************************************************************/
void distrFree(
	       DISTR *distr,   /* pointer to first struct */
	       int first,      /* first struct that was initialized */
	       long n          /* number of structs to allocate memory for */
	       )
{
    
  int distrNum;          /* index for iterating over structs */

  if (distr==NULL) return;

  for (distrNum = first; distrNum < n; distrNum++){
    myfree(distr[distrNum].d);
    myfree(distr[distrNum].cdf);
    myfree(distr[distrNum].range);
    myfree(distr[distrNum].offset);

  }
    
  myfree(distr);

}


/**********************************************************************/
/*
  donlpFree

  Free memory allocated for donlp

*/
/**********************************************************************/
void donlpFree(
	       DATASET *dataSet
	       )
{

  long iterma;
  long nstep;
  long ndualm;
  long mdualm;

  iterma =  30 * dataSet->maxN; 
  nstep  =  20;
  
  ndualm = 2*dataSet->maxN + dataSet->maxNlin + dataSet->maxNonlin;
  mdualm = 2*(dataSet->maxN + dataSet->maxNlin + dataSet->maxNonlin);
  
  global_mem_free(dataSet->maxN, dataSet->maxNlin, dataSet->maxNonlin, iterma, ndualm, mdualm);

}

/**********************************************************************/
/*
  tmFree

  Free memory allocated for transition matrices

*/
/**********************************************************************/
void tmFree(
	    DATASET *dataSet
	    )
{

  int order;
  
  for (order=0; order<=dataSet->maxOrder; order++){
    myfree(dataSet->transMats[order]);
  }
  myfree(dataSet->transMats);

}
  
