/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains routines for initializing cosmo                  *
*                                                                            *
*****************************************************************************/
 
#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "general.h"        /* for DO_STANDARD_COMMAND_LINE */
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "init.h"
#include "io.h"
#include "matLib.h"
#include "string.h"


/**********************************************************************/
/*
  setDefaults

  Initializes default values for the dataset 
	
*/
/**********************************************************************/
int setDefaults(
		 DATASET *dataSet     /* the dataset */
		 )
{


  int i;

  /* Input sequences */
  dataSet->seqFileName       = NULL;
  dataSet->seqFile           = NULL;
  dataSet->samples           = NULL;
  dataSet->numSeqs           = 0;
  dataSet->maxSeqLength      = 0;      
  dataSet->totalNucs         = 0;
  dataSet->nucFreq           = DoubleMalloc(4);
  dataSet->maxNameLength     = 0;

  /* Basic model parameters */
  dataSet->aLength           = 4;
  dataSet->revComp           = TRUE;
  dataSet->R                 = FALSE;
  
  dataSet->minWidth          = 6;
  dataSet->maxWidth          = 15;
  dataSet->numWidths         = 10;

  dataSet->minNumSites       = 0;
  dataSet->zoopsMaxNumSites  = 0;
  dataSet->tcmMaxNumSites    = 0;
  dataSet->findBestNumSites  = TRUE;
  dataSet->maxIntensity      = FALSE;

  /* The candidate model types */
  dataSet->mTypes            = NULL;
  dataSet->numModTypes       = 1;
  dataSet->oops              = FALSE;
  dataSet->zoops             = FALSE;
  dataSet->tcm               = TRUE;
  dataSet->approx            = OVER;
  dataSet->modTypeName       = CharMalloc(10);
  dataSet->cutFactor         = 5;
  dataSet->numCuts           = NULL;
  dataSet->cutLengths        = NULL;

  dataSet->distrs            = NULL;
  dataSet->ndistrs           = -1;
  dataSet->len               = NULL;

  dataSet->IP                = NULL;
  dataSet->logP              = NULL;
  dataSet->minI              = NULL;
  dataSet->maxI              = NULL;

  /* The constraints */
  dataSet->conFileName       = NULL;
  dataSet->conFile           = NULL;
  dataSet->numConSets        = 0;
  dataSet->cSets             = NULL;
  dataSet->hasConstraints    = FALSE;
  dataSet->addFree           = FALSE;
  dataSet->maxNumProfParms   = 0;
  dataSet->maxNumInts        = 0;
  dataSet->mod               = NULL;
  dataSet->numModels         = 0;
  dataSet->effNumModels      = 0;
  dataSet->intNumBasePairs   = NULL;
  dataSet->intFirst          = NULL;  

  dataSet->intLast           = NULL; 
  dataSet->line              = CharMalloc(MAXNUMCHARS);               
  dataSet->word              = CharMalloc(MAXNUMCHARS);               
  dataSet->first             = CharMalloc(MAXNUMCHARS);               
  dataSet->maxN              = 0;
  dataSet->maxNlin           = 0;
  dataSet->maxNonlin         = 0;


  /* The background model */
  dataSet->hasBackFile       = FALSE;
  dataSet->backFileName      = NULL;
  dataSet->backFile          = NULL;
  dataSet->backSamples       = NULL;
  dataSet->numBackSeqs       = 0;
  dataSet->backProb          = DoubleMalloc(4);

  dataSet->orderByCV         = TRUE;  
  dataSet->order             = 0;
  dataSet->backFold          = 5;
  dataSet->maxOrder          = 6;
  dataSet->backKL            = DoubleMalloc(dataSet->maxOrder+1);
  for (i=0; i<=dataSet->maxOrder; i++) dataSet->backKL[i] = -1;

  dataSet->markovDesc        = CharMalloc(100);

  dataSet->readTransMat      = FALSE;
  dataSet->transMatFileName  = NULL;  
  dataSet->transMatFile      = NULL;

  dataSet->readBfile         = FALSE;
  dataSet->bfileName         = NULL;
  dataSet->bfile             = NULL;
  dataSet->fields            = pCharMalloc(2);
  dataSet->fields[0]         = CharMalloc(MAXNUMCHARS);
  dataSet->fields[1]         = CharMalloc(MAXNUMCHARS);


  dataSet->transMats         = NULL;



  /* Additional output */
  dataSet->xml                = FALSE;
  dataSet->printStatus        = FALSE;

  dataSet->output4sim         = FALSE;
  dataSet->simFileName        = NULL; 
  dataSet->simFile            = NULL;
  dataSet->elapsedTime        = 0;
  
  dataSet->outputProbs        = FALSE;
  dataSet->probFileName       = NULL; 
  dataSet->probFile           = NULL;

  dataSet->outputROC          = FALSE;
  dataSet->rocFileName        = NULL; 
  dataSet->rocFile            = NULL;


  /* Variables for finding starting values */
  dataSet->numStarts          = 5;
  dataSet->likStarts          = FALSE;
  dataSet->hiProb             = 0;
  dataSet->lowProb            = 0;
  dataSet->svSpace            = NULL;
  dataSet->cvSvSpace1         = NULL;
  dataSet->cvSvSpace2         = NULL;
  dataSet->cvStarts           = FALSE;

  dataSet->intMotProbs1       = NULL;
  dataSet->intMotProbs2       = NULL;
  dataSet->maxMotProbs        = NULL;

  /* Variables needed for cross-validation */
  dataSet->logLiks            = NULL;
  dataSet->klQuant            = 0;
  dataSet->fold               = 0;
  dataSet->refThetas          = NULL;
  dataSet->refThetaConv       = NULL;
  dataSet->nRefThetas         = 0;

  /* Various small scratch spaces */
  dataSet->pwm1               = NULL;
  dataSet->pwm2               = NULL;
  dataSet->lPwm1              = NULL;
  dataSet->lPwm2              = NULL;
  dataSet->posGrad            = DoubleMalloc(4);
  dataSet->cumProb            = NULL;
  dataSet->motifCounts        = NULL;
  dataSet->svTheta            = NULL;
  dataSet->theta              = NULL;
  dataSet->theta1             = NULL;
  dataSet->theta2             = NULL;

  /* Choice of model fine-tuning parameters */
  dataSet->crits.width        = 0;
  dataSet->crits.conSetNum    = 0;
  dataSet->crits.modTypeNum   = 0;
  dataSet->crits.mType        = 0;

  dataSet->wCrit.crit         = bic;
  dataSet->wCrit.fold         = 5;
  dataSet->wCrit.trunc        = FALSE;
  dataSet->wCrit.quant        = 1;
  dataSet->wDesc              = CharMalloc(100);

  dataSet->modCrit.crit       = lik;
  dataSet->modCrit.fold       = 5;
  dataSet->modCrit.trunc      = FALSE;
  dataSet->modCrit.quant      = 1;
  dataSet->modDesc            = CharMalloc(100);

  dataSet->conCrit.crit       = likCV;
  dataSet->conCrit.fold       = 5;
  dataSet->conCrit.trunc      = TRUE;
  dataSet->conCrit.quant      = 0.9;
  dataSet->conDesc            = CharMalloc(100);

  dataSet->intCrit.crit       = lik;
  dataSet->intCrit.fold       = 5;
  dataSet->intCrit.trunc      = FALSE;
  dataSet->intCrit.quant      = 1;


  dataSet->fxUpdated          = FALSE;

  /* Posterior probabilities */
  dataSet->postProbs          = NULL;
  dataSet->maxPostProbs       = NULL;


dataSet->version = "1.0.10 (Release date: 2/23/2007)";
 dataSet->reference          = "If you use this software in your research, please cite\n\nO. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection of Conserved\nMotifs in DNA Sequences with cosmo. Statistical Applications in Genetics and\nMolecular Biology: Vol. 6  : Iss. 1, Article 8. Available at\nhttp://www.bepress.com/sagmb/vol6/iss1/art8.";

  dataSet->donlpReference = "cosmo makes use of the donlp2() function by Peter Spellucci. The use of donlp2\nmust be acknowledged, in any publication which contains results obtained with\ncosmo or parts of it. Citation of the author's name and netlib-source is\nsuitable.";

  return(1);
}


/**********************************************************************/
/*
  readArgs

  Reads in the command line arguments to modify the default values. 

*/
/**********************************************************************/
int readArgs(
	      int argc,            /* number of command line arguments */
	      char **argv,         /* array of command line argumens */
	      DATASET *dataSet     /* the dataset */
	      )
{

  int ALLOPTS;          /* give access to all options from command line? */
  int i, j, k, l, pos;
  int len;              /* length of option */
  BOOLEAN setIntFold;   /* changed fold for intensity CV?*/
  BOOLEAN setIntTrunc;  /* changed truncation percentile for intensity CV?*/
  BOOLEAN setWFold;     /* changed fold for width CV?*/
  BOOLEAN setWTrunc;    /* changed truncation percentile for width CV? */
  BOOLEAN setModFold;   /* changed fold for model type CV? */
  BOOLEAN setModTrunc;  /* changed truncation percentile for model type CV?*/
  BOOLEAN setConFold;   /* changed fold for constraint set CV?*/
  BOOLEAN setConTrunc;  /* changed truncation percentile for constraint set CV? */
  BOOLEAN orderGiven;   /* changed order? */

  ALLOPTS      = TRUE;
  setIntFold   = FALSE;
  setIntTrunc  = FALSE;
  setWFold     = FALSE;
  setWTrunc    = FALSE;
  setModFold   = FALSE;
  setModTrunc  = FALSE;
  setConFold   = FALSE;
  setConTrunc  = FALSE;
  orderGiven   = FALSE;
  

  /* save command line */
  dataSet->command = NULL;
  argv[0] = "cosmo";
  for (i=pos=len=0; i<argc; i++)  len += strlen(argv[i])+1;			/* +1 for space following */
  dataSet->command = CharMalloc(len+2);
  for (i=pos=len=0; i<argc; i++){ 
    len += strlen(argv[i])+1;
    strcpy((dataSet->command)+pos, argv[i]);
    dataSet->command[len-1] = ' ';
    dataSet->command[len] = '\0';
    pos = len;
  }


  /* print the command line */
  i = 1;
  j = 1;
  k = 1;
  l = 1;
  argv[0] = "";
  DO_STANDARD_COMMAND_LINE(1,
    USAGE(USAGE:\n\tcosmo\t<dataset> [optional arguments]\n);
    NON_SWITCH(1, <dataset> \t\tfile containing sequences in FASTA format\n,
      switch (i++) {
        case 1: dataSet->seqFileName = _OPTION_; break;
        default: COMMAND_LINE_ERROR;
      });
     FLAG_OPTN(1, h, \t\t\tprint this message, USAGE_MESSAGE);
     DATA_OPTN(1, w, <w>, \t\tmotif width,
	       dataSet->minWidth = atoi(_OPTION_);
	       dataSet->maxWidth = atoi(_OPTION_));
     DATA_OPTN(1, minw, <minw>, \t\tminumum motif width,
       dataSet->minWidth = atoi(_OPTION_));
     DATA_OPTN(1, maxw, <maxw>, \t\tmaximum motif width,
       dataSet->maxWidth = atoi(_OPTION_));
     FLAG_OPTN(1, posonly, \t\tconsider sites only on the given strand,
       dataSet->revComp = FALSE);
     DATA_OPTN(1, minsites, <minsites>, 
       \tminimum number of sites for each motif, dataSet->minNumSites=atoi(_OPTION_));
     DATA_OPTN(1, maxsites, <maxsites>, \tmaximum number of sites for each motif,
       dataSet->zoopsMaxNumSites=atoi(_OPTION_);
       dataSet->tcmMaxNumSites=atoi(_OPTION_));
     FLAG_OPTN(1, oops, \t\t\tconsider OOPS model,
      switch (k++) {
        case 1: dataSet->tcm = FALSE; 
        default: dataSet->oops = TRUE;
      });
     FLAG_OPTN(1, zoops, \t\tconsider ZOOPS model,
      switch (k++) {
        case 1: dataSet->tcm = FALSE; 
        default: dataSet->zoops = TRUE;
      });
     FLAG_OPTN(1, tcm, \t\t\tconsider TCM model,
       dataSet->tcm = TRUE;k++);
     FLAG_OPTN(1, exact, \t\tuse full likelihood for TCM model instead of approximation by cutting model with overlaps,
	       switch (j++) {
	       case 1: dataSet->approx = TCM;break;
	       default: Rprintf( "Only one of -cut and -exact may be given.\n");return(0);
	       });
     FLAG_OPTN(1, cut, \t\t\tapproximate TCM model by cutting model without overlaps,
	       switch (j++) {
	       case 1: dataSet->approx = CUT;break;
	       default: Rprintf( "Only one of -cut and -exact may be given.\n");return(0);
	       });
     DATA_OPTN(1, cutfac, <factor>, \tfor cutting and overlap approximations sequences are cut into pieces of length <factor>*width,
       dataSet->cutFactor = atoi(_OPTION_));			   
     DATA_OPTN(1, con, <file>, \t\tfile containing definition of constraints,
       dataSet->conFileName = _OPTION_; dataSet->hasConstraints = TRUE);
     FLAG_OPTN(1, addfree, \t\tadd empty constraint set to given list ,
       dataSet->addFree = TRUE);
     DATA_OPTN(1, wcrit, <crit>, \t\tcriterion for choosing the motif width (lik, aic, bic, eval, or likCV),
 	       if (!strcmp(_OPTION_,"lik")){
		 dataSet->wCrit.crit = lik;
	       }else if (!strcmp(_OPTION_,"aic")){
		 dataSet->wCrit.crit = aic;
	       }else if (!strcmp(_OPTION_,"bic")){
		 dataSet->wCrit.crit = bic;
	       }else if (!strcmp(_OPTION_,"eval")){
		 dataSet->wCrit.crit = eval;
	       }else if (!strcmp(_OPTION_,"likCV")){
		 dataSet->wCrit.crit = likCV;
	       }else{
		 Rprintf("Option %s for wCrit is not valid. Exiting...\n", _OPTION_);
		 return(0);
	       });
     DATA_OPTN(1, wfold, <fold>, \t\tfold for CV in choosing the motif width,
       dataSet->wCrit.fold = atoi(_OPTION_); setWFold = TRUE);
     DATA_OPTN(1, wtrunc, <perc>, \ttruncation percentile for CV in choosing the motif width (0-100),
       dataSet->wCrit.quant = atof(_OPTION_)/100;
	       dataSet->wCrit.trunc = TRUE;
	       setWTrunc = TRUE);
     DATA_OPTN(1, modcrit, <crit>, \tcriterion for choosing the model type (lik, aic, bic, eval, or likCV),
	       if (!strcmp(_OPTION_,"lik")){
		 dataSet->modCrit.crit = lik;
	       }else if (!strcmp(_OPTION_,"aic")){
		 dataSet->modCrit.crit = aic;
	       }else if (!strcmp(_OPTION_,"bic")){
		 dataSet->modCrit.crit = bic;
	       }else if (!strcmp(_OPTION_,"eval")){
		 dataSet->modCrit.crit = eval;
	       }else if (strstr(_OPTION_,"likCV")){
		 dataSet->modCrit.crit = likCV;
	       }else{
		 Rprintf("Option %s for modCrit is not valid. Exiting...\n", _OPTION_);
		 return(0);
	       });
     DATA_OPTN(1, modfold, <fold>, \tfold for CV in choosing the model type,
       dataSet->modCrit.fold = atoi(_OPTION_); setModFold = TRUE);
     DATA_OPTN(1, modtrunc, <perc>, \ttruncation percentile for CV in choosing the model type (0-100),
	       dataSet->modCrit.quant = atof(_OPTION_)/100;
	       dataSet->modCrit.trunc = TRUE;
	       setModTrunc = TRUE);
     DATA_OPTN(1, concrit, <crit>, \tcriterion for choosing the constraint set (lik, eval, likCV, or pwmCV),
	       if (!strcmp(_OPTION_,"lik")){
		 dataSet->conCrit.crit = lik;
	       }else if (!strcmp(_OPTION_,"eval")){
		 dataSet->conCrit.crit = eval;
	       }else if (!strcmp(_OPTION_,"likCV")){
		 switch (l++) {
		 case 1: dataSet->conCrit.crit = likCV; dataSet->conCrit.fold = 5; break;
		 default: dataSet->conCrit.crit = likCV;
		 }
	       }else if (!strcmp(_OPTION_,"pwmCV")){
		 switch (l++) {
		 case 1: dataSet->conCrit.crit = pwmCV; dataSet->conCrit.fold = 2; break;
		 default: dataSet->conCrit.crit = pwmCV;
		 }
	       }else{
		 Rprintf("Option %s for conCrit is not valid. Exiting...\n", _OPTION_);
		 return(0);
	       });
     DATA_OPTN(1, confold, <fold>, \tfold for CV in choosing the constraint set,
	       dataSet->conCrit.fold = atoi(_OPTION_);
	       setConFold = TRUE;
	       l++);
     DATA_OPTN(1, contrunc, <perc>, \ttruncation percentile for CV in choosing the constraint set (0-100),
	       dataSet->conCrit.trunc = TRUE;
	       dataSet->conCrit.quant =  atof(_OPTION_)/100;
	       setConTrunc = TRUE);
     DATA_OPTN(1, intcrit, <crit>, \tcriterion for choosing the intensity parameter (lik, eval, or likCV), 
	       if (!strcmp(_OPTION_,"lik")){
		 dataSet->intCrit.crit = lik;
	       }else if (!strcmp(_OPTION_,"eval")){
		 dataSet->intCrit.crit = eval;
	       }else if (!strcmp(_OPTION_,"likCV")){
		 dataSet->intCrit.crit = likCV;
	       }else{
		 Rprintf("Option %s for intCrit is not valid. Exiting...\n", _OPTION_);
		 return(0);
	       });
     DATA_OPTN(1, intfold, <fold>, \tfold for CV in choosing the intensity parameter,
       dataSet->intCrit.fold = atoi(_OPTION_); setIntFold = TRUE);
     DATA_OPTN(1, inttrunc, <perc>, \ttruncation percentile for CV in choosing the intensity parameter (0-100),
               dataSet->intCrit.quant = atof(_OPTION_)/100; 
	       dataSet->intCrit.trunc = TRUE;
               setIntTrunc = TRUE);
     FLAG_OPTN(1, maxint, \t\tmaximize intensity parameter during constrained optimization,
	       dataSet->maxIntensity = TRUE); 
     DATA_OPTN(1, starts, <starts>, 
       \tnumber of starting values to use for maximization, 
	       dataSet->numStarts=atoi(_OPTION_));
     FLAG_OPTN(1, lstarts, \t\tuse likelihood-based starting values,
	       dataSet->likStarts = TRUE; 
	       dataSet->cvStarts=FALSE);
     DATA_OPTN(1, back, <file>, \t\tsequence file to estimate background model from,
       dataSet->backFileName = _OPTION_; dataSet->hasBackFile = TRUE);
     DATA_OPTN(1, backfold, <fold>, \tfold for likelihood-based cross-validation for selecting order of background Markov model,
       dataSet->backFold = atoi(_OPTION_));			   
     DATA_OPTN(1, bfile, <file>, \t\tMEME-style background file,
       dataSet->bfileName = _OPTION_; dataSet->readBfile=TRUE; dataSet->orderByCV=FALSE);
     DATA_OPTN(1, trans, <file>, \t\ttransition matrix file,
       dataSet->transMatFileName = _OPTION_; dataSet->readTransMat=TRUE; dataSet->orderByCV=FALSE);
     DATA_OPTN(1, order, <order>, \torder of background Markov model to be estimated from sequences,
       dataSet->order = atoi(_OPTION_); dataSet->orderByCV = FALSE; orderGiven=TRUE);
     FLAG_OPTN(1, xml, \t\t\tgenerate XML output rather than text output,
       dataSet->xml = TRUE);

     FLAG_OPTN(ALLOPTS, status, \t\tprint simplified status output,
	       dataSet->printStatus = TRUE);
     DATA_OPTN(ALLOPTS, simout, <file>, \toutput results for simulation purposes to <file>,
	       dataSet->output4sim = TRUE;
	       dataSet->simFileName    = _OPTION_);
     DATA_OPTN(ALLOPTS, roc, <file>, \t\toutput for purposes of calculating ROC to <file>,
	       dataSet->outputROC  = TRUE;
	       dataSet->rocFileName    = _OPTION_);
     DATA_OPTN(ALLOPTS, prob, <file>, \t\toutput posterior probabilities to <file>,
	       dataSet->outputProbs = TRUE;
	       dataSet->probFileName = _OPTION_);
     DATA_OPTN(ALLOPTS, debug, <level>, \tset debugging message level,
	       COSMO_MSG_LEVEL = atoi(_OPTION_));
  )


  /* Check selection criteria to make sure that CV options are only entered for valid CV criteria */
  if (dataSet->intCrit.crit!=likCV && setIntTrunc){
    Rprintf("\n********************************* WARNING *****************************************************\n");
    Rprintf("Truncation percentile for intensity criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("***********************************************************************************************\n\n");
  }

  if (dataSet->intCrit.crit!=likCV && setIntFold){
    Rprintf("\n********************************* WARNING ************************************\n");
    Rprintf("Fold for intensity criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("******************************************************************************\n\n");
  }

  if (dataSet->wCrit.crit!=likCV && setWTrunc){
    Rprintf("\n********************************* WARNING *******************************************************\n");
    Rprintf("Truncation percentile for motif width criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("*************************************************************************************************\n\n");
  }

  if (dataSet->wCrit.crit!=likCV && setWFold){
    Rprintf("\n********************************* WARNING *************************************\n");
    Rprintf("Fold for motif width criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("********************************************************************************\n\n");
  }
 
  if (dataSet->modCrit.crit!=likCV && setModTrunc){
    Rprintf("\n********************************* WARNING ******************************************************\n");
    Rprintf("Truncation percentile for model type criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("************************************************************************************************\n\n");
  }

  if (dataSet->modCrit.crit!=likCV && setModFold){
    Rprintf("\n********************************* WARNING *************************************\n");
    Rprintf("Fold for model type criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("*******************************************************************************\n\n");
  }


  if (dataSet->conCrit.crit!=likCV && setConTrunc){
    Rprintf("\n********************************* WARNING **********************************************************\n");
    Rprintf("Truncation percentile for constraint set criterion CV only applies to likelihood-based CV criterion.\n");
    Rprintf("****************************************************************************************************\n\n");
  }

  if (dataSet->conCrit.crit!=likCV && dataSet->conCrit.crit!=pwmCV && setConFold){
    Rprintf("\n********************************* WARNING **********************************************\n");
    Rprintf("Fold for constraint set CV only applies to likelihood-based CV or PWM-based CV criterion.\n");
    Rprintf("*****************************************************************************************\n\n");
  }

  /* order is ignored if transition matrix file or MEME-style bfile are given */
  if (orderGiven && (dataSet->readTransMat || dataSet->readBfile)){
    Rprintf("\n********************************* WARNING ********************************************************\n");
    Rprintf("Given order %d for background Markov model will be ignored since entire background model is given.\n", 
	    dataSet->order);
    Rprintf("**************************************************************************************************\n\n");
  }
  

  return(1);
}


/**********************************************************************/
/*
  processArgs

  Arguments are checked for consistency and memory is allocated 
  accordingly.

*/
/**********************************************************************/
int processArgs(
		 DATASET *dataSet     /* the dataset */
		 )
{

  int numSelType;       /* counter for setting up model types */

   
  /* Check arguments */  

  /* minWidth must be at least 1 */
  if (dataSet->minWidth<1){
    Rprintf("minWidth must be at least 1. Exiting ...\n");
    return(0);
  }

  /* maxWidth cannot be greater than 30 */
  if (dataSet->maxWidth>30){
    Rprintf("maxWidth cannot be more than 30. Exiting ...\n");
    return(0);
  }
  
  /* minWidth cannot be greater than maxWidth */
  if (dataSet->maxWidth < dataSet->minWidth){
    Rprintf("maxWidth cannot be less than minWidth. Exiting ...\n");
    return(0);
  }

  /* maximum number of sites cannot be less than minimum number of sites */
  if ((dataSet->zoopsMaxNumSites && dataSet->minNumSites) && (dataSet->zoopsMaxNumSites < dataSet->minNumSites) ){
    Rprintf("maxNumSites cannot be less than minNumSites. Exiting ...\n");
    return(0);
  }


  /* only one of transition matrix file or MEME-style bfile can be given */
  if (dataSet->readTransMat && dataSet->readBfile){
    Rprintf("Only one of the options -trans and -bfile can be given. Exiting ...\n");
    return(0);
  }

  

  /* order for background Markov model must be between 0 and 6 */
  if (dataSet->orderByCV == FALSE && (dataSet->order < 0 || dataSet->order > 6)){
    Rprintf("Order must be between 0 and 6 inclusive. Got %d. Exiting ...\n", dataSet->order);
    return(0);
  }

  /* only E-value starts can be found based on the training sample alone */
  if (dataSet->cvStarts && dataSet->likStarts){
    Rprintf( "Starting values from the training sample alone can only be found based on the E-value. Exiting...\n");
    return(0);
  }


  /* find best number of sites only for E-value based estimator */
  if(dataSet->intCrit.crit!=eval) dataSet->findBestNumSites = FALSE;

  dataSet->numWidths   = dataSet->maxWidth - dataSet->minWidth + 1;
  

  /* Set model types appropriately */
  dataSet->numModTypes = dataSet->oops + dataSet->zoops + dataSet->tcm;


  /* Approximations by CUT or OVER only make sense if TCM model is considered */
  if ((dataSet->approx!=OVER) && (!dataSet->tcm)){
    Rprintf("\n********************************* WARNING **************************************\n");
    Rprintf("processArgs: Option -exact or -cut will be ignored  since TCM is not considered.\n");
    Rprintf("********************************************************************************\n\n");
  }


  /* allocate memory that depends on the number of model types */
  dataSet->mTypes = mTypeMalloc(dataSet->numModTypes);

  numSelType = 0;
  if (dataSet->oops){
    dataSet->mTypes[numSelType] = OOPS;
    numSelType++;
  }
  if (dataSet->zoops){
    dataSet->mTypes[numSelType] = ZOOPS;
    numSelType++;
  }
  if (dataSet->tcm){
    dataSet->mTypes[numSelType] = TCM;
    numSelType++;
  }

  if (COSMO_MSG_LEVEL>=3){
    for (numSelType=0; numSelType<dataSet->numModTypes; numSelType++){
      Rprintf("Model type %d: %d\n", numSelType + 1, dataSet->mTypes[numSelType]);
    }
  }



  /* allocate memorty depending on candidate motif widths */
  dataSet->pwm1        = DoubleMalloc(4*(dataSet->maxWidth));
  dataSet->pwm2        = DoubleMalloc(4*(dataSet->maxWidth));
  dataSet->lPwm1       = IntMalloc(4*(dataSet->maxWidth));
  dataSet->lPwm2       = IntMalloc(4*(dataSet->maxWidth));
  dataSet->motifCounts = DoubleMalloc(4*(dataSet->maxWidth));
  dataSet->numCuts     = IntMalloc(dataSet->numWidths);
  dataSet->cutLengths  = IntMalloc(dataSet->numWidths);

  return(1);
}


/**********************************************************************/
/*
  cutSequences

  Creates subdivision of sequences into shorter cuts of length
  dataSet->cutFactor * width
	
*/
/**********************************************************************/
int cutSequences(
		  DATASET *dataSet     /* the dataset */
		  )
{

  int seqNum;             /* sequence number */
  int numSeqs;            /* number of sequences in dataset */
  int seqLength;          /* length of current sequence */
  int cutLength;          /* length of cut for given motif width */
  int width;              /* current motif width */
  int width0;             /* motif width - minimum motif width */


  numSeqs = dataSet->numSeqs;

  for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){    
    width0 = width - dataSet->minWidth;

    cutLength =  dataSet->cutFactor*width;


    dataSet->cutLengths[width0] = cutLength;
    dataSet->numCuts[width0] = 0;
    for (seqNum=0;seqNum<numSeqs;seqNum++){
      
      seqLength = dataSet->samples[seqNum].length;
      dataSet->samples[seqNum].numCuts[width0] = (long) ceil((double)seqLength/(double)cutLength);

      /* we have the maximum number of cuts for the smallest width */
      if (width0==0) dataSet->samples[seqNum].cutLRs  = DoubleMalloc(dataSet->samples[seqNum].numCuts[width0]);
   
      dataSet->numCuts[width0] += dataSet->samples[seqNum].numCuts[width0];

    }
  }

  if (COSMO_MSG_LEVEL>=3) Rprintf("cutSequences: Maximum number of cuts: %d\n", dataSet->numCuts[0]);

  return(1);
}


