/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains various input/output functions for cosmo         *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "eval.h"
#include "io.h"
#include "matLib.h"
#include "posterior.h"
#include "xml.h"
  

/**********************************************************************/
/*  openFiles

  Open all relevant files to make sure they exist.

  Exits if one of the files cannot be opened.

*/
/**********************************************************************/
int openFiles(
	       DATASET *dataSet   /* the dataset */
	       )
{

  if (!dataSet->R){
    if((dataSet->seqFile = fopen(dataSet->seqFileName,"r")) == NULL){
      Rprintf("main: Can't open input sequence file.  Exiting ...\n");
      return(0);
    }
  }

  if (dataSet->hasBackFile){
    if((dataSet->backFile = fopen(dataSet->backFileName,"r")) == NULL){
      Rprintf("main: Can't open background file.  Exiting ...\n");
      return(0);
    }
  }
    

  if (dataSet->hasConstraints){
    if((dataSet->conFile = fopen(dataSet->conFileName,"r")) == NULL){
      Rprintf("main: Can't open constraint file.  Exiting ...\n");
      return(0);
    }
  }

  if (dataSet->readTransMat){
    if((dataSet->transMatFile = fopen(dataSet->transMatFileName,"r")) == NULL){
      Rprintf("main: Can't open background transition matrix file.  Exiting ...\n");
      return(0);
    }
  }

  if (dataSet->readBfile){
    if((dataSet->bfile = fopen(dataSet->bfileName,"r")) == NULL){
      Rprintf("main: Can't open MEME-style background file.  Exiting ...\n");
      return(0);
    }
  }

  if (dataSet->outputProbs){
    if((dataSet->probFile = fopen(dataSet->probFileName,"w")) == NULL){
      Rprintf("main: Can't open posterior probability output file.  Exiting ...\n");
      return(0);
    }
  }
  
  if (dataSet->output4sim){
    if((dataSet->simFile = fopen(dataSet->simFileName,"w")) == NULL){
      Rprintf("main: Can't open simulation output file.  Exiting ...\n");
      return(0);
    }
  }

  if (dataSet->outputROC){
    if((dataSet->rocFile = fopen(dataSet->rocFileName,"w")) == NULL){
      Rprintf("main: Can't open ROC output file.  Exiting ...\n");
      return(0);
    }
  }
    
  return(1);
}



/**********************************************************************/
/*
  readTransMatFile

  Reads in transition matrix from file.

*/
/**********************************************************************/
int readTransMatFile(
		      DATASET *dataSet   /* the dataset */
		      )
{
  int row;              /* row in transition matrix */
  int col;              /* column in transition matrix */
  int maxOrder;         /* maximum order to consider */
  int order;            /* current order */
  int tryOrder;         /* candidate order */
  int subOrder;         /* lower order */
  FILE *transMatFile;   /* transition matrix file */
  char *line;           /* current line */
  long orderNumRows;    /* number of rows in matrix for this order */
  long numRows;         /* number of rows altogether */
  long index;           /* index into matrix */
  double *transMat;     /* transition matrix */
  
  maxOrder     = 10;
  line         = dataSet->line;
  transMatFile = dataSet->transMatFile;
  numRows      = 0;

  /* get number of rows in transition matrix file */
  while (fgets(line, MAXNUMCHARS, transMatFile)!=NULL) numRows++;

  /* find corresponding order */
  order = -1;
  orderNumRows = 0;
  for (tryOrder=0; tryOrder<=maxOrder; tryOrder++){
    orderNumRows += pow(4,tryOrder);
    if (orderNumRows==numRows) order = tryOrder;
  }

  /* number of rows doesn't correspond to any order */
  if (order==-1){
    Rprintf("readTransMatFile: Number of rows in transition matrix file (%ld) doesn't correspond to any order. Exiting...\n", numRows);
    return(0);
  }

  dataSet->order = order;
  dataSet->maxOrder = order;

  /*rewind the file*/
  fseek(transMatFile,(long)0,0);


  dataSet->transMats = p2DoubleMalloc(order+1);
  for (subOrder=0; subOrder<=order; subOrder++) dataSet->transMats[subOrder] = DoubleMalloc(pow(4,subOrder+1));


  if (COSMO_MSG_LEVEL>=2) Rprintf("readTransMat: reading transition matrix for background Markov model with order = %d\n",  order);

  for (subOrder=0; subOrder<=order; subOrder++){

    transMat = dataSet->transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      for (col=0; col<4; col++){
	index = 4*row + col;
	fscanf(transMatFile, "%lf", transMat + 4*row + col);

	/* transform to logs */
	transMat[4*row+col] = LOG(transMat[4*row+col]);
      }
    }
    if (COSMO_MSG_LEVEL>=3){
      Rprintf("transMat for order = %d\n", subOrder);
      PrintDoubleMatrix2File(stderr,transMat,4,numRows);
      Rprintf("\n");
    }
  }

  return(1);
}

/**********************************************************************/
/*
  readBfile

  Reads in transition matrix from MEME-style background file.

*/
/**********************************************************************/
int readBfile(
	      DATASET *dataSet   /* the dataset */
	      )
{
  FILE *bfile;          /* MEME-style background file */
  char *line;           /* current line */
  char **fields;        /* array of strings that hold individual fields */
  char *tuple;          /* tuple */
  double prob;          /* corresponding probability */
  int order;            /* order of Markov model */
  int subOrder;         /* order between 0 and order of Markov model */
  long index;           /* index of cell in transtion matrix */
  long line_no;         /* number of current line */
  double *transMat;     /* transition matrix */
  long numRows;         /* number of rows in transition matrix */
  long row;             /* row in transition matrix */
  int col;              /* column in transition matrix */
  double psum;          /* sum of probabilities in one row */

  bfile        = dataSet->bfile;
  line         = dataSet->line;
  fields       = dataSet->fields;
  order        = 0;

  while (fgets(line, MAXNUMCHARS, bfile)!=NULL){
    if (line[0] == '#') continue;		/* skip comments */
    Split(line, fields, 2);       		/* get tuple and prob */
    tuple = fields[0];
    order = MAX(order, strlen(tuple)-1);
  }

  dataSet->order = order;
  dataSet->maxOrder = order;
 
  /*rewind the file*/
  fseek(bfile,(long)0,0);


  dataSet->transMats = p2DoubleMalloc(order+1);
  for (subOrder=0; subOrder<=order; subOrder++) dataSet->transMats[subOrder] = DoubleMalloc(pow(4,subOrder+1));

  if (COSMO_MSG_LEVEL>=3) 
    Rprintf("readBfile: reading transition matrix for background Markov model with order = %d\n",  order);
  
  line_no = 0;
  while (fgets(line, MAXNUMCHARS, bfile)!=NULL){
    line_no++;

    if (line[0] == '#') continue;		/* skip comments */
    Split(line, fields, 2);       		/* get tuple and prob */
    tuple = fields[0];
    prob = atof(fields[1]);
    if (prob<0 || prob>1){
      Rprintf( "Illegal probability in bfile line %d: %s", line_no, line);
      return(0);
    }
    subOrder = strlen(tuple)-1;
    index = s2i(tuple);
    if (index == -1) return(0);
    dataSet->transMats[subOrder][index] = prob;
  }

  /* normalize and log transform */
  for (subOrder=0; subOrder<=order; subOrder++){

    transMat = dataSet->transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      psum = 0;
      for (col=0; col<4; col++){
	psum += transMat[4*row+col];
      }
      for (col=0; col<4; col++){
	transMat[4*row+col] = LOG(transMat[4*row+col]/psum);
      }

    }
    if (COSMO_MSG_LEVEL>=3){
      Rprintf("transMat for order = %d\n", subOrder);
      PrintDoubleMatrix2File(stderr,transMat,4,numRows);
      Rprintf("\n");
    }
  }

  return(1);
}

/***************************************************************************/
/*
	s2i

	Convert a string to an array index.

	Returns the index or -1 if the string contains illegal characters.
*/
/***************************************************************************/
int s2i(
	char *string					/* the string to index */
	)
{
  int i;
  int index;					/* array index */
  int len, fac;
  int oneNumber;
  char oneLetter;

  index = 0;
  len = strlen(string);
  fac = 1;

  /* move from last position backwards */
  for (i=len-1; i>=0; fac*=4, i--){
    oneLetter = string[i];
    if(oneLetter=='A' || oneLetter=='a'){
      oneNumber=0;
    }else if(oneLetter=='C' || oneLetter=='c'){
      oneNumber=1;
    }else if(oneLetter=='G' || oneLetter=='g'){
      oneNumber=2;
    }else if(oneLetter=='T' || oneLetter=='t'){
      oneNumber=3;
    }else{
      /* Everything else is illegal*/
      Rprintf( "The following tuple in bfile contains an illegal character: %s\n", string);
      return(-1);
    }
    index += fac*oneNumber;
  }
    
  return(index);
} 




/**********************************************************************/
/*
  readSeqFile

  Reads in sequences and sequence names from FASTA file.

*/
/**********************************************************************/
int readSeqFile(
		 DATASET *dataSet  /* the dataset */
		 )
{

  int seqNum;           /* number of current sequence */
  SAMPLE *sample;        /* current sample */

  /* check FASTA format and get number of sequences */
  if (!checkFASTA(dataSet,FALSE)) return(0);

  /* allocate memory for samples */
  dataSet->samples = sampleMalloc(dataSet->numSeqs);

  /* read sequence lengths */
  if (!getSeqLengths(dataSet)) return(0);

  /* allocate memory for sequences and names */
  if (!initSamples(dataSet)) return(0);

 
  /* read in data */
  if (!readData(dataSet)) return(0);

  /* get background frequencies */
  if (!dataSet->hasBackFile){
    if (!getBackFreqs(dataSet, dataSet->samples, dataSet->numSeqs)) return(0);
    getIrange(dataSet);

    /* Initialize the probability tables for the objective function.  */
    init_llr_pv_tables(dataSet, MIN(2, dataSet->minNumSites), MAX(dataSet->zoopsMaxNumSites,dataSet->tcmMaxNumSites), dataSet->nucFreq);
  }


 
  if (COSMO_MSG_LEVEL>=4){
    Rprintf("readSeqFile:\n");
    Rprintf("numSeqs: %d\n",dataSet->numSeqs);  
    Rprintf("maxSeqLength: %d\n", dataSet->maxSeqLength);
	
    for (seqNum=0;seqNum<dataSet->numSeqs;seqNum++){
      sample = &(dataSet->samples[seqNum]);

      Rprintf("Name for Sequence %d: ", seqNum);
      PrintCharMatrix2File(stderr,sample->name,1,sample->nameLength);

      Rprintf("Sequence %d:\n", seqNum);
      Rprintf("Length: %ld\n", sample->length);
      PrintIntMatrix2File(stderr,sample->seq,1,sample->length);

      if (dataSet->revComp){
	Rprintf("reverse complement: ");
	PrintIntMatrix2File(stderr,sample->rcSeq,1,sample->length);
      }
    }
  }

  return(1);
}

/**********************************************************************/
/*
  readBackSeqFile

  Reads in background sequences from FASTA file.

*/
/**********************************************************************/
int readBackSeqFile(
		     DATASET *dataSet  /* the dataset */
		     )
{

  int seqNum;           /* number of current sequence */
  SAMPLE *sample;        /* current sample */

  /* check FASTA format and get number of sequences */
  if (!checkFASTA(dataSet,TRUE)) return(0);

  /* allocate memory for samples */
  dataSet->backSamples = sampleMalloc(dataSet->numBackSeqs);

  /* allocate memory for sequences and names */
  if(!initBackSamples(dataSet)) return(0);

  /* read in data */
  if (!readBackData(dataSet)) return(0);

  /* get background frequencies */
  if (!getBackFreqs(dataSet, dataSet->backSamples, dataSet->numBackSeqs)) return(0);
  getIrange(dataSet);

  /* Initialize the probability tables for the objective function.  */
  init_llr_pv_tables(dataSet, MIN(2, dataSet->minNumSites), MAX(dataSet->zoopsMaxNumSites,dataSet->tcmMaxNumSites), dataSet->nucFreq);

 
  if (COSMO_MSG_LEVEL>=4){
    Rprintf("readBackSeqFile:\n");
    Rprintf("numSeqs: %d\n",dataSet->numBackSeqs);  
	
    for (seqNum=0;seqNum<dataSet->numBackSeqs;seqNum++){
      sample = &(dataSet->backSamples[seqNum]);

      Rprintf("Sequence %d:\n", seqNum);
      Rprintf("Length: %ld\n", sample->length);
      PrintIntMatrix2File(stderr,sample->seq,1,sample->length);

    }
  }

  return(1);
}

/**********************************************************************/
/*
  checkFASTA

  Check if sequences or background sequences are in FASTA format.

  Exits if this is not the case.

*/
/**********************************************************************/
int checkFASTA(
		DATASET *dataSet,   /* the dataset */
		BOOLEAN back        /* check background sequences rather than samples? */
		)
{

  char oneLetter;             /* single character */
  int moreInput;              /* boolean: is there more input to be read in? */
  int moreSequence;           /* boolean: are there more nucleotides for this sequence? */
  FILE *seqFile;              /* the sequence file */
  int *numSeqs;               /* number of sequences */
  long seqLength;             /* length of current sequence */

  seqFile = back ? dataSet->backFile : dataSet->seqFile;
  numSeqs = back ? &(dataSet->numBackSeqs) : &(dataSet->numSeqs);
  moreInput=1;
  seqLength = 0;
  oneLetter=fgetc(seqFile);
  if (COSMO_MSG_LEVEL>=3) Rprintf("%c", oneLetter);

  while (moreInput){
	
    /* Found header */
    if (oneLetter=='>'){

      if (((*numSeqs)>0) & (seqLength == 0)){
	Rprintf( "ERROR: %s sequence file is not in FASTA format. Sequence %d is of length zero.\n", 
		back ? "Background" : "Input", *numSeqs);
	return(0);
      }


      (*numSeqs)++;
      seqLength = 0;

      /* read name until we encounter first space or newline */
      while(oneLetter!='\n'){
	oneLetter=fgetc(seqFile);
	if (COSMO_MSG_LEVEL>=3) Rprintf("%c", oneLetter);
	if (oneLetter=='>'){
	  Rprintf("checkFASTA: sequence file is not in FASTA format: '>' appears twice in header. Exiting ...\n");
	  return(0);
	}
      }

    /* Found sequence ? */
    }else if (oneLetter!='\n' && oneLetter!=EOF){	 

      moreSequence=1;

      while (moreSequence){
	if((oneLetter=='\n') || (oneLetter=='\r')) moreSequence=0;
	if(oneLetter==EOF){
	  moreInput=0;
	  moreSequence=0;
	}
	if (moreSequence){
	  /*
	  if (oneLetter!='A' && oneLetter!='C' && oneLetter!='G' && oneLetter!='T' && 
	      oneLetter!='a' && oneLetter!='c' && oneLetter!='g' && oneLetter!='t'){
	    
	    Rprintf("checkFASTA: sequence file is not in FASTA format: Character %c encountered in sequence.", oneLetter);
	    Rprintf("Exiting ...\n");
	    return(0);
	  }
	  */

	  seqLength++;
	  oneLetter=fgetc(seqFile);
	  if (COSMO_MSG_LEVEL>=3) Rprintf("%c", oneLetter);
	}
      }

    }else if(oneLetter==EOF){
      moreInput=0;
    }

    oneLetter=fgetc(seqFile);

  }

  if (seqLength == 0){
    Rprintf( "ERROR: %s sequence file is not in FASTA format. Sequence %d is of length zero.\n", 
	    back ? "Background" : "Input", *numSeqs);
    return(0);
  }


  /*rewind the file*/
  fseek(seqFile,(long)0,0);


  if (COSMO_MSG_LEVEL>=3) Rprintf("numSeqs = %d\n", *numSeqs);

  return(1);
}


/**********************************************************************/
/*
  getSeqLengths

  Get sequence lengths and store them in dataSet->samples[seqNum].length

*/
/**********************************************************************/
int getSeqLengths(
		  DATASET *dataSet   /* the dataset */
		  )
{

  long seqLength;        /* length of current sequence */
  int seqNum;            /* number of current sequence */
  long nameLength;       /* length of current sequence's name */
  char oneLetter;        /* single character */
  int moreInput;         /* boolean: more input to be read in? */
  int moreSequence;      /* boolean: more nucleotides to be read in? */
  int moreName;          /* boolean: more letters for sequence name to be read in? */
  int minWidth;          /* minimum width */
  int maxWidth;          /* maximum width */
  FILE *seqFile;         /* sequence file */
    
  seqFile   = dataSet->seqFile;
  minWidth  = dataSet->minWidth;
  maxWidth  = dataSet->maxWidth;
  seqNum    = -1;
  seqLength = 0;
  moreInput = 1;
  oneLetter =fgetc(seqFile);

  while (moreInput){

    /* found header */
    if (oneLetter=='>'){
      seqNum++;
      moreName = 1;
      nameLength = 0;
      seqLength = 0;

      if (COSMO_MSG_LEVEL>=4) Rprintf("readSeqFile: Getting Sequence Name\n");

      while(oneLetter!='\n'){
	oneLetter=fgetc(seqFile);
	if ((oneLetter==' ') || (oneLetter=='\r') || (oneLetter=='\t') || (oneLetter=='\n')){
	  moreName = 0;
	  dataSet->samples[seqNum].nameLength  = nameLength;
	  dataSet->maxNameLength               = MAX(dataSet->maxNameLength, nameLength);
	}
	if (moreName==1){
	  nameLength++;
	}
      }

      /* found sequence? */
    }else if (oneLetter!='\n' && oneLetter!=EOF){
      moreSequence=1;

      while (moreSequence){
	if((oneLetter=='\n') || (oneLetter=='\r') || (oneLetter==EOF)) moreSequence=0;
	if(oneLetter==EOF) moreInput = 0;
	
	/* convert nucleotides to numbers */
	if (moreSequence){
	  seqLength++;
	  oneLetter=fgetc(seqFile);
	}
      }
      dataSet->samples[seqNum].length = seqLength;
      dataSet->maxSeqLength           = MAX(dataSet->maxSeqLength, seqLength);
      
      /* make sure that sequence length is >= max motif width specified */
      if (seqLength < dataSet->maxWidth){
	Rprintf("Sequence %d has length %ld, which is less than the maximum motif width %d. Exiting...\n", 
		seqNum+1, seqLength, dataSet->maxWidth);
	return(0);
      }


    }else if(oneLetter==EOF){
      moreInput=0;
    }
    oneLetter=fgetc(seqFile);
  }

  /*rewind the file*/
  fseek(seqFile,(long)0,0);

  return(1);
}

/**********************************************************************/
/*
  initSamples

  Initialize samples and allocate memory for their components 

*/
/**********************************************************************/
int initSamples(
		 DATASET *dataSet   /* the dataset */
 		 )
{

  long seqLength;        /* length of current sequence */
  int seqNum;            /* number of current sequence */
  long nameLength;       /* length of current sequence's name */
  int minWidth;          /* minimum width */
  int maxWidth;          /* maximum width */

  minWidth  = dataSet->minWidth;
  maxWidth  = dataSet->maxWidth;

  if (dataSet->numSeqs == 1){

    dataSet->orderByCV = FALSE;
    dataSet->order = 0;
    
    if (dataSet->oops || dataSet->zoops || (!dataSet->tcm)){
      Rprintf( "WARNING: Since the dataset contains only one sequence, only the TCM model will be considered\n");
      dataSet->oops = FALSE;
      dataSet->zoops = FALSE;
      dataSet->tcm = TRUE;

      dataSet->numModTypes = 1;
      dataSet->mTypes[0] = TCM;

    }

    if (dataSet->intCrit.crit==likCV || dataSet->intCrit.crit==pwmCV){
      Rprintf( "WARNING: Since the dataset contains only one sequence, the intensity parameter cannot be chosen by cross-validation.\n");
      Rprintf( "         Using the profile-likelihood criterion instead.\n");
      dataSet->intCrit.crit = lik;
    }


    if (dataSet->wCrit.crit==likCV || dataSet->wCrit.crit==pwmCV){
      Rprintf( "WARNING: Since the dataset contains only one sequence, the motif width cannot be chosen by cross-validation.\n");
      Rprintf( "         Using BIC instead.\n");

      dataSet->wCrit.crit = bic;
    }

    if (dataSet->modCrit.crit==likCV || dataSet->modCrit.crit==pwmCV){
      Rprintf( "WARNING: Since the dataset contains only one sequence, the model type cannot be chosen by cross-validation.\n");
      Rprintf( "         Using the likelihood criterion instead.\n");
      dataSet->modCrit.crit = lik;
    }

    if (dataSet->conCrit.crit==likCV || dataSet->conCrit.crit==pwmCV){
      Rprintf( "WARNING: Since the dataset contains only one sequence, the constraint set cannot be chosen by cross-validation.\n");
      Rprintf( "         Using the likelihood criterion instead.\n");
      dataSet->conCrit.crit = lik;
    }



  }

  /* set minNumSites and maxNumSites if necessary */
  if (!dataSet->minNumSites) dataSet->minNumSites = 2;
  if (!dataSet->zoopsMaxNumSites){
    dataSet->zoopsMaxNumSites = dataSet->numSeqs;
  }else{
    if ((dataSet->zoops) & (dataSet->zoopsMaxNumSites>dataSet->numSeqs)){
      Rprintf( "WARNING: For the ZOOPS model, maxNumSites will be set to the number of sequences, %d\n",
	      dataSet->numSeqs);
      dataSet->zoopsMaxNumSites = dataSet->numSeqs;
      if (!dataSet->tcm){
	Rprintf( "WARNING: Since the TCM model is not used, maxNumSites for the TCM model will also be set to the number of sequences, %d\n",dataSet->numSeqs);
	dataSet->tcmMaxNumSites = dataSet->numSeqs;
      }
    }
  }
  if (!dataSet->tcmMaxNumSites) dataSet->tcmMaxNumSites = MIN(50, 5*dataSet->numSeqs);

  if (dataSet->zoops && (dataSet->zoopsMaxNumSites < dataSet->minNumSites)){
    Rprintf("zoopsMaxNumSites cannot be less than minNumSites. Exiting ...\n");
    return(0);
  }

  if (dataSet->tcmMaxNumSites < dataSet->minNumSites){
    Rprintf("tcmMaxNumSites cannot be less than minNumSites. Exiting ...\n");
    return(0);
  }


  /* set Folds */
  dataSet->backFold     = MIN(dataSet->backFold, dataSet->numSeqs);
  dataSet->intCrit.fold = MIN(dataSet->intCrit.fold, dataSet->numSeqs);
  dataSet->wCrit.fold   = MIN(dataSet->wCrit.fold, dataSet->numSeqs);
  dataSet->modCrit.fold = MIN(dataSet->modCrit.fold, dataSet->numSeqs);
  dataSet->conCrit.fold = MIN(dataSet->conCrit.fold, dataSet->numSeqs);
  if (dataSet->conCrit.crit==pwmCV)   dataSet->conCrit.fold = 2;

 
  /* allocate memory for cumProb */
  dataSet->cumProb   = DoubleMalloc(dataSet->maxSeqLength+1);
  
  /* allocate memory for postProbs and maxPostProbs */
  dataSet->postProbs    = postProbMalloc(dataSet->numSeqs*dataSet->maxSeqLength);
  dataSet->maxPostProbs = postProbMalloc(dataSet->numSeqs*dataSet->maxSeqLength);

  /* allocate memory for intMotProbs */
  dataSet->intMotProbs1 = intMotProbMalloc((dataSet->numSeqs)*(dataSet->maxSeqLength)*(1 + dataSet->revComp));
  dataSet->intMotProbs2 = intMotProbMalloc((dataSet->numSeqs)*(dataSet->maxSeqLength)*(1 + dataSet->revComp));
  dataSet->maxMotProbs  = intMotProbMalloc((dataSet->numSeqs)*(dataSet->maxSeqLength)*(1 + dataSet->revComp));

  /* allocate memory for logLiks */
  dataSet->logLiks      = DoubleMalloc(dataSet->numSeqs);

  /* allocate memory for len */
  dataSet->len          = IntMalloc(dataSet->numSeqs);

  /* allocate memory for scratch spaces */
  dataSet->IP     = IntMalloc(4*(MAX(dataSet->tcmMaxNumSites, dataSet->zoopsMaxNumSites)+1));
  dataSet->minI   = IntMalloc(MAX(dataSet->tcmMaxNumSites, dataSet->zoopsMaxNumSites)+1);
  dataSet->maxI   = IntMalloc(MAX(dataSet->tcmMaxNumSites, dataSet->zoopsMaxNumSites)+1);
  dataSet->logP   = DoubleMalloc(4*(MAX(dataSet->tcmMaxNumSites, dataSet->zoopsMaxNumSites)+1));


  /* allocate memory for various arrays in samples */
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    seqLength           = dataSet->samples[seqNum].length;
    nameLength          = dataSet->samples[seqNum].nameLength;
    dataSet->totalNucs += seqLength;

    if (COSMO_MSG_LEVEL>=3){
      Rprintf("initSamples: seq %d seqLength = %ld\n", seqNum, seqLength);
    }

    dataSet->samples[seqNum].name           = CharMalloc(nameLength+1);

    dataSet->samples[seqNum].numCuts        = IntMalloc(maxWidth-minWidth+1);
    dataSet->samples[seqNum].seq            = IntMalloc(seqLength);
    dataSet->samples[seqNum].rcSeq          = IntMalloc(seqLength);

    dataSet->samples[seqNum].logCumBack     = DoubleMalloc(seqLength+1);
    dataSet->samples[seqNum].backLR         = DoubleMalloc(seqLength*(maxWidth-minWidth+1));
    dataSet->samples[seqNum].cumProbBack    = DoubleMalloc(seqLength*(maxWidth-minWidth+1));


    dataSet->samples[seqNum].mProbs1        = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].rcmProbs1      = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].mProbs2        = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].rcmProbs2      = DoubleMalloc(seqLength);


    dataSet->samples[seqNum].seqProbs       = DoubleMalloc(seqLength);
    dataSet->samples[seqNum].rcSeqProbs     = DoubleMalloc(seqLength);

  }



  return(1);
}


/**********************************************************************/
/*
  initBackSamples

  Initialize samples to estimate background distribution from and 
  allocate memory for their components. 

*/
/**********************************************************************/
int initBackSamples(
		     DATASET *dataSet  /* the dataset */
		     )
{

  long seqLength;        /* length of current sequence */
  int seqNum;            /* number of current sequence */
  long nameLength;       /* length of current sequence's name */
  char oneLetter;        /* single character */
  int moreInput;         /* boolean: more input to be read in? */
  int moreSequence;      /* boolean: more nucleotides to be read in? */
  int moreName;          /* boolean: more letters for sequence name to be read in? */
  FILE *seqFile;         /* background sequence file */
    
  seqFile   = dataSet->backFile;
  seqNum    = -1;
  seqLength = 0;
  moreInput = 1;
  oneLetter =fgetc(seqFile);

  while (moreInput){

    /* found header */
    if (oneLetter=='>'){
      seqNum++;
      moreName = 1;
      nameLength = 0;
      seqLength = 0;

      if (COSMO_MSG_LEVEL>=4) Rprintf("readSeqFile: Getting Sequence Name\n");

      while(oneLetter!='\n'){
	oneLetter=fgetc(seqFile);
      }

      /* found sequence? */
    }else if (oneLetter!='\n' && oneLetter!=EOF){
      moreSequence=1;

      while (moreSequence){
	if((oneLetter=='\n') || (oneLetter=='\r') || (oneLetter==EOF)) moreSequence=0;
	if(oneLetter==EOF) moreInput = 0;
	
	/* convert nucleotides to numbers */
	if (moreSequence){
	  seqLength++;
	  oneLetter=fgetc(seqFile);
	}
      }
      dataSet->backSamples[seqNum].length = seqLength;

    }else if(oneLetter==EOF){
      moreInput=0;
    }
    oneLetter=fgetc(seqFile);
  }


  /* allocate memory for various arrays in samples */
  for (seqNum=0; seqNum<dataSet->numBackSeqs; seqNum++){
    seqLength  = dataSet->backSamples[seqNum].length;

    if (COSMO_MSG_LEVEL>=4){
      Rprintf("initBackSamples: seq %d seqLength = %ld\n", seqNum, seqLength);
    }

    dataSet->backSamples[seqNum].seq            = IntMalloc(seqLength);
  }

  /*rewind the file*/
  fseek(seqFile,(long)0,0);

  return(1);
}

/**********************************************************************/
/*
  readData

  Read in sequence samples and also create reverse complement if
  desired.

*/
/**********************************************************************/
int readData(
	      DATASET *dataSet  /* the dataset */
	      )
{
  long seqPos;           /* position in current sequence */
  long seqNum;           /* number of current sequence */
  long namePos;          /* position in current sequence's name */
  long pos;              /* position in sequence */
  char oneLetter;        /* single character */
  int moreInput;         /* boolean: more input to be read in? */
  int moreSequence;      /* boolean: more nucleotides to be read in? */
  int moreName;          /* boolean: more letters for sequence name to be read in? */
  int oneNumber;         /* single number */
  SAMPLE *sample;        /* current sample */
  FILE *seqFile;         /* sequence file */

  seqFile   = dataSet->seqFile;
  seqNum    = -1;
  moreInput = 1;
  seqPos    = 0;
  oneNumber = 0;
  oneLetter = fgetc(seqFile);
    

  while (moreInput){

    /* found header */
    if (oneLetter=='>'){
      seqNum++;
      moreName   = 1;
      namePos    = 0;
      seqPos     = 0;


      if (COSMO_MSG_LEVEL>=4) Rprintf("readSeqFile: Getting Sequence Name\n");
      while(oneLetter!='\n'){
	oneLetter=fgetc(seqFile);
	if  (namePos<MSN){
	  if ((oneLetter=='\n') || (oneLetter=='\t') || (oneLetter=='\r') || (oneLetter==' ')) moreName = 0;

	  if (moreName==1){
	    dataSet->samples[seqNum].name[namePos] = oneLetter;
	    namePos++;
	  }
	}
      }

      /* found sequence? */
    }else if (oneLetter!='\n' && oneLetter!=EOF){
      moreSequence = 1;

      while (moreSequence){
	if((oneLetter=='\n')|| (oneLetter=='\r') || (oneLetter==EOF)) moreSequence=0;
	if(oneLetter==EOF) moreInput = 0;
	
	/* convert nucleotides to numbers */
	if (moreSequence){
	  if(oneLetter=='A' || oneLetter=='a'){
	    oneNumber=0;
	  }else if(oneLetter=='C' || oneLetter=='c'){
	    oneNumber=1;
	  }else if(oneLetter=='G' || oneLetter=='g'){
	    oneNumber=2;
	  }else if(oneLetter=='T' || oneLetter=='t'){
	    oneNumber=3;
	  }else{
	    /* Everything else is treated as unknown 'X' */
	    if (COSMO_MSG_LEVEL>=3) Rprintf( "Got letter %c\n",oneLetter);
	    oneNumber=4;
	  }

		
	  dataSet->samples[seqNum].seq[seqPos] = oneNumber;
	  seqPos++;
	  oneLetter=fgetc(seqFile);
	}
      }
    }else if(oneLetter==EOF){
      moreInput=0;
    }
    oneLetter=fgetc(seqFile);
  }

  /* get reverse complement if necessary */
  if (dataSet->revComp){
    for (seqNum=0;seqNum<dataSet->numSeqs;seqNum++){
      sample = &(dataSet->samples[seqNum]);

      for (pos=0;pos<sample->length;pos++){
	if (sample->seq[sample->length - pos - 1] < 4){
	  sample->rcSeq[pos] = 3 - sample->seq[sample->length - pos - 1];
	}else{
	  sample->rcSeq[pos] = 4;
	}
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  readBackData

  Read in sequence samples from which to estimate the backgrounde
  distribution.

*/
/**********************************************************************/
int readBackData(
		  DATASET *dataSet  /* the dataset */
		  )
{
  long seqPos;           /* position in current sequence */
  long seqNum;           /* number of current sequence */
  long namePos;          /* position in current sequence's name */
  char oneLetter;        /* single character */
  int moreInput;         /* boolean: more input to be read in? */
  int moreSequence;      /* boolean: more nucleotides to be read in? */
  int moreName;          /* boolean: more letters for sequence name to be read in? */
  int oneNumber = 0;     /* single number */
  FILE *seqFile;         /* background sequence file */

  seqFile   = dataSet->backFile;
  seqNum    = -1;
  moreInput = 1;
  seqPos    = 0;
  oneLetter = fgetc(seqFile);
    

  while (moreInput){

    /* found header */
    if (oneLetter=='>'){
      seqNum++;
      moreName   = 1;
      namePos    = 0;
      seqPos     = 0;


      if (COSMO_MSG_LEVEL>=4) Rprintf("readSeqFile: Getting Sequence Name\n");
      while(oneLetter!='\n'){
	oneLetter=fgetc(seqFile);
      }

      /* found sequence? */
    }else if (oneLetter!='\n' && oneLetter!=EOF){
      moreSequence = 1;

      while (moreSequence){
	if((oneLetter=='\n') || (oneLetter=='\r') || (oneLetter==EOF)) moreSequence=0;
	if(oneLetter==EOF) moreInput = 0;
	
	/* convert nucleotides to numbers */
	if (moreSequence){
	  if(oneLetter=='A' || oneLetter=='a'){
	    oneNumber=0;
	  }else if(oneLetter=='C' || oneLetter=='c'){
	    oneNumber=1;
	  }else if(oneLetter=='G' || oneLetter=='g'){
	    oneNumber=2;
	  }else if(oneLetter=='T' || oneLetter=='t'){
	    oneNumber=3;
	  }else{
	    /* everything else is treated as unknown 'X' */
	    oneNumber=4;
	  }

		
	  dataSet->backSamples[seqNum].seq[seqPos] = oneNumber;
	  seqPos++;
	  oneLetter=fgetc(seqFile);
	}
      }

    }else if(oneLetter==EOF){
      moreInput=0;
    }
    oneLetter=fgetc(seqFile);
  }

  return(1);
}


/**********************************************************************/
/*
  getBackFreqs

  Get empirical proportions of nucleotide frequencies among samples.
  
  These are assigned to dataSet->nucFreq.

*/
/**********************************************************************/
int getBackFreqs(
		  DATASET *dataSet,   /* the dataset */
		  SAMPLE *samples,    /* samples to estimate backfreqs from */
                  int numSeqs         /* number of samples */
		  )
{

  int seqNum;           /* sequence number */
  int pos;              /* position in sequence */
  int nuc;              /* nucleotide */
  long seqLength;       /* sequence length */
  int *seq;             /* current sequence */
  double *nucFreq;      /* nucleotide frequencies */
  double count;         /* total number of non-X characters */
			
  nucFreq = dataSet->nucFreq;

  for (nuc=0; nuc<4; nuc++) nucFreq[nuc] = 0;

  count = 0;
  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seqLength = samples[seqNum].length;
    seq       = samples[seqNum].seq;
      
    for (pos=0; pos<seqLength;pos++){
      if (seq[pos]<4){
	nucFreq[seq[pos]]++;
	count++;
      }
    }
  }

  for (nuc=0; nuc<4; nuc++) nucFreq[nuc] = ((double)nucFreq[nuc])/count;

  if (COSMO_MSG_LEVEL>=4){
    for (nuc=0; nuc<4; nuc++) Rprintf("freq = %lf\n", nucFreq[nuc]);
  }

  return(1);
}


/**********************************************************************/
/*
  getIrange

  Find the quantity Irange.

  Returns Irange.

*/
/**********************************************************************/
void getIrange(
	       DATASET *dataSet
	       )
{

  double *nucFreq;      /* nucleotide frequencies */
  int *IP;		/* I'_i[n] */
  int *minI;		/* minimum intermediate value of I */
  int *maxI;	        /* maximum intermediate value of I */
  double *logP; 	/* log P_i[n] */
  int maxN;             /* maximum number of sites */
  int N;                /* number of sites */
  double alpha;
  double min_dd;
  double max, min;
  long Irange;
  int i, k, n;
  int A;

  if (COSMO_MSG_LEVEL>=3) Rprintf( "Finding Irange...");
  nucFreq = dataSet->nucFreq;


  IP   = dataSet->IP;
  logP = dataSet->logP;
  minI = dataSet->minI; 
  maxI = dataSet->maxI;
  maxN = MAX(dataSet->zoopsMaxNumSites, dataSet->tcmMaxNumSites);
  A    = 4;

  Irange = 0;
  for (N=MIN(2,dataSet->minNumSites); N<=maxN; N++){

    /* get estimates of minimum and miximum values of llr */
    for (i=0, min_dd=1; i<A; i++) min_dd = MIN(min_dd, nucFreq[i]);
    max = NINT(-N * log(min_dd));
    for (i=min=0; i<A; i++) min += nucFreq[i]*N*(log(nucFreq[i]) - log(nucFreq[i]));
    min = NINT(min);
    if (COSMO_MSG_LEVEL>=3) Rprintf("min = %f max = %f\n", min, max);
    
    /* set alpha to achieve the desired range */
    alpha = LLR_RANGE/((max-min));

    /* compute I', P, minI and maxI */ 
    for (n=0; n<=N; n++) minI[n] = maxI[n] = 0;
    for (i=0; i<A; i++) {				/* index over alphabet */
      double logdd = LOG(nucFreq[i]);			/* log(nucFreq[i]) */
      IP[II(i,0)] = 0; 				/* I'_i(0) */
      logP[LL(i,0)] = 0;				/* log P_i(0) */
      for (n=1; n<=N; n++) {			/* index over samples */
	IP[II(i,n)] = NINT(alpha*n*log(n/(N*nucFreq[i]))); 	   /* I'_i(n) */
	logP[LL(i,n)] = logP[LL(i,n-1)] + logdd - log(n);	   /* log P_i(n) */
	for (k=1; k<=n; k++) {			/* index over samples of new */
	  minI[n] = MIN(minI[n], minI[n-k] + IP[II(i,k)]);
	  maxI[n] = MAX(maxI[n], maxI[n-k] + IP[II(i,k)]);
	}
      }
    }
    
    /* get overall minI and maxI */
    for (n=1; n<=N; n++) {
      /*Rprintf("minI[%d] %d maxI[%d] %d\n", n, minI[n], n, maxI[n]);*/
      minI[0] = MIN(minI[0], minI[n]);		/* min for intermediates */
      maxI[0] = MAX(maxI[0], maxI[n]);		/* max for intermediates */
      minI[n] = LOGZEROI;
      maxI[n] = 0;
    }

    if (Irange < maxI[0] - minI[0] + 2){
      Irange = maxI[0] - minI[0] + 2;
    }
  }

  /* create script_P arrays with enough space for intermediate calculations */
  /*dataSet->logSP = DoubleMalloc((N+1)*(Irange+1));*/
  dataSet->Irange = Irange;
  dataSet->logSP  = DoubleMalloc((MAX(dataSet->tcmMaxNumSites, dataSet->zoopsMaxNumSites)+1)*(Irange+1));

  if (COSMO_MSG_LEVEL>=3) Rprintf( "Irange = %ld\n", Irange);

}

/**********************************************************************/
/*
  output

  Print results to screen.

*/
/**********************************************************************/
int output(
	    DATASET *dataSet   /* the dataset */
	    )
{
  int i;                     /* index */
  char *stars;               /* line of stars */
  char *dashes;              /* line of dashes */
  int pageWidth;             /* page width */


  /* generate xml output? */
  if (dataSet->xml) return(xmlOutput(dataSet));
  

  pageWidth     = 80; 
  stars         = CharMalloc(pageWidth+1);
  dashes        = CharMalloc(pageWidth+1);
  
  for (i=0;i<pageWidth;i++) stars[i]='*';
  stars[pageWidth] = '\0';

  for (i=0;i<pageWidth;i++) dashes[i]='-';
  dashes[pageWidth] = '\0';
  
  /* print information about cosmo */
  if (!printCosmo(dataSet,stars)) return(0);

  /* print reference information */
  if (!printReference(dataSet,stars)) return(0);
  
  /* print summary of command line arguments */
  if (!printCommandSummary(dataSet,stars)) return(0);

  /* print summary of training set */
  if (!printTrainingSet(dataSet,stars)) return(0);

  /* print original constraint file */
  if (!printConstraintFile(dataSet,stars)) return(0);

  /* print interpreted constraints */
  if (!printConstraints(dataSet,stars)) return(0);

  /* print description of background Markov model */
  if (!printTransMat(dataSet, stars)) return(0);

  /* print summary of selected fine-tuning parameters */
  if (!printModelSelection(dataSet, stars)) return(0);

  /* print description of motif */
  if (!printMotif(dataSet, stars)) return(0);

  myfree(stars);
  myfree(dashes);

  return(1);
}


/**********************************************************************/
/*
  printConstraintFile

  Print the original constraint file.

*/
/**********************************************************************/
int printConstraintFile(
			 DATASET *dataSet,   /* the datset */
			 char *stars         /* one line of stars */
			 )
{

  FILE *conFile;
  char *line;

  conFile = dataSet->conFile;
  line    = dataSet->line;

  Rprintf("\n%s\nOriginal constraint file\n%s\n", stars,stars);  

  if (dataSet->hasConstraints){
    
    /* rewind file */
    fseek(conFile,(long) 0,0);
    while (fgets(line,MAXNUMCHARS,conFile)!=NULL){
      Rprintf("%s", line);
    }


  }else{
    Rprintf("No constraints were given.\n");
  }

  Rprintf("\n");

  return(1);
}

/**********************************************************************/
/*
  printConstraints

  Print the interpreted constraints

*/
/**********************************************************************/
int printConstraints(
		      DATASET *dataSet,   /* the datset */
		      char *stars         /* one line of stars */
		      )
{
  int cSetNum;             /* index of current constraint set */
  int intNum;              /* current interval */
  int i;                   /* index */
  CONSET *cSet;            /* current constraint set */
  INTFREQCON *intFreqCon;  /* interval frequency constraint */
  POSFREQCON *posFreqCon;  /* position frequency constraint */
  LINCON *linCon;          /* linear constraint */

  Rprintf("\n%s\nInterpreted constraints\n%s\n", stars,stars);  

  cSetNum = dataSet->hasConstraints ? 1 : 0;

  for (;cSetNum < dataSet->numConSets; cSetNum++){
    cSet = dataSet->cSets + cSetNum;
    
    Rprintf("@ Constraint set %d\n\n", (dataSet->hasConstraints ? cSetNum : cSetNum+1));

    /* Interval setup */
    for (intNum=0; intNum < cSet->numInts; intNum++){
      Rprintf("Length: ");
      switch(cSet->intCons[intNum].lenType){
      case bp:
	Rprintf("%d bp\n", cSet->intCons[intNum].numBasePairs);
	break;
      case prop:
	Rprintf("%.3lf\n", 100*cSet->intCons[intNum].prop);
	break;
      case var:
	Rprintf("variable\n");
	break;
      default:
	Rprintf( "printConstraints: Length type %d not recognized. Exiting...\n", cSet->intCons[intNum].lenType);
	return(0);
      }
    }
    Rprintf("\n");

    /* Interval constraints */
    
    for (intNum=0; intNum < cSet->numInts; intNum++){

      if (cSet->intCons[intNum].hasBoundCon){
	Rprintf(">ICBounds\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Bounds: %.3lf to %.3lf\n\n", cSet->intCons[intNum].boundCon.low, cSet->intCons[intNum].boundCon.up);
      }

      if (cSet->intCons[intNum].hasShapeCon){
	Rprintf(">ICShape\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Shape: %s\n", (cSet->intCons[intNum].shapeCon.shape == lin ? "Linear" : (cSet->intCons[intNum].shapeCon.shape == monInc ? "MonotoneIncreasing" : "Monotone Decreasing"))); 
	Rprintf("LeftBounds: %.3lf to %.3lf\n", cSet->intCons[intNum].shapeCon.startLow, cSet->intCons[intNum].shapeCon.startUp);
	Rprintf("RightBounds: %.3lf to %.3lf\n", cSet->intCons[intNum].shapeCon.endLow, cSet->intCons[intNum].shapeCon.endUp);
	Rprintf("ErrorTol: %.3lf\n\n", cSet->intCons[intNum].shapeCon.error);
      }
      
      intFreqCon = cSet->intCons[intNum].intFreqCons;
      while (intFreqCon!=NULL){
	Rprintf(">NucFreq\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Pos: all\n");
	Rprintf("Nuc: %s\n", Convert2String(intFreqCon->nuc));
	Rprintf("LowerBound: %.3lf\n\n", intFreqCon->low);
	intFreqCon = intFreqCon->next;
      }

      posFreqCon = cSet->intCons[intNum].posFreqCons;
      while (posFreqCon!=NULL){
	Rprintf(">NucFreq\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Pos: %d\n", posFreqCon->pos);
	Rprintf("Nuc: %s\n", Convert2String(posFreqCon->nuc));
	Rprintf("LowerBound: %.3lf\n\n", posFreqCon->low);
	posFreqCon = posFreqCon->next;
      }
    }

    if (cSet->hasPalCon){
      Rprintf(">Palindrome\n");
      Rprintf("Intervals: %d and %d\n", cSet->palCon.int1+1, cSet->palCon.int2+1);
      Rprintf("ErrorTol: %.3lf\n\n", cSet->palCon.error);
    }

    if (cSet->hasSubCon){
      Rprintf(">SubMotif:\n");
      Rprintf("Motif: ");
      for (i=0; i< cSet->subCon.length;i++) Rprintf("%c",Convert2Char(cSet->subCon.motif[i]));
      Rprintf("\nMinFreq: %.3lf\n\n", cSet->subCon.cutOff);
    }

    linCon = cSet->linCons;
    while (linCon!=NULL){
      Rprintf(">ParmDiff\n");
      Rprintf("Parameters: %d%c - %d%c\n", linCon->int1 +1, (linCon->pos1==start ? 'a' : 'b'), linCon->int2 + 1,  
	     (linCon->pos2==start ? 'a' : 'b'));
      Rprintf("Bounds: %.3lf to %.3lf\n\n", linCon->low, linCon->up);
      linCon = linCon->next;
    }

  }
  Rprintf("\n");

  return(1);

}

/**********************************************************************/
/*
  printMotif

  Print description of the motif that was found

*/
/**********************************************************************/
int printMotif(
		DATASET *dataSet,   /* the datset */
		char *stars         /* one line of stars */
		)
{

  int width;                 /* motif width */
  MTYPE mType;               /* model type */
  int conSetNum;             /* constraint number */
  THETA *theta;              /* final parameter estimate */
  int maxNameLength;         /* maximum sequence name length */
  int pos;                   /* position in motif */
  int nuc;                   /* nucleotide */
  int seqNum;                /* sequence number */
  int numSeqs;               /* number of sequences */
  long seqLength;            /* sequence length */
  int *seq;                  /* sequence */
  int *rcSeq;                /* reverse complement of sequence */
  double prob;               /* posterior probability */
  char *name;                /* sequence name */
  int nameLen;               /* length of that name */
  int i;                     /* index */
  int start;                 /* start of motif */
  int leftOff;               /* number of characters to include on the left */
  int rightOff;              /* number of characters to include on the right */
  int sideLen;               /* number of characters to print on sides of motif */
  char oneChar;              /* a single character */
  long maxSeqLength;         /* maximum sequence length */
  int siteNum;               /* index of current site */
  int numSites;              /* number of sites */
  POSTPROB *postProbs;       /* posterior probabilities */
  POSTPROB *maxPostProbs;    /* maximum posterior probabilities */
  BOOLEAN onRevComp;         /* is the site on the reverse complement? */
  BOOLEAN global;            /* sort posterior probabilities globally? */


  sideLen       = 10;
  numSeqs       = dataSet->numSeqs;
  maxNameLength = dataSet->maxNameLength;
  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;
  maxSeqLength  = dataSet->maxSeqLength;
  width         = dataSet->crits.width;
  mType         = dataSet->crits.mType;
  conSetNum     = dataSet->crits.conSetNum;
  theta         = dataSet->theta;
  numSites      = theta->numSites;
  global        = theta->mType==TCM;



  /* get posterior probabilities */
  getPostProbs(theta, dataSet);


  /* sort posterior probabilities */
  sortPostProbs(dataSet, width, global);

  /* print position weight matrix */
  Rprintf("\n%s\nEstimated position weight matrix\n%s\n\n", stars,stars);

  /* print header */
  Rprintf("%7s", "Nuc\\Pos");
  for (pos=0; pos<width;pos++) Rprintf("%6d", pos+1);
  Rprintf("\n");
  for(nuc=0;nuc<4;nuc++){
    Rprintf("   %c    ", Convert2Char(nuc));
    for(pos=0;pos<width;pos++){
      Rprintf("%.3lf ", theta->pwm[PWM(pos,nuc)]);
    }
    Rprintf("\n");
  }
  
  Rprintf("\n%s\nInformation Content Profile\n%s\n\n", stars, stars);
  Rprintf("%7s", "    Pos");
  for (pos=0; pos<width;pos++) Rprintf("%6d", pos+1);
  Rprintf("\n IC     ");
  for(pos=0;pos<width;pos++) Rprintf("%.3lf ", theta->posIC[pos]);
  Rprintf("\n");
  
  
  /*Extract the actual sequences*/
  Rprintf("\n%s\nAlignment of discovered sites (E-value = %8.3g)\n%s\n\n", stars, exp(theta->logEval), stars);

  /* print header */
  fprintf(stdout, "%-*.*s%s ", MSN, MSN, "Sequence name", 
    " Strand");
  fprintf(stdout, "%6s %7s %10s %*sSite%*s\n", 
    "Start", "Prob", "", width/2 - 1, "", width - width/2 - 4, "");
  fprintf(stdout, "%-*.*s%s ", MSN, MSN, "-------------", 
     " ------");
  fprintf(stdout, "%6s %6s %10s  ", "-----", "-------", "");
  for (i=0; i<width; i++) fputc('-', stdout);
  fputc('\n', stdout);

  
  for (siteNum=0; siteNum<numSites;siteNum++){
    
    start      = maxPostProbs[siteNum].start;
    seqNum     = maxPostProbs[siteNum].seqNum;
    onRevComp  = maxPostProbs[siteNum].revComp;
    prob       = maxPostProbs[siteNum].prob;
    seq        = dataSet->samples[seqNum].seq;
    rcSeq      = dataSet->samples[seqNum].rcSeq;
    name       = dataSet->samples[seqNum].name;
    nameLen    = dataSet->samples[seqNum].nameLength;
    seqLength  = dataSet->samples[seqNum].length;
    
    
    /* print sequence name */
    /* print name and strand */
    Rprintf("%-*.*s%s ", MSN, MSN, name, onRevComp ? "     -" : "     +");

    /* print position and position p-value */
    Rprintf("%6d  %7.4lf  ", start+1, prob);

    /* 
       number of nucleotides to the left and right that will be printed
       check for end of sequence here 
    */
    leftOff = MIN(sideLen,start);
    rightOff = MIN(sideLen,seqLength-start-width);
    
    for (i=0;i<sideLen-leftOff;i++){
      Rprintf(" ");
    }
    
    /* print nucleotides to left of motif */
    for (pos=start-leftOff;pos<start;pos++){
      oneChar = Convert2LowChar(seq[pos]);
      Rprintf("%c",oneChar);
    }
    Rprintf(" ");

    /* print actual motif */
    for (pos=0;pos<width;pos++){
      oneChar=Convert2Char(seq[start + pos]);
      Rprintf("%c",oneChar);
    }
    Rprintf(" ");    

    /* print nucleotides to the right of motif */
    for (pos=start+width;pos<start+width+rightOff;pos++){
      oneChar = Convert2LowChar(seq[pos]);
      Rprintf("%c",oneChar);
    }
    Rprintf("\n");
  }
  Rprintf("\n");

  Rprintf("%s\nTime: %8.2g secs\n%s\n", stars, dataSet->elapsedTime,stars);

  return(1);
}

/**********************************************************************/
/*
  printCosmo

  Print information about cosmo to stdout.

*/
/**********************************************************************/
int printCosmo(
		DATASET *dataSet,   /* the dataset */
		char *stars         /* one line of stars */
		)
{
  Rprintf("\n%s\nCOSMO - Constrained search for motifs\n%s\n", stars,stars);
  Rprintf("COSMO version %s\n\n", dataSet->version);
  Rprintf("For further information on how to interpret this output, please access\nhttp://cosmoweb.berkeley.edu.\n");

  return(1);
}


/**********************************************************************/
/*
  printReference

  Print reference information to stdout.

*/
/**********************************************************************/
int printReference(
		   DATASET *dataSet,
		   char *stars
		   )
{

  Rprintf("\n%s\nReference\n%s\n", stars,stars);
  Rprintf("%s\n\n", dataSet->reference);
  Rprintf("%s\n", dataSet->donlpReference);

  return(1);  
}


/**********************************************************************/
/*
  printTrainingSet

  Print summary of training set to stdout.

*/
/**********************************************************************/
int printTrainingSet(
		      DATASET *dataSet,  /* the dataset */
		      char *stars        /* one line of stars */
		      )
{

  int i, pcol, w;               /* various indices */
  
  w = MSN + 15;

  Rprintf("\n%s\nTraining set\n%s\n", stars,stars);
  Rprintf("Datafile: %s\n\n", dataSet->seqFileName);

  /*
    print a table of sequence lengths 
  */

  /*   print table header */
  for (pcol = w; pcol < 80; pcol += w) {
    Rprintf("%-*.*s %6s %2s", MSN, MSN, 
      "Sequence name", "Length", " ");
  }
  Rprintf("\n");
  for (pcol = w; pcol < 80; pcol += w) {
    Rprintf("%-*.*s %6s %2s", MSN, MSN, 
      "-------------", "------", " ");
  }
  Rprintf("\n");

  /*   print table columns */
  pcol = 0; 
  for (i=0; i<dataSet->numSeqs; i++) {
    SAMPLE *sample = dataSet->samples + i ;
    char *sample_name = sample->name;
    int lseq = sample->length;
    /* print the sample name and its length */
    pcol += w;          /* new line for print out? */
    if (pcol >= 80) { Rprintf("\n"); pcol = w; }
    Rprintf("%-*.*s %6d%2s", MSN, MSN, sample_name, lseq, " ");
  }
  Rprintf("\n");

  return(1);
}


/**********************************************************************/
/*
  printCommandLineSummary

  Print summary of command line to stdout.

*/
/**********************************************************************/
int printCommandSummary(
			 DATASET *dataSet,    /* the dataset */
			 char *stars          /* one line of stars */
			 )
{

  int i;                /* index */
  BOOLEAN nonEmpty;     /* any non-empty constraint sets?*/

  /* are there any non-empty constraint sets? */
  i             = dataSet->hasConstraints ? 1 : 0;
  nonEmpty      = FALSE;
  for (;i<dataSet->numConSets; i++) nonEmpty = nonEmpty || (!dataSet->cSets[i].isEmpty);

  fprintf(stdout, "\n%s\nCall summary\n%s\n"
	  "This information can also be useful in the event you wish to report a\n"
	  "problem with the cosmo software.\n\n"
	  "command: %s\n\n"
	  "cons:   number=   %8d    nonempty= %8s\n"
	  "        crit=     %8s    ",
	  stars, stars,
	  dataSet->command,
	  dataSet->numConSets - 1, (nonEmpty ? "TRUE" : "FALSE"),
	  (dataSet->conCrit.crit==lik ? "lik" : (dataSet->conCrit.crit==aic ? "aic" : (dataSet->conCrit.crit==bic ? "bic" : (dataSet->conCrit.crit==eval ? "eval" : (dataSet->conCrit.crit==likCV ? "likCV" : "pwmCV"))))));
  if (dataSet->conCrit.crit==likCV || dataSet->conCrit.crit==pwmCV){
    fprintf(stdout, 
	    "fold=     %8d    trunc=    %8d",
	    dataSet->conCrit.fold, (int) (100*dataSet->conCrit.quant));
  }
  fprintf(stdout,"\n");
  
  fprintf(stdout,
	  "model:  oops=     %8s    zoops=    %8s    tcm=      %8s\n",
	  (dataSet->oops ? "TRUE" : "FALSE"), (dataSet->zoops ? "TRUE" : "FALSE") , (dataSet->tcm  ? "TRUE" : "FALSE"));
  fprintf(stdout,
	  "        approx=   %8s    cutfac=   %8d\n",
	  (dataSet->approx==TCM ? "TCM" : (dataSet->approx==OVER ? "OVER" : "CUT")), dataSet->cutFactor); 
  fprintf(stdout,
	  "        crit=     %8s    ",
	  (dataSet->modCrit.crit==lik ? "lik" : (dataSet->modCrit.crit==aic ? "aic" : (dataSet->modCrit.crit==bic ? "bic" : (dataSet->modCrit.crit==eval ? "eval" : "likCV")))));
  if (dataSet->modCrit.crit==likCV){
    fprintf(stdout, 
	    "fold=     %8d    trunc=    %8d",	    dataSet->modCrit.fold, (int) (100*dataSet->modCrit.quant));
  }
  fprintf(stdout,"\n");


  fprintf(stdout,
	  "width:  minw=     %8d    maxw=     %8d\n"
	  "        crit=     %8s    ",
	  dataSet->minWidth, dataSet->maxWidth,
	  (dataSet->wCrit.crit==lik ? "lik" : (dataSet->wCrit.crit==aic ? "aic" : (dataSet->wCrit.crit==bic ? "bic" : (dataSet->wCrit.crit==eval ? "eval" : "likCV")))));
  if (dataSet->wCrit.crit==likCV){
    fprintf(stdout, 
	    "fold=     %8d    trunc=    %8d",
	    dataSet->wCrit.fold, (int) (100*dataSet->wCrit.quant));
  }
  fprintf(stdout,"\n");

  fprintf(stdout,  
	  "nsites: minsites= %8d    zmaxsites= %7d   tmaxnumsite = %5d\n"
	  "        crit=     %8s    ",
	  dataSet->minNumSites, dataSet->zoopsMaxNumSites, dataSet->tcmMaxNumSites,
	  (dataSet->intCrit.crit==lik ? "lik" : (dataSet->intCrit.crit==eval ? "eval" : "likCV"))); 
  if (dataSet->intCrit.crit==likCV){
    fprintf(stdout,
	    "\fold=     %8d    trunc=    %8d"
	    "likCV",  dataSet->intCrit.fold,(int) (100 * dataSet->intCrit.quant));
  }  
  fprintf(stdout,"\n");

  fprintf(stdout,
	  "starts: type=     %8s    number=   %8d\n",
	  (dataSet->likStarts ? "lik" : "eval"), dataSet->numStarts);

  fprintf(stdout,
	  "back:   source=   %8s    cvorder=  %8s    fold=     %8d\n",
	  (dataSet->hasBackFile ? "other" : (dataSet->readTransMat ? "matrix" : (dataSet->readBfile ? "bfile" : "same"))), 
	  (dataSet->orderByCV ? "TRUE" : "FALSE"), dataSet->backFold);


  fprintf(stdout,
	  "data:   n=        %8ld    N=        %8d    revcomp=  %8s\n", 
	  dataSet->totalNucs, dataSet->numSeqs, (dataSet->revComp ? "TRUE" : "FALSE")
	  );

  return(1);
}


/**********************************************************************/
/*
  printTransmat

  Print out the transition matrix

*/
/**********************************************************************/
int printTransMat(
		   DATASET *dataSet,    /* the dataset */
		   char *stars          /* one line of stars */
		   )
{

  int i,j,k,l;         /* various indices */
  double *transMat;    /* current transition matrix */
  char *prefix;        /* the last k nucletides */
  int sum;             /* sum of all higher powers for prefix */
  int part;            /* part of this power for prefix */
  
  prefix = CharMalloc(dataSet->maxOrder+1);

  Rprintf("\n%s\nBackground Markov model\n%s\n\n", stars, stars);  

  if ((!dataSet->orderByCV) || (dataSet->readTransMat)){
    Rprintf("Using order %d for background Markov model.\nOrder was specified by user or dataset contains only one sequence.\n", 
	   dataSet->order);



  }else{

    Rprintf("Order of background Markov model chosen by likelihood-based CV: %d\n", dataSet->order);
    Rprintf("\nKullback-Leibler divergences for candidate orders 0 to %d:\n\n", dataSet->maxOrder);
    Rprintf("Order    KL-divergence\n");
    for (i=0; i<=dataSet->maxOrder; i++){
      if (dataSet->backKL[i]==-1){
	Rprintf("%5d       %10s\n", i, "NA");
      }else if (dataSet->backKL[i]==0){
	Rprintf("%5d       %10s\n", i, "inf");
      }else{
	Rprintf("%5d       %10.5g\n", i, dataSet->backKL[i]);
      }
    }
  }

  for (i=0; i<=MIN(2,dataSet->order);i++){
    transMat = dataSet->transMats[i];
    if (dataSet->readTransMat){
      Rprintf("\n\nUser-supplied transition matrix for order %d:\n\n", i);
    }else{
      Rprintf("\n\nEstimated transition matrix for order %d:\n\n", i);
    }
    Rprintf("Prefix       A         C         G         T\n");

    for (j=0; j < pow(4,i);j++){

      /* get prefix */
      sum = 0;
      l = 0;
      for (k=i-1;k>=0;k--,l++){
	part = (int)((j-sum)/(pow(4,k)));
	sum += part*pow(4,k);
	prefix[l] = Convert2Char(part);
	
      }
      prefix[i] = '\0';
      

      Rprintf("%6s %7.4lf   %7.4lf   %7.4lf   %7.4lf\n", prefix, exp(transMat[4*j]), exp(transMat[4*j+1]), exp(transMat[4*j+2]), 
	     exp(transMat[4*j+3]));
    }
	   
  }

  if (dataSet->order>2) Rprintf("\nEstimated transition matrices for orders greater than 2 are omitted for convenience.\n");

  return(1);
}



/**********************************************************************/
/*
  printModelSelection

  Print summary of selected fine-tuning parameters to stdout.

*/
/**********************************************************************/
int printModelSelection(
			 DATASET *dataSet,   /* the dataset */
			 char *stars         /* one line of stars */
			 )
{

  CRIT *crits;             /* array of stored model selection criteria */
  int cSetNum;             /* current constraint set number */
  int modTypeNum;          /* current model type number */
  int width;               /* current width */
  int index;               /* index into array of criteria */

  crits   = dataSet->allCrits;


  /* evalue is stored as log(evalue) */
  if (dataSet->intCrit.crit==eval) dataSet->crits.intCrit = exp(dataSet->crits.intCrit);
  if (dataSet->wCrit.crit==eval)   dataSet->crits.wCrit = exp(dataSet->crits.wCrit);
  if (dataSet->modCrit.crit==eval) dataSet->crits.modCrit = exp(dataSet->crits.modCrit); 
  if (dataSet->conCrit.crit==eval) dataSet->crits.conCrit = exp(dataSet->crits.conCrit);

  cSetNum = dataSet->hasConstraints ? 1 : 0;
  for (;cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){
	index = MOD(width, modTypeNum, cSetNum);

	if (dataSet->wCrit.crit==eval)   crits[index].wCrit = exp(crits[index].wCrit);
	if (dataSet->modCrit.crit==eval) crits[index].modCrit = exp(crits[index].modCrit);
	if (dataSet->conCrit.crit==eval) crits[index].conCrit = exp(crits[index].conCrit);
      }
    }
  }


  /* likelihood criterion is used as -log-likelihood */
  if (dataSet->intCrit.crit==lik) dataSet->crits.intCrit *= -1;
  if (dataSet->wCrit.crit==lik)   dataSet->crits.wCrit *= -1;
  if (dataSet->modCrit.crit==lik) dataSet->crits.modCrit *= -1;
  if (dataSet->conCrit.crit==lik) dataSet->crits.conCrit  *= -1;


  cSetNum = dataSet->hasConstraints ? 1 : 0;
  for (;cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){
	index = MOD(width, modTypeNum, cSetNum);

	if (dataSet->wCrit.crit==lik)   crits[index].wCrit *= -1;
	if (dataSet->modCrit.crit==lik) crits[index].modCrit *= -1;
	if (dataSet->conCrit.crit==lik) crits[index].conCrit *= -1;
      }
    }
  }


  Rprintf("\n%s\nSummary of candidate models\n%s\n\n", stars, stars);  

  Rprintf("Constraint set   Model type   Width   widthCrit    modCrit    conCrit\n");
  cSetNum = dataSet->hasConstraints ? 1 : 0;
  for (;cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){

	index = MOD(width, modTypeNum, cSetNum);
	Rprintf("%14d   %10s   %5d", cSetNum, getModName(dataSet,dataSet->mTypes[modTypeNum],0), width);
	if (crits[index].wInit){
	  Rprintf("   %9.3g",  crits[index].wCrit);
	}else{
	  Rprintf("         ---");
	}
	if (crits[index].modInit){
	  Rprintf("   %8.3g", crits[index].modCrit);
	}else{
	  Rprintf("        ---");
	}
	if (crits[index].conInit){
	  Rprintf("   %8.3g\n",  crits[index].conCrit);
	}else{
	  Rprintf("        ---\n");
	}
      }
    }
  }

  Rprintf("\n%s\nSelected model\n%s\n\n", stars, stars);  


  if (dataSet->numConSets>2){
    Rprintf("Constraints:  choice= %8d    crit= %8.3g\n", dataSet->crits.conSetNum, dataSet->crits.conCrit);
  }else{
    Rprintf("Constraints:  choice= %8d    crit=      ---\n", dataSet->crits.conSetNum);
  }
  if (dataSet->numModTypes>1){
      Rprintf("Model:        choice= %8s    crit= %8.3g\n", getModName(dataSet,dataSet->crits.mType,0), dataSet->crits.modCrit);    
  }else{
      Rprintf("Model:        choice= %8s    crit=      ---\n", getModName(dataSet,dataSet->crits.mType,0));
  }
  if (dataSet->numWidths>1){
    Rprintf("Width:        choice= %8d    crit= %8.3g\n", dataSet->crits.width, dataSet->crits.wCrit);
  }else{
    Rprintf("Width:        choice= %8d    crit=      ---\n", dataSet->crits.width);
  }
  Rprintf("Sites:        choice= %8d    crit= %8.3g\n", dataSet->theta->numSites, dataSet->crits.intCrit);

  return(1);
}


/**********************************************************************/
/*
  output2ProbFile

  Print posterior probabilities to file.
  Posterior probabilities are negative if motif is thought to be on
  the reverse complement strand.

*/
/**********************************************************************/
int output2ProbFile(
		     DATASET *dataSet /* the dataset */
		     )
{

  int seqNum;        /* sequence number */
  int numSeqs;       /* number of sequences */
  int pos;           /* position in motif */
  FILE *probFile;    /* posterior probability file */
  THETA *theta;      /* final parameter estimates */
  double prob;       /* posterior probability */
  long seqLength;    /* length of current sequence */
  long maxSeqLength; /* maximum sequence length */
  int rcFac;         /* make probabilities negative if motif is thought to be on reverse complement */
 
  probFile     = dataSet->probFile;
  numSeqs      = dataSet->numSeqs;
  theta        = dataSet->theta;
  maxSeqLength = dataSet->maxSeqLength;

  /* calculate posterior probabilities to be safe */
  if (!getPostProbs(theta,dataSet)) return(0);
    
  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seqLength = dataSet->samples[seqNum].length;
    
    for (pos=0; pos<seqLength; pos++){
      prob  = dataSet->postProbs[PP(seqNum,pos)].prob;
      rcFac = dataSet->postProbs[PP(seqNum,pos)].revComp ? -1 : 1;
      fprintf(probFile, "%g\t", rcFac*prob);
    }
    fprintf(probFile, "\n");
  }

  return(1);
}



/**********************************************************************/
/*
  output2SimFile

  Print output for simulation purposes.


*/
/**********************************************************************/
int output2SimFile(
		   DATASET *dataSet
		   )
{

  FILE *simOut;           /* simulation output file */
  THETA *theta;           /* final parameter estimates */
  int seqNum;             /* sequence number */
  int start;              /* starting position in sequence */
  int siteNum;            /* number of current site */
  int numSites;           /* number of sites */
  POSTPROB *maxPostProbs; /* maximum posterior probabilities */
  POSTPROB *postProbs;    /* posterior probabilities */
  BOOLEAN global;         /* sort posterior probabilities over all sequences */
  long maxSeqLength;      /* maximum sequence length */


  simOut        = dataSet->simFile;
  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;
  theta         = dataSet->theta;
  numSites      = theta->numSites;
  global        = theta->mType==TCM || theta->mType==OVER || theta->mType==CUT;
  maxSeqLength  = dataSet->maxSeqLength;

  /* get posterior probabilities */
  if (!getPostProbs(theta, dataSet)) return(0);

  /* sort posterior probabilities */
  if (!sortPostProbs(dataSet, theta->width, global)) return(0);


  fprintf(simOut, "time = %lf sec\n", dataSet->elapsedTime);
  fprintf(simOut, "conSet = %d\n", dataSet->crits.conSetNum);
  fprintf(simOut, "model = %d\n", dataSet->crits.mType);
  fprintf(simOut, "width = %d\n", theta->width);


  for (siteNum=0; siteNum<numSites;siteNum++){
    
    start      = maxPostProbs[siteNum].start;
    seqNum     = maxPostProbs[siteNum].seqNum;
    
    fprintf(simOut,"%d %d\n",seqNum+1, start+1);
  }

  fprintf(simOut,"\n");

  return(1);
}


/**********************************************************************/
/*
  output2ROCFile

  Output all sites in order of posterior probability for calculating ROC.
  Order all candidate sites by posterior probability even if model type 
  is OOPS or ZOOPS

*/
/**********************************************************************/
int output2ROCFile(
		    DATASET *dataSet  /* the dataset */
		    )
{

  FILE *rocOut;           /* ROC output file */
  THETA *theta;           /* final parameter estimates */
  int seqNum;             /* sequence number */
  int numSeqs;            /* number of sequences */
  int start;              /* starting position in sequence */
  int siteNum;            /* number of current site */
  int numSites;           /* number of sites */
  POSTPROB *postProbs;    /* posterior probabilities */
  int width;              /* width of motif */
  double prob;            /* posterior probability */
  long maxSeqLength;      /* maximum sequence length */


  rocOut        = dataSet->rocFile;
  theta         = dataSet->theta;
  width         = theta->width;
  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  postProbs     = dataSet->postProbs;


  /* total number of candidate sites */
  numSites      = 0;
  for (seqNum=0; seqNum<numSeqs;seqNum++) numSites += dataSet->samples[seqNum].length - width + 1;

  /* get posterior probabilities */
  if (!getPostProbs(theta, dataSet)) return(0);


  /* sort posterior probabilities */
  qsort((void*) postProbs,maxSeqLength*numSeqs,sizeof(POSTPROB),compPostProbs);
  
  /* output to file */
  for (siteNum=0; siteNum<numSites; siteNum++){
    start      = postProbs[siteNum].start;
    seqNum     = postProbs[siteNum].seqNum;
    prob       = postProbs[siteNum].prob;

    fprintf(rocOut,"%d %d\n",seqNum+1, start+1);
  }

  
  fprintf(rocOut,"\n");

  return(1);
}


/**********************************************************************/
/*
  getModName

  Get name of model type as string.

*/
/**********************************************************************/
char *getModName(
		 DATASET *dataSet,   /* the dataset */
		 MTYPE mType,        /* model type for which to get string name */
		 BOOLEAN pad         /* pad model names to all take up five spaces? */
		 )
{

  char *name;

  name = dataSet->modTypeName;

  switch(mType){
  case OOPS:
    name = pad ? " OOPS" : "OOPS";
    break;
  case ZOOPS:
    name = "ZOOPS";
    break;
  case TCM:
    name = pad ? "  TCM" : "TCM";
    break;
  case OVER:
    name = pad ? " OVER" : "OVER";
    break;
  case CUT:
    name = pad ? "  CUT" : "CUT";
    break;
  default:
    Rprintf( "getModName: Didn't recognize model type %d. Exiting...\n", mType);
    return(0);
  }

  return(name);
}


/**********************************************************************/
/*
  getCritDesc

  Get strings that describe what model fine-tuning parameters are used
  in the final model and how they were selected.
  Stores these strings in dataSet->wDesc, dataSet->modDesc, 
  dataSet->conDesc, and dataSet->markovDesc.

*/
/**********************************************************************/
int getCritDesc(
		 DATASET *dataSet  /* the dataset */
		 )
{

  if (dataSet->minWidth==dataSet->maxWidth){
    sprintf(dataSet->wDesc, "Width of motif:             %d (specified by user)",dataSet->crits.width);
  }else{
    switch(dataSet->wCrit.crit){
    case lik:
      sprintf(dataSet->wDesc, "Width of motif:             %d (chosen by maximum-likelihood)",dataSet->crits.width);
      break;
    case aic:
      sprintf(dataSet->wDesc, "Width of motif:             %d (chosen by AIC)",dataSet->crits.width);
      break;
    case bic:
      sprintf(dataSet->wDesc, "Width of motif:             %d (chosen by BIC)",dataSet->crits.width);
      break;
    case eval:
      sprintf(dataSet->wDesc, "Width of motif:             %d (chosen by E-value)",dataSet->crits.width);
      break;
    case likCV:
      if (dataSet->wCrit.trunc){
	sprintf(dataSet->wDesc, "Width of motif:             %d (chosen by %d-fold likelihood-based CV with truncation at %dth percentile)", dataSet->crits.width, dataSet->wCrit.fold, (int) (dataSet->wCrit.quant*100));
      }else{
	sprintf(dataSet->wDesc,"Width of motif:             %d (chosen by %d-fold likelihood-based CV)", dataSet->crits.width, dataSet->wCrit.fold);
      }
      break;
    default:
      Rprintf( "getCritDesc: Didn't recognize width criterion %d. Exiting...\n", dataSet->wCrit.crit);
      return(0);
    }
  }

  if (dataSet->numModTypes==1){
    sprintf(dataSet->modDesc, "Model type:                 %s (specified by user)", getModName(dataSet,dataSet->crits.mType,0));
  }else{
    switch(dataSet->modCrit.crit){
    case lik:   
      sprintf(dataSet->modDesc, "Model type:                 %s (chosen by maximum-likelihood)", getModName(dataSet, dataSet->crits.mType,0));
      break;
    case aic:
      sprintf(dataSet->modDesc, "Model type:                 %s (chosen by AIC)", getModName(dataSet, dataSet->crits.mType,0));
      break;
    case bic:
      sprintf(dataSet->modDesc, "Model type:                 %s (chosen by BIC)", getModName(dataSet, dataSet->crits.mType,0));
      break;
    case eval:
      sprintf(dataSet->modDesc, "Model type:                 %s (chosen by E-value)", getModName(dataSet, dataSet->crits.mType,0));
      break;
    case likCV:
      if (dataSet->modCrit.trunc){
	sprintf(dataSet->modDesc, "Model type:                 %s (chosen by %d-fold likelihood-based CV with truncation at %dth percentile)",getModName(dataSet, dataSet->crits.mType,0), dataSet->modCrit.fold, (int) (dataSet->modCrit.quant*100));
      }else{
	sprintf(dataSet->modDesc,"Model type:                 %s (chosen by %d-fold likelihood-based CV)",getModName(dataSet, dataSet->crits.mType,0), dataSet->modCrit.fold);
      }
      break;
    default:
      Rprintf( "getCritDesc: Didn't recognize model type criterion %d. Exiting...\n", dataSet->modCrit.crit);
      return(0);
    }
  }

  if (dataSet->numConSets==1){
    sprintf(dataSet->conDesc, "Constraints:                None (specified by user)");
  }else if(dataSet->numConSets==2){
    sprintf(dataSet->conDesc, "Constraints:                Set 1 (specified by user)");
  }else{

    if (dataSet->cSets[dataSet->crits.conSetNum].isEmpty){
      switch(dataSet->conCrit.crit){
      case lik:
	sprintf(dataSet->conDesc, "Constraints:                None (chosen by maximum-likelihood)");
	break;
      case aic:
	sprintf(dataSet->conDesc, "Constraints:                None (chosen by AIC)");
	break;
      case bic:
	sprintf(dataSet->conDesc, "Constraints:                None (chosen by BIC)");
	break;
      case eval:
	sprintf(dataSet->conDesc, "Constraints:                None (chosen by E-value)");
	break;
      case likCV:
	if (dataSet->conCrit.trunc){
	  sprintf(dataSet->conDesc, "Constraints:                None (chosen by %d-fold likelihood-based CV with truncation at %dth percentile)", dataSet->conCrit.fold, (int) (dataSet->conCrit.quant*100));
	}else{
	  sprintf(dataSet->conDesc,"Constraints:                None (chosen by %d-fold likelihood-based CV)", dataSet->conCrit.fold);
	}
	break;
      case pwmCV:
	if (dataSet->conCrit.trunc){
	  sprintf(dataSet->conDesc, "Constraints:                None (chosen by %d-fold PWM-based CV with truncation at %dth percentile)", dataSet->conCrit.fold, (int) (dataSet->conCrit.quant*100));
	}else{
	  sprintf(dataSet->conDesc,"Constraints:                None (chosen by %d-fold PWM-based CV)", dataSet->conCrit.fold);
	}
	break;
      default:
	Rprintf( "getCritDesc: Didn't recognize constraint set criterion %d. Exiting...\n", dataSet->conCrit.crit);
	return(0);
      }

    }else{

      switch(dataSet->conCrit.crit){
      case lik:
	sprintf(dataSet->conDesc, "Constraints:                Set %d (chosen by maximum-likelihood)",dataSet->crits.conSetNum);
	break;
      case aic:
	sprintf(dataSet->conDesc, "Constraints:                Set %d (chosen by AIC)",dataSet->crits.conSetNum);
	break;
      case bic:
	sprintf(dataSet->conDesc, "Constraints:                Set %d (chosen by BIC)",dataSet->crits.conSetNum);
	break;
      case eval:
	sprintf(dataSet->conDesc, "Constraints:                Set %d (chosen by E-value)",dataSet->crits.conSetNum);
	break;
      case likCV:
	if (dataSet->conCrit.trunc){
	  sprintf(dataSet->conDesc, "Constraints:                Set %d (chosen by %d-fold likelihood-based CV with truncation at %dth percentile)", dataSet->crits.conSetNum, dataSet->conCrit.fold, (int) (dataSet->conCrit.quant*100));
	}else{
	  sprintf(dataSet->conDesc,"Constraints:                Set %d (chosen by %d-fold likelihood-based CV)", dataSet->crits.conSetNum, dataSet->conCrit.fold);
	}
	break;
      case pwmCV:
	if (dataSet->conCrit.trunc){
	  sprintf(dataSet->conDesc, "Constraints:                Set %d (chosen by %d-fold PWM-based CV with truncation at %dth percentile)", dataSet->crits.conSetNum, dataSet->conCrit.fold, (int) (dataSet->conCrit.quant*100));
	}else{
	  sprintf(dataSet->conDesc,"Constraints:                Set %d (chosen by %d-fold PWM-based CV)", dataSet->crits.conSetNum, dataSet->conCrit.fold);
	}
	break;

      default:
	Rprintf( "getCritDesc: Didn't recognize constraint set criterion %d. Exiting...\n", dataSet->conCrit.crit);
	return(0);
      }
    }
  }

  if (dataSet->orderByCV){
    sprintf(dataSet->markovDesc,"Order of Markov model:      %d (chosen by %d-fold likelihood-based CV)", dataSet->order, MIN(5,dataSet->numSeqs));
  }else{
    sprintf(dataSet->markovDesc,"Order of Markov model:      %d (specified by user)", dataSet->order);
  }


  return(1);
}






