/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains functions for finding starting values based      *
*      on the likelhood function.                                            *
*                                                                            *
*****************************************************************************/
 
#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "estarts.h"
#include "lstarts.h"
#include "matLib.h"
#include "theta.h"
#include "constraints.h"
  
#define SCALE_LOGS 1e6                  /* don't use larger! */
#define INT_LOG(X) ((int)(SCALE_LOGS * LOG(X)))
#define RNDEPS 1e-12



/**********************************************************************/
/*
  getSvLogLR

  Pre-calculate log likelihood ratios for all possible starting
  values for the postion weight matrix that can be extracted from
  the oligos of appropriate size in the actual sequences in the
  following way: Assign a probablity 'hiProb' to the nucleotide
  that is observed in the oligo in a given position, and assign 
  'lowProb' = (1-hiProb)/3 to all other nucleotides.

  This function makes use of the structure of these starting
  values. It first calculates the log LR for the first starting 
  value of a given width from sequence n. The logLR for the
  starting values derived from the same position of successively
  longer motif widths are derived from this log LR. Next we 
  consider the oligo obtained from the very first lowest width
  oligo by shifting the motif one position ahead. The log LR
  for this starting value is again obtained from the log LR
  of the previous oligo. The log LR for the starting values
  of successively larger widths are again derived from the
  smaller widths.

  This approach saves a considerable amount of time since
  the calculation of starting value log LRs becomes the bottleneck
  with larger sized problems.
	
*/
/************************************************************************/
int getSvLogLR(
		DATASET *dataSet   /* the dataset */
		)
{

  double *pwm;            /* space for a position weight matrix */
  double *prevPwm;        /* the position weight matrix derived from the previous oligo */
  double hiProb;          /* probability to assigne to an observed nucleotide */
  int width;              /* current width */
  int minWidth;           /* minimum width */
  int maxWidth;           /* maximum width */
  int oligoSeqNum;        /* index of sequence from which oligo is derived */
  int oligoStart;         /* position in sequence at which oligo starts */
  long maxSeqLength;      /* maximum sequence length */
  int *oligoSeq;          /* oligo sequence */
  long oligoSeqLength;    /* length of oligo sequence */
  BOOLEAN oneWidth;       /* is there only one width? */
  int numSeqs;            /* number of sequences */

  minWidth       = dataSet->minWidth;
  maxWidth       = dataSet->maxWidth;
  maxSeqLength   = dataSet->maxSeqLength;
  pwm            = dataSet->pwm1;
  prevPwm        = dataSet->pwm2;
  hiProb         = dataSet->hiProb;
  oneWidth       = (dataSet->maxWidth==dataSet->minWidth);
  numSeqs        = dataSet->numSeqs;


  /* initialize arrays to 0 */
  if (!initSvLogLR(dataSet)) return(0);

  for (oligoSeqNum=0;oligoSeqNum<numSeqs;oligoSeqNum++){

    if(COSMO_MSG_LEVEL>=2 && !dataSet->printStatus){
      Rprintf( "\rgetSvLogLR: Extracting starting values from sequence %d/%d", oligoSeqNum+1,numSeqs);
    }

    /* print first status message if desired */
    if (dataSet->printStatus) Rprintf("%d/%d   0/%d\n", oligoSeqNum+1, dataSet->numSeqs, dataSet->effNumModels);
    

    oligoSeq       = dataSet->samples[oligoSeqNum].seq;
    oligoSeqLength = dataSet->samples[oligoSeqNum].length;

    /* loop over start positions for oligo */
    for (oligoStart=0;oligoStart<oligoSeqLength-minWidth + 1; oligoStart++){

      /* set up with width = minWidth*/
      width = minWidth;

      /* Create pwm from new oligo */
      seq2pwm(pwm, oligoSeq + oligoStart, width, hiProb);

      if (oligoStart!=0){

	/* Create pwm from previous oligo */
	seq2pwm(prevPwm, oligoSeq + oligoStart - 1, width, hiProb);

	/* calculate mProbs1 from previous mProbs*/
	if (!getOverMProbs(dataSet, pwm, prevPwm, width, 1,oneWidth)) return(0);

      }else{
  
	/* calculate mProbs1 fresh*/
	if (!getNewMProbs(dataSet, pwm, width, 1)) return(0);
      }

      
      /* evaluate likelihood under all the desired models */
      if (!evalModLiks(dataSet, width, oligoSeqNum, oligoStart, 1)) return(0);

      /* copy of mProbs1 to mProbs2*/
      if (!oneWidth){
	if (!copyMProbs(dataSet,1,2)) return(0);
      }


      /* fill in for oligoStart = 0 and minWidth<width<=maxWidth */
      /* be careful to respect oligoStart<oligoSeqLength-width+1 */
      for (width = minWidth+1;width<=MIN(maxWidth,oligoSeqLength-oligoStart);width++){
	  
	/* add column to pwm */
	addCol2pwm(pwm, oligoSeq, oligoStart, width, hiProb);

	/* update mProbs2 */
	if (!getHigherMProbs(dataSet, pwm, width, 2)) return(0);

	/* copy mProbs2 to mProbs1 for getOverMProbs */
	if (width==minWidth+1){
	  if (!copyMProbs(dataSet,2,1)) return(0);
	}

	/* evaluate likelihood under all the desired models */
	if (!evalModLiks(dataSet, width, oligoSeqNum, oligoStart, 2)) return(0);

      }
    }	    
  }

  if (COSMO_MSG_LEVEL>=2) Rprintf( "\n");

  return(1);
}

 
/**********************************************************************/
/*
  initSvLogLR

  Initialize arrays of svLogLR to zero.
	
*/
/************************************************************************/
int initSvLogLR(
		 DATASET *dataSet   /* the dataset */
		 )
{

  double *svLogLR;        /* array of logLR for all possible starting values */
  int numModType;         /* index of current model type */
  int width;              /* current width */
  int minWidth;           /* minimum width */
  int maxWidth;           /* maximum width */
  int oligoSeqNum;        /* index of sequence from which oligo is derived */
  int oligoStart;         /* position in sequence at which oligo starts */
  int numSeqs;            /* number of sequences */
  long maxSeqLength;      /* maximum sequence length */
  int nnSites;            /* number of candidates for the expected number of motif occurrences */
  int nSitesNum;          /* index of current candidate */


  minWidth       = dataSet->minWidth;
  maxWidth       = dataSet->maxWidth;
  maxSeqLength   = dataSet->maxSeqLength;
  numSeqs        = dataSet->numSeqs;
  
  /* initialize arrays to 0 */
  for (numModType=0; numModType< dataSet->numModTypes; numModType++){
    svLogLR = dataSet->svSpace[numModType].svLogLR;
    nnSites = dataSet->svSpace[numModType].nnSites;

    for (width=minWidth;width<=maxWidth;width++){
      for (oligoSeqNum=0;oligoSeqNum<numSeqs;oligoSeqNum++){
	for (oligoStart=0;oligoStart<maxSeqLength;oligoStart++){
	  for (nSitesNum=0; nSitesNum<nnSites; nSitesNum++){
	    svLogLR[LR(width,oligoSeqNum,oligoStart,nSitesNum)] = 0;
	  }
	}
      }
    }
  }

  return(1);
}

/**********************************************************************/
/*
  seq2pwm

  Convert an oligo into a position weight matrix for a candidate 
  starting value by assigning the observed nucleotides a probability
  'hiProb' and the remaining ones the probability (1-hiProb)/3 for
  all positions in which a regular letter occurs, and by assigning
  1/alen to all nucleotides in positions with unknown character 'X'.
  
  The position weight matrix is log-transformed and stored in integer
  format by calling the macro INT_LOG.

  The position weight matrix is written to pwm.


	
*/
/************************************************************************/
int seq2pwm(
	     double *pwm,   /* on exit: the PWM */
	     int *seq,      /* the oligo sequence */
	     int width,     /* width of oligo */
	     double hiProb  /* probability to assign to observed nucleotide */
	     )
{

  double lowProb;    /* the probability to assign to unobserved nucleotides */
  int pos;           /* index for looping over positions of motif */ 
  int nuc;           /* index for looping over nucleotides */

  lowProb = (1-hiProb)/3;

  for (pos=0;pos<width;pos++){
    if (seq[pos]<4){
      for (nuc=0;nuc<4;nuc++){
	pwm[PWM(pos,nuc)] = lowProb;
      }
      pwm[PWM(pos,seq[pos])] = hiProb;
    }else{
      for (nuc=0;nuc<4;nuc++){
	pwm[PWM(pos,nuc)] = 1.0/4;
      }
    }
  }

  return(1);
} 


/**********************************************************************/
/*
  addCol2pwm
   
  Add another column to a log-transformed position weight matrix.

 
*/
/**********************************************************************/
int addCol2pwm(
		double *pwm,   /* the PWM to which we are adding a column */
		int *seq,      /* the sequence from which the PWM was constructed */
		int start,     /* the starting position of the corresponding oligo */
		int newWidth,  /* the new width of the PWM */
		double hiProb  /* the probability to assign to observed nucleotides */
		)
{

  double lowProb;    /* probability to assign to unobserved nucleotides */
  int lastCol;       /* last column of motif */
  int nuc;           /* nucleotide */

  lowProb = (1-hiProb)/3;

  /* add new column to PWM */
  lastCol = newWidth - 1;
  if (seq[start+lastCol]<4){
    for (nuc=0;nuc<4;nuc++){
      pwm[PWM(lastCol,nuc)] = lowProb;
    }
    pwm[PWM(lastCol,seq[start+lastCol])] = hiProb;
  }else{
    for (nuc=0;nuc<4;nuc++){
      pwm[PWM(lastCol,nuc)] = 1.0/4;
    }
  }


  return(1);
}

/**********************************************************************/
/*
  printMProbs
   
  Print out motif probabilities stored in mProbs1 or mProbs2.

 
*/
/**********************************************************************/
int printMProbs(
		 DATASET *dataSet,   /* the dataset */
		 int width,          /* width of motif */
		 int spaceNum        /* are motif probs stored in mProbs1 or mProbs2? */
		 )
{

  int seqNum;         /* current sequence number */
  int numSeqs;        /* number of sequences */
  int seqLength;      /* length of current sequence */
  double *mProbs;     /* motif probs on forward strand */
  double *rcmProbs;   /* motif probs on reverse strand */


  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("printMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  Rprintf("printMProbs;\n");

  numSeqs = dataSet->numSeqs;
  for (seqNum=0;seqNum<numSeqs;seqNum++){


    seqLength = dataSet->samples[seqNum].length;
    if (spaceNum==1){
      mProbs   = dataSet->samples[seqNum].mProbs1;
      rcmProbs = dataSet->samples[seqNum].rcmProbs1;
    }else{
      mProbs   = dataSet->samples[seqNum].mProbs2;
      rcmProbs = dataSet->samples[seqNum].rcmProbs2;
    }

    PrintDoubleMatrixE(mProbs, 1, seqLength - width + 1);

    if (dataSet->revComp){
      PrintDoubleMatrixE(mProbs, 1, seqLength - width + 1);
    }

  }

  return(1);
}

/**********************************************************************/
/*
  getNewMProbs
  
  Get motif probabilities by calculation from scratch for a given PWM.

*/
/**********************************************************************/ 
int getNewMProbs(
		  DATASET *dataSet,    /* the dataset */
		  double *pwm,         /* PWM for which to calculate motif probabilities */ 
		  int width,           /* width of this PWM */
		  int spaceNum         /* store mProbs in mProbs1 or mProbs2 ? */
		  )
{

  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int start;                    /* staring position of current motif site */
  int seqLength;                /* length of current sequence */
  double mProb;                 /* log-transformed motif probability */
  int pos;                      /* position within the motif */
  int *seq;                     /* current sequence */
  int *rcSeq;                   /* reverse complement of current sequence */
  double *mProbs;               /* motif probabilities on forward strand */
  double *rcmProbs;             /* motif probabilities on reverse strand */
 
  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("getNewMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numSeqs = dataSet->numSeqs;
 

  for (seqNum=0;seqNum<numSeqs;seqNum++){

    /* skip this sequence? */
    if (dataSet->samples[seqNum].exclude){
      if (COSMO_MSG_LEVEL>=5) Rprintf("getNewMProbs: skipping %d\n", seqNum);
      continue;
    }

    seq       = dataSet->samples[seqNum].seq;
    rcSeq     = dataSet->samples[seqNum].rcSeq;
    seqLength = dataSet->samples[seqNum].length;
    if (spaceNum==1){
      mProbs   = dataSet->samples[seqNum].mProbs1;
      rcmProbs = dataSet->samples[seqNum].rcmProbs1;
    }else{
      mProbs   = dataSet->samples[seqNum].mProbs2;
      rcmProbs = dataSet->samples[seqNum].rcmProbs2;
    }

    for (start=0;start<seqLength - width + 1;start++){
      mProb = 1;
      for (pos=0;pos<width;pos++){
	if (seq[start+pos]<4){
	  mProb *= pwm[PWM(pos, seq[start+pos])];
	}else{
	  mProb *= 1.0/4;
	}
      }
      mProbs[start] = mProb;
    }

    /* store probability of motif in opposite orientation */
    if (dataSet->revComp){
      for (start=0;start<seqLength - width + 1;start++){
	mProb = 1;
	for (pos=0;pos<width;pos++){
	  if (rcSeq[start+pos]<4){
	    mProb *= pwm[PWM(pos, rcSeq[start+pos])];
	  }else{
	    mProb *= 1.0/4;
	  }
	}
	rcmProbs[start] = mProb;

      }
    }

  }

  return(1);
}

/**********************************************************************/
/*
  getHigherMProbs
  
  Calculate motif probabilities for width W from the stored motif 
  probabilities for width W-1, which are stored either in mProbs1 or
  mProbs2.

  New motif probabilities are stored in the same array.

*/
/**********************************************************************/
int getHigherMProbs(
		     DATASET *dataSet,    /* the dataset */
		     double *pwm,         /* PWM for which motif probabilities are desired */
		     int width,           /* width of this PWM */
		     int spaceNum         /* store mProbs in mProbs1 or mProbs2 ? */
		     )
{

  int lastCol;                  /* last position in motif */
  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int start;                    /* staring position of current motif site */
  int seqLength;                /* length of current sequence */
  int *seq;                     /* current sequence */
  int *rcSeq;                   /* reverse complement of current sequence */
  double *mProbs;               /* motif probabilities on forward strand */
  double *rcmProbs;             /* motif probabilities on reverse strand */


  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("getHigherMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numSeqs = dataSet->numSeqs;
  lastCol = width - 1;

  for (seqNum=0; seqNum<numSeqs; seqNum++){

    seq       = dataSet->samples[seqNum].seq;
    rcSeq     = dataSet->samples[seqNum].rcSeq;
    seqLength = dataSet->samples[seqNum].length;
    if (spaceNum==1){
      mProbs   = dataSet->samples[seqNum].mProbs1;
      rcmProbs = dataSet->samples[seqNum].rcmProbs1;
    }else{
      mProbs   = dataSet->samples[seqNum].mProbs2;
      rcmProbs = dataSet->samples[seqNum].rcmProbs2;
    }

    for (start=0;start<seqLength - width + 1;start++){
      if (seq[start+lastCol]<4){
	mProbs[start] *= pwm[PWM(lastCol,seq[start+lastCol])];
      }else{
	mProbs[start] *= 1.0/4;
      }
    }

    /* store probability of motif in opposite orientation, starting in same position */
    /* changed this one 3/27/06 */
    if (dataSet->revComp){
      for (start=0;start<seqLength - width + 1;start++){
	if (rcSeq[start+lastCol]<4){
	  rcmProbs[start] *= pwm[PWM(lastCol, rcSeq[start + lastCol])];
	}else{
	  rcmProbs[start] *= 1.0/4;
	}
      }
    }

  }

  return(1);
}


/**********************************************************************/
/*
  getOverMProbs
  
  Calculate motif probabilities for width W and oligo at position l from 
  the stored motif probabilities for width W+1 and oligo at position l+1, 
  which are stored either in mProbs1 or mProbs2.

  New motif probabilities are stored in the same array.

*/
/**********************************************************************/
int getOverMProbs(

		   DATASET *dataSet,  /* the dataset */
		   double *pwm,       /* the PWM for which to find the motif probabilities */
		   double *prevPwm,   /* the PWM for width+1 starting in the previous position */
		   int width,         /* the width of the current PWM */
		   int spaceNum,      /* store mProbs in mProbs1 or mProbs2 ? */
		   BOOLEAN oneWidth   /* is there only one candidate width? */
		   )
{

  double mProb;                 /* motif probability */
  int pos;                      /* index for looping over columns in motif */
  int firstCol;                 /* first position in motif */
  int lastCol;                  /* last position in motif */
  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int start;                    /* staring position of current motif site */
  int seqLength;                /* length of current sequence */
  int *seq;                     /* current sequence */
  int *rcSeq;                   /* reverse complement of current sequence */
  double *mProbs;               /* motif probabilities on forward strand */
  double *rcmProbs;             /* motif probabilities on reverse strand */


  if (spaceNum!=1 && spaceNum!=2){
    Rprintf("getHigherMProbs: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }

  numSeqs  = dataSet->numSeqs;
  lastCol  = width - 1;
  firstCol = 0;


  if (oneWidth){
    for (seqNum=0; seqNum<numSeqs; seqNum++){

      seq       = dataSet->samples[seqNum].seq;
      rcSeq     = dataSet->samples[seqNum].rcSeq;
      seqLength = dataSet->samples[seqNum].length;
      if (spaceNum==1){
	mProbs   = dataSet->samples[seqNum].mProbs1;
	rcmProbs = dataSet->samples[seqNum].rcmProbs1;
      }else{
	mProbs   = dataSet->samples[seqNum].mProbs2;
	rcmProbs = dataSet->samples[seqNum].rcmProbs2;
      }

      for (start = seqLength - width;start>0;start--){
	if ((seq[start+lastCol]<4) && (seq[start-1]<4)){
	  mProbs[start] = mProbs[start-1]*pwm[PWM(lastCol,seq[start+lastCol])]/prevPwm[PWM(firstCol,seq[start-1])];
	}else if (seq[start+lastCol]<4){
	  mProbs[start] = mProbs[start-1]*pwm[PWM(lastCol,seq[start+lastCol])]/(1.0/4);
	}else if (seq[start-1]<4){
	  mProbs[start] = mProbs[start-1]*(1.0/4)/prevPwm[PWM(firstCol,seq[start-1])];
	}else{
	  mProbs[start] = mProbs[start-1];
	}
      }

      /* do start=0 last */
      start = 0;
      mProb = 1;
      for (pos=0;pos<width;pos++){
	if (seq[start+pos]<4){
	  mProb *= pwm[PWM(pos, seq[start+pos])];
	}else{
	  mProb *= 1.0/4;
	}
      }
      mProbs[start] = mProb;


      /* store probability of motif in opposite orientation */
      if (dataSet->revComp){

	for (start = seqLength - width;start>0;start--){
	  if ((rcSeq[start+lastCol]<4) && (rcSeq[start-1]<4)){
	    rcmProbs[start] = rcmProbs[start-1]*pwm[PWM(lastCol,rcSeq[start+lastCol])]/prevPwm[PWM(firstCol,rcSeq[start-1])];
	  }else if (rcSeq[start+lastCol]<4){
	    rcmProbs[start] = rcmProbs[start-1]*pwm[PWM(lastCol,rcSeq[start+lastCol])]/(1.0/4);
	  }else if (rcSeq[start-1]<4){
	    rcmProbs[start] = rcmProbs[start-1]*(1.0/4)/prevPwm[PWM(firstCol,rcSeq[start-1])];
	  }else{
	    rcmProbs[start] = rcmProbs[start-1];
	  }
	}

	/* do start=0 last */
	start = 0;
	mProb = 1;
	for (pos=0;pos<width;pos++){
	  if (seq[start+pos]<4){
	    mProb *= pwm[PWM(pos, seq[start+pos])];
	  }else{
	    mProb *= 1.0/4;
	  }
	}
	rcmProbs[start] = mProb;

      }

      
    }
  }else{
    for (seqNum=0; seqNum<numSeqs; seqNum++){

      seq       = dataSet->samples[seqNum].seq;
      rcSeq     = dataSet->samples[seqNum].rcSeq;
      seqLength = dataSet->samples[seqNum].length;
      if (spaceNum==1){
	mProbs   = dataSet->samples[seqNum].mProbs1;
	rcmProbs = dataSet->samples[seqNum].rcmProbs1;
      }else{
	mProbs   = dataSet->samples[seqNum].mProbs2;
	rcmProbs = dataSet->samples[seqNum].rcmProbs2;
      }

      for (start = seqLength - width;start>0;start--){
	if (seq[start-1]<4){
	  mProbs[start] = mProbs[start-1]/prevPwm[PWM(firstCol,seq[start-1])];
	}else{
	  mProbs[start] = mProbs[start-1]/(1.0/4);
	}
      }

      /* do start=0 last */
      start = 0;
      mProb = 1;
      for (pos=0;pos<width;pos++){
	if (seq[start+pos]<4){
	  mProb *= pwm[PWM(pos, seq[start+pos])];
	}else{
	  mProb *= 1.0/4;
	}
      }
      mProbs[start] = mProb;


      /* store probability of motif in opposite orientation, starting in same position */
      if (dataSet->revComp){

	for (start = seqLength - width;start>0;start--){
	  if (rcSeq[start-1]<4){
	    rcmProbs[start] = rcmProbs[start-1]/prevPwm[PWM(firstCol,rcSeq[start-1])];
	  }else{
	    rcmProbs[start] = rcmProbs[start-1]/(1.0/4);
	  }
	}
	
	/* do start=0 last */
	start = 0;
	mProb = 1;
	for (pos=0;pos<width;pos++){
	  if (rcSeq[start+pos]<4){
	    mProb *= pwm[PWM(pos, rcSeq[start+pos])];
	  }else{
	    mProb *= 1.0/4;
	  }
	}
	rcmProbs[start] = mProb;
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  evalModLiks
  

  After having computed the motif probabilities, evaluate the likelihood
  function for the different model types and candidate values for the
  expected number of motif occurrences.

  Assigns the log-likelihood ratio to the arrays svLogLR.

*/
/**********************************************************************/
int evalModLiks(
		 DATASET *dataSet,      /* the dataset */
		 int width,             /* motif width */
		 int oligoSeqNum,       /* sequence from oligo was derived */
		 int oligoStart,        /* position in sequence at which oligo starts */
		 int spaceNum           /* are motif probs stored in mProbs1 or mProbs2? */
		 )

{

  double *svLogLR;         /* array of logLR for all possible starting values */
  double intensity;        /* intensity parameter */
  int numModType;          /* index of current model type */
  int minWidth;            /* minimum motif width */
  int maxWidth;            /* maximum motif width */
  int nnSites;             /* number of candidate values for the expected number of motif occurrences */
  int nSitesNum;           /* index of current candidate value */
  long maxSeqLength;       /* maximum sequence length */
  int seqNum;              /* index of current sequence */
  int numSeqs;             /* number of sequences */ 
  int numSites;            /* number of sites */
  MTYPE mType;             /* model type */

  if(spaceNum!=1 && spaceNum!=2){
    Rprintf("seqLR: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    return(0);
  }
    

  minWidth       = dataSet->minWidth;
  maxWidth       = dataSet->maxWidth;
  maxSeqLength   = dataSet->maxSeqLength;
  numSeqs        = dataSet->numSeqs;

  for (numModType=0; numModType< dataSet->numModTypes; numModType++){

    nnSites     = dataSet->svSpace[numModType].nnSites;
    svLogLR     = dataSet->svSpace[numModType].svLogLR;
    mType       = dataSet->mTypes[numModType];
 	
    for (nSitesNum=0;nSitesNum<nnSites;nSitesNum++){
      numSites = dataSet->svSpace[numModType].numSites[nSitesNum];
      intensity = getIntensity(mType, width, numSites, dataSet);

      for (seqNum=0;seqNum<numSeqs;seqNum++){
	svLogLR[LR(width,oligoSeqNum,oligoStart,nSitesNum)] += seqLLR(dataSet,seqNum,mType,intensity,width,spaceNum);
      }	  
    }
  }

  return(1);
}


/**********************************************************************/
/*
  copyMProbs
  

  Copy motif probabilities from mProbs1 to mProbs2 or the other way
  around.


*/
/**********************************************************************/
int copyMProbs(
		DATASET *dataSet,    /* the dataset */
		int fromSpace,       /* copy from this space */
		int toSpace          /* to this space */
		)
{

  int seqNum;          /* number of current sequence */
  int start;           /* position in sequence */
  int seqLength;       /* length of current sequence */
  double *mProbsFrom;  /* input motif probabilities */
  double *mProbsTo;    /* output motif probabilities */

  if (fromSpace!=1 && fromSpace!=2){
    Rprintf("copyMProbs: fromSpace must be 1 or 2. Got %d. Exiting ...\n", fromSpace);
    return(0);
  }

  if (toSpace!=1 && toSpace!=2){
    Rprintf("copyMProbs: toSpace must be 1 or 2. Got %d. Exiting ...\n", toSpace);
    return(0);
  }

  if (fromSpace==toSpace){
    Rprintf("copyMProbs: toSpace and fromSpace cannot be the same. Got %d for both. Exiting ...\n", toSpace);
    return(0);
  }


  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    seqLength  = dataSet->samples[seqNum].length;
    mProbsTo   = (toSpace==1) ? dataSet->samples[seqNum].mProbs1 : dataSet->samples[seqNum].mProbs2;
    mProbsFrom = (toSpace==1) ? dataSet->samples[seqNum].mProbs2 : dataSet->samples[seqNum].mProbs1;
    
    for (start=0; start<seqLength; start++){
      mProbsTo[start] = mProbsFrom[start];
    }
  }
  
  if(dataSet->revComp){
    for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
      seqLength  = dataSet->samples[seqNum].length;
      mProbsTo   = (toSpace==1) ? dataSet->samples[seqNum].rcmProbs1 : dataSet->samples[seqNum].rcmProbs2;
      mProbsFrom = (toSpace==1) ? dataSet->samples[seqNum].rcmProbs2 : dataSet->samples[seqNum].rcmProbs1;
      
      for (start=0; start<seqLength; start++){
	mProbsTo[start] = mProbsFrom[start];
      }
    }
  }

  return(1);
}



/**********************************************************************/
/*
  xLikStarts

  Extract starting values from pre-calculated logLR's of all the possible 
  starting values. 

  Starting values are written into the svTheta arrays. 
	
*/
/**********************************************************************/
int xLikStarts(
		DATASET *dataSet  /* dataset */
		)
{


  double *svLogLR;        /* array of logLR for all possible starting values broken down by validation samples */
  double *pwm;            /* space for a position weight matrix */
  double intensity;       /* intensity parameter */
  int modTypeNum;         /* index of current model type */
  int numModTypes;        /* number of model types */
  int width;              /* current width */
  int oligoSeqNum;        /* index of sequence from which oligo is derived */
  int oligoStart;         /* position in sequence at which oligo starts */
  int nnSites;            /* number of candidates for the expected number of motif occurrences */
  int nSitesNum;          /* index of current candidate */
  int *oligoSeq;          /* oligo sequence */
  MTYPE mType;            /* model type */
  int numSeqs;            /* number of sequences */
  SVSPACE *svSpace;       /* starting value space */
  int numSv;              /* number of starting values for each candidate value of expected motif occurrences */
  long seqLength;         /* length of sequence */
  int conSetNum;          /* number of current constraint set */
  int numConSets;         /* number of current constraint sets */
  MODEL *mod;             /* interpreted constraint set */
  double logLR;           /* log-likelihood ratio */
  THETA *svTheta;         /* starting value */
  int numSites;           /* numSites */


  numSeqs        = dataSet->numSeqs;
  pwm            = dataSet->pwm1;
  numModTypes    = dataSet->numModTypes;
  numConSets     = dataSet->numConSets;

  for (modTypeNum=0; modTypeNum<numModTypes; modTypeNum++){
    svSpace   = dataSet->svSpace + modTypeNum;
    nnSites   = svSpace->nnSites;
    mType     = svSpace->mType;
    numSv     = svSpace->numSv;
    svLogLR   = svSpace->svLogLR;

    for (width=dataSet->minWidth; width<=dataSet->maxWidth;width++){
      
      /* iterate over oligoSeqNum and oligoStart */
      for (oligoSeqNum=0;oligoSeqNum<numSeqs;oligoSeqNum++){
	
	
	seqLength   = dataSet->samples[oligoSeqNum].length;
	oligoSeq    = dataSet->samples[oligoSeqNum].seq;
	
	for (oligoStart=0;oligoStart<seqLength-width+1;oligoStart++){
	  
	  /* create PWM */		
	  seq2pwm(pwm, oligoSeq + oligoStart, width, dataSet->hiProb);
	  
	  for (conSetNum=0; conSetNum<numConSets; conSetNum++){
	    
	    /* set model for chosen width, modelType, and constraint set */
	    mod = dataSet->mod + MOD(width, modTypeNum, conSetNum);
	    
	    /* check constraints */
	    if (!satisfiesConstraints(pwm, width, mod)) continue;
	    
	    for (nSitesNum=0;nSitesNum<nnSites;nSitesNum++){
	      
	      numSites       =  svSpace->numSites[nSitesNum];
	      intensity      =  getIntensity(mType, width, numSites, dataSet);
	      logLR          =  svLogLR[LR(width,oligoSeqNum,oligoStart,nSitesNum)];


	      svTheta = svSpace->svTheta + SV(width, conSetNum, nSitesNum, 0);	      	     
	      if (COSMO_MSG_LEVEL>=3 && nSitesNum==nnSites-1) Rprintf("score = %lf vs. score3 = %lf\n", logLR, svTheta[numSv-1].score);
	      
	      if (logLR > svTheta[numSv-1].score){
		
		if (!isDup(pwm, width, svTheta, numSv)){

		  if (!insertSv(svTheta, numSv-1, pwm, logLR, width, numSites)) return(0);
		  qsort((void*) svTheta, numSv,sizeof(THETA), compTheta); 	    

		}
	      }
	    }
	  }
	}
      }
    }    
  }

  return(1);
}
  





 
