/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains functions for interacting with donlp2() to       *
*      maximize the likelihood function under constraints.                   *
*      Before calling donlp2() for the first time, a pointer to the dataset  *
*      must be copied to the static variable myDataSet by calling            *
*      init_NLP().                                                           *
*      When optimization routine donlp2() is called, starting value must be  *
*      stored in dataSet->svTheta and sequences must have their ".exclude"   *
*      entry updated correctly.
*      On exit, final parameter estimates will be stored in                  * 
*      dataSet->svTheta.                                                     *
*                                                                            *
*****************************************************************************/

#include "o8para.h"
#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "lstarts.h"  
#include "matLib.h"
#include "optimize.h"
#include "theta.h"
#include "donlp.h"
  
static DATASET *myDataSet;
static int *liste_loc;


/**********************************************************************/
/*

  donlpMalloc

  Allocate memory for donlp function

*/
/**********************************************************************/
int donlpMalloc(
		DATASET *dataSet
		) 
{

  /* problem dimension n = dim(x), nlin=number of linear constraints
     nonlin = number of nonlinear constraints  */

  MODEL *mod;         /* one candidate model */
  int modTypeNum;     /* number of current model */
  int width;          /* motif width */
  int conSetNum;      /* constraint set number */
  long iterma;        /* max valur for o8iterma */
  long nstep;         /* max value for 08nstep */
  long ndualm;        /* max value for o8ndualm */
  long mdualm;        /* max value for o8mdualm */

  dataSet->maxN      = 0;
  dataSet->maxNlin   = 0;
  dataSet->maxNonlin = 0;
  for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){

    /* loop over all widths */
    for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){

      /* loop over constraint sets */
      for (conSetNum=0; conSetNum<dataSet->numConSets; conSetNum++){
	
	/* set model for chosen width, modelType, and constraint set */
	mod = dataSet->mod + MOD(width, modTypeNum, conSetNum);
  
	dataSet->maxN      = MAX(dataSet->maxN, mod->n); 
	dataSet->maxNlin   = MAX(dataSet->maxNlin, mod->nlin);
	dataSet->maxNonlin = MAX(dataSet->maxNonlin, mod->nonlin);
      }
    }
  }

  iterma =  30 * dataSet->maxN; 
  nstep  =  20;

  ndualm = 2*dataSet->maxN + dataSet->maxNlin + dataSet->maxNonlin;
  mdualm = 2*(dataSet->maxN + dataSet->maxNlin + dataSet->maxNonlin);

  global_mem_malloc(dataSet->maxN, dataSet->maxNlin, dataSet->maxNonlin, iterma, nstep, ndualm, mdualm); 

  if (COSMO_MSG_LEVEL>=3) Rprintf( "maxN = %ld, maxNlin = %ld, maxNonlin = %ld\n", dataSet->maxN, dataSet->maxNlin, dataSet->maxNonlin);

  return(1);
}

 

/**********************************************************************/
/*
  user_init_size

  Initialize size of the problem by copying information out of
  dataSet->svTheta->conMod.

*/
/**********************************************************************/
void user_init_size(
		    void
		    ) 
{
#define  X extern
#include "o8comm.h"
#include "o8fint.h"
#include "o8cons.h"
#undef   X

  /* problem dimension n = dim(x), nlin=number of linear constraints
     nonlin = number of nonlinear constraints  */

  MODEL *mod;

  mod      =  myDataSet->svTheta->conMod; 
  o8n      =  mod->n; 
  o8nlin   =  mod->nlin;
  o8nonlin =  mod->nonlin;
  o8iterma =  30 * o8n; /* changed this one from 30*/
  o8nstep  =  20;

  if (COSMO_MSG_LEVEL>=3) Rprintf("n = %d, nlin = %d, nonlin = %d, mType = %d, intensity = %lf, iterma = %d\n", o8n, o8nlin, o8nonlin, myDataSet->svTheta->mType, myDataSet->svTheta->intensity, o8iterma);
  
}


/**********************************************************************/
/*
  user_init

  Standard setup for donlp2-intv.
  Copy starting values out of dataSet->svTheta.

*/
/**********************************************************************/
void user_init(
	       void
	       ) 
{
#define  X extern
#include "o8comm.h"
#include "o8fint.h"
#include "o8cons.h"
#undef   X
    
  static INTEGER  i,j;   /* indices */
  long pos;              /* position in motif */
  long nuc;              /* nucleotide */
  MODEL *mod;            /* interpreted constraint set */
  THETA *svTheta;        /* starting value */
  int width;             /* width of motif */
  int seqNum;            /* current sequence number */


  svTheta   = myDataSet->svTheta;
  mod       = svTheta->conMod;
  width     = svTheta->width;

  /* fx has not yet been called */
  myDataSet->fxUpdated = FALSE;

  /* name is ident of the example/user and can be set at users will       */
  /* the first static character must be alphabetic. 40 characters maximum */
  strcpy(o8name,"maximOut");
  o8silent = TRUE;

  /* x is initial guess and also holds the current solution */
  /* problem dimension n = dim(x), nlin=number of linear constraints
     nonlin = number of nonlinear constraints  */

  /* analytical gradients for all models except for TCM */
  o8analyt =  (svTheta->mType==TCM) ? FALSE : TRUE;

  o8epsdif = 1.e-16;   /* gradients exact to machine precision */
  /* if you want numerical differentiation being done by donlp2 then:*/
  o8epsfcn   = 1.e-16; /* function values exact to machine precision */

  o8taubnd   = 1.e-6; 
  /* bounds may be violated at most by taubnd in finite differencing */
  /*  bloc    = TRUE; */
  /* if one wants to evaluate all functions  in an independent process */
  o8difftype = 1; /* the most accurate and most expensive choice */
    
  o8nreset = o8n;
    
  o8del0 = 1e-2;   /* Fiddle with this one! */
  o8tau0 = 1e-2;   /* Fiddle with this one! */
  o8tau  = 0.1e0;

  liste_loc = IntMalloc(mod->n+1);

  /* Fill in starting value */
  i = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      o8x[i] = svTheta->pwm[PWM(pos,nuc)];
      i++;
    }
  }

  if (svTheta->mType!=OOPS && myDataSet->maxIntensity){
    o8x[i] = svTheta->intensity;
    i++;
  }


  /* profile shape parameters */
  for (j=0;j<mod->numProfParms;j++){
    o8x[i] = svTheta->profParms[j];
    if (COSMO_MSG_LEVEL >=4) Rprintf("x[%d] is ProfParm[%d] and = %lf\n", i,j,o8x[i]);
    i++;
  }

    
  /*  set lower and upper bounds */
  o8big = 1.e20;
  for (i=0;i<mod->n + mod->nlin + mod->nonlin;i++){
    o8low[i+1] = mod->low[i];
    o8up[i+1]  = mod->up[i];
  }

  if (COSMO_MSG_LEVEL>=3){
    Rprintf("bounds: low, up, start\n");
    PrintDoubleMatrix2File(stderr,o8low,1,mod->n+mod->nlin+mod->nonlin+1);
    Rprintf("\n");
    PrintDoubleMatrix2File(stderr,o8up,1,mod->n+mod->nlin+mod->nonlin+1);
    Rprintf("\n");
    PrintDoubleMatrix2File(stderr,o8x,1,mod->n+mod->nlin+mod->nonlin+1);
    Rprintf("\n");
  }

  /* store coefficients of linear constraints directly in gres */
  for ( i = 1 ; i<= o8nlin ; i++ ){ 
    for  ( j = 1 ; j <= o8n ; j++ ){
      o8gres[j][i] = mod->linConMatrix[LCM(i-1,j-1)];
    }
  }
  
  if (COSMO_MSG_LEVEL>=4){
    Rprintf("user_init: Matrix A:\n");
    for ( i = 1 ; i<= o8nlin ; i++ ){ 
      for  ( j = 1 ; j <= o8n ; j++ ){
	Rprintf("%d ", (int)o8gres[j][i]);
      }
      Rprintf("\n");
    }
  }

  
  if (COSMO_MSG_LEVEL>=3){
    for (seqNum=0; seqNum<myDataSet->numSeqs;seqNum++){
      Rprintf( "sequence %d exclude %d\n", seqNum, myDataSet->samples[seqNum].exclude);
    }
  }

  return;
}

/**********************************************************************/
/*
  setup

  Set special global parameters here.

*/
/**********************************************************************/
void setup(
	   void
	   ) 
{
#define  X extern
#include "o8comm.h"
#include "o8fuco.h"
#undef   X

  /* o8epsphi = 1e-10;*/  /* changed this one */
  o8te0=FALSE;       /* print detailed output? */
  o8te2=TRUE;
  o8te3=TRUE;

  o8numSinguls=0;
  return;
}

/**********************************************************************/
/*
  solchk

  At end of algorithm, copy final parameter estimates to 
  dataSet->svTheta. Check if algorithm converged or not.

*/
/**********************************************************************/
void solchk(
	    void
	    ) 
{
#define  X extern
#include "o8comm.h"
#undef   X
#include "o8cons.h"
   
  int k;          /* an index */
  int ppNum;      /* an index for looping over profile parameters */
  int pos;        /* position in motif */
  int nuc;        /* nucleotide */
  int width;      /* width of motif */
  THETA *svTheta; /* will contain the final parameter estimates on exit */
  MODEL *mod;     /* interpreted constraint set */

  svTheta   = myDataSet->svTheta;  
  width     = svTheta->width;
  mod       = svTheta->conMod;

  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      svTheta->pwm[PWM(pos,nuc)] = o8x[k];
      k++;
    }
  }

  if (svTheta->mType!=OOPS && myDataSet->maxIntensity){
    svTheta->intensity = o8x[k];
    k++;
  }


  /* profile shape parameters */
  for (ppNum=0;ppNum<mod->numProfParms;ppNum++){
    svTheta->profParms[ppNum] = o8x[k];
    if (COSMO_MSG_LEVEL >=4) Rprintf("x[%d] is ProfParm[%d] and = %lf\n", k,ppNum,o8x[k]);
    k++;
  }



  /* 
     if no feasible solution is found, algorithm returns o8fx = 0 
      o8optite gives the termination reason: 
      0,1,2:  good reasons 
      8:      newc termintated algorithm 
      3:      relaxed criteria met     
      -2:     no acceptable stepsize
      -3:     hit max iterations
  */

  svTheta->converged = (o8fx!=0.0) && ((o8optite>=0.0) || (o8optite==-2.0)) ;
  svTheta->logLR     = -o8fx;

  if (COSMO_MSG_LEVEL>=3){
    Rprintf("\n\noptimize:\n");
    Rprintf("after %d iterations, termination: %lf => conv = %d\n", o8itstep, o8optite, svTheta->converged);

    Rprintf("LLR = %e\n", -o8fx);
    PrintDoubleMatrix2File(stderr,svTheta->pwm,4,svTheta->width);    
  }

  if (!svTheta->converged & (COSMO_MSG_LEVEL>=3)){
    Rprintf( "donlp() failed to converge. Termination reason: %lf\n", o8optite);
  }

  if (!myDataSet->R) myfree(liste_loc);

  return;
}

/**********************************************************************/
/*
  ef

  Evaluate the objective function and assign value to *o8fx

*/
/**********************************************************************/
void ef(
	DOUBLE o8x[],
	DOUBLE *o8fx
	) 
{
#define  X extern
#include "o8fuco.h"
#undef   X
   

  int k;                /* an index */
  int nuc;              /* nucleotide */
  int pos;              /* position in motif */
  int seqNum;           /* number of current sequence */
  int numSeqs;          /* number of sequences */
  THETA *svTheta;       /* parameter estimates */
  double intensity;     /* intensity parameter */
  int width;            /* motif width */
  MTYPE mType;          /* model type */

  numSeqs     = myDataSet->numSeqs;
  svTheta     = myDataSet->svTheta;
  width       = svTheta->width;
  mType       = svTheta->mType;

  /* copy current parameter estimates into svTheta->pwm */
  k = 1; 
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      svTheta->pwm[PWM(pos,nuc)]=o8x[k];
      k++;
    }
  }

  if (svTheta->mType!=OOPS && myDataSet->maxIntensity){
    intensity = o8x[k];
    k++;;
  }else{
    intensity   = svTheta->intensity;
  }

  /* Calculate mProbs1 from scratch */
  getNewMProbs(myDataSet,svTheta->pwm,width,1);
  
  /* evaluate objective function */
  *o8fx = 0;    
  for (seqNum=0;seqNum<numSeqs;seqNum++){

    /* skip this sequence if necessary */
    if (myDataSet->samples[seqNum].exclude)  continue;

    *o8fx -= seqLLR(myDataSet, seqNum, mType, intensity, width, 1);


  }

  myDataSet->fxUpdated = TRUE;

  if (COSMO_MSG_LEVEL>=4) Rprintf( "fx = %e\n", *o8fx);
}

/**********************************************************************/
/*
  egradf

  Evaluate the gradient of objective function                     
	
  This function assumes that ef() was called earlier so that 
  contributions to likelihood function for each possible motif start
  site are stored in dataSet->seqProbs.

*/
/**********************************************************************/
void egradf(
	    DOUBLE o8x[],
	    DOUBLE o8gradf[]
	    ) 
{
#define  X extern
#include "o8fuco.h"
#undef   X
  
  int k;                /* an index */
  int nuc;              /* nucleotide */
  int pos;              /* position in motif */
  int seqNum;           /* number of current sequence */
  int cutNum;           /* number of current cut */
  int start;            /* motif start position */
  int first;            /* first postion in cut */
  int last;             /* last position in cut */
  int numSeqs;          /* number of sequences */
  THETA *svTheta;       /* parameter estimates */
  double intensity;     /* intensity parameter */
  int width;            /* motif width */
  MTYPE mType;          /* model type */
  int cutLength;        /* length of cut pieces */
  BOOLEAN revComp;      /* consider motifs on reverse complement? */
  double *posGrad;      /* gradient w.r.t. one entry of the PWM */
  SAMPLE *sample;       /* current sample */
  int *seq;             /* current sequence */
  int *rcSeq;           /* reverse complement of current sequence */
  long seqLength;       /* length of current sequence */
  int numCuts;          /* number of cuts in current sequence */
  double *seqProbs;     /* contribution to likelihood function of each possible motif start site on the forward strand */
  double *rcSeqProbs;   /* contribution to likelihood function of each possible motif start site on the reverse strand */
  double *cutLRs;       /* likelihood ratios for the separate cuts */
  double seqLR;         /* likelihood ratio for the whole sequence */
  double cutFac;        /* factor for integrating out over different possible motif start sites in cut model */
  double seqFac;        /* factor for integrating out over different possible motif start sites */

 
  numSeqs     = myDataSet->numSeqs;
  svTheta     = myDataSet->svTheta;
  width       = svTheta->width;
  mType       = svTheta->mType;
  revComp     = myDataSet->revComp;
  posGrad     = myDataSet->posGrad;

  /* copy parameter estimates into position weight matrix of svTheta */
  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      svTheta->pwm[PWM(pos,nuc)]=o8x[k];
      o8gradf[k] = 0;
      k++;
    }
  }

  if (svTheta->mType!=OOPS && myDataSet->maxIntensity){
    intensity = o8x[k];
    o8gradf[k] = 0;
    k++;;
  }else{
    intensity   = svTheta->intensity;
  }


  if(COSMO_MSG_LEVEL>=4){
    Rprintf("Entering gradf():\n");
    Rprintf("fx has been called: %d\n", myDataSet->fxUpdated);
    PrintDoubleMatrix2File(stderr,svTheta->pwm,4,svTheta->width);
    for (seqNum=0;seqNum<numSeqs; seqNum++){
      sample = myDataSet->samples + seqNum;
      seqProbs   = sample->seqProbs;
      seqLength  = sample->length;
      seqLR      = sample->seqLR;
      Rprintf("seq %d with seqLR = %lf\n", seqNum, seqLR);
    }
  }

  switch(mType){

  case OOPS:

    /* grad wrt pwm parm for X_i = -(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	seqLR      = sample->seqLR;

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;
	
	for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;
	
	/* factor for integrating over different starting positions and possible reverse complement */
	seqFac  = 1 / ((double) (1 + revComp)*(seqLength - width + 1));
	
	for (start=0;start<seqLength-width+1;start++){
	  nuc             = seq[start + pos];
	  if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	}
	
	if (revComp){
	  for (start=0;start<seqLength-width+1;start++){
	    nuc           = rcSeq[RS(start) + pos];
	    if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	  }
	}
	
	for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += seqFac*posGrad[nuc]/seqLR;
      }
    }
   

    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] /= svTheta->pwm[PWM(pos,nuc)];
      }
    }

    break;

  case ZOOPS:

    /* grad wrt pwm parm for X_i = -intensity*(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	seqLR      = sample->seqLR;

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;
	
	for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;
	
	/* factor for integrating over different starting positions and possible reverse complement */
	seqFac  = 1 / ((double) (1 + revComp)*(seqLength - width + 1));
	
	for (start=0;start<seqLength-width+1;start++){
	  nuc             = seq[start + pos];
	  if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	}
	
	if (revComp){
	  for (start=0;start<seqLength-width+1;start++){
	    nuc           = rcSeq[RS(start) + pos];
	    if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	  }
	}
	
	for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += seqFac*posGrad[nuc]/seqLR;
      }
    }
     
    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] *= intensity/svTheta->pwm[PWM(pos,nuc)];
      }
    }

    if (svTheta->mType!=OOPS && myDataSet->maxIntensity){

      for (seqNum=0;seqNum<numSeqs;seqNum++){

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;

	sample     = myDataSet->samples + seqNum;
	seqLR      = sample->seqLR;

	o8gradf[4*width+1] -= (seqLR - 1)/(seqLR*intensity);
      }
    }
    
    break;


  case OVER:

    cutLength   = myDataSet->cutLengths[width-myDataSet->minWidth];

    /* grad wrt pwm parm for X_i = -intensity*(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	numCuts    = sample->numCuts[width-myDataSet->minWidth];
	cutLRs     = sample->cutLRs;

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;

	for (cutNum=0;cutNum<numCuts;cutNum++){
		
	  for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;

	  /* be careful not to go over the end of the sequence here */		
	  first   = cutNum*cutLength;
	  last    = MIN((cutNum+1)*cutLength-1, seqLength - width);

	  if (last < first) continue;

	  /* factor for integrating over different starting positions and possible reverse complement */
	  cutFac  = 1 / ((double) (1 + revComp)*(last - first + 1));

	  for (start=first;start<=last;start++){
	    nuc             = seq[start + pos];
	    if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	  }

	  if (revComp){
	    for (start=first;start<=last;start++){
	      nuc           = rcSeq[RS(start) + pos];
	      if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	    }
	  }
		
	  for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += cutFac*posGrad[nuc]/cutLRs[cutNum];
				
	}
      }
    }

    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] *= intensity/svTheta->pwm[PWM(pos,nuc)];
      }
    }

    if (svTheta->mType!=OOPS && myDataSet->maxIntensity){

      for (seqNum=0;seqNum<numSeqs;seqNum++){

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;

	sample     = myDataSet->samples + seqNum;
	numCuts    = sample->numCuts[width-myDataSet->minWidth];
	cutLRs     = sample->cutLRs;

	for (cutNum=0;cutNum<numCuts;cutNum++){
	  o8gradf[4*width+1] -=  (cutLRs[cutNum] - 1)/(intensity *cutLRs[cutNum]);
	}

      }
    }


    break;

  case CUT:

    cutLength   = myDataSet->cutLengths[width-myDataSet->minWidth];

    /* grad wrt pwm parm for X_i = -intensity*(sum_(hasThisNuc) rcFac/numStart *mProb*backLR)/(LR*pwmVal) */
    for (pos=0;pos<width;pos++){
      for (seqNum=0;seqNum<numSeqs;seqNum++){

	sample     = myDataSet->samples + seqNum;
	seq        = sample->seq;
	rcSeq      = sample->rcSeq;
	seqLength  = sample->length;
	seqProbs   = sample->seqProbs;
	rcSeqProbs = sample->rcSeqProbs;
	numCuts    = sample->numCuts[width-myDataSet->minWidth];
	cutLRs     = sample->cutLRs;

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;

	for (cutNum=0;cutNum<numCuts;cutNum++){
		
	  for (nuc=0;nuc<4;nuc++) posGrad[nuc] = 0;

	  /* be careful not to go over the end of the sequence here */		
	  first   = cutNum*cutLength;
	  last    = MIN((cutNum+1)*cutLength - width + 1, seqLength - width);

	  if (last < first) continue;

	  /* factor for integrating over different starting positions and possible reverse complement */
	  cutFac  = 1 / ((double) (1 + revComp)*(last - first + 1));

	  for (start=first;start<=last;start++){
	    nuc             = seq[start + pos];
	    if (nuc<4) posGrad[nuc]   -= seqProbs[start];
	  }

	  if (revComp){
	    for (start=first;start<=last;start++){
	      nuc           = rcSeq[RS(start) + pos];
	      if (nuc<4) posGrad[nuc] -= rcSeqProbs[start];
	    }
	  }
		
	  for (nuc=0;nuc<4;nuc++) o8gradf[PWM(pos,nuc)+1] += cutFac*posGrad[nuc]/cutLRs[cutNum];
				
	}
      }
    }

    for (pos=0;pos<width;pos++){
      for (nuc=0;nuc<4;nuc++){
	if (svTheta->pwm[PWM(pos,nuc)]!=0) o8gradf[PWM(pos,nuc)+1] *= intensity/svTheta->pwm[PWM(pos,nuc)];
      }
    }

    if (svTheta->mType!=OOPS && myDataSet->maxIntensity){

      for (seqNum=0;seqNum<numSeqs;seqNum++){

	/* skip this sequence if necessary */
	if (myDataSet->samples[seqNum].exclude) continue;

	sample     = myDataSet->samples + seqNum;
	numCuts    = sample->numCuts[width-myDataSet->minWidth];
	cutLRs     = sample->cutLRs;

	for (cutNum=0;cutNum<numCuts;cutNum++){
	  o8gradf[4*width+1] -=  (cutLRs[cutNum] - 1)/(intensity *cutLRs[cutNum]);

	}

      }
    }


    break;

  default:
    Rprintf("gradf: mType %d not recognized. Exiting ...\n", mType);
  }

  if(COSMO_MSG_LEVEL>=4){
    Rprintf( "gradient:\n");
    PrintDoubleMatrix2File(stderr, o8gradf, 4,8);
  }


  myDataSet->fxUpdated = FALSE;

}

/**********************************************************************/
/*
  econ

  Compute the i-th equality constaint, value is hxi                 
	
*/
/**********************************************************************/
void econ(
	  INTEGER type,
	  INTEGER liste[], 
	  DOUBLE o8x[],
	  DOUBLE con[], 
	  LOGICAL err[]
	  ) 
{
#define  X extern
#include "o8fuco.h"
#undef   X

  int i,j,k;            /* various indices */
  int pos;              /* position in motif */
  int nuc;              /* nucleotide */
  int start;            /* motif start site */
  double prob;          /* probablity */
  int width;            /* width of motif */
  MODEL *mod;           /* interpreted constraints */
  double *pwm;          /* position weight matrix */
  double logTwo;        /* constant: log(2) */
  double eps;           /* small error tolerance */
  double subMin;        /* minimum submotif penalty for any window so far */
  double subScore;      /* submotif penalty for this window so far */
  double penalty;       /* submotif penalty for one position */
   


  width       = myDataSet->svTheta->width;
  mod         = myDataSet->svTheta->conMod;
  pwm         = myDataSet->svTheta->pwm;
  logTwo      = log(2);
  eps         = 1e-8;

  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      pwm[PWM(pos,nuc)]=o8x[k];
      k++;
    }
  }

  if (type==1){
    liste_loc[0] = mod->nonlin;
    if (COSMO_MSG_LEVEL>4) Rprintf("liste_loc[0] = %d\n", liste_loc[0]);
    for (i=1;i<=mod->nonlin;i++){
      liste_loc[i] = i;
    }
  }else{
    liste_loc[0] = liste[0];
    for (i=1;i<=liste[0];i++){
      liste_loc[i] = liste[i];
    }
  }

  for (j=1;j<=liste_loc[0];j++){
    i = liste_loc[j] - 1; /*back to zero-based arrays! */
    if (COSMO_MSG_LEVEL>4) Rprintf("econ: index i = %d\n",i);

    con[i+1] = 0;
    switch(mod->nlCons[i].pCon.conType){
    case bdd:

      /* 2 + sum(f_i*log(f_i)/log(2)) */
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[PWM(mod->nlCons[i].pCon.pos,nuc)];
	if (prob<eps) continue;
	con[i+1] += log(prob)*prob;
      }
      con[i+1] /= logTwo;
      con[i+1] += 2;
      break;

    case lin:
      /* 2 + sum(f_i*log(f_i)/log(2)) - (parm1 + (parm2-parm1)*xPos) */
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[PWM(mod->nlCons[i].pCon.pos,nuc)];
	if (prob<eps) continue;
	con[i+1] += log(prob)*prob;
      }
      con[i+1] /= logTwo;
      con[i+1] += 2;
      con[i+1] -= o8x[mod->nlCons[i].pCon.parm1 + 1] + (o8x[mod->nlCons[i].pCon.parm2 + 1] - o8x[mod->nlCons[i].pCon.parm1 + 1])*mod->nlCons[i].pCon.xPos;

      break;

    case sub:
      subMin = mod->nlCons[i].subCon.length;
      for (start=0;start< mod->width - mod->nlCons[i].subCon.length + 1; start++){
	subScore = 0;
	for (pos = 0; pos< mod->nlCons[i].subCon.length;pos++){
	  /* penalty = exp(-5*prob) */
	  penalty = exp(-5*pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  if (COSMO_MSG_LEVEL>=3) Rprintf("Adding penalty %lf for freq %lf\n",
			       penalty, pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  subScore += penalty;
	}
	if (COSMO_MSG_LEVEL>=3) Rprintf("subScore for position %d: %lf\n", start, subScore);
	subMin = MIN(subMin,subScore);
      }
      if (COSMO_MSG_LEVEL>=3) Rprintf("subMin = %lf\n", subMin);
      con[i+1] = subMin;		
      break;

    default:
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[PWM(mod->nlCons[i].pCon.pos,nuc)];
	if (prob<eps) continue;
	con[i+1] += log(prob)*prob;
      }	

      switch(mod->nlCons[i].pCon.ref){
      case prevPos:
	/* sum_i(f_ij*log(f_ij)/log(2)) - sum_i(f_i(j-1)*log(f_i(j-1))/log(2)) */
	for (nuc=0;nuc<4;nuc++){
	  prob = pwm[PWM(mod->nlCons[i].pCon.pos-1,nuc)];
	  if (prob<eps) continue;
	  con[i+1] -= log(prob)*prob;
	}
	con[i+1] /= logTwo;
	break;

      case prevBrk:
	/* 2 + sum_i(f_ij*log(f_ij)/log(2)) - parm1 */
	con[i+1] /= logTwo;
	con[i+1] += 2 - o8x[mod->nlCons[i].pCon.parm1 + 1];
	break;

      default:
	/* parm2 - (2 + sum_i(f_ij*log(f_ij)/log(2))) */
	con[i+1] /= logTwo;
	con[i+1] = o8x[mod->nlCons[i].pCon.parm2 + 1] -(con[i+1] + 2);
      }
    }
    
    if (COSMO_MSG_LEVEL>4) Rprintf("con[%d] = %lf\n",i+1,con[i+1]);
  }

  return;
}

/**********************************************************************/
/*
  econgrad

  Compute the gradient of the i-th equality constraint           
	
*/
/**********************************************************************/
void econgrad(
	      INTEGER liste[] ,
	      INTEGER shift ,  
	      DOUBLE o8x[],
	      DOUBLE **grad
	      ) 
{
#define  X extern
#include "o8fuco.h"
#undef   X
   
  int i,j,k;            /* various indices */
  int pos;              /* position in motif */
  int nuc;              /* nucleotide */
  int start;            /* motif start site */
  double prob;          /* probablity */
  int width;            /* width of motif */
  MODEL *mod;           /* interpreted constraints */
  double *pwm;          /* position weight matrix */
  double logTwo;        /* constant: log(2) */
  double eps;           /* small error tolerance */
  double small;         /* a small number */

  int minStart;         /* start site for submotif that incurs minimum penalty */
  double subMin;        /* minimum submotif penalty for any window so far */
  double subScore;      /* submotif penalty for this window so far */
  double penalty;       /* submotif penalty for one position */


  width       = myDataSet->svTheta->width;
  mod         = myDataSet->svTheta->conMod;
  pwm         = myDataSet->svTheta->pwm;
  logTwo      = log(2);
  eps         = 1e-8;
  small       = 1e-20;

  k = 1;
  for(pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      pwm[4*pos + nuc]=o8x[k];
      k++;
    }
  }

  liste_loc[0] = liste[0];
  for (i=1;i<=liste[0];i++){
    liste_loc[i] = liste[i];
  }
    
  for (j=1;j<=liste_loc[0];j++){
    i = liste_loc[j] - 1; /*back to zero-based arrays!*/
    if (COSMO_MSG_LEVEL>4) Rprintf("econgrad: index i = %d\n",i);

    /* Initialize gradient to zero */
    for (k=1;k<=mod->n;k++){
      grad[k][i+1+shift] = 0.0;
    }

	
    switch(mod->nlCons[i].pCon.conType){
    case bdd:
      /* grad wrt pwm parm at X_i = (log(pwmParm) + 1)/log(2)*/
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[4*mod->nlCons[i].pCon.pos + nuc];
	grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = prob<eps ? small : (log(prob) + 1)/logTwo;
      }
      break;

    case lin:
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[4*mod->nlCons[i].pCon.pos + nuc];
	grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = prob<eps ? small : (log(prob) + 1)/logTwo;
      }
      grad[mod->nlCons[i].pCon.parm1 + 1][i+1+shift] = -1 + mod->nlCons[i].pCon.xPos; 
      grad[mod->nlCons[i].pCon.parm2 + 1][i+1+shift] = -mod->nlCons[i].pCon.xPos; 
      break;

    case sub:

      /* First find in which window the minimum occurs */
      subMin = mod->nlCons[i].subCon.length;
      minStart = 0;
      for (start=0;start<mod->width - mod->nlCons[i].subCon.length + 1; start++){
	subScore = 0;
	if (COSMO_MSG_LEVEL>=3) Rprintf("egrad: start = %d\n", start);
	for (pos = 0; pos<mod->nlCons[i].subCon.length;pos++){
	  penalty = exp(-5*pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  if (COSMO_MSG_LEVEL>=3) Rprintf("egrad: penalty %lf for freq %lf\n",
			       penalty, pwm[PWM(start+pos,mod->nlCons[i].subCon.motif[pos])]);
	  subScore += penalty;
	}
	if (subScore<subMin){
	  subMin   = subScore;
	  minStart = start;
	}
      }

      /* Now get gradient as grad = -5*exp(-5*prob) */
      for (pos = 0; pos<mod->nlCons[i].subCon.length;pos++){
	grad[4*(minStart+pos) + mod->nlCons[i].subCon.motif[pos] + 1][i+1+shift] = 
	  -5*exp(-5*pwm[PWM(minStart+pos,mod->nlCons[i].subCon.motif[pos])]);
      }
      break;
		
    default:
      for (nuc=0;nuc<4;nuc++){
	prob = pwm[4*mod->nlCons[i].pCon.pos + nuc];
	grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = prob<eps ? small : (log(prob) + 1)/logTwo;
      }
		 
      switch(mod->nlCons[i].pCon.ref){
      case prevPos:
	for (nuc=0;nuc<4;nuc++){
	  prob = pwm[4*(mod->nlCons[i].pCon.pos-1) + nuc];
	  grad[4*(mod->nlCons[i].pCon.pos-1) + nuc + 1][i+1+shift] = prob<eps ? small : - (log(prob) + 1 )/logTwo;
	}
	break;

      case prevBrk:
	grad[mod->nlCons[i].pCon.parm1 + 1][i+1+shift] = -1; 
	break;

      default:
	for (nuc=0;nuc<4;nuc++){
	  grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift] = -grad[4*mod->nlCons[i].pCon.pos + nuc + 1][i+1+shift]; 		
	}
	grad[mod->nlCons[i].pCon.parm2 + 1][i+1+shift] = 1; 
      }
    }
	
    if (COSMO_MSG_LEVEL>4){
      for (k=0;k<mod->n;k++){
	Rprintf("%5.3lf ",grad[k+1][i+1+shift]);
      }
      Rprintf("\n");
    }

  }
	    
  return;
}

/**********************************************************************/
/*
  eval_extern

  User functions that is called if bloc==TRUE                     
	
*/
/**********************************************************************/
void eval_extern(
		 INTEGER mode
		 ) 
{
  return;
}



/**********************************************************************/
/*
  initNLP

  Get a pointer to the dataset and store in the static variable 
  myDataSet.
	
*/
/**********************************************************************/
void initNLP(
	     DATASET *dataSet  /* the dataset */
	     )
{
  myDataSet = dataSet;
}
