/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
*                                                                            *
*      This module contains various functions for computing posterior        * 
*      probabilities                                                         *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "lstarts.h"
#include "matLib.h"
#include "posterior.h"


/**********************************************************************/
/*
  getPostProbs

  Get posterior probabilities for a motif starting in each of the 
  eligible positions given the parameter estimates.

*/
/**********************************************************************/
int getPostProbs(
		  THETA *theta,     /* parameter estimates */
		  DATASET *dataSet  /* the dataset */
		  )
{

  int seqNum;            /* number of current sequence */
  int numSeqs;           /* number of sequences */
  long maxSeqLength;     /* maximum sequence length */
  int minWidth;          /* minimum motif width */
  long seqLength;        /* length of current sequence */
  int revComp;           /* consider reverse complement? */
  int cutNum;            /* index of current cut */
  int numCuts;           /* number of cuts in this sequence */
  int cutLength;         /* length of cut pieces */
  int first;             /* first position in this cut */ 
  int last;              /* last position in this cut */
  SAMPLE *sample;        /* the current sample */
  int *seq;              /* the current sequence */
  double pi;             /* intensity parameter in the ZOOPS model */
  double lambda;         /* intensity parameter in the TCM model */
  double *mProbs;        /* motif probabilities on forward strand */
  double *rcmProbs;      /* motif probabilities on reverse strand */
  double *seqProbs;      /* sequence probabilities */
  double *cumProb;       /* cumulative probabilities up to current position */
  double *cumProbBack;   /* cumulative probabilities under background model */
  double *backLR;        /* contributions of background to likelihood ratio */
  POSTPROB *postProbs;   /* posterior probabilities */
  int start;             /* position in sequence */
  double back;           /* likelihood under background model */
  double seqLR;          /* likelihood ratio for this sequence */
  double cutLR;          /* likelihood ratio for this cut */
  double *pwm;           /* position weight matrix */
  int width;             /* width of motif */
  MTYPE mType;           /* model type */
  double add;
  
  pwm           = theta->pwm;
  width         = theta->width;  
  mType         = theta->mType;
  maxSeqLength  = dataSet->maxSeqLength;
  minWidth      = dataSet->minWidth;
  numSeqs       = dataSet->numSeqs;
  revComp       = dataSet->revComp;
  cumProb       = dataSet->cumProb;
  postProbs     = dataSet->postProbs;
  cutLength     = dataSet->cutLengths[width-minWidth];


  /* calculate motif probabilities from scratch */
  getNewMProbs(dataSet, pwm, width, 1);

  switch (mType){
    
  case OOPS:

    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;
      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1;
      seqProbs      = sample->seqProbs;

      seqLR = 0;
      
      if (revComp){
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start] = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 	  
	  seqLR          += seqProbs[start];
	}
      }else{
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start]  = mProbs[start] * backLR[(width-minWidth)*seqLength + start];
	  seqLR           += seqProbs[start];
	}
      }
    
      seqLR /= seqLength - width + 1;

      for (start=0;start<seqLength-width+1;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = (seqProbs[start]/(seqLength - width + 1))/seqLR;
	postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;
      }

      for (start=seqLength-width+1;start<seqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

      for (start=seqLength;start<maxSeqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = -1;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

    }

    return(1);

  case ZOOPS:

    pi = theta->intensity;

    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;

      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      seqProbs      = sample->seqProbs;

      seqLR = 0;
      
      if (revComp){
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start]  = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 
	  seqLR           += seqProbs[start];
	}
      }else{
	for (start=0;start<seqLength-width+1;start++){
	  seqProbs[start]  = mProbs[start] * backLR[(width-minWidth)*seqLength + start];
	  seqLR           += seqProbs[start];
	}
      }

      seqLR = pi/(seqLength - width + 1) * seqLR + (1 - pi);
 
      for (start=0;start<seqLength-width+1;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = pi/(seqLength - width + 1) * seqProbs[start]/seqLR;
	postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;
      }

      for (start=seqLength-width+1;start<seqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

      for (start=seqLength;start<maxSeqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = -1;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

    }

    return(1);

  case TCM:

    lambda       = theta->intensity;
    cumProb      = dataSet->cumProb;

    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;

      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      cumProbBack   = sample->cumProbBack;


      for (start=0;start<seqLength+1;start++){
	cumProb[start] = 0;
      }
      
      cumProb[0] = 1;
      
      if (revComp){
	for (start=0;start<seqLength - width + 1;start++){
	  cumProb[start+1] += cumProb[start] * (1 - lambda);
	  
	  back = cumProbBack[CPB(width,start)];
	  cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;		
	}
      }else{
	for (start=0;start<seqLength - width + 1;start++){
	  cumProb[start+1] += cumProb[start] * (1 - lambda);
	  
	  back = cumProbBack[CPB(width,start)];
	  cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;		
	}
      }
      
      for (start=seqLength - width + 1;start<seqLength;start++){
	cumProb[start+1] += cumProb[start];
      }
      

      seqLR = cumProb[seqLength];

      for (start=0;start<seqLength-width+1;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	add = getCondTcmLik(start,seqNum,lambda,width,dataSet);
	if (COSMO_MSG_LEVEL>=3){
	  Rprintf( "seq %d start %d: add = %e and seqLR = %e -> %e\n", seqNum, start, add, seqLR, add/seqLR);
	}
	postProbs[PP(seqNum,start)].prob    = add/seqLR;
	postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;

     }

      for (start=seqLength-width+1;start<seqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = start;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }

      for (start=seqLength;start<maxSeqLength;start++){
	postProbs[PP(seqNum,start)].seqNum  = seqNum;
	postProbs[PP(seqNum,start)].start   = -1;
	postProbs[PP(seqNum,start)].prob    = 0;
	postProbs[PP(seqNum,start)].revComp = 0;
      }


    }

    return(1);


  case OVER:

    pi       = theta->intensity;

    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;

      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      seqProbs      = sample->seqProbs;
      numCuts       = sample->numCuts[width-minWidth];

      for (cutNum=0;cutNum <numCuts;cutNum++){
	
	/* cutLR here is the probability over one cut sequence */
	cutLR = 0;
	
	first = cutNum*cutLength;
	
	/* be careful not to go over the end of the sequence here */		    
	last = MIN((cutNum+1)*cutLength-1, seqLength - width + 1);
 
	/* changed this on 1/13/06 */
	/* if cutLength < width, we can't consider the last few cuts */	    
	if (last<first) continue;
	
	if (revComp){
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 
	    cutLR            += seqProbs[start]; 
	  }
	}else{
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = mProbs[start]* backLR[(width-minWidth)*seqLength + start];
	    cutLR            += seqProbs[start];        
	  }
	}
	
	cutLR  = pi/(last - first + 1) * cutLR + (1 - pi);
	
	for (start=first;start<=last;start++){
	  postProbs[PP(seqNum,start)].seqNum  = seqNum;
	  postProbs[PP(seqNum,start)].start   = start;
	  postProbs[PP(seqNum,start)].prob    = pi/(last - first + 1) * seqProbs[start]/cutLR;
	  postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;	  
	}

	if (last == seqLength - width + 1){
	  for (start=seqLength-width+1;start<seqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = start;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	  for (start=seqLength;start<maxSeqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = -1;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	}
      }
    }

    return(1);

  case CUT:

    pi       = theta->intensity;

    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;

      sample        = dataSet->samples + seqNum;
      seq           = sample->seq;
      seqLength     = sample->length;
      backLR        = sample->backLR;
      mProbs        = sample->mProbs1;
      rcmProbs      = sample->rcmProbs1; 
      seqProbs      = sample->seqProbs;
      numCuts       = sample->numCuts[width-minWidth];

      for (cutNum=0;cutNum <numCuts;cutNum++){
	
	/* cutLR here is the probability over one cut sequence */
	cutLR = 0;
	
	first = cutNum*cutLength;
	
	/* be careful not to go over the end of the sequence here */		    
	last = MIN((cutNum+1)*cutLength - width + 1, seqLength - width + 1);	
 
	/* changed this on 1/13/06 */
	/* if cutLength < width, we can't consider the last few cuts */	    
	if (last<first) continue;
	
	if (revComp){
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = 0.5*(mProbs[start] + rcmProbs[RS(start)])* backLR[(width-minWidth)*seqLength + start]; 
	    cutLR            += seqProbs[start]; 
	  }
	}else{
	  for (start=first; start <= last; start++){
	    seqProbs[start]   = mProbs[start]* backLR[(width-minWidth)*seqLength + start];
	    cutLR            += seqProbs[start];        
	  }
	}
	
	cutLR  = pi/(last - first + 1) * cutLR + (1 - pi);
	
	for (start=first;start<=last;start++){
	  postProbs[PP(seqNum,start)].seqNum  = seqNum;
	  postProbs[PP(seqNum,start)].start   = start;
	  postProbs[PP(seqNum,start)].prob    = pi/(last - first + 1) * seqProbs[start]/cutLR;
	  postProbs[PP(seqNum,start)].revComp = (revComp && (rcmProbs[RS(start)] > mProbs[start])) ? 1 : 0;	  
	}

	if (last == seqLength - width + 1){
	  for (start=seqLength-width+1;start<seqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = start;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	  for (start=seqLength;start<maxSeqLength;start++){
	    postProbs[PP(seqNum,start)].seqNum  = seqNum;
	    postProbs[PP(seqNum,start)].start   = -1;
	    postProbs[PP(seqNum,start)].prob    = 0;
	    postProbs[PP(seqNum,start)].revComp = 0;
	  }

	}
      }
    }

    return(1);
    
  default:
    Rprintf("getPostProbs: mType %d not recognized. Exiting ...\n", mType);
    return(0);
  }

}
 
/**********************************************************************/
/*
	getCondTcmLik

	Calculate likelihood ratio of one sequence under TCM model, 
	assuming that a motif starts at position motifStart

*/
/**********************************************************************/
double getCondTcmLik(
		     int motifStart,    /* motif start position */
		     int seqNum,        /* sequence for which to calculate conditional likelihood */
		     double lambda,     /* intensity parameter */
		     int width,         /* width of motif */
		     DATASET *dataSet   /* the dataset */
 		     )
{
  
  long maxSeqLength;     /* maximum sequence length */
  int minWidth;          /* minimum motif width */
  BOOLEAN revComp;       /* consider reverse complement? */
  SAMPLE *sample;        /* the current sample */
  int *seq;              /* the current sequence */
  long seqLength;        /* length of current sequence */
  double *mProbs;        /* motif probabilities on forward strand */
  double *rcmProbs;      /* motif probabilities on reverse strand */
  double *cumProb;       /* cumulative probabilities up to current position */
  double *cumProbBack;   /* cumulative probabilities under background model */
  int start;             /* position in sequence */
  double back;           /* likelihood under background model */


  maxSeqLength  = dataSet->maxSeqLength;
  minWidth      = dataSet->minWidth;
  revComp       = dataSet->revComp;
  cumProb       = dataSet->cumProb;

  sample        = dataSet->samples + seqNum;
  seq           = sample->seq;
  seqLength     = sample->length;
  mProbs        = sample->mProbs1;
  rcmProbs      = sample->rcmProbs1; 
  cumProbBack   = sample->cumProbBack;

  for (start=0;start<seqLength+1;start++){
    cumProb[start] = 0;
  }

  cumProb[0] = 1;

  if (revComp){

    for (start=0;start<motifStart - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;				
    }

    for (start=MAX(0, motifStart-width+1);start<motifStart;start++){
      cumProb[start+1] += cumProb[start]*(1-lambda);
    }

    back = cumProbBack[CPB(width,motifStart)];
    cumProb[motifStart+width] += 0.5*cumProb[motifStart]*lambda*(mProbs[motifStart]+rcmProbs[RS(motifStart)])/back;	 

    for (start=motifStart+width;start<seqLength - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;				
    }
    
    for (start=MAX(seqLength - width + 1,motifStart+width);start<seqLength;start++){
      cumProb[start+1] += cumProb[start];
    }

  }else{

    for (start=0;start<motifStart - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;		
    }

    for (start=MAX(0, motifStart-width+1);start<motifStart;start++){
      cumProb[start+1] += cumProb[start]*(1-lambda);
    }
    
    back = cumProbBack[CPB(width,motifStart)];
    cumProb[motifStart+width] += cumProb[motifStart]*lambda*mProbs[motifStart]/back; 

  
    for (start=motifStart+width;start<seqLength - width + 1;start++){
      
      cumProb[start+1] += cumProb[start]*(1-lambda);
      
      back = cumProbBack[CPB(width,start)];
      cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;				
    }
    
    for (start=MAX(seqLength - width + 1,motifStart+width);start<seqLength;start++){
      cumProb[start+1] += cumProb[start];
    }
  }

  return cumProb[seqLength];
    
}



/**********************************************************************/
/*
   compPostProbs

   Compare to postProbs structs on the basis of their posterior
   probabilities

*/
/**********************************************************************/
int compPostProbs(
		  const void *x,   /* first posterior probability struct */
		  const void *y    /* second posterior probability struct */
		 )
{

  POSTPROB *p1, *p2;

  p1 = (POSTPROB *) x;
  p2 = (POSTPROB *) y;

  if (p1->prob > p2->prob) return -1;
  if (p2->prob > p1->prob) return 1;
  return 0;
}

/**********************************************************************/
/*
  sortPostProbs
  
  Sort the posterior probabilites and find maxima.
  If global=TRUE, all motif proababilites are sorted, under the constraint
  that no consecutive W nucleotides may contain two maxima.
  If global=FALSE, the maximum motif probability of each sequence is
  found.

  Returns the number of maxima found.
 
*/
/**********************************************************************/
int sortPostProbs(
		  DATASET *dataSet,   /* the dataset */
		  int width,          /* width of motif */
		  BOOLEAN global      /* global sort or within sequences? */
		  )
{

  int seqNum;                   /* index of current sequence */
  int numSeqs;                  /* total number of sequences */
  int seqLength;                /* length of current sequence */

  int pos;                      /* index for looping over columns in motif */
  POSTPROB *postProbs;          /* array of posterior probabilities, one for each possible motif site */
  POSTPROB *maxPostProbs;       /* sorted maximum posterior probabilities */
  long maxSeqLength;            /* maximum sequence length */

  double prob;                  /* current probability */
  double maxProb;               /* current maximum probability */
  int lastPos;                  /* last position in sequence */
  int nextPos;                  /* next eligible position from last maximum */
  int numMax;                   /* number of maxima found so far */

  numMax        = 0;
  numSeqs       = dataSet->numSeqs;
  maxSeqLength  = dataSet->maxSeqLength;
  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;

  
  /* find maximum motif probabilities on all sequences such that no maxima are within W nucleotides of each other */
  if (global){

    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;
 
      seqLength = dataSet->samples[seqNum].length;
      lastPos   = seqLength - width;
      maxProb   = postProbs[PP(seqNum, 0)].prob;

      if (seqLength < width) continue;

      maxPostProbs[numMax].seqNum  = seqNum;
      maxPostProbs[numMax].start   = 0;
      maxPostProbs[numMax].prob    = maxProb;
      maxPostProbs[numMax].revComp = postProbs[PP(seqNum, 0)].revComp;
      
      nextPos                      = MIN(width,lastPos+1);
      
      for (pos=0; pos<=lastPos; pos++){
	prob = postProbs[PP(seqNum, pos)].prob;
	
	if (pos==nextPos) numMax++;


	if (pos==nextPos || prob > maxProb){
	  maxProb = prob;
	  maxPostProbs[numMax].seqNum  = seqNum;
	  maxPostProbs[numMax].start   = pos;
	  maxPostProbs[numMax].prob    = prob;
	  maxPostProbs[numMax].revComp = postProbs[PP(seqNum, pos)].revComp;

	  nextPos = MIN(pos+width,lastPos+1);

	  
	}
      }
      numMax++;

    }	
  }else{

    /* find maximum motif probabilities for each sequence separately */
    for (seqNum=0; seqNum<numSeqs; seqNum++){

      if (dataSet->samples[seqNum].exclude) continue;

      seqLength = dataSet->samples[seqNum].length;
      
      if (seqLength < width) continue;

      maxPostProbs[numMax].prob     = postProbs[PP(seqNum, 0)].prob;
      maxPostProbs[numMax].revComp  = postProbs[PP(seqNum, 0)].revComp;
      maxPostProbs[numMax].seqNum   = seqNum;
      maxPostProbs[numMax].start    = 0;	  

      for (pos=0; pos<seqLength - width + 1; pos++){

	if (postProbs[PP(seqNum, pos)].prob > maxPostProbs[numMax].prob){
	  maxPostProbs[numMax].prob        = postProbs[PP(seqNum, pos)].prob;
	  maxPostProbs[numMax].revComp     = postProbs[PP(seqNum, pos)].revComp;
	  maxPostProbs[numMax].seqNum      = seqNum;
	  maxPostProbs[numMax].start       = pos;	  

	}
      }	

      numMax++;
    }

  }

  if (COSMO_MSG_LEVEL>=4) Rprintf("sortPostProbs: numMax = %d and global = %d\n", numMax, global);
  
  /* finally sort the identified maxima */
  qsort((void*) maxPostProbs,numMax,sizeof(POSTPROB),compPostProbs);



  return numMax;

}

