/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/



/*****************************************************************************
*                                                                            *
*      This module contains various input/output functions for cosmo for     *
*      interfacing with R.                                                   *
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "rio.h"
#include "matLib.h"
#include "eval.h"
#include "io.h"
#include "posterior.h"

  
/**********************************************************************/
/*
  readSeqsFromVec

  Read sequences in from matrix supplied as input from R
  Sequences are passed in stacked in one long vector, stored in the 
  format A=1, C=2, G=3, T=4, with each sequence allotted the space
  of the longest sequence and additional entries filled with 0s.
  

*/
/**********************************************************************/
int readSeqsFromVec(
		    int *seqs,          /* vector that stores sequences */
		    DATASET *dataSet    /* the dataset */
		    )
{
  int seqNum;        /* number of current sequence */
  int numSeqs;       /* number of sequences */
  long pos;          /* position in sequence */
  long seqLength;    /* sequence length */
  long maxSeqLength; /* length of longest sequence */
  int *seq;          /* pointer to sequence */
  int minWidth;      /* minimum motif width to consider */
  int maxWidth;      /* maximum motif width to consider */
  SAMPLE *sample;    /* current sample */

  minWidth                = dataSet->minWidth;
  maxWidth                = dataSet->maxWidth;
  numSeqs                 = dataSet->numSeqs;
  maxSeqLength            = dataSet->maxSeqLength;

  /* allocate memory for samples */
  dataSet->samples = sampleMalloc(dataSet->numSeqs);


  /* get sequence lengths: sequence ends at first 0 entry in vector */
  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seq       = seqs + seqNum*maxSeqLength;
    pos       = 0;

    while (pos < maxSeqLength && seq[pos]!=0) pos++;

    if (pos == 0){
      Rprintf( "ERROR: Input sequence file is not in FASTA format. Sequence %d has length zero.\n", seqNum+1);
      return(0);
    }else{
      dataSet->samples[seqNum].length  = pos;
    }
  }

  if (!initSamples(dataSet)) return(0);

  /* read in sequences */
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    sample     = dataSet->samples + seqNum;
    seqLength  = sample->length;
    seq        = sample->seq;

    /* Sequences stored internally as 0-4 but read in as 1-5 */
    /* 4/5 is the unknown character 'X' */
    for (pos=0; pos<seqLength;pos++) seq[pos] = seqs[seqNum*maxSeqLength + pos] - 1;
  }

  /* get reverse complement if necessary */
  if (dataSet->revComp){
    for (seqNum=0;seqNum<dataSet->numSeqs;seqNum++){
      sample = &(dataSet->samples[seqNum]);

      for (pos=0;pos<sample->length;pos++){
	if (sample->seq[sample->length - pos - 1]<4){
	  sample->rcSeq[pos] = 3 - sample->seq[sample->length - pos - 1];
	}else{
	  /* unknown character 'X' */
	  sample->rcSeq[pos] = 4;
	}
      }
    }
  }

  if (dataSet->numBackSeqs == 0){
    /* get background frequencies */
    if (!getBackFreqs(dataSet, dataSet->samples, dataSet->numSeqs)) return(0);
    getIrange(dataSet);

    /* Initialize the probability tables for the objective function.  */
    /* Get for 2 and up because weighting may cause < min_nsites sites */
    init_llr_pv_tables(dataSet, MIN(2, dataSet->minNumSites), MAX(dataSet->zoopsMaxNumSites,dataSet->tcmMaxNumSites), dataSet->nucFreq);    
  }


 
  if (COSMO_MSG_LEVEL>=3){
    Rprintf("readSeqsFromVec:\n");
    Rprintf("numSeqs: %ld\n",dataSet->numSeqs);  
	
    for (seqNum=0;seqNum<dataSet->numSeqs;seqNum++){
      sample = &(dataSet->samples[seqNum]);

      Rprintf("Sequence %ld: ", seqNum);
      Rprintf("Length: %d\n", sample->length);
      PrintIntMatrix(sample->seq,1,sample->length);

      if (dataSet->revComp){
	Rprintf("reverse complement: ");
	PrintIntMatrix(sample->rcSeq,1,sample->length);
      }
    }
  }

  return(1);
}


/**********************************************************************/
/*
  readBackSeqsFromVec

  Read background sequences in from matrix supplied as input from R

  Sequences are passed in stacked in one long vector, stored in the 
  format A=1, C=2, G=3, T=4, with each sequence allotted the space
  of the longest sequence and additional entries filled with 0s.

*/
/**********************************************************************/
int readBackSeqsFromVec(
			int *seqs,           /* vector of sequences */
			int pval,            /* initialize p-value tables? */
			DATASET *dataSet     /* the dataset */			
			)
{
  int seqNum;           /* number of current sequence */
  int numSeqs;          /* number of background sequences */
  long pos;             /* position in current sequence */ 
  long seqLength;       /* length of sequence */
  long maxSeqLength;    /* length of longest background sequence */
  int *seq;             /* pointer to sequence */
  int minWidth;         /* minimum motif width to consider */
  int maxWidth;         /* maximum motif width to consider */
  SAMPLE *sample;       /* current sample */

  minWidth                = dataSet->minWidth;
  maxWidth                = dataSet->maxWidth;
  numSeqs                 = dataSet->numBackSeqs;
  maxSeqLength            = dataSet->maxBackSeqLength;

  /* allocate memory for samples */
  dataSet->backSamples = sampleMalloc(dataSet->numBackSeqs);

  /* get sequence lengths: sequence ends at first 0 entry in vector */
  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seq       = seqs + seqNum*maxSeqLength;
    pos       = 0;

    while (pos < maxSeqLength && seq[pos]!=0) pos++;

    dataSet->backSamples[seqNum].length  = pos;

    /* allocate memory for sequence */
    if (pos == 0){
      Rprintf( "ERROR: Background sequence file is not in FASTA format. Sequence %d has length zero.\n", seqNum+1);
      return(0);
    }else{
      dataSet->backSamples[seqNum].seq            = IntMalloc(pos);
    }
  }



  /* read in sequences */
  for (seqNum=0; seqNum<dataSet->numBackSeqs; seqNum++){
    sample     = dataSet->backSamples + seqNum;
    seqLength  = sample->length;
    seq        = sample->seq;

    /* Sequences stored internally as 0-4 but read in as 1-5 */
    for (pos=0; pos<seqLength;pos++) seq[pos] = seqs[seqNum*maxSeqLength + pos] - 1;
    
  }      

  if (pval){
    /* get background frequencies */
    if (!getBackFreqs(dataSet, dataSet->backSamples, dataSet->numBackSeqs)) return(0);
    getIrange(dataSet);
    
    /* Initialize the probability tables for the objective function.  */
    /* Get for 2 and up because weighting may cause < min_nsites sites */
    init_llr_pv_tables(dataSet, MIN(2, dataSet->minNumSites), MAX(dataSet->zoopsMaxNumSites,dataSet->tcmMaxNumSites), 
		       dataSet->nucFreq);       
  }

  return(1);
}

/**********************************************************************/
/*
  readTransMatFromVec

  Reads in transition matrix from vector as input from R.
  Transition matrices are stored stacked in one vector, in increasing
  order of the order. 


*/
/**********************************************************************/
int readTransMatFromVec(
			double *transMats, /* transition matrices */
			DATASET *dataSet   /* the dataset */
		       )
{
  int row;              /* row in transition matrix */
  int col;              /* column in transition matrix */
  int order;            /* current order */
  int subOrder;         /* lower order */
  long numRows;         /* number of rows altogether */
  long index;           /* index into matrix */
  double *transMat;     /* transition matrix */
  long addRows;         /* number of rows to add for previous orders */


  order = dataSet->order;

  dataSet->transMats = p2DoubleMalloc(order+1);
  for (subOrder=0; subOrder<=order; subOrder++) dataSet->transMats[subOrder] = DoubleMalloc(pow(4,subOrder+1));

  if (COSMO_MSG_LEVEL>=3) Rprintf("readTransMatFromVec: reading transition matrix for background Markov model with order = %d\n",  order);
  for (subOrder=0, addRows=0; subOrder<=order; subOrder++, addRows+=numRows){

    transMat = dataSet->transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      for (col=0; col<4; col++){
	index = 4*row + col;
	transMat[index] = transMats[4*addRows + index];

	/* transform to logs */
	transMat[index] = LOG(transMat[index]);
      }
    }

    if (COSMO_MSG_LEVEL>=3){
      Rprintf("transMat for order = %d\n", subOrder);
      PrintDoubleMatrix2File(stderr,transMat,4,numRows);
      Rprintf("\n");
    }
  }

  return(1);
}

/**********************************************************************/
/*
  output results to various vectors to be passed to R

*/
/**********************************************************************/
int output2vec(
	       DATASET *dataSet,      /* the dataset */
	       int *selOrder,         /* chosen order of background model */
	       double *klDivs,        /* Kullback-Leibler divergences for candidate orders of background model */
	       double *estTransMat,   /* estimated transition matrices */
	       double *selIntCrit,    /* intensity criterion of selected model */
	       double *wCrits,        /* width criteria of candidate models */
	       double *modCrits,      /* model criteria of candidate models */
	       double *conCrits,      /* constraint set criteria of candidate models */
	       int *wInit,            /* was the width criterion evaluated for this model? */
	       int *modInit,          /* was the model type criterion evaluated for this model? */
	       int *conInit,          /* was the constraint set criterion evaluated for this model? */
	       int *selWidth,         /* chosen motif width */
 	       int *selModel,         /* chosen model type */
	       int *selConSet,        /* chosen constraint set number */
	       int *selNumSites,      /* chosen number of motif occurrences */
	       int *numModels,        /* number of candidate models */
	       double *pwm,           /* estimated position weight matrix */
	       double *motProbs,      /* posterior probabilities of motif occurrences */
	       int *motRevComp,       /* is posterior probability greater on reverse strand? */
	       int *alignSeqs,        /* sequences in which the motifs can be found */
	       int *alignStarts,      /* starting positions of the the motifs */
	       double *eVal           /* the evalue of the alignment */
	       )
{
  THETA *theta;            /* parameter estimates */
  int pos;                 /* position */
  int nuc;                 /* nucleotide */
  int seqNum;              /* number of sequence */
  int index;               /* index */
  int i;                   /* index */
  int subOrder;            /* a particular order */
  POSTPROB *postProbs;     /* posterior probabilities */
  POSTPROB *maxPostProbs;  /* sorted maximal posterior probabilities */
  double *transMat;        /* transition matrix for given order */
  int addRows;             /* number of rows to add for previous transition matrices */
  int numRows;             /* number of rows in this transition matrix */
  int row;                 /* current row */
  int col;                 /* current column */
  int width;               /* motif width */
  long maxSeqLength;       /* maximum sequence length */
  BOOLEAN global;          /* do posterior probabilities need to be sorted globally or for each sequence separately? */
  int cSetNum;             /* constraint set number */
  int modTypeNum;          /* model type number */

  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;
  theta         = dataSet->theta;
  maxSeqLength  = dataSet->maxSeqLength;
  global        = theta->mType==TCM || theta->mType==OVER || theta->mType==CUT;

  *selOrder     = dataSet->order;
  *selWidth     = theta->width;
  *selModel     = dataSet->crits.mType + 1;
  *selConSet    = dataSet->crits.conSetNum;

  *selNumSites  = theta->numSites;
  *eVal         = exp(theta->logEval);
 
  /* background Kullback-Leibler divergences */
  for (i=0; i<=dataSet->maxOrder; i++) klDivs[i] = dataSet->backKL[i];

  /* intensity criterion */
  selIntCrit[0] = dataSet->crits.intCrit;
  if (dataSet->intCrit.crit==eval){
    selIntCrit[0] = exp(selIntCrit[0]);
  }else if (dataSet->intCrit.crit==lik){
    selIntCrit[0] *= -1;
  }

  /* model selection criteria */
  cSetNum = dataSet->hasConstraints ? 1 : 0;
  for (i=0; cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){

	index = MOD(width, modTypeNum, cSetNum);
	wInit[i] =  dataSet->allCrits[index].wInit;
	if (dataSet->wCrit.crit==eval){
	  wCrits[i]   =  dataSet->allCrits[index].wInit  ? exp(dataSet->allCrits[index].wCrit)     : 0;
	}else if (dataSet->wCrit.crit==lik){
	  wCrits[i]   = dataSet->allCrits[index].wInit   ? -dataSet->allCrits[index].wCrit     : 0;
	}else{
	  wCrits[i]   = dataSet->allCrits[index].wInit   ? dataSet->allCrits[index].wCrit     : 0;
	}
	
	modInit[i] =  dataSet->allCrits[index].modInit;
	if (dataSet->modCrit.crit==eval){
	  modCrits[i]   = dataSet->allCrits[index].modInit   ? exp(dataSet->allCrits[index].modCrit)     : 0;
	}else if (dataSet->modCrit.crit==lik){
	  modCrits[i]   = dataSet->allCrits[index].modInit   ? -dataSet->allCrits[index].modCrit     : 0;
	}else{
	  modCrits[i]   = dataSet->allCrits[index].modInit   ? dataSet->allCrits[index].modCrit     : 0;
	}
	
	conInit[i] =  dataSet->allCrits[index].conInit;
	if (dataSet->conCrit.crit==eval){
	  conCrits[i]   = dataSet->allCrits[index].conInit   ? exp(dataSet->allCrits[index].conCrit)     : 0;
	}else if (dataSet->conCrit.crit==lik){
	  conCrits[i]   = dataSet->allCrits[index].conInit   ? -dataSet->allCrits[index].conCrit     : 0;
	}else{
	  conCrits[i]   = dataSet->allCrits[index].conInit   ? dataSet->allCrits[index].conCrit     : 0;
	}
	i++;
      }
    }
  }
  *numModels    = i;

 
  /* print position weight matrix */
  for(nuc=0;nuc<4;nuc++){
    for(pos=0;pos<(*selWidth);pos++){
      /* Rprintf("Entering pos %d nuc %d of PWM\n", pos, nuc); */
      pwm[PWM(pos,nuc)] = theta->pwm[PWM(pos,nuc)];
    }
  }

  /* estimated transition matrices */
  for (subOrder=0, addRows=0; subOrder<=dataSet->order; subOrder++, addRows+=numRows){

    transMat = dataSet->transMats[subOrder];
    numRows  = pow(4,subOrder);

    for (row=0; row<numRows; row++){
      for (col=0; col<4; col++){
	estTransMat[4*(addRows+row) + col] = exp(transMat[4*row+col]);
      }
    }
  }

  
  /* get posterior probabilities */
  if (!getPostProbs(theta, dataSet)) return(0);

  index = 0;
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    for (pos=0; pos<maxSeqLength;pos++){
      motProbs[index] = postProbs[index].prob;
      motRevComp[index] = postProbs[index].revComp;
      index++;
    }
  }

  
  if (!sortPostProbs(dataSet, *selWidth, global)) return(0);
  for (index=0; index<*selNumSites;index++){
    alignStarts[index]  = maxPostProbs[index].start + 1;
    alignSeqs[index]    = maxPostProbs[index].seqNum + 1;
  }


  return(1);
}


