/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/



/*****************************************************************************
*                                                                            *
*      This module contains functions for evaluating certain summary         *
*      measures or criteria for a parameter estimate theta.                  *  
*                                                                            *
*****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "eval.h"
#include "lstarts.h"
#include "matLib.h"
#include "posterior.h"
#include "theta.h"
  

/**********************************************************************/
/*
  getAIC

  Calculate Akaike's Information Criterion from parameter estimates 
  and log-likelihood as

  -2*log-likelihood + 2*numberOfFreeParameters
  
  Assigns AIC to theta->AIC

*/
/**********************************************************************/
void getAIC(
	    THETA *theta,             /* parameter estimates */
	    DATASET *dataSet          /* the dataset */
	    )
{
  int numParms;   /* number of free parameters */
 
  numParms  = 3*theta->width + (theta->mType!=OOPS);

  theta->aic = -2*theta->logLik + 2*numParms;
}

/**********************************************************************/
/*
  getBIC

  Calculate Bayesian Information Criterion from parameter estimates 
  and log-likelihood as

  -2*log-likelihood + numberOfFreeParameters*log(number of observations)
  
  Assigns BIC to theta->BIC

*/
/**********************************************************************/
void getBIC(
	    THETA *theta,
	    DATASET *dataSet
	    )
{
  int nobs;       /* number of 'independent' observations */
  int numParms;   /* number of free parameters */
  int seqNum;     /* sequence number */
  

  numParms  = 3*theta->width + (theta->mType!=OOPS);

  nobs = 0;  
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++) if (!dataSet->samples[seqNum].exclude) nobs++;
  
  theta->bic = -2*theta->logLik + numParms*log(nobs);

}

/**********************************************************************/
/*
  getLLR
  
  Calculate log-likelihood ratio for given parameter estimates by 
  calculating motif probs from scratch.

  Assigns the log-likelihood ratio to theta->logLR.
  

*/
/**********************************************************************/
void getLLR(
	    THETA *theta,           /* parameter estimates */
	    DATASET *dataSet        /* the dataset */
	    )
{
  int width;              /* motif width */
  double *pwm;            /* position weight matrix estimates */
  int seqNum;             /* current sequence */
  int numSeqs;            /* number of sequences */
  double llr;             /* log likelihood ratio */
  MTYPE mType;            /* model type */
  double intensity;       /* intensity parameter */
  int numSites;           /* number of sites */

  numSeqs   = dataSet->numSeqs;
  width     = theta->width;
  pwm       = theta->pwm;
  mType     = theta->mType;
  intensity = theta->intensity;
  numSites  = theta->numSites;
  
  getNewMProbs(dataSet, pwm, width, 1);

  llr = 0;
  for (seqNum=0;seqNum<numSeqs;seqNum++){
    if (!dataSet->samples[seqNum].exclude)  llr += seqLLR(dataSet,seqNum,mType,intensity,width,1);
  }	  

  theta->logLR = llr;
  
}

/**********************************************************************/
/*
  getLogLik
  
  Calculate log-likelihood for given parameter estimates by 
  calculating motif probs from scratch.

  Assigns the log-likelihood to theta->logLik.
  

*/
/**********************************************************************/
void getLogLik(
	       THETA *theta,           /* parameter estimates */
	       DATASET *dataSet        /* the dataset */
	       )
{
  int width;              /* motif width */
  double *pwm;            /* position weight matrix estimates */
  int seqNum;             /* current sequence */
  int numSeqs;            /* number of sequences */
  double lik;             /* log likelihood */
  MTYPE mType;            /* model type */
  double intensity;       /* intensity parameter */
  SAMPLE *sample;         /* current sample */
  int numSites;           /* number of sites */

  numSeqs   = dataSet->numSeqs;
  width     = theta->width;
  pwm       = theta->pwm;
  mType     = theta->mType;
  intensity = theta->intensity;
  numSites  = theta->numSites;

  /* calculate motif probabilities from scratch */
  getNewMProbs(dataSet, pwm, width, 1);

  lik = 0;
  for (seqNum=0;seqNum<numSeqs;seqNum++){
    sample = dataSet->samples + seqNum;
    if (!dataSet->samples[seqNum].exclude){
      lik += seqLLR(dataSet,seqNum,mType,intensity,width,1) + sample->logCumBack[sample->length];
    }
  }	  

  theta->logLik = lik;
  
}



/**********************************************************************/
/*
  getLogEval
	

  Calculate the log-transformed E-value of the multiple alignment 
  corresponding to the estimated position weight matrix stored in 
  theta.


*/
/**********************************************************************/
void getLogEval(
		THETA *theta,      /* parameter estimates */
		DATASET *dataSet   /* the dataset */
)
{
  
  int width;                /* motif width */
  MTYPE mType;              /* model type */
  int pos;                  /* position in motif */
  int nuc;                  /* nucleotide */
  int start;                /* start site in sequence */
  int seqNum;               /* number of current sequence */
  int numSeqs;              /* number of sequences in dataset */
  long seqLength;           /* length of current sequence */
  int *site;                /* site that is to be aligned */
  double *motifCounts;      /* nucleotide counts in the alignment so far */
  double *backFreq;         /* nucleotide frequencies in the background */
  int siteNum;              /* index for looping over sites to be aligned */
  int numSites;             /* current number of aligned sites */
  POSTPROB *postProbs;      /* posterior probabilities to be sorted */
  POSTPROB *maxPostProbs;   /* sorted posterior probabilities */
  BOOLEAN onRevComp;        /* is site on the reverse complement? */
  BOOLEAN global;           /* sort motif probabilities over all sequences or separately for each sequence? */
  double log_pop;           /* log-transformed product of column p-values */
  double log_pv;            /* log-transformed p-value for column-specific hypothesis test */
  double log_nalign;        /* log number of possible alignments */
  int jj;                   /* index */

  numSeqs       = dataSet->numSeqs;
  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;
  motifCounts   = dataSet->motifCounts;
  backFreq      = dataSet->nucFreq;

  width         = theta->width;
  mType         = theta->mType;
  numSites      = theta->numSites;
  global        = mType==TCM || mType==OVER || mType==CUT;

  
  /* get posterior probabilities */
  getPostProbs(theta, dataSet);


  /* sort posterior probabilities */
  sortPostProbs(dataSet, width, global);

  /* initialize to zero */
  for (pos=0;pos<width;pos++){
    for (nuc=0;nuc<4;nuc++){
      motifCounts[PWM(pos,nuc)] = 0;
    }
  }



  /* create matrix of observed nucleotide counts 
     by aligning sites in order of their posterior probabilities */
  for (siteNum=0;siteNum<numSites;siteNum++){
    start      = maxPostProbs[siteNum].start;
    seqNum     = maxPostProbs[siteNum].seqNum;
    onRevComp  = maxPostProbs[siteNum].revComp;
    seqLength  = dataSet->samples[seqNum].length;
    site       = onRevComp ? dataSet->samples[seqNum].rcSeq + RS(start) : dataSet->samples[seqNum].seq + start;


    if (COSMO_MSG_LEVEL>=3){
      Rprintf("site %d with prob = %lf\n", siteNum+1, maxPostProbs[siteNum].prob);
      PrintIntMatrix2File(stderr,site, 1, width);
    }

    for (pos=0;pos<width;pos++){
      nuc = site[pos];
      if (COSMO_MSG_LEVEL>=3) Rprintf("nuc = %d\n",nuc);
      if (nuc<4){
	/* regular characters */
	motifCounts[PWM(pos,nuc)] ++;
      }else{
	/* unknown characters */
	for (jj=0; jj<4; jj++) motifCounts[PWM(pos,nuc)] += dataSet->nucFreq[jj];
      }
    }
  }
  if (COSMO_MSG_LEVEL>=4) PrintDoubleMatrix2File(stderr,motifCounts,4,width);

  /* get product of column p-values */
  log_pop = get_log_pop(dataSet, motifCounts, width, numSites, backFreq);
  
  /* added this 10/14/05 */
  /* p-value of product of p-values */
  log_pv = log_qfast(width, log_pop);	

  
  /* (log) number of possible alignments of this width and model type */
  log_nalign = get_log_nalign(mType, theta->width, theta->numSites, dataSet);
  
  if (COSMO_MSG_LEVEL>=4) Rprintf("log_pop = %lf, log_pv = %lf, log_nalign = %lf, logEval = %lf\n", 
			log_pop, log_pv, log_nalign, log_pv+log_nalign);


  /* added this 06/02/06 */
  theta->logEval = log_pv+log_nalign; 

}



  
/**********************************************************************/
/*
  euclDist

  Calculate the Euclidean distance between two PWM estimates

  Returns Euclidean distance 

*/
/**********************************************************************/
double euclDist(
		double *pwm1,            /* position weight matrix 1*/
		double *pwm2,            /* position weight matrix 2*/
		int width                /* motif width */ 
		)
{
  
  
  int pos;       /* position in the motif */
  int nuc;       /* nucleotide at that position */    
  double diff;   /* difference between nucleotide frequencies at a given position */
  double dist;   /* Euclidean distance between two PWMs */
  double col;

  dist = 0;
  if (COSMO_MSG_LEVEL>=3) Rprintf("\n"); 
  for (pos=0;pos<width;pos++){
    col = 0;
    for (nuc=0;nuc<4;nuc++){
      diff = pwm1[PWM(pos,nuc)] - pwm2[PWM(pos,nuc)];
      dist += diff*diff;
      col += diff*diff;
    }
    if (COSMO_MSG_LEVEL>=3) Rprintf("%4.3lf ", col); 

  } 
  if (COSMO_MSG_LEVEL>=3) Rprintf("\n"); 
  dist = sqrt(dist);
  return dist;
}


/**********************************************************************/
/*
	getICProfile

	calculate the information content profile for a given position
	weight matrix as

	IC(w) = sum_{j=1}^{4} p_{wj} * log_2(p_{wj})

	Assigns positional information content to theta->posIC and
	average information content to theta->IC.

*/
/**********************************************************************/
void getIcProfile(
		  THETA *theta  /* parameter estimates */
		  )
{

  double *pwm;        /* position weight matrix */
  int width;          /* width of PWM */
  double *posIC;      /* information content over the different positions */
  int pos;            /* position in motif */
  int nuc;            /* nucleotide */
  double IC;          /* information content */

  pwm   = theta->pwm;
  width = theta->width;
  posIC = theta->posIC;


  IC = 0;
  for (pos=0;pos<width;pos++){
    posIC[pos]=2;
    for(nuc=0;nuc<4;nuc++){
      posIC[pos] += pwm[PWM(pos,nuc)]==0 ? 0 : pwm[PWM(pos,nuc)]*LOG(pwm[PWM(pos,nuc)])/log(2);
    }
    IC += posIC[pos];
  }

  theta->IC = (IC/width);
}


/**********************************************************************/
/*
	getPwmKlDiv

	Calculates Kullback-Leibler divergence for colums of motif 
	between estimated position weight matrix and background
	distribution.

	Positional KL divergences are assigned to theta->posKL,
	average KL over the whole motif is assigned to theta->KL.

*/
/**********************************************************************/
void getPwmKlDiv(
		 THETA *theta,          /* parameter estimates */
		 double *backFreq       /* background frequencies */
		 )
{

  int pos;       /* position in motif */
  int nuc;       /* nucleotide */
  double KL;     /* Kullback-Leibler divergence */
  double *pwm;   /* position weight matrix */
  double *posKL; /* KL divergence at individual positions */
  int width;     /* width of motif */


  pwm   = theta->pwm;
  posKL = theta->posKL;
  width = theta->width;

  KL = 0;
  for (pos=0; pos < width; pos++){

    posKL[pos] = 0;
    for (nuc=0; nuc<4; nuc++){
      KL         += (pwm[PWM(pos,nuc)]!=0) ? pwm[PWM(pos,nuc)] * log(pwm[PWM(pos,nuc)]/backFreq[nuc]) : 0;
      posKL[pos] += (pwm[PWM(pos,nuc)]!=0) ? pwm[PWM(pos,nuc)] * log(pwm[PWM(pos,nuc)]/backFreq[nuc]) : 0;
    }
  }

  theta->KL  = KL /(double) width; 

}

/**********************************************************************/
/*
  seqLLR
  

  After having computed the motif probabilities, evaluate the likelihood
  function for a single sequence for a given model type.

  Contributions to the likelihood for each possible motif site are
  stored in the array dataSet->seqProbs for later calculation of 
  the gradient of the objective function.

  Returns the log-likelihood ratio.	
*/
/**********************************************************************/
double seqLLR(
	      DATASET *dataSet,   /* the dataset */
	      int seqNum,         /* number of sequence for which to evaluate likelihood */
	      MTYPE mType,        /* model type */
	      double intensity,   /* intensity parameter */
	      int width,          /* width of motif */
	      int spaceNum        /* are motif probs stored in mProbs1 or mProbs2? */
	      )
{


  int numSeqs;          /* number of sequences */
  long maxSeqLength;    /* maximum sequence length */
  int minWidth;         /* minimum width */
  long seqLength;       /* length of this sequence */
  int *seq;             /* sequence */
  int cutNum;           /* index of current cut */
  int numCuts;          /* number of cuts in this sequence */
  int cutLength;        /* length of cut pieces */
  int first;            /* first position in cut */
  int last;             /* last position in cut */

  double pi;            /* intensity parameter in ZOOPS model */
  double lambda;        /* intensity parameter in TCM model */
  double *mProbs;       /* motif probabilities of forward strand */
  double *rcmProbs;     /* motif probabilities of reverse complement */
  double *cumProb;      /* array of cumulative TCM probabilities */
  double *cumProbBack;  /* array of cumulative probabilities under background model */
  double *backLR;       /* contribution of background to likelihood ratio */

  int start;            /* index over starting positions */
  double back;          /* probability under background model */
  double lr;            /* likelihood ratio */
  double llr;           /* log-likelihood ratio */ 
  double cutProb;       /* probability of current cut */

  SAMPLE *sample;       /* current sample */
  double *seqProbs;     /* array of contributions to likelihood for each possible motif start site on forward strand */
  double *rcSeqProbs;   /* array of contributions to likelihood for each possible motif start site on reverse strand */
  double *cutLRs;       /* likelihood ratios for each of the separate cuts */
  BOOLEAN revComp;      /* consider motifs on the reverse complement strand? */


  if(spaceNum!=1 && spaceNum!=2){
    Rprintf("seqLR: spaceNum must be 1 or 2. Got %d. Exiting ...\n", spaceNum);
    exit(1);
  }
  
  sample        = dataSet->samples + seqNum;
  seq           = sample->seq;
  seqLength     = sample->length;
  maxSeqLength  = dataSet->maxSeqLength;
  minWidth      = dataSet->minWidth;
  numSeqs       = dataSet->numSeqs;
  mProbs        = (spaceNum==1) ? dataSet->samples[seqNum].mProbs1 : dataSet->samples[seqNum].mProbs2;
  rcmProbs      = (spaceNum==1) ? dataSet->samples[seqNum].rcmProbs1 : dataSet->samples[seqNum].rcmProbs2;
  seqProbs      = sample->seqProbs;
  rcSeqProbs    = sample->rcSeqProbs;
  cutLRs        = sample->cutLRs;
  revComp       = dataSet->revComp;
  backLR        = sample->backLR;

  
  switch (mType){
    
  case OOPS:

    lr = 0;

    if (revComp){
      for (start=0;start<seqLength-width+1;start++){
	seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	rcSeqProbs[start] = rcmProbs[RS(start)]*backLR[(width-minWidth)*seqLength + start];	

	lr += 0.5*(seqProbs[start] + rcSeqProbs[start]);
      }
    }else{
      for (start=0;start<seqLength-width+1;start++){
	seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	lr += seqProbs[start];
      }
    }


    lr /= seqLength - width + 1;
    sample->seqLR = lr;

    return LOG(lr);

  case ZOOPS:

    pi = intensity;
    lr = 0;

    if (revComp){
      for (start=0;start<seqLength-width+1;start++){
	seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
 	rcSeqProbs[start] = rcmProbs[RS(start)]*backLR[(width-minWidth)*seqLength + start];	
	lr += 0.5*(seqProbs[start] + rcSeqProbs[start]);
      }
    }else{
      for (start=0;start<seqLength-width+1;start++){
	seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	lr += seqProbs[start];
      }
    }
    lr /= seqLength - width + 1;
    sample->seqLR = pi*lr + (1-pi);

    return LOG(pi*lr + (1-pi));

  case TCM:

    lambda       = intensity;
    cumProb      = dataSet->cumProb;
    cumProbBack  = sample->cumProbBack;

    for (start=0;start<seqLength+1;start++){
      cumProb[start] = 0;
    }
    
    cumProb[0] = 1;

    if (revComp){
      for (start=0;start<seqLength - width + 1;start++){
	cumProb[start+1] += cumProb[start] * (1 - lambda);
	
	back = cumProbBack[CPB(width,start)];
	cumProb[start+width] += 0.5*cumProb[start]*lambda*(mProbs[start]+rcmProbs[RS(start)])/back;		
      }
    }else{
      for (start=0;start<seqLength - width + 1;start++){
	cumProb[start+1] += cumProb[start] * (1 - lambda);
	
	back = cumProbBack[CPB(width,start)];
	cumProb[start+width] += cumProb[start]*lambda*mProbs[start]/back;		
      }
    }

    for (start=seqLength - width + 1;start<seqLength;start++){
      cumProb[start+1] += cumProb[start];
    }

    return LOG(cumProb[seqLength]);    

  case OVER:

    pi         = intensity;
    numCuts    = sample->numCuts[width-dataSet->minWidth];
    cutLength  = dataSet->cutLengths[width-dataSet->minWidth];

    llr = 0;
    for (cutNum=0;cutNum <numCuts;cutNum++){
      
      /* cutProb here is the probability over one cut sequence */
      cutProb = 0;
      
      first = cutNum*cutLength;
      
      /* be careful not to go over the end of the sequence here */		    
      last = MIN((cutNum+1)*cutLength-1, seqLength - width);
      
      /* if cutLength < width, we can't consider the last few cuts */	    
      if (last<first) continue;      

      
      if (revComp){
	for (start=first; start <= last; start++){
	  seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	  rcSeqProbs[start] = rcmProbs[RS(start)]*backLR[(width-minWidth)*seqLength + start];	

	  cutProb += 0.5*(seqProbs[start] + rcSeqProbs[start]);
	  
	}
      }else{
	for (start=first; start <= last; start++){
	  seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	  cutProb += seqProbs[start];
	}
      }

      cutProb         /= (double)(last-first+1);

      cutLRs[cutNum]   = pi*cutProb + (1-pi);
      llr             += LOG(pi*cutProb + (1-pi));  
    }

    
    return llr;

  case CUT:

    pi         = intensity;
    numCuts    = sample->numCuts[width-dataSet->minWidth];
    cutLength  = dataSet->cutLengths[width-dataSet->minWidth];

    llr = 0;
    for (cutNum=0;cutNum <numCuts;cutNum++){
      
      /* cutProb here is the probability over one cut sequence */
      cutProb = 0;
      
      first = cutNum*cutLength;
      
      /* be careful not to go over the end of the sequence here */		    
      last = MIN((cutNum+1)*cutLength - width + 1, seqLength - width);
      
      /* if cutLength < width, we can't consider the last few cuts */	    
      if (last<first) continue;
      
      if (revComp){
	for (start=first; start <= last; start++){
	  seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	  rcSeqProbs[start] = rcmProbs[RS(start)]*backLR[(width-minWidth)*seqLength + start];	

	  cutProb += 0.5*(seqProbs[start] + rcSeqProbs[start]);

	}
      }else{
	for (start=first; start <= last; start++){
	  seqProbs[start]   = mProbs[start]*backLR[(width-minWidth)*seqLength + start];
	  cutProb += seqProbs[start];
	}
      }

      cutProb         /= (double)(last-first+1);
      cutLRs[cutNum]   = pi*cutProb + (1-pi);
      llr             += LOG(pi*cutProb + (1-pi));  
    }

    return llr;
    
  default:
    Rprintf("optimSeqLLR: mType %d not recognized. Exiting ...\n", mType);
    exit(1);
  }

}


/**********************************************************************/
/*
  copyTheta
  
  Copy one THETA struct into another THETA struct.

*/
/**********************************************************************/
void copyTheta(
	       THETA *from,    /* copy from this struct */
	       THETA *to       /* into this struct */
	       ) 
{

  int pos, nuc, ppNum;

  to->mType          = from->mType;
  to->width          = from->width;
  to->intensity      = from->intensity;
  to->numSites       = from->numSites;
  to->numProfParms   = from->numProfParms;
  to->converged      = from->converged;

  to->logLR          = from->logLR;
  to->logLik         = from->logLik;
  to->KL             = from->KL;
  to->IC             = from->IC;
  to->logEval        = from->logEval;
  to->aic            = from->aic;
  to->bic            = from->bic;
  to->score          = from->score;

  to->conSetNum      = from->conSetNum;
  to->conMod         = from->conMod;


  /* copy PWM */
  for (pos=0; pos< from->width; pos++){
    for (nuc=0; nuc<4; nuc++){
      to->pwm[PWM(pos,nuc)] = from->pwm[PWM(pos,nuc)];
    }
  }

  /* copy ic profile */
  for (pos=0; pos< from->width; pos++) to->posIC[pos] = from->posIC[pos];

  /* copy KL profile */
  for (pos=0; pos< from->width; pos++) to->posKL[pos] = from->posKL[pos];

  /* copy profile parmaters */
  for (ppNum=0; ppNum<from->numProfParms; ppNum++) to->profParms[ppNum] = from->profParms[ppNum];

}




/**********************************************************************/
/*
  getKlDiv
  
  Calculate Kullback-Leibler divergence for given parameter estimates.
  Only sequences in given validation group will be used.
  If desired KL divergence will be truncated at given quantile.
  This quantile is estimated as the corresponding quantile among
  the negative log-likelihood values among all sequences.

  The sequences must have the ".exclude" entry properly updated. This
  must be set to TRUE for sequences in the training sample and to FALSE
  for sequences in the validation sample.

  Returns KL divergence.

*/
/**********************************************************************/
double getKlDiv(
		DATASET *dataSet,   /* the dataset */
		THETA *theta,       /* parameter estimates */
		BOOLEAN trunc,      /* truncate KL divergence? */
		double quant        /* truncate to this quantile \in [0,1] */
		)
{

  double klDiv;        /* Kullback-Leibler divergence */
  int numValSeqs;      /* number of sequences in validation sample */
  int numSeqs;         /* number of sequences */
  int seqNum;          /* current sequence number */
  int width;           /* motif width */
  int numSites;        /* number of sites */
  MTYPE mType;         /* model type */
  double intensity;    /* intensity parameter */
  double loss;         /* value of loss function */
  SAMPLE *sample;      /* current sample */
  double bound;        /* bound by which to truncate */
  double *logLiks;     /* array of log likeihoods */

  numSeqs     = dataSet->numSeqs;
  mType       = theta->mType;
  width       = theta->width;
  intensity   = theta->intensity;
  numSites    = theta->numSites;
  logLiks     = dataSet->logLiks;


  /* calculate motifProbs from scratch */
  getNewMProbs(dataSet,theta->pwm, width, 1);
  

  /* get truncation threshold as quantile among log-likelihoods of ALL sequences */

  /* compute all sequence log-likelihoods */
  for (seqNum=0;seqNum<numSeqs;seqNum++){
    sample = dataSet->samples + seqNum;
    logLiks[seqNum] = seqLLR(dataSet, seqNum, mType, intensity, width, 1) + sample->logCumBack[sample->length]; 

    if (COSMO_MSG_LEVEL>=4) Rprintf("logLik for seq %d = %lf and mType = %d and intensity = %lf and logCumBack = %lf\n", 
			  seqNum, logLiks[seqNum], mType, intensity, sample->logCumBack[sample->length]);
  }
  
  /* sort all log-likelihoods */
  qsort((void*) logLiks ,numSeqs,sizeof(double),compDouble);

  /* get truncation quantile */
  bound =  - logLiks[MAX((int) (quant*numSeqs)-1,0)];
  if (COSMO_MSG_LEVEL>=4) Rprintf("getKlDiv: threshold = %lf\n", bound);

  
  
  /* compute truncated KL divergence */
  klDiv       = 0;
  numValSeqs  = 0;
  for (seqNum=0;seqNum<numSeqs;seqNum++){
    sample = dataSet->samples + seqNum;

    if (sample->exclude){
      if (COSMO_MSG_LEVEL>=4) Rprintf("getKlDiv: skipping %d\n", seqNum);
      continue;
    }
 
    /* loss for one sequence */
    loss = - seqLLR(dataSet, seqNum, mType, intensity, width, 1) - sample->logCumBack[sample->length]; 
    if (COSMO_MSG_LEVEL>=4) Rprintf("loss for seq %d = %lf and mType = %d and intensity = %lf\n", seqNum, loss, mType, intensity);

    /* truncate if necessary */
    if (trunc & (loss > bound)){
      if (COSMO_MSG_LEVEL>=3) Rprintf("\nloss for seq %d = %lf: truncated to %lf\n", seqNum, loss, bound);
      loss = bound;
    }

    klDiv += loss;
    numValSeqs ++;
  }

  if (COSMO_MSG_LEVEL>=3) Rprintf( "klDiv = %lf, nseqs = %d\n", klDiv, numValSeqs);

  /* return average loss function */
  return klDiv/numValSeqs;
}



/**********************************************************************/
/*
  compDouble
  
  Compare two doubles. 
  Returns -1 if first double is greater
           1 if first double is less
	   0 otherwise

*/
/**********************************************************************/
int compDouble(
	       const void *x,  /* first double */
	       const void *y   /* second double */
	       )
{

  const double *d1;
  const double *d2;

  d1 = (const double *) x;
  d2 = (const double *) y;

  if (*d1 > *d2) return -1;
  if (*d2 > *d1) return 1;
  return 0;
}


/**********************************************************************/
/*
  get_log_pop

  Calculate product of column p-values for hypotheses that columns
  each are generated from background rather than motif.

  Return product of p-values.


*/
/**********************************************************************/
double get_log_pop(
		   DATASET *dataSet,         /* the dataset */
		   double *motifCounts,      /* the nucleotide counts in the alignment */
		   int width,                /* the width of the alignment */
		   int numSites,             /* the number of sites in the alignment */
		   double *backFreq          /* the four background letter frequencies */
		   )
{

  double log_pop;   /* log p-value of product */
  double log_pv;    /* log-transformed p-value for column-specific hypothesis test */
  double llr;       /* log likelihood ratio */
  int pos;          /* position in PWM */
  int nuc;          /* nucleotide at given position */
  double f;         /* estimated nucleotide frequency from PWM */
  double p;         /* background nucleotide frequency */

  /* compute log likelihood ratio */
  log_pop = 0;				
  for (pos=0;pos<width;pos++){
    
    /* compute log likelihood for position pos */
    llr = 0;
    for (nuc=0;nuc<4;nuc++){
      f = (motifCounts[PWM(pos,nuc)])/((double) numSites);
      p = backFreq[nuc];
      llr += (f&&p) ? f *(log(f) - log(p)) : 0; 
    }
    llr *= numSites;				        /* convert entropy to ll */ 
    RND(llr, RNDDIG, llr);       	       	        /* round to RNDDIG places */

    log_pv = get_llr_pv(dataSet, llr, numSites, 1, LLR_RANGE, 1.0, 4, backFreq); 
    
    if (COSMO_MSG_LEVEL>=4) Rprintf("column p-value:%lf\n",log_pv);
    log_pop += log_pv;	

  }
  
  return(log_pop);

}
