#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include "o8para.h"              

void user_eval(DOUBLE xvar[],INTEGER mode) {
/* ************************************************************************ */
/* called only if bloc = TRUE                                               */
/* interface to a users evaluation code of functions                        */
/* which define the optimization problem                                    */
/* mode = -1 : evaluate function values at xvar only, store them in fu      */
/* mode = 1 : evaluate f , grad(f), con[], grad(con[]) and store            */
/*             them in  fu and fugrad                                       */
/* mode = 2  : evaluate gradients only at xvar and store them in fugrad     */
/* The users evaluation routine                                             */
/*             eval_extern(mode)                                            */
/* is called with two settings of its mode only:                            */
/* on one variable only (gunit[1][i] = 1).                                  */
/* mode = 1: set function value                                             */
/* fu using xtr, mode = 2 set function values fu and gradients fugrad using */
/* xtr                                                                      */
/* Since donlp2 uses internally a scaled value of x, the external value xtr */
/* is obtained by multiplying with xsc. In evaluating the gradients,        */
/* the user must not take care of this scaling. Scaling of the gradients    */
/* is done by donlp2 internally.                                            */
/* that means xtr and the function values returned below to the unscaled    */
/* original problem                                                         */
/* If the user has set analyt = TRUE, he is responsible for computing       */
/* the gradients himself correctly                                          */
/* otherwise, if analyt = FALSE, then this routine does numerical           */
/* differentiation according to the method defined by difftype:             */
/* difftype = 1   ordinary forward difference quotient, using               */
/*                min(0.1*sqrt(epsfcn),1.e-2) as stepsize                   */
/*                epsfcn is the assumed precision of function values        */
/*                and must be set by the user in setup0                     */
/* difftype = 2   central difference quotient with stepsize                 */
/*                min(0.1*pow(epsfcn,2/3),1.e-2) as stepsize                */
/* difftype = 3   6-th order Richardson extrapolation method using          */
/*                6 function values for each component of grad,             */
/*                with stepsize min(0.1*pow(epsfcn,1/7),0.01)               */
/* This is done by several calls of eval_extern                             */
/* xtr is the current design to be used by the external routine             */
/* Values must be returned in the arrays fu (and fugrad).                   */
/*  The order of information is as follows:                                 */
/* fu[0]  objective function value                                          */
/* fugrad[i][0] : objective function gradient, i = 1,...,o8n the components   */
/* fu[1],...,fu[nonlin]  the values of c  (if any)                          */
/* fugrad[i][1],.., fugrad[i][nonlin] gradients of  constraint functions c  */
/* i=1,..,o8n                                                                 */
/* bound constraints and linear constraints must not be evaluated here      */
/* The method uses the parameters                                           */
/*      epsfcn  =  relative accuracy of function values                     */
/*      taubnd  =  amount by which bound constraints may be violated        */
/*                 during finite differencing                               */
/*      difftype   (see above)                                              */
/*      analyt     (see above)                                              */
/* These must be set prior in user_init                                     */
/* ************************************************************************ */
    
    #define  X extern
    #include "o8comm.h"
    #include "o8fint.h"
    #undef   X
    #include "o8cons.h"
    
    void    eval_extern(INTEGER mode);
    DOUBLE* d1_malloc(INTEGER size1, INTEGER init); 

    static DOUBLE   *fusave,xhelp,xincr,sd1,sd2,sd3,d1,d2,d3;
    static INTEGER  i,j;
    static LOGICAL  changed;
    static INTEGER  firstTime = 1;
    
    if (firstTime) {
        firstTime = 0; 
	fusave = d1_malloc(o8nlin+o8nonlin+1, 1);
    }	
	
    
    if ( mode < -1 || mode  > 2 || mode == 0 ) { 
        Rprintf("donlp2: call of user_eval with undefined mode\n");
        exit(1);
    }
    if ( mode == -1 ) {
        changed = FALSE;
        for (i = 1 ; i <= o8n ; i++) {
            if ( o8xtr[i] != o8xsc[i]*xvar[i] ) changed = TRUE;
        }
        if ( changed || ! o8valid ) {
            for (i = 1 ; i <= o8n ; i++) {
                o8xtr[i] = o8xsc[i]*xvar[i];
            }
            eval_extern(1);
            
            o8valid = TRUE;
        }
        return;
    }
    if ( mode >= 1 ) {
    
        /* evaluate functions and its gradients */
        
        for (i = 1 ; i <= o8n ; i++) {
            o8xtr[i] = o8xsc[i]*xvar[i];
        }
        /* even if o8xtr did not change we must evaluate externally, since now */
        /* gradients are required                                            */
        
        if ( o8analyt ) {
        
            eval_extern(2);
            
            o8valid = TRUE;
            
            return;
            
        } else {
            if ( mode == 1 ) eval_extern(1);
        }
        /* continue with computing the gradients : mode = 1 or mode = 2 */
        /* require gradients                                            */
        
        if ( o8difftype == 1 ) {
            o8deldif = min(o8tm1*sqrt(o8epsfcn),o8tm2);
            
            if ( mode == 2 ) eval_extern(1);
            
            for (j = 0 ; j <= o8nonlin ; j++) {
                fusave[j] = o8fu[j];
            }
            for (i = 1 ; i <= o8n ; i++) {
                xhelp = o8xtr[i];
                xincr = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd);
                if ( xhelp >= o8zero ) {
                    o8xtr[i] = xhelp+xincr;
                } else {
                    o8xtr[i] = xhelp-xincr;
                }
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fugrad[i][j] = (o8fu[j]-fusave[j])/(o8xtr[i]-xhelp);
                }
                o8xtr[i] = xhelp;
            }
            for (j = 0 ; j <= o8nonlin ; j++) {
                o8fu[j] = fusave[j];
            }
            o8valid = TRUE;
            
            return;
            
        } else if ( o8difftype == 2 ) {
            o8deldif = min(o8tm1*pow(o8epsfcn,o8one/o8three),o8tm2);
            for (j = 0 ; j <= o8nonlin ; j++) {
                fusave[j] = o8fu[j];
            }
            for (i = 1 ; i <= o8n ; i++) {
                xhelp  = o8xtr[i];
                xincr  = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd);
                o8xtr[i] = xhelp+xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][1] = o8fu[j];
                }
                o8xtr[i] = xhelp-xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][2] = o8fu[j];
                }
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fugrad[i][j] = (o8fud[j][1]-o8fud[j][2])/(xincr+xincr);
                }
                o8xtr[i] = xhelp;
            }
            for (j = 0 ; j <= o8nonlin ; j++) {
                o8fu[j] = fusave[j];
            }
            o8valid = TRUE;
            
            return;
            
        } else {
            o8deldif = min(o8tm1*pow(o8epsfcn,o8one/o8seven),o8tm2);
            for (j = 0 ; j <= o8nonlin ; j++) {
                fusave[j] = o8fu[j];
            }
            for (i = 1 ; i <= o8n ; i++) {
                xhelp  = o8xtr[i];
                xincr  = min(min(o8tm2,o8deldif*fabs(xhelp)+o8deldif),o8taubnd/o8four);
                o8xtr[i] = xhelp-xincr;
                
                eval_extern(1);
        
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][1] = o8fu[j];
                }
                o8xtr[i] = xhelp+xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][2] = o8fu[j];
                }
                xincr  = xincr+xincr;
                d1     = xincr;
                o8xtr[i] = xhelp-xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][3] = o8fu[j];
                }
                o8xtr[i] = xhelp+xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][4] = o8fu[j];
                }
                xincr  = xincr+xincr;
                d2     = xincr;
                o8xtr[i] = xhelp-xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][5] = o8fu[j];
                }
                o8xtr[i] = xhelp+xincr;
                
                eval_extern(1);
                
                for (j = 0 ; j <= o8nonlin ; j++) {
                    o8fud[j][6] = o8fu[j];
                }
                o8xtr[i] = xhelp;
                d3     = xincr+xincr;
                for (j = 0 ; j <= o8nonlin ; j++) {
                    sd1 = (o8fud[j][2]-o8fud[j][1])/d1;
                    sd2 = (o8fud[j][4]-o8fud[j][3])/d2;
                    sd3 = (o8fud[j][6]-o8fud[j][5])/d3;
                    sd3 = sd2-sd3;
                    sd2 = sd1-sd2;
                    sd3 = sd2-sd3;
                    o8fugrad[i][j] = sd1+o8p4*sd2+sd3/o8c45;
                }
            }
            for (j = 0 ; j <= o8nonlin ; j++) {
                o8fu[j] = fusave[j];
            }
            o8valid = TRUE;
            
            return;
        }
    }
}


