/******************************************************************************
*                                                                             *
*      cosmo - Constrained search for motifs in DNA sequences                 *
*      Author: Oliver Bembom                                                  *
*                                                                             *
*      This file is part of cosmo.                                            *
*                                                                             *
*      Conditions of use:                                                     *
*                                                                             *
*    1. cosmo is under the exclusive copyright of Oliver Bembom,              *
*       (bembom@berkeley.edu), the Regents of the University of California,   *
*       and Peter Spellucci (spellucci@mathematik.tu-darmstadt.de).           *
*                                                                             *
*    2. Permission to use, copy, modify, and distribute any part of this      *
*       software for research purposes, without fee and without a written     *
*       agreement, is hereby granted, provided that this license appears in   *
*       all copies. Commercial uses require permission and licensing from     *
*       Oliver Bembom, the Regents of the University of California, and       *
*       Peter Spellucci.                                                      *
*                                                                             *
*    3. If you use this software in your research, please cite                *
*                                                                             *
*       O. Bembom, S. Keles, M.J. van der Laan (2007). Supervised Detection   *
*       of Conserved Motifs in DNA Sequences with cosmo. Statistical          *
*       Applications in Genetics and Molecular Biology: Vol. 6 : Iss. 1,      *
*       Article 8. http://www.bepress.com/sagmb/vol6/iss1/art8.               *
*                                                                             *
*       cosmo makes use of the donlp2() function by Peter Spellucci.          *
*       The use of donlp2() must be acknowledged in any publication which     *
*       contains results obtained with cosmo or parts of it. Citation of the  *
*       author's name and netlib-source is suitable.                          *
*                                                                             *
*    4. In no event shall the author be liable to any party for direct, in-   *
*       direct, special, incidental, or consequential damages, including lost *
*       profits, arising out of the use of this software, even if the author  *
*       has been advised of the possibility of such damage.                   *
*                                                                             *
*    5. cosmo is provided on an 'as is' basis, and the author has no obli-    *
*       gations to provide maintenance, support, updates, enhancements, or    *
*       modifications. The author makes no representations and extends no     *
*       warranties of any kind, either expressed or implied, including, but   *
*       not limited to, the implied warranties of merchantability or fitness  *
*       for a particular purpose, or that the use of the material will not    *
*       infringe any patent, trademark or other rights.                       *
*                                                                             *
*                                                                             *
******************************************************************************/


/*****************************************************************************
 *                                                                           *
 *      This module contains various creating XML output                     *
 *                                                                           *
 ****************************************************************************/

#include <R_ext/PrtUtil.h>
#include <R_ext/Memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "macros.h"
#include "cosmo.h"
#include "alloc.h"
#include "eval.h"
#include "xml.h"
#include "matLib.h"
#include "posterior.h"
#include "io.h"
  
/**********************************************************************/
/*
  xmlOutput

  Print results in XML format to screen.

*/
/**********************************************************************/
int xmlOutput(
	       DATASET *dataSet   /* the dataset */
	       )
{

  /* start XML document */
  Rprintf("<?xml version=\"1.0\"?>\n");
  Rprintf("<results>\n");

  /* print information about cosmo */
  if (!xmlPrintCosmo(dataSet)) return(0);

  /* print reference information */
  if (!xmlPrintReference(dataSet)) return(0);
  
  /* print summary of command line arguments */
  if (!xmlPrintCommandSummary(dataSet)) return(0);

  /* print summary of training set */
  if (!xmlPrintTrainingSet(dataSet)) return(0);

  /* print given constraint file */
  if (!xmlPrintConstraintFile(dataSet)) return(0);

  /* print interpreted constraint sets */
  if (!xmlPrintConstraints(dataSet)) return(0);

  /* print background model */
  if (!xmlPrintTransMat(dataSet)) return(0);	

  /* print summary of selected fine-tuning parameters */
  if (!xmlPrintModelSelection(dataSet)) return(0);

  /* print description of motif */
  if (!xmlPrintMotif(dataSet)) return(0);

  /* print posterior probabilities */
  if (!xmlPrintProbs(dataSet)) return(0);  

  /* print sequences in numerical format */
  if (!xmlPrintSeqs(dataSet)) return(0); 

  /* print computing time */
  if (!xmlPrintTime(dataSet)) return(0);

  /* end XML document */
  Rprintf("</results>\n");

  return(1);
}

/**********************************************************************/
/*
  xmlPrintCosmo

  Print information about cosmo to stdout.

*/
/**********************************************************************/
int xmlPrintCosmo(
		   DATASET *dataSet   /* the dataset */
		   )
{
  Rprintf("<cosmo>\n");
  Rprintf("<section>COSMO - Constrained search for motifs</section>\n");
  Rprintf("<version>COSMO version %s</version>\n", dataSet->version);
  Rprintf("<website>http://cosmoweb.berkeley.edu</website>\n");
  Rprintf("</cosmo>\n");

  return(1);
}



/**********************************************************************/
/*
  xmlPrintReference

  Print reference information to stdout.

*/
/**********************************************************************/
int xmlPrintReference(
		       DATASET *dataSet   /* the dataset */
		       )
{

  Rprintf("<reference>\n");
  Rprintf("<section>Reference</section>");
  Rprintf("<cite>%s\n\n%s</cite>", dataSet->reference, dataSet->donlpReference);
  Rprintf("</reference>\n");
  
  return(1);
}

/**********************************************************************/
/*
  xmlPrintCommandLineSummary

  Print summary of command line to stdout.

*/
/**********************************************************************/
int xmlPrintCommandSummary(
			    DATASET *dataSet    /* the dataset */
			    )
{

  int i;                /* index */
  BOOLEAN nonEmpty;     /* any non-empty constraint sets?*/

  /* are there any non-empty constraint sets? */
  i             = dataSet->hasConstraints ? 1 : 0;
  nonEmpty      = FALSE;
  for (;i<dataSet->numConSets; i++) nonEmpty = nonEmpty || (!dataSet->cSets[i].isEmpty);

  Rprintf("<commandline>\n");
  Rprintf("<section>Call summary</section>\n");
  Rprintf("<intro>This information can also be useful in the event you wish to report a problem with the cosmo software.</intro>\n");
  Rprintf("<command>%s</command>\n", dataSet->command);
  Rprintf("<consets>%d</consets>\n", dataSet->numConSets - 1);
  Rprintf("<nonempty>%s</nonempty>\n", (nonEmpty ? "TRUE" : "FALSE"));
  Rprintf("<concrit>%s</concrit>\n", 	  (dataSet->conCrit.crit==lik ? "lik" : (dataSet->conCrit.crit==aic ? "aic" : (dataSet->conCrit.crit==bic ? "bic" : (dataSet->conCrit.crit==eval ? "eval" : (dataSet->conCrit.crit==likCV ? "likCV" : "pwmCV"))))));
  Rprintf("<confold>%d</confold>\n", dataSet->conCrit.fold);
  Rprintf("<contrunc>%d</contrunc>\n", (int) (100*dataSet->conCrit.quant));
  Rprintf("<oops>%s</oops>\n", (dataSet->oops ? "TRUE" : "FALSE"));
  Rprintf("<zoops>%s</zoops>\n", (dataSet->zoops ? "TRUE" : "FALSE"));
  Rprintf("<tcm>%s</tcm>\n", (dataSet->tcm ? "TRUE" : "FALSE"));
  Rprintf("<approx>%s</approx>\n", (dataSet->approx==TCM ? "TCM" : (dataSet->approx==OVER ? "OVER" : "CUT")));
  Rprintf("<cutfac>%d</cutfac>\n", dataSet->cutFactor);
  Rprintf("<modcrit>%s</modcrit>\n", 	  (dataSet->modCrit.crit==lik ? "lik" : (dataSet->modCrit.crit==aic ? "aic" : (dataSet->modCrit.crit==bic ? "bic" : (dataSet->modCrit.crit==eval ? "eval" : "likCV")))));
  Rprintf("<modfold>%d</modfold>\n", dataSet->modCrit.fold);
  Rprintf("<modtrunc>%d</modtrunc>\n", (int) (100*dataSet->modCrit.quant));
  Rprintf("<minw>%d</minw>\n", dataSet->minWidth);
  Rprintf("<maxw>%d</maxw>\n", dataSet->maxWidth);
  Rprintf("<wcrit>%s</wcrit>\n", 	  (dataSet->wCrit.crit==lik ? "lik" : (dataSet->wCrit.crit==aic ? "aic" : (dataSet->wCrit.crit==bic ? "bic" : (dataSet->wCrit.crit==eval ? "eval" : "likCV")))));
  Rprintf("<wfold>%d</wfold>\n", dataSet->wCrit.fold);
  Rprintf("<wtrunc>%d</wtrunc>\n", (int) (100*dataSet->wCrit.quant));
  Rprintf("<minsites>%d</minsites>\n", dataSet->minNumSites);
  Rprintf("<zoopsmaxsites>%d</zoopsmaxsites>\n", dataSet->zoopsMaxNumSites);
  Rprintf("<tcmmaxsites>%d</tcmmaxsites>\n", dataSet->tcmMaxNumSites);
  Rprintf("<intcrit>%s</intcrit>\n",  (dataSet->intCrit.crit==lik ? "lik" : (dataSet->intCrit.crit==eval ? "eval" : "likCV")));
  Rprintf("<intfold>%d</intfold>\n", dataSet->intCrit.fold);
  Rprintf("<inttrunc>%d</inttrunc>\n", (int) (100*dataSet->intCrit.quant));
  Rprintf("<starts>%s</starts>\n", (dataSet->likStarts ? "lik" : "eval"));
  Rprintf("<nstarts>%d</nstarts>\n", dataSet->numStarts);
  Rprintf("<backsource>%s</backsource>\n",(dataSet->hasBackFile ? "other" : (dataSet->readTransMat ? "matrix" : 
									    (dataSet->readBfile ? "bfile" : "same"))));
  Rprintf("<cvorder>%s</cvorder>\n", (dataSet->orderByCV ? "TRUE" : "FALSE"));
  Rprintf("<backfold>%d</backfold>\n", dataSet->backFold);
  Rprintf("<nnucs>%ld</nnucs>\n", dataSet->totalNucs);
  Rprintf("<nseqs>%d</nseqs>\n", dataSet->numSeqs);
  Rprintf("<rev>%s</rev>\n", (dataSet->revComp ? "TRUE" : "FALSE"));

  Rprintf("</commandline>\n");

  return(1);
}


int xmlPrintTrainingSet(
			 DATASET *dataSet  /* the dataset */
			 )
{

  int i,  w;               /* various indices */
  
  w = MSN + 15;

  Rprintf("<trainset>\n");
  Rprintf("<section>Training set</section>\n");
  Rprintf("<seqfile>%s</seqfile>\n", dataSet->seqFileName);

  for (i=0; i<dataSet->numSeqs; i++) Rprintf("<name>%s</name>\n", dataSet->samples[i].name);
  for (i=0; i<dataSet->numSeqs; i++) Rprintf("<len>%ld</len>\n", dataSet->samples[i].length);

  Rprintf("</trainset>\n");

  return(1);
}

/**********************************************************************/
/*
  xmlPrintConstraints

  Print the original constraint file.

*/
/**********************************************************************/
int xmlPrintConstraintFile(
			    DATASET *dataSet   /* the datset */
			    )
{

  FILE *conFile;
  char *line;

  conFile = dataSet->conFile;
  line    = dataSet->line;

  Rprintf("<confile>\n");
  Rprintf("<section>Original constraint file</section>\n");  

  Rprintf("<constraints>\n");
  if (dataSet->hasConstraints){
    
    /* rewind file */
    fseek(conFile,(long) 0,0);
    while (fgets(line,MAXNUMCHARS,conFile)!=NULL){
      Rprintf("%s", line);
    }


  }else{
    Rprintf("No constraint file was specified");
  }

  Rprintf("\n");
  Rprintf("</constraints>\n");

  Rprintf("</confile>\n");

  return(1);
}



/**********************************************************************/
/*
  xmlPrintModelSelection

  Print summary of selected fine-tuning parameters to stdout.

*/
/**********************************************************************/
int xmlPrintModelSelection(
			    DATASET *dataSet   /* the dataset */
			    )
{

  CRIT *crits;             /* array of stored model selection criteria */
  int cSetNum;             /* current constraint set number */
  int modTypeNum;          /* current model type number */
  int width;               /* current width */
  int index;               /* index into array of criteria */

  crits   = dataSet->allCrits;
  cSetNum = dataSet->hasConstraints ? 1 : 0;


  /* evalue is stored as log(evalue) */
  if (dataSet->intCrit.crit==eval) dataSet->crits.intCrit = exp(dataSet->crits.intCrit);
  if (dataSet->wCrit.crit==eval)   dataSet->crits.wCrit = exp(dataSet->crits.wCrit);
  if (dataSet->modCrit.crit==eval) dataSet->crits.modCrit = exp(dataSet->crits.modCrit); 
  if (dataSet->conCrit.crit==eval) dataSet->crits.conCrit = exp(dataSet->crits.conCrit);

  for (;cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){
	index = MOD(width, modTypeNum, cSetNum);

	if (dataSet->wCrit.crit==eval)   crits[index].wCrit = exp(crits[index].wCrit);
	if (dataSet->modCrit.crit==eval) crits[index].modCrit = exp(crits[index].modCrit);
	if (dataSet->conCrit.crit==eval) crits[index].conCrit = exp(crits[index].conCrit);
      }
    }
  }



  Rprintf("<models>\n");
  Rprintf("<section>Summary of candidate models</section>\n");  


  cSetNum = dataSet->hasConstraints ? 1 : 0;
  for (;cSetNum<dataSet->numConSets; cSetNum++){
    for (modTypeNum=0; modTypeNum<dataSet->numModTypes; modTypeNum++){
      for (width=dataSet->minWidth; width<=dataSet->maxWidth; width++){

	index = MOD(width, modTypeNum, cSetNum);
	Rprintf("<candidate>\n");
	Rprintf("<cset>%d</cset>\n", cSetNum);
	Rprintf("<mtype>%s</mtype>\n", getModName(dataSet, dataSet->mTypes[modTypeNum], 0));
	Rprintf("<width>%d</width>\n", width);
	Rprintf("<wcrit>");
	if (crits[index].wInit){
	  Rprintf("%.3g",  crits[index].wCrit);
	}else{
	  Rprintf("---");
	}
	Rprintf("</wcrit>\n");

	Rprintf("<modcrit>");
	if (crits[index].modInit){
	  Rprintf("%.3g", crits[index].modCrit);
	}else{
	  Rprintf("---");
	}
	Rprintf("</modcrit>");

	Rprintf("<concrit>");
	if (crits[index].conInit){
	  Rprintf("%.3g\n",  crits[index].conCrit);
	}else{
	  Rprintf("---\n");
	}
	Rprintf("</concrit>\n");
	Rprintf("</candidate>\n");
      }
    }
  }
  Rprintf("</models>\n");


  Rprintf("<themodel>\n");
  Rprintf("<section>Selected model</section>\n");  
  
  Rprintf("<selection>\n");
  Rprintf("<parm>Constraints</parm>\n");
  Rprintf("<choice>%d</choice>\n", dataSet->crits.conSetNum);
  Rprintf("<crit>");
  if (dataSet->numConSets>2){
    Rprintf("%s", (dataSet->conCrit.crit==lik ? "Likeliood" : (dataSet->conCrit.crit==aic ? "AIC" : (dataSet->conCrit.crit==bic ? "BIC" : (dataSet->conCrit.crit==eval ? "E-value" : (dataSet->conCrit.crit==likCV ? "Likelihood-based CV" : "PWM-based CV"))))));
  }else{
    Rprintf("---");
  }
  Rprintf("</crit>");
  Rprintf("<critval>");
  if (dataSet->numConSets>2){
    Rprintf("%.3g", dataSet->crits.conCrit);
  }else{
    Rprintf("---");
  }
  Rprintf("</critval>\n");
  Rprintf("</selection>\n");


  Rprintf("<selection>\n");
  Rprintf("<parm>Distribution</parm>\n");
  Rprintf("<choice>%s</choice>\n", getModName(dataSet,dataSet->crits.mType,0));
  Rprintf("<crit>");
  if (dataSet->numModTypes>1){
      Rprintf("%s",   (dataSet->modCrit.crit==lik ? "Likelihood" : (dataSet->modCrit.crit==aic ? "AIC" : (dataSet->modCrit.crit==bic ? "BIC" : (dataSet->modCrit.crit==eval ? "E-value" : "Likelihood-based CV")))));
  }else{
      Rprintf("---");
  }
  Rprintf("</crit>\n");
  Rprintf("<critval>");
  if (dataSet->numModTypes>1){
      Rprintf("%.3g", dataSet->crits.modCrit);    
  }else{
      Rprintf("---");
  }
  Rprintf("</critval>\n");

  Rprintf("</selection>\n");

  Rprintf("<selection>\n");
  Rprintf("<parm>Width</parm>\n");
  Rprintf("<choice>%d</choice>\n", dataSet->crits.width);
  Rprintf("<crit>");
  if (dataSet->numWidths>1){
    Rprintf("%s", (dataSet->wCrit.crit==lik ? "Likelihood" : (dataSet->wCrit.crit==aic ? "AIC" : (dataSet->wCrit.crit==bic ? "BIC" : (dataSet->wCrit.crit==eval ? "E-value" : "Likelihood-based CV")))));
  }else{
    Rprintf("---");
  }

  Rprintf("</crit>\n");
  Rprintf("<critval>");
  if (dataSet->numWidths>1){
    Rprintf("%.3g",dataSet->crits.wCrit);
  }else{
    Rprintf("---");
  }
  Rprintf("</critval>\n");
  Rprintf("</selection>\n");
  
  Rprintf("<selection>\n");
  Rprintf("<parm>NumSites</parm>\n");
  Rprintf("<choice>%d</choice>\n", dataSet->theta->numSites);
  Rprintf("<crit>%s</crit>\n", (dataSet->intCrit.crit==lik ? "Likelihood" : (dataSet->intCrit.crit==eval ? "E-value" : "Likelihood-based CV")));
  Rprintf("<critval>%.3g</critval>\n", dataSet->crits.intCrit);
  Rprintf("</selection>\n");

  Rprintf("</themodel>\n");

  return(1);
}


/**********************************************************************/
/*
  xmlPrintMotif

  Print description of the motif that was found

*/
/**********************************************************************/
int xmlPrintMotif(
		DATASET *dataSet   /* the datset */
		)
{

  int width;                 /* motif width */
  MTYPE mType;               /* model type */
  int conSetNum;             /* constraint number */
  THETA *theta;              /* final parameter estimate */
  int maxNameLength;         /* maximum sequence name length */
  int pos;                   /* position in motif */
  int nuc;                   /* nucleotide */
  int seqNum;                /* sequence number */
  int numSeqs;               /* number of sequences */
  long seqLength;            /* sequence length */
  int *seq;                  /* sequence */
  int *rcSeq;                /* reverse complement of sequence */
  double prob;               /* posterior probability */
  char *name;                /* sequence name */
  int nameLen;               /* length of that name */
  int i;                     /* index */
  int start;                 /* start of motif */
  int leftOff;               /* number of characters to include on the left */
  int rightOff;              /* number of characters to include on the right */
  int sideLen;               /* number of characters to print on sides of motif */
  char oneChar;              /* a single character */
  long maxSeqLength;         /* maximum sequence length */
  int siteNum;               /* index of current site */
  int numSites;              /* number of sites */
  POSTPROB *postProbs;       /* posterior probabilities */
  POSTPROB *maxPostProbs;    /* maximum posterior probabilities */
  BOOLEAN onRevComp;         /* is the site on the reverse complement? */
  BOOLEAN global;            /* sort posterior probabilities globally? */


  sideLen       = 10;
  numSeqs       = dataSet->numSeqs;
  maxNameLength = dataSet->maxNameLength;
  postProbs     = dataSet->postProbs;
  maxPostProbs  = dataSet->maxPostProbs;
  maxSeqLength  = dataSet->maxSeqLength;
  width         = dataSet->crits.width;
  mType         = dataSet->crits.mType;
  conSetNum     = dataSet->crits.conSetNum;
  theta         = dataSet->theta;
  numSites      = theta->numSites;
  global        = theta->mType==TCM;

  /* get posterior probabilities */
  if (!getPostProbs(theta, dataSet)) return(0);

  /* sort posterior probabilities */
  if (!sortPostProbs(dataSet, width, global)) return(0);
    

  
  /* print position weight matrix */
  Rprintf("<pwm>\n");
  Rprintf("<section>Estimated position weight matrix</section>\n");
  Rprintf("<nrows>4</nrows>\n");
  Rprintf("<ncols>%d</ncols>\n", width);
  for(nuc=0;nuc<4;nuc++) Rprintf("<nuc>%c</nuc>\n",Convert2Char(nuc));

  for(nuc=0;nuc<4;nuc++){
    Rprintf("<row>\n");
    for(pos=0;pos<width;pos++){
      Rprintf("<elem>%.3lf</elem>\n", theta->pwm[PWM(pos,nuc)]);
    }
    Rprintf("</row>\n");
  }
  Rprintf("</pwm>\n");

  Rprintf("<icprofile>\n");
  Rprintf("<section>Information Content Profile</section>\n");
  for(pos=0;pos<width;pos++) Rprintf("<ic>%.3lf</ic>\n", theta->posIC[pos]);
  Rprintf("</icprofile>\n");
  
  /*Extract the actual sequences*/
  Rprintf("<alignment>\n");
  Rprintf("<section>Alignment of discovered sites</section>\n");
  Rprintf("<eval>%.3g</eval>\n", exp(theta->logEval));

  
  for (siteNum=0; siteNum<numSites;siteNum++){
    
    start      = maxPostProbs[siteNum].start;
    seqNum     = maxPostProbs[siteNum].seqNum;
    onRevComp  = maxPostProbs[siteNum].revComp;
    prob       = maxPostProbs[siteNum].prob;
    seq        = dataSet->samples[seqNum].seq;
    rcSeq      = dataSet->samples[seqNum].rcSeq;
    name       = dataSet->samples[seqNum].name;
    nameLen    = dataSet->samples[seqNum].nameLength;
    seqLength  = dataSet->samples[seqNum].length;
    


    Rprintf("<site>\n");

    /* print sequence name */
    /* print name and strand */
    Rprintf("<seqname>%s</seqname>\n", name);
    Rprintf("<strand>%s</strand>\n", onRevComp ? "-" : "+");
    Rprintf("<start>%d</start>\n", start+1);
    Rprintf("<prob>%7.4lf</prob>\n", prob);


    /* 
       number of nucleotides to the left and right that will be printed
       check for end of sequence here 
    */
    leftOff = MIN(sideLen,start);
    rightOff = MIN(sideLen,seqLength-start-width);
    
    for (i=0;i<sideLen-leftOff;i++){
      Rprintf(" ");
    }
    
    /* print nucleotides to left of motif */
    Rprintf("<pre>");
    for (pos=start-leftOff;pos<start;pos++){
      oneChar = Convert2LowChar(seq[pos]);
      Rprintf("%c",oneChar);
    }
    Rprintf("</pre>\n");

    /* print actual motif */
    Rprintf("<motif>");
    for (pos=0;pos<width;pos++){
      oneChar=Convert2Char(seq[start + pos]);
      Rprintf("%c",oneChar);
    }
    Rprintf("</motif>\n");    

    /* print nucleotides to the right of motif */
    Rprintf("<post>");
    for (pos=start+width;pos<start+width+rightOff;pos++){
      oneChar = Convert2LowChar(seq[pos]);
      Rprintf("%c",oneChar);
    }
    Rprintf("</post>\n");

    Rprintf("</site>\n");
  }

  Rprintf("</alignment>\n");

  return(1); 
}

/**********************************************************************/
/*
  xmlPrintProbs

  Print posterior probabilities.
  Posterior probabilities are negative if motif is thought to be on
  the reverse complement strand.

*/
/**********************************************************************/
int xmlPrintProbs(
		   DATASET *dataSet /* the dataset */
		   )
{

  int seqNum;        /* sequence number */
  int numSeqs;       /* number of sequences */
  int pos;           /* position in motif */
  THETA *theta;      /* final parameter estimates */
  double prob;       /* posterior probability */
  long seqLength;    /* length of current sequence */
  long maxSeqLength; /* maximum sequence length */
  int rcFac;         /* make probabilities negative if motif is thought to be on reverse complement */
 
  numSeqs      = dataSet->numSeqs;
  theta        = dataSet->theta;
  maxSeqLength = dataSet->maxSeqLength;

  /* calculate posterior probabilities to be safe */
  if (!getPostProbs(theta,dataSet)) return(0);
    
  Rprintf("<probs>\n");

  for (seqNum=0; seqNum<numSeqs; seqNum++){
    seqLength = dataSet->samples[seqNum].length;
    
    for (pos=0; pos<seqLength; pos++){
      prob  = dataSet->postProbs[PP(seqNum,pos)].prob;
      rcFac = dataSet->postProbs[PP(seqNum,pos)].revComp ? -1 : 1;
      Rprintf("%g\t", rcFac*prob);
    }
    Rprintf("\n");
  }

  Rprintf("</probs>\n");

  return(1);
}

/**********************************************************************/
/*
  xmlPrintSeqs

  Print input sequences in numerical format (A=1, C=2, G=3, T=4)
  one sequence per line.


*/
/**********************************************************************/
int xmlPrintSeqs(
		  DATASET *dataSet /* the dataset */
		  )
{

 
  int seqNum;        /* sequence number */
  int pos;           /* position in sequence */
    

  Rprintf("<rseqs>\n");
  for (seqNum=0; seqNum<dataSet->numSeqs; seqNum++){
    for (pos=0;pos<dataSet->samples[seqNum].length;pos++) Rprintf("%d ", dataSet->samples[seqNum].seq[pos] + 1);
    Rprintf("\n");
  }

  Rprintf("</rseqs>\n");

  return(1);
}


/**********************************************************************/
/*
  xmlPrintConstraints

  Print the interpreted constraints

*/
/**********************************************************************/
int xmlPrintConstraints(
		      DATASET *dataSet   /* the datset */
		      )
{
  int cSetNum;             /* index of current constraint set */
  int intNum;              /* current interval */
  int i;                   /* index */
  CONSET *cSet;            /* current constraint set */
  INTFREQCON *intFreqCon;  /* interval frequency constraint */
  POSFREQCON *posFreqCon;  /* position frequency constraint */
  LINCON *linCon;          /* linear constraint */

  Rprintf("<interpreted>\n");
  Rprintf("<section>Interpreted constraints</section>\n");
  Rprintf("<constraints>\n");

  cSetNum = dataSet->hasConstraints ? 1 : 0;

  for (;cSetNum < dataSet->numConSets; cSetNum++){
    cSet = dataSet->cSets + cSetNum;
    
    Rprintf("@ Constraint set %d\n\n", (dataSet->hasConstraints ? cSetNum : cSetNum+1));

    /* Interval setup */
    Rprintf(">IntervalSetup\n");
    for (intNum=0; intNum < cSet->numInts; intNum++){
      Rprintf("Length: ");
      switch(cSet->intCons[intNum].lenType){
      case bp:
	Rprintf("%d bp\n", cSet->intCons[intNum].numBasePairs);
	break;
      case prop:
	Rprintf("%.3lf\n", 100*cSet->intCons[intNum].prop);
	break;
      case var:
	Rprintf("variable\n");
	break;
      default:
	Rprintf( "printConstraints: Length type %d not recognized. Exiting...\n", cSet->intCons[intNum].lenType);
	return(0);
      }
    }
    Rprintf("\n");

    /* Interval constraints */
    
    for (intNum=0; intNum < cSet->numInts; intNum++){

      if (cSet->intCons[intNum].hasBoundCon){
	Rprintf(">ICBounds\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Bounds: %.3lf to %.3lf\n\n", cSet->intCons[intNum].boundCon.low, cSet->intCons[intNum].boundCon.up);
      }

      if (cSet->intCons[intNum].hasShapeCon){
	Rprintf(">ICShape\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Shape: %s\n", (cSet->intCons[intNum].shapeCon.shape == lin ? "Linear" : (cSet->intCons[intNum].shapeCon.shape == monInc ? "MonotoneIncreasing" : "Monotone Decreasing"))); 
	Rprintf("LeftBounds: %.3lf to %.3lf\n", cSet->intCons[intNum].shapeCon.startLow, cSet->intCons[intNum].shapeCon.startUp);
	Rprintf("RightBounds: %.3lf to %.3lf\n", cSet->intCons[intNum].shapeCon.endLow, cSet->intCons[intNum].shapeCon.endUp);
	Rprintf("ErrorTol: %.3lf\n\n", cSet->intCons[intNum].shapeCon.error);
      }
      
      intFreqCon = cSet->intCons[intNum].intFreqCons;
      while (intFreqCon!=NULL){
	Rprintf(">NucFreq\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Pos: all\n");
	Rprintf("Nuc: %s\n", Convert2String(intFreqCon->nuc));
	Rprintf("LowerBound: %.3lf\n\n", intFreqCon->low);
	intFreqCon = intFreqCon->next;
      }

      posFreqCon = cSet->intCons[intNum].posFreqCons;
      while (posFreqCon!=NULL){
	Rprintf(">NucFreq\n");
	Rprintf("Interval: %d\n", intNum+1);
	Rprintf("Pos: %d\n", posFreqCon->pos);
	Rprintf("Nuc: %s\n", Convert2String(posFreqCon->nuc));
	Rprintf("LowerBound: %.3lf\n\n", posFreqCon->low);
	posFreqCon = posFreqCon->next;
      }
    }

    if (cSet->hasPalCon){
      Rprintf(">Palindrome\n");
      Rprintf("Intervals: %d and %d\n", cSet->palCon.int1+1, cSet->palCon.int2+1);
      Rprintf("ErrorTol: %.3lf\n\n", cSet->palCon.error);
    }

    if (cSet->hasSubCon){
      Rprintf(">SubMotif:\n");
      Rprintf("Motif: ");
      for (i=0; i< cSet->subCon.length;i++) Rprintf("%c",Convert2Char(cSet->subCon.motif[i]));
      Rprintf("\nMinFreq: %.3lf\n\n", cSet->subCon.cutOff);
    }

    linCon = cSet->linCons;
    while (linCon!=NULL){
      Rprintf(">ParmDiff\n");
      Rprintf("Parameters: %d%c - %d%c\n", linCon->int1 +1, (linCon->pos1==start ? 'a' : 'b'), linCon->int2 + 1,  
	     (linCon->pos2==start ? 'a' : 'b'));
      Rprintf("Bounds: %.3lf to %.3lf\n\n", linCon->low, linCon->up);
      linCon = linCon->next;
    }

  }
  Rprintf("</constraints>\n");
  Rprintf("</interpreted>\n");

  return(1);
}

/**********************************************************************/
/*
  xmlPrintTime

  Print the computing time in XML format

*/
/**********************************************************************/
int xmlPrintTime(
		  DATASET *dataSet   /* the dataset */
		  )
{
  double time;       /* elapsed time in seconds */
  int minutes;       /* full minutes in elapsed time */
  int hours;         /* full hours in elapsed time */

  time    = dataSet->elapsedTime;
  minutes = (int) (dataSet->elapsedTime/60);
  hours   = (int) (dataSet->elapsedTime/3600);

  if (time<60){
    Rprintf("<time>%.2g seconds</time>\n", time);
  }else if (time <3600){
    Rprintf("<time>%d minutes %.2g seconds</time>\n", minutes, time - 60*minutes);
  }else{
    Rprintf("<time>%d hours %d minutes %.2g seconds</time>\n", hours, minutes, time - 3600*hours - 60*minutes);
  }

  return(1);
}


/**********************************************************************/
/*
  xmlPrintTransmat

  Print out the transition matrix

*/
/**********************************************************************/
int xmlPrintTransMat(
		      DATASET *dataSet    /* the dataset */
		      )
{

  int i,j,k,l;         /* various indices */
  double *transMat;    /* current transition matrix */
  char *prefix;        /* the last k nucletides */
  int sum;             /* sum of all higher powers for prefix */
  int part;            /* part of this power for prefix */
  
  prefix = CharMalloc(dataSet->maxOrder+1);

  
  Rprintf("<background>\n");  

  Rprintf("<section>Estimated background model</section>\n");

  Rprintf("<order>%d</order>\n", dataSet->order);

  for (i=0; i<=dataSet->maxOrder; i++){
    if (dataSet->backKL[i]==-1){
      Rprintf("<kldiv>NA</kldiv>\n");
    }else if (dataSet->backKL[i]==0){
      Rprintf("<kldiv>inf</kldiv>\n");
    }else{
      Rprintf("<kldiv>%.5g</kldiv>\n", dataSet->backKL[i]);
    }
  }

  for (i=0; i<=MIN(dataSet->order,2);i++){
    Rprintf("<transmat>\n");

    transMat = dataSet->transMats[i];
    for (j=0; j < pow(4,i);j++){

      /* get prefix */
      sum = 0;
      l = 0;
      for (k=i-1;k>=0;k--,l++){
	part = (int)((j-sum)/(pow(4,k)));
	sum += part*pow(4,k);
	prefix[l] = Convert2Char(part);
	
      }
      prefix[i] = '\0';
      
      Rprintf("<prefix>%s</prefix>\n", prefix);

    }
    for (j=0; j < pow(4,i);j++){
      for (k=0; k<4;k++) Rprintf("<prob>%5.4lf</prob>\n",exp(transMat[4*j+k]));
    }

    Rprintf("</transmat>\n");
	   
  }

  Rprintf("</background>\n");

  return(1);
}

